/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Definitions of an Error and associated functions.
 */


#include <jwsc/config.h>

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <assert.h>
#include <inttypes.h>

#include "jwsc/base/error.h"

#ifdef JWSC_HAVE_DMALLOC
#include <dmalloc.h>
#endif


extern int errno;


/** @brief Number of errors created and not handled (freed). */
static uint32_t num_unhandled_errors = 0;


/** @brief Used for special memory Exception. */
static Error no_memory = 
{ 
    ERROR_NO_MEMORY, 
    "No memory available.",
    NULL,
    0,
    0
};


/** Global pointer to the static Error @em no_memory. */
Error* error_no_memory = &no_memory;


/**
 * Any and all of the const char* parameters can be NULL. If non-NULL, a copy 
 * of the the const char* parameters is allocated. The Error::code is set to 0.
 *
 * @param  type  Error_type.
 * @param  file  File where the Error occurred.
 * @param  line  Line near the offending code.
 * @param  msg   Message for the Error.
 *
 * @returns An Error is always returned. If the Error could not be created 
 * because of memory exhaustion, error_no_memory is returned. 
 */
Error* create_error
(
    Error_type   type, 
    const char*  file,
    unsigned int line,
    const char*  msg
)
{
    size_t len;
    const char* ptr;
    Error* e;

    if ((e = malloc(sizeof(Error))) == NULL)
    {
        return error_no_memory;
    }

    num_unhandled_errors++;

    e->type = type;
    e->file = NULL;
    e->line = line;
    e->msg  = NULL;
    e->code = 0;

    if (file != NULL)
    {
        len = strlen(file);
        ptr = file + len;
        while (ptr != file && *ptr != '/') ptr--;
        if (ptr != file) ptr++;
        len = strlen(ptr);

        if ((e->file = malloc((len+1)*sizeof(char))) == NULL)
        {
            free_error(e);
            return error_no_memory;
        }
        strcpy(e->file, ptr);
    }
    if (msg != NULL)
    {
        if ((e->msg = malloc((strlen(msg)+1)*sizeof(char))) == NULL)
        {
            free_error(e);
            return error_no_memory;
        }
        strcpy(e->msg, msg);
    }

    return e;
}


/** @param  e  Error to free. */
void free_error(Error* e)
{
    if (e == NULL || e == error_no_memory) return;

    num_unhandled_errors--;

    if (e->file != NULL)
    {
        free(e->file);
    }
    if (e->msg != NULL)
    {
        free(e->msg);
    }

    free(e);
}


/**
 * Prints all the non-NULL fields of the Error in the format
 *
 *   @code 'prefix: file: line: msg\n' @endcode
 *
 * If @em prefix == NULL, just the error message is printed. The Error @em e 
 * must not be NULL.
 *
 * @param  prefix   Prefix to print before the error message.
 * @param  e        Error to print.
 */
void print_error(const char* prefix, const Error* e)
{
    int had_prefix;

    assert(e != NULL);
    had_prefix = 0;

    if (prefix != NULL)
    {
        fprintf(stderr, "%s", prefix);
        had_prefix = 1;
    }
    if (e->file != NULL)
    {
        if (had_prefix)
        {
            fprintf(stderr, ": %s", e->file);
        }
        else
        {
            fprintf(stderr, "%s", e->file);
            had_prefix = 1;
        }
    }
    if (had_prefix)
    {
        fprintf(stderr, ": line %u", e->line);
    }
    else
    {
        fprintf(stderr, "line %u", e->line);
        had_prefix = 1;
    }
    if (e->msg != NULL)
    {
        if (had_prefix)
        {
            fprintf(stderr, ": %s", e->msg);
        }
        else
        {
            fprintf(stderr, "%s", e->msg);
        }
    }
    fprintf(stderr, "\n");
}


void print_error_exit(const char* prefix, const Error* e)
{
    print_error(prefix, e);
    exit(EXIT_FAILURE);
}


/**
 * Prints the Error in the format
 *
 *   @code 'prefix: msg\n' @endcode
 *
 * If @em prefix == NULL, just the error message is printed. If @em msg is NULL,
 * no new-line is printed. This is useful for printing multi-prefixed error
 * messages.
 *
 * @param  prefix   Prefix to print before the error message.
 * @param  msg      Error message to print.
 */
void print_error_msg(const char* prefix, const char* msg)
{
    if (prefix != NULL)
    {
        fprintf(stderr, "%s: ", prefix);
    }

    if (msg != NULL)
    {
        fprintf(stderr, "%s\n", msg);
    }
}


/**
 * Prints the Error in the format
 *
 *   @code 'prefix: msg\n' @endcode
 *
 * If @em prefix == NULL, just the error message is printed. If @em msg is NULL,
 * no new-line is printed. This is useful for printing multi-prefixed error
 * messages.
 *
 * The exit status is EXIT_FAILURE.
 *
 * @param  prefix   Prefix to print before the error message.
 * @param  msg      Error message to print.
 */
void print_error_msg_exit(const char* prefix, const char* msg)
{
    print_error_msg(prefix, msg);
    exit(EXIT_FAILURE);
}


/**
 * Prints the Error in the format
 *
 *   @code 'prefix: errno_msg\n' @endcode
 *
 * The standard function @em strerror is used to get the message for @em errno.
 *
 * If @em prefix == NULL, just the error message is printed.
 *
 * @param  prefix  Prefix to print before the error message.
 */
void print_errno(const char* prefix)
{
    if (prefix != NULL)
    {
        fprintf(stderr, "%s: ", prefix);
    }

    fprintf(stderr, "%s\n", strerror(errno));
}


/**
 * Prints the Error in the format
 *
 *   @code 'prefix: errno_msg\n' @endcode
 *
 * The standard function @em strerror is used to get the message for @em errno.
 *
 * If @em prefix == NULL, just the error message is printed.
 *
 * The exit status is EXIT_FAILURE.
 *
 * @param  prefix  Prefix to print before the error message.
 */
void print_errno_exit(const char* prefix)
{
    print_errno(prefix);
    exit(EXIT_FAILURE);
}


/** @return The number of errors created and never handled (freed). */
uint32_t get_num_unhandled_errors()
{
    return num_unhandled_errors;
}
