/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Declarations of an Error and associated functions.
 */


#ifndef ERROR_H
#define ERROR_H


#include <jwsc/config.h>

#include <stdlib.h>
#include <inttypes.h>


#ifdef __cplusplus
namespace jwsc {
extern "C" {
#endif


/** @brief Error types. */
typedef enum
{
    ERROR_NO_MEMORY,     /*!< No memory available. */
    ERROR_IO,            /*!< Input/output problem. */
    ERROR_INV_ARG,       /*!< Argument is not in valid range. */
    ERROR_CALC,          /*!< Calculated value outside error tolerance. */
    ERROR_NOT_READY,     /*!< Non-blocking Result is not yet ready. */
    ERROR_SOCKET,        /*!< Error creating a socket. */
    ERROR_CONNECT        /*!< Error connecting a socket. */
}
Error_type;


/** @brief Error structure used for error handling. */
typedef struct
{
    /** @brief Error type. */
    Error_type type;

    /** @brief Description message. */
    char* msg;

    /** @brief File name where the Error occurred. */
    char* file;

    /** @brief Line number where the Error occurred. */
    unsigned int line;

    /** @brief Specific code for the Error. */
    int code;
}
Error;


/**
 * @brief Special Error for memory allocation errors.
 *
 * This is a very special Error: it is static and must never be freed.
 */
extern Error* error_no_memory;


/** @brief Creates a new Error. */
Error* create_error
(
    Error_type   type, 
    const char*  file,
    unsigned int line,
    const char*  msg
);


/** @brief Frees an Error. */
void free_error(Error* e);


/** @brief Prints an Error on @em stderr. */
void print_error(const char* prefix, const Error* e);


/** @brief Prints an Error on @em stderr and exits. */
void print_error_exit(const char* prefix, const Error* e);


/** @brief Prints an error message on @em stderr. */
void print_error_msg(const char* prefix, const char* msg);


/** @brief Prints an error message on @em stderr and exits. */
void print_error_msg_exit(const char* prefix, const char* msg);


/**
 * @brief Prints the error message associated with the current value of @em
 * errno on @em stderr.
 */
void print_errno(const char* prefix);


/**
 * @brief Prints the error message associated with the current value of @em 
 * errno on @em stderr and exits.
 */
void print_errno_exit(const char* prefix);


/** @brief Returns the number of errors created and never handled (freed). */
uint32_t get_num_unhandled_errors();


/**
 * @brief Creates a new error caused by invalid argument.
 *
 * Macro shortcut to include file name and line number when creating an error.
 *
 * @param  msg  Error message.
 *
 * @return The error.
 */
#define  JWSC_EARG(msg)  create_error(ERROR_INV_ARG, __FILE__, __LINE__, msg)


/**
 * @brief Creates a new error caused by invalid I/O.
 *
 * Macro shortcut to include file name and line number when creating an error.
 *
 * @param  msg  Error message.
 *
 * @return The error.
 */
#define  JWSC_EIO(msg)  create_error(ERROR_IO, __FILE__, __LINE__, msg)


/**
 * @brief Creates a new error caused by a calculation.
 *
 * Macro shortcut to include file name and line number when creating an error.
 *
 * @param  msg  Error message.
 *
 * @return The error.
 */
#define  JWSC_ECALC(msg)  create_error(ERROR_CALC, __FILE__, __LINE__, msg)


#ifdef __cplusplus
}
}
#endif


#endif
