/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Definitions for file I/O utility functions.
 */


#include <jwsc/config.h>

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <assert.h>
#include <inttypes.h>

#include "jwsc/base/file_io.h"


/**
 * @brief Skips the @em fp ahead over blank space and lines beginning with '#'.
 *
 * @param  fp  File to skip comments and blank lines in.
 *
 * @return 0 if @em fp is positioned at the EOF after skipping over all
 * comments; 1 otherwise.
 */
int skip_fp_spaces_and_comments(FILE* fp)
{
    char c;
    int done;

    done = 0;

    while (!done && (c = fgetc(fp)) != EOF)
    {
        if (c == '#')
        {
            while ((c = fgetc(fp)) != EOF && c != '\n') 
                ;
            if (c == EOF) 
                done = 1;
        }
        else if (!isspace(c))
        {
            ungetc(c, fp);
            done = 1;
        }
    }

    return (c != EOF);
}


/**
 * Skips the file pointer ahead, one position past the newline character '\\n'.
 *
 * @param  fp  File to skip a line of input from.
 *
 * @return 0 if @em fp is positioned at the EOF after skipping a line; 1
 * otherwise.
 */
int skip_line(FILE* fp)
{
    char c;

    while ((c = fgetc(fp)) != EOF && c != '\n') ;

    return (c != EOF);
}


/**
 * The file pointer is advanced to either just after the newline character or
 * EOF.
 *
 * @param  fp  File to count the columns in.
 *
 * @return Number of columns on the line.
 */
int count_fp_line_columns(FILE* fp)
{
    int  count;
    int  done;
    int  in_column;
    char c;

    count     = 0;
    done      = 0;
    in_column = 0;

    while(!done)
    {
        if ((c = fgetc(fp)) == EOF || c == '\n')
        {
            done = 1;
        }
        else if (isspace(c))
        {
            in_column = 0;
        }
        else if (!in_column)
        {
            count++;
            in_column = 1;
        }
    }

    return count;
}
