/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Provides program option support via the command-line or an options 
 * file.
 *
 * A program can use this option support by defining a set of Option_with_arg
 * and Option_no_arg structures and processing the command-line arguments or an
 * option file with them.
 *
 * The options can be represented by a long Option_string and/or a single 
 * character Option_flag. Both can take an argument by defining an 
 * Option_with_arg. If no argument is called for, use Option_no_arg.
 */


#ifndef OPTION_H
#define OPTION_H


#include <jwsc/config.h>

#include <stdlib.h>
#include <stdio.h>

#include "jwsc/base/error.h"


#ifdef __cplusplus
namespace jwsc {
extern "C" {
#endif


/** @brief String representing the long name of an option. */
typedef const char* Option_string;

/** @brief Character representing the short name (flag) of an option. */
typedef char Option_flag;

/** @brief String containing a brief option description. */
typedef const char* Option_desc;

/** @brief Argument to an option. */
typedef const char* Option_arg;

/** @brief Callback function for processing an option with an argument. */
typedef Error* (*Option_func_with_arg)(Option_arg);

/** @brief Callback function for processing an option with no argument. */
typedef Error* (*Option_func_no_arg)(void);


/**
 * @brief Data structure for an option that takes an argument.
 *
 * An option is specified on the command-line as follows
 *
 * @code
 *   --option-string=arg
 *   -O arg
 * @endcode
 *
 * where 
 *
 * @code
 *   Option_with_arg::string = "option-string"
 *   Option_with_arg::flag = 'O'
 *   Option_arg = "arg"
 * @endcode
 *
 * Specifying an option in a file is similar, but only the @link
 * Option_with_arg::string string @endlink can be used, for instance 
 *
 * @code
 *   option-string=arg
 * @endcode
 */
typedef struct
{
    /** 
     * @brief Long name of the option. 
     *
     * If not used, set to NULL.
     */
    Option_string string;

    /** 
     * @brief Character flag for the option. 
     *
     * If not used, set to 0 or '\\0'.
     */ 
    Option_flag flag;

    /** 
     * @brief String containing a short description of the option.
     *
     * If not used, set to NULL.
     */
    Option_desc desc;

    /** @brief Function for processing the option and its argument. */
    Option_func_with_arg func;
}
Option_with_arg;


/**
 * @brief Data structure for an option that does not take an argument.
 *
 * An option is specified on the command-line as follows
 *
 * @code
 *   --option-string
 *   -O
 * @endcode
 *
 * where 
 *
 * @code
 *   Option_no_arg::string = "option-string"
 *   Option_no_arg::flag = 'O'
 * @endcode
 *
 * Specifying an option in a file is similar, but only the @link
 * Option_with_arg::string string @endlink can be used.
 */
typedef struct
{
    /** 
     * @brief Long name of the option.
     *
     * If not used, set to NULL.
     */
    Option_string string;

    /** 
     * @brief Character flag for the option.
     *
     * If not used, set to 0 or '\\0'.
     */
    Option_flag flag;

    /** 
     * @brief String containing a short description of the option.
     *
     * If not used, set to NULL.
     */
    Option_desc desc;

    /** @brief Function for processing the option, no argument. */
    Option_func_no_arg func;
}
Option_no_arg;


/**
 * @brief Initializes an Option_with_arg.
 */
void init_option_with_arg
(
    Option_with_arg*     opt,
    Option_string        str, 
    Option_flag          flag, 
    Option_desc          desc,
    Option_func_with_arg func
);


/**
 * @brief Initializes an Option_no_arg.
 */
void init_option_no_arg
(
    Option_no_arg*     opt,
    Option_string      str, 
    Option_flag        flag, 
    Option_desc        desc,
    Option_func_no_arg func
);


/**
 * @brief Processes an Option_with_arg from the command-line arguments.
 */
Error* process_option_with_arg
(
    int              argc, 
    const char**     argv, 
    int*             argi,
    Option_with_arg* opt
);


/**
 * @brief Processes an Option_no_arg from the command-line arguments.
 */
Error* process_option_no_arg
(
    int            argc, 
    const char**   argv, 
    int*           argi,
    Option_no_arg* opt
);


/**
 * @brief Processes a set of Option_with_arg and Option_no_arg from the 
 * command-line arguments.
 */
Error* process_options
(
    int              argc, 
    const char**     argv, 
    int*             argi,
    int              num_opts_no_arg,
    Option_no_arg*   opts_no_arg,
    int              num_opts_with_arg,
    Option_with_arg* opts_with_arg
);


/**
 * @brief Processes a set of Option_with_arg and Option_no_arg from an option
 * file.
 */
Error* process_options_from_file
(
    const char*      fname,
    int              num_opts_no_arg,
    Option_no_arg*   opts_no_arg,
    int              num_opts_with_arg,
    Option_with_arg* opts_with_arg
);


/**
 * @brief Prints an Option_with_arg to a file stream.
 */
void print_option_with_arg
(
    FILE*            fp,
    int              opt_width,
    Option_with_arg* opt
);


/**
 * @brief Prints an Option_no_arg to a file stream.
 */
void print_option_no_arg
(
    FILE*          fp,
    int            opt_width,
    Option_no_arg* opt
);


/**
 * @brief Prints options to a file stream.
 */
void print_options
(
    FILE*            fp,
    int              opt_width,
    int              num_opts_no_arg,
    Option_no_arg*   opts_no_arg,
    int              num_opts_with_arg,
    Option_with_arg* opts_with_arg
);


#ifdef __cplusplus
}
}
#endif


#endif
