/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Utility functions to track process time.
 */


#include <jwsc/config.h>

#include <stdlib.h>
#include <stdio.h>
#include <inttypes.h>
#include <unistd.h>
#include <sys/times.h>

#include "jwsc/base/time.h"


/** 
 * @param  t  Timer structure to initialize.
 */
void start_timer(Timer* t)
{
    times(&(t->time));
}


/** 
 * @param  t  Timer structure to stop.
 */
void stop_timer(Timer* t)
{
    struct tms now;

    times(&now);

    t->time.tms_utime  = now.tms_utime  - t->time.tms_utime;
    t->time.tms_stime  = now.tms_stime  - t->time.tms_stime;
    t->time.tms_cutime = now.tms_cutime - t->time.tms_cutime;
    t->time.tms_cstime = now.tms_cstime - t->time.tms_cstime;
}


/**
 * @param  t_1  Timer to add to.
 * @param  t_2  Timer to add from.
 */
void add_time_to_timer(Timer* t_1, const Timer* t_2)
{
    t_1->time.tms_utime  += t_2->time.tms_utime;
    t_1->time.tms_stime  += t_2->time.tms_stime;
    t_1->time.tms_cutime += t_2->time.tms_cutime;
    t_1->time.tms_cstime += t_2->time.tms_cstime;
}


/** @param  t  Timer to zero. */
void zero_timer(Timer* t)
{
    t->time.tms_utime  = 0;
    t->time.tms_stime  = 0;
    t->time.tms_cutime = 0;
    t->time.tms_cstime = 0;
}


/**
 * @name user_time
 *
 * Returns the user time of a Timer.
 * @{
 */

double user_time_s(const Timer* t)
{
    return (double)t->time.tms_utime / (double)sysconf(_SC_CLK_TCK);
}


double user_time_ms(const Timer* t)
{
    return (double)(1000 * t->time.tms_utime) / 
           (double)sysconf(_SC_CLK_TCK);
}

/*@}*/


/**
 * @name sys_time
 *
 * Returns the system time of a Timer.
 * @{
 */

double sys_time_s(const Timer* t)
{
    return (double)t->time.tms_stime / (double)sysconf(_SC_CLK_TCK);
}


double sys_time_ms(const Timer* t)
{
    return (double)(1000 * t->time.tms_stime) / 
           (double)sysconf(_SC_CLK_TCK);
}

/*@}*/


/**
 * @name proc_user_time
 *
 * Returns the current ellapsed process user time.
 * @{
 */

double proc_user_time_s()
{
    struct tms now;

    times(&now);

    return (double)now.tms_utime / (double)sysconf(_SC_CLK_TCK);
}


double proc_user_time_ms()
{
    struct tms now;

    times(&now);

    return (double)(1000 * now.tms_utime) / (double)sysconf(_SC_CLK_TCK);
}

/*@}*/


/**
 * @name proc_sys_time
 *
 * Returns the current ellapsed process system time.
 * @{
 */

double proc_sys_time_s()
{
    struct tms now;

    times(&now);

    return (double)now.tms_stime / (double)sysconf(_SC_CLK_TCK);
}


double proc_sys_time_ms()
{
    struct tms now;

    times(&now);

    return (double)(1000 * now.tms_stime) / (double)sysconf(_SC_CLK_TCK);
}

/*@}*/
