/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Utility functions to track process time.
 */


#ifndef TIME_H
#define TIME_H


#include <jwsc/config.h>

#include <stdlib.h>
#include <stdio.h>
#include <inttypes.h>
#include <sys/times.h>


#ifdef __cplusplus
namespace jwsc {
extern "C" {
#endif


/**
 * @brief Timer structure.
 *
 * Holds the time of initialization and can be reset to the current time. Used
 * to measure the length of ellapsed time.
 *
 * @note Does not require dynamic allocation, should be an automatic.
 */
typedef struct
{
    /** @brief System dependent time structure. */
    struct tms time;
}
Timer;


/** @brief Starts the timer by initializing it to the current process time. */
void start_timer(Timer* t);


/** 
 * @brief Stops the timer by differencing a started timer with the current
 * process time.
 */
void stop_timer(Timer* t);


/** @brief Adds time from one timer to another. */
void add_time_to_timer(Timer* t_1, const Timer* t_2);


/** @brief Initializes a timer to zero. */
void zero_timer(Timer* t);


/**
 * @name user_time
 *
 * Returns the user time of a Timer.
 * @{
 */

/** @brief Returns the user time in seconds. */
double user_time_s(const Timer* t);


/** @brief Returns the user time in milliseconds. */
double user_time_ms(const Timer* t);

/*@}*/


/**
 * @name sys_time
 *
 * Returns the system time of a Timer.
 * @{
 */

/** @brief Returns the system time of a Timer in seconds. */
double sys_time_s(const Timer* t);


/** @brief Returns the system time of a Timer in milliseconds. */
double sys_time_ms(const Timer* t);

/*@}*/


/**
 * @name proc_user_time
 *
 * Returns the current ellapsed process user time.
 * @{
 */

/** @brief Returns the current ellapsed process user time in seconds. */
double proc_user_time_s();


/** @brief Returns the current ellapsed process user time in milliseconds. */
double proc_user_time_ms();

/*@}*/


/**
 * @name proc_sys_time
 *
 * Returns the current ellapsed process system time.
 * @{
 */

/** @brief Returns the current ellapsed process system time in seconds. */
double proc_sys_time_s();


/** @brief Returns the system time of a Timer in milliseconds. */
double proc_sys_time_ms();

/*@}*/


#ifdef __cplusplus
}
}
#endif


#endif
