/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Read, write, and access XML documents.
 *
 * The XML data structures are documented at http://xmlsoft.org/html
 */


#include <jwsc/config.h>

#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <inttypes.h>

#ifdef JWSC_HAVE_LIBXML2
#include <libxml/tree.h>
#include <libxml/parser.h>
#include <libxml/valid.h>
#endif

#include "jwsc/base/error.h"
#include "jwsc/base/xml.h"

#ifdef JWSC_HAVE_DMALLOC
#include <dmalloc.h>
#endif


/** @param  doc_out  An empty XML document. */
void create_xml_doc(xmlDoc** doc_out)
{
    if (*doc_out)
    {
        free_xml_doc(*doc_out);
    }
    *doc_out = xmlNewDoc((xmlChar*)"1.0");
}


/** @param  doc  The XML document to free. */
void free_xml_doc(xmlDoc* doc)
{
    xmlFreeDoc(doc);
}


/** 
 * @param  doc_out  The XML document.
 * @param  fname    The file to read from. 
 */
Error* read_xml_doc(xmlDoc** doc_out, const char* fname)
{
    xmlParserCtxt* ctx;
    Error* e = NULL;

    create_xml_doc(doc_out);
    assert(ctx = xmlNewParserCtxt());

    if (*doc_out)
    {
        free_xml_doc(*doc_out);
    }

    if (!(*doc_out = xmlCtxtReadFile(ctx, fname, NULL, 0)))
    {
        e = JWSC_EARG("Could not parse file");
    }

    xmlFreeParserCtxt(ctx);

    return e;
}


/** 
 * @param  dtd_out  The XML DTD.
 * @param  fname    The file to read from.
 */
Error* read_xml_dtd(xmlDtd** dtd_out, const char* fname)
{
    Error* e = NULL;

    if (!(*dtd_out = xmlParseDTD(NULL, (xmlChar*)fname)))
    {
        e = JWSC_EARG("Could not parse file");
    }

    return e;
}


/** @param  dtd  The XML DTD. */
void free_xml_dtd(xmlDtd* dtd)
{
    xmlFreeDtd(dtd);
}


/** 
 * @param  doc  The XML document to validate. 
 * @param  dtd  The DTD to validate against. 
 *
 * @return 1 if valid; 0 otherwise.
 */
int validate_xml_doc(xmlDoc* doc, xmlDtd* dtd)
{
    xmlValidCtxt* ctx;
    int valid;

    assert(ctx = xmlNewValidCtxt());

    valid = xmlValidateDtd(ctx, doc, dtd);

    xmlFreeValidCtxt(ctx);

    return valid;
}
