/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Declarations for 2D filters.
 */


#ifndef TWOD_H
#define TWOD_H


#include <jwsc/config.h>

#include <stdlib.h>
#include <inttypes.h>

#include "jwsc/base/error.h"
#include "jwsc/matrix/matrix.h"


#ifdef __cplusplus
namespace jwsc {
extern "C" {
#endif


/**
 * @name create_sobel_x_filter
 *
 * Creates a Sobel operator for the @em x direction.
 * @{
 */

/** @brief Creates a single precision Sobel operator for the @em x direction. */
void create_sobel_x_filter_f(Matrix_f** sobel_out);

/** @brief Creates a double precision Sobel operator for the @em x direction. */
void create_sobel_x_filter_d(Matrix_d** sobel_out);

/*@}*/




/**
 * @name create_sobel_y_filter
 *
 * Creates a Sobel operator for the @em y direction.
 * @{
 */

/** @brief Creates a single precision Sobel operator for the @em y direction. */
void create_sobel_y_filter_f(Matrix_f** sobel_out);

/** @brief Creates a double precision Sobel operator for the @em y direction. */
void create_sobel_y_filter_d(Matrix_d** sobel_out);

/*@}*/




/**
 * @name create_LoG_filter
 *
 * Creates a Laplacian of Gaussian filter.
 * @{
 */

/** @brief Creates a single precision Laplacian of Gaussian filter. */
Error* create_LoG_filter_f(Matrix_f** LoG_out, float sigma);

/** @brief Creates a double precision Laplacian of Gaussian filter. */
Error* create_LoG_filter_d(Matrix_d** LoG_out, double sigma);

/*@}*/




/**
 * @name create_auto_2d_gaussian_filter
 *
 * Creates a 2D Gaussian filter automatically sized.
 * @{
 */

/** 
 * @brief Creates a single precision 2D Gaussian filter automatically sized.
 */
Error* create_auto_2d_gaussian_filter_f
(
    Matrix_f** h_out, 
    float      row_sigma, 
    float      col_sigma
);

/** 
 * @brief Creates a double precision 2D Gaussian filter automatically sized.
 */
Error* create_auto_2d_gaussian_filter_d
(
    Matrix_d** h_out, 
    double     row_sigma, 
    double     col_sigma
);

/*@}*/




/**
 * @name create_2d_gaussian_filter
 *
 * Creates a 2D Gaussian filter.
 * @{
 */

/** 
 * @brief Creates a single precision 2D Gaussian filter.
 */
Error* create_2d_gaussian_filter_f
(
    Matrix_f** h_out, 
    float      row_sigma,
    float      col_sigma,
    uint32_t   num_rows,
    uint32_t   num_cols
);

/** 
 * @brief Creates a double precision 2D Gaussian filter.
 */
Error* create_2d_gaussian_filter_d
(
    Matrix_d** h_out, 
    double     row_sigma,
    double     col_sigma,
    uint32_t   num_rows,
    uint32_t   num_cols
);

/*@}*/




/**
 * @name create_auto_2d_gaussian_dx_filter
 *
 * Creates a 2D Gaussian filter with a partial derivative along columns 
 * automatically sized.
 * @{
 */

/** 
 * @brief Creates a single precision 2D Gaussian filter with a partial 
 * derivative in the x-direction (column) automatically sized.
 */
Error* create_auto_2d_gaussian_dx_filter_f
(
    Matrix_f** h_out, 
    float      row_sigma,
    float      col_sigma
);

/** 
 * @brief Creates a double precision 2D Gaussian filter with a partial 
 * derivative in the x-direction (column) automatically sized.
 */
Error* create_auto_2d_gaussian_dx_filter_d
(
    Matrix_d** h_out, 
    double     row_sigma,
    double     col_sigma
);

/*@}*/




/**
 * @name create_2d_gaussian_dx_filter
 *
 * Creates a 2D Gaussian filter with a partial derivative along columns.
 * @{
 */

/** 
 * @brief Creates a single precision 2D Gaussian filter with a partial 
 * derivative in the x-direction (column).
 */
Error* create_2d_gaussian_dx_filter_f
(
    Matrix_f** h_out, 
    float      row_sigma,
    float      col_sigma,
    uint32_t   num_rows,
    uint32_t   num_cols
);

/** 
 * @brief Creates a double precision 2D Gaussian filter with a partial 
 * derivative in the x-direction (column).
 */
Error* create_2d_gaussian_dx_filter_d
(
    Matrix_d** h_out, 
    double     row_sigma,
    double     col_sigma,
    uint32_t   num_rows,
    uint32_t   num_cols
);

/*@}*/




/**
 * @name create_auto_2d_gaussian_dy_filter
 *
 * Creates a 2D Gaussian filter with a partial derivative along rows 
 * automatically sized to sqrt(2)*6*sigma.
 * @{
 */

/** 
 * @brief Creates a single precision 2D Gaussian filter with a partial 
 * derivative in the y-direction (row) automatically sized.
 */
Error* create_auto_2d_gaussian_dy_filter_f
(
    Matrix_f** h_out, 
    float      row_sigma,
    float      col_sigma
);

/** 
 * @brief Creates a double precision 2D Gaussian filter with a partial 
 * derivative in the y-direction (row) automatically sized.
 */
Error* create_auto_2d_gaussian_dy_filter_d
(
    Matrix_d** h_out, 
    double     row_sigma,
    double     col_sigma
);

/*@}*/




/**
 * @name create_2d_gaussian_dy_filter
 *
 * Creates a 2D Gaussian filter with a partial derivative along rows.
 * @{
 */

/**
 * @brief Creates a single precision 2D Gaussian filter with a partial 
 * derivative in the y direction (row).
 */
Error* create_2d_gaussian_dy_filter_f
(
    Matrix_f** h_out, 
    float      row_sigma,
    float      col_sigma,
    uint32_t   num_rows,
    uint32_t   num_cols
);

/**
 * @brief Creates a double precision 2D Gaussian filter with a partial 
 * derivative in the y direction (row).
 */
Error* create_2d_gaussian_dy_filter_d
(
    Matrix_d** h_out, 
    double     row_sigma,
    double     col_sigma,
    uint32_t   num_rows,
    uint32_t   num_cols
);

/*@}*/


#ifdef __cplusplus
}
}
#endif


#endif
