/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Declarations for an image edge detector.
 *
 * The edge detection algorithm is a variation of the standard Canny algorithm:
 * the image is convolved with the gradiant of Gaussian, then non-maximal 
 * suppression and hysterisis are applied to the resulting gradient map.
 */


#ifndef EDGE_H 
#define EDGE_H 


#include <jwsc/config.h>

#include <stdlib.h>
#include <inttypes.h>

#include "jwsc/base/error.h"
#include "jwsc/image/image.h"
#include "jwsc/matrix/matrix.h"


#ifdef __cplusplus
namespace jwsc {
extern "C" {
#endif


/** @brief Edge point in an image or matrix (single-sampled image). */
typedef struct
{
    /** @brief Column the point lies in. */
    uint32_t col;

    /** @brief Row the point lies in. */
    uint32_t row;

    /** @brief Rate of change in brightness along the columns at the point. */
    float dcol;

    /** @brief Rate of change in brightness along the rows at the point. */
    float drow;

    /** @brief Magnitude of the gradient. */
    float mag;
}
Edge_point_f;



/** @brief Set of edges composed of edge points. */
typedef struct
{
    /** @brief Number of rows in the input image. */
    uint32_t num_rows;

    /** @brief Number of columns in the input image. */
    uint32_t num_cols;

    /** @brief Number of edges. */
    uint32_t num_edges;

    /** @brief Total number of points in all edges. */
    uint32_t total_num_pts;

    /** @brief Number of points in each edge. */
    uint32_t* num_pts;

    /** @brief Set of edges and edge points. */
    Edge_point_f** pts;
}
Edge_set_f;




/** 
 * @name detect_image_edge_set
 *
 * Detects a set of edges in each sample of an image.
 * @{
 */

/** @brief Detects a set of edges in each sample of a single precision image. */
Error* detect_image_edge_set_f
(
    Edge_set_f**   edges_out,
    const Image_f* img,
    float          sigma, 
    float          begin_thresh, 
    float          end_thresh,
    uint32_t       padding
);

/** @brief Detects a set of edges in each sample of a single precision image. */
Error* DEPRECATED_detect_image_edge_set_f
(
    Edge_set_f**   r_edges_out,
    Edge_set_f**   g_edges_out,
    Edge_set_f**   b_edges_out,
    const Image_f* img,
    float          sigma, 
    float          begin_thresh, 
    float          end_thresh,
    uint32_t       padding
);

/*@}*/




/** 
 * @name detect_matrix_edge_set
 *
 * Detects a set of edges in a matrix (single-sample image).
 * @{
 */

/** 
 * @brief Detects a set of edges in a single precision matrix (single-sample 
 * image).
 */
Error* detect_matrix_edge_set_f
(
    Edge_set_f**    edges_out,
    const Matrix_f* m,
    float           sigma, 
    float           begin_thresh, 
    float           end_thresh,
    uint32_t        padding
);

/*@}*/




/** 
 * @name free_edge_set
 *
 * Frees a set of edges.
 * @{
 */

/** @brief Frees a set of single precision edges. */
void free_edge_set_f(Edge_set_f* edges);

/*@}*/




/** 
 * @name remove_short_edges
 *
 * Removes edges with a number of points less than a specified minimum.
 * @{
 */

/** 
 * @brief Removes single precision edges with a number of points less than a
 * specified minimum. 
 */
void remove_short_edges_f(Edge_set_f* edges, uint32_t min_len);

/*@}*/




/** 
 * @name break_edges_at_corners
 *
 * Finds the edge point with the largest gradient difference on either side,
 * breaks the edge, and recurses on the two newly created edges.
 * @{
 */

/** 
 * @brief Breaks apart single precision edges where the gradient direction
 * changes significantly.
 */
void break_edges_at_corners_f
(
    Edge_set_f* edges, 
    float       thresh, 
    uint32_t    num_avg
);

/*@}*/




/** 
 * @name sample_edge_set
 *
 * Randomly samples edge points from a set of them.
 * @{
 */

/** @brief Randomly samples single precision edge points from a set of them. */
void sample_edge_set_f
(
    Edge_set_f**      edges_out, 
    const Edge_set_f* edges_in, 
    float             p
);

/*@}*/




/** 
 * @name color_edge_set
 *
 * Colors a set of edges in an image.
 * @{
 */

/** @brief Colors single precision set of edges in an image. */
Error* color_edge_set_f
(
    Image_f**         img_out,
    const Image_f*    img_in,
    const Edge_set_f* edges, 
    const Pixel_f*    pxl
);

/*@}*/




/** 
 * @name randomly_color_edge_set
 *
 * Colors a set of edges in an image with randomly selected colors.
 * @{
 */

/** 
 * @brief Colors single precision set of edges in an image with randomly
 * selected colors. 
 */
Error* randomly_color_edge_set_f
(
    Image_f**         img_out,
    const Image_f*    img_in,
    const Edge_set_f* edges
);

/*@}*/




/** 
 * @name color_edge_points
 *
 * Colors edge points in an image.
 * @{
 */

/** @brief Colors single precision edge points in an image. */
Error* color_edge_points_f
(
    Image_f**           img_out,
    const Image_f*      img_in,
    const Edge_point_f* pts, 
    uint32_t            num_pts,
    const Pixel_f*      pxl
);

/*@}*/




/** 
 * @name randomly_color_edge_points
 *
 * Colors edge points in an image with a randomly selected color.
 * @{
 */

/**
 * Colors single precision edge points in an image with a randomly selected 
 * color.
 */
Error* randomly_color_edge_points_f
(
    Image_f**           img_out,
    const Image_f*      img_in,
    const Edge_point_f* pts, 
    uint32_t            num_pts
);

/*@}*/




/** 
 * @name read_edge_set
 *
 * Reads a set of edges in ASCII format from a file.
 * @{
 */

/** @brief Reads a set of single precision edges in ASCII format from a file. */
Error* read_edge_set_f
(
    Edge_set_f** edges_out,
    const char*  fname
);

/*@}*/




/** 
 * @name write_edge_set
 *
 * Writes a set of edges in ASCII format to a file.
 * @{
 */

/** @brief Writes a set of single precision edges in ASCII format to a file. */
Error* write_edge_set_f
(
    const Edge_set_f* edges,
    const char*       fname
);

/*@}*/


#ifdef __cplusplus
}
}
#endif


#endif 
