/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/** 
 * @file 
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Definitions for an image and associated functions.
 *
 * The image holds single precision floating point pixel elements.
 *
 * Although the image is indexed by row and column, the pixels are
 * allocated as a 1D array in row-major order. So iterating over the 
 * pixels can be done inside a single loop.
 */


#include <jwsc/config.h>

#include <stdlib.h>
#include <inttypes.h>
#include <math.h>
#include <assert.h>

#include "jwsc/base/error.h"
#include "jwsc/matrix/matrix.h"
#include "jwsc/image/image.h"

#ifdef JWSC_HAVE_DMALLOC
#include <dmalloc.h>
#endif


/** 
 * @name create_image
 *
 * Creates an image.
 * @{
 */

/**
 * @param  img_out   Result parameter. If @em *img_out is NULL, an image is
 *                   allocated; otherwise its space is re-used.
 * @param  num_rows  Number of rows in the image.
 * @param  num_cols  Number of columns in the image.
 *
 * @note Dimensions must be > 0.
 */
void create_image_f
(
    Image_f** img_out, 
    uint32_t  num_rows, 
    uint32_t  num_cols
)
{
    uint32_t row;
    Pixel_f* pxls;
    Image_f* img;

    assert(num_rows > 0 && num_cols > 0);

    if (*img_out == NULL)
    {
        assert(*img_out = malloc(sizeof(Image_f)));
        (*img_out)->num_rows = 0;
        (*img_out)->num_cols = 0;
        (*img_out)->pxls     = NULL;
    }
    img = *img_out;

    if (img->num_rows != num_rows || img->num_cols != num_cols)
    {
        pxls = (img->pxls != NULL) ? *(img->pxls) : NULL;
        pxls = realloc(pxls, num_rows*num_cols*sizeof(Pixel_f));
        assert(pxls);

        img->pxls = realloc(img->pxls, num_rows*sizeof(Pixel_f*));
        assert(img->pxls);

        for (row = 0; row < num_rows; row++)
        {
            img->pxls[ row ] = &(pxls[ row*num_cols ]);
        }
        img->num_rows = num_rows;
        img->num_cols = num_cols;
    }
}

/*@}*/




/**
 * @name create_init_image
 *
 * Creates and initializes an image.
 * @{
 */

/**
 * @param  img_out   Result parameter. If @em *img_out is NULL, an image is
 *                   allocated; otherwise its space is re-used.
 * @param  num_rows  Number of rows in the image.
 * @param  num_cols  Number of columns in the image.
 * @param  val       Pixel value to initialize all pixels of the image with.
 *
 * @note Dimensions must be > 0.
 */
void create_init_image_f
(
    Image_f** img_out, 
    uint32_t  num_rows, 
    uint32_t  num_cols,
    Pixel_f   val
)
{
    uint32_t num_pxls;
    uint32_t pxl;
    Pixel_f* img_pxls;

    create_image_f(img_out, num_rows, num_cols);
    img_pxls = *((*img_out)->pxls);
    num_pxls = num_rows*num_cols;

    for (pxl = 0; pxl < num_pxls; pxl++)
    {
        img_pxls[ pxl ] = val;
    }
}

/*@}*/




/**
 * @name create_zero_image
 *
 * Creates an image and initializes it with zeros.
 * @{
 */

/**
 * @param  img_out   Result parameter. If @em *img_out is NULL, an image is
 *                   allocated; otherwise its space is re-used.
 * @param  num_rows  Number of rows in the image.
 * @param  num_cols  Number of columns in the image.
 *
 * @note Dimensions must be > 0.
 */
void create_zero_image_f
(
    Image_f** img_out, 
    uint32_t  num_rows, 
    uint32_t  num_cols
)
{
    Pixel_f p = {0};

    create_init_image_f(img_out, num_rows, num_cols, p);
}

/*@}*/




/**
 * @name create_random_image
 *
 * Creates an image initializes it with random values.
 * @{
 */

/**
 * Calls rand() to generate the random numbers.
 *
 * @param  img_out   Result parameter. If @em *img_out is NULL, an image is
 *                   allocated; otherwise its space is re-used.
 * @param  num_rows  Number of rows in the image.
 * @param  num_cols  Number of columns in the image.
 * @param  min       Minimum random value.
 * @param  max       Maximum random value.
 *
 * @note Dimensions must be > 0.
 *
 * @note Since Pixel_f samples range [0,1], @em min and @em max should be in 
 * that range as well so a meaningful image is created.
 *
 * @warning Does not check @em min and @em max to see if they would cause
 * overflow.
 */
void create_random_image_f
(
    Image_f** img_out, 
    uint32_t  num_rows, 
    uint32_t  num_cols,
    Pixel_f   min,
    Pixel_f   max
)
{
    uint32_t num_pxls;
    uint32_t pxl;
    Pixel_f* img_pxls;
    float    x;

    create_image_f(img_out, num_rows, num_cols);
    img_pxls = *((*img_out)->pxls);
    num_pxls = num_rows*num_cols;

    for (pxl = 0; pxl < num_pxls; pxl++)
    {
        x = (float)rand() / (float)RAND_MAX;
        img_pxls[ pxl ].r = min.r + (max.r - min.r)*x;
        x = (float)rand() / (float)RAND_MAX;
        img_pxls[ pxl ].g = min.g + (max.g - min.r)*x;
        x = (float)rand() / (float)RAND_MAX;
        img_pxls[ pxl ].b = min.b + (max.b - min.b)*x;
    }
}

/*@}*/




/** 
 * @name create_image_from_matrices
 *
 * Creates an image from a three matrices.
 * @{
 */

/**
 * @param  img_out  Result parameter. If @em *img_out is NULL, an image is
 *                  allocated; otherwise its space is re-used.
 * @param  m_r      Matrix to copy into the red samples of the image.
 * @param  m_g      Matrix to copy into the green samples of the image.
 * @param  m_b      Matrix to copy into the blue samples of the image.
 *
 * @return On success, NULL is returned. On error, an Error is returned and @em
 * *img_out is freed and set to NULL.
 * - @link Error_types::ERROR_INV_ARG ERROR_INV_ARG @endlink 
 *   Matrices do not have the same dimensions.
 */
Error* create_image_from_matrices_f
(
    Image_f**   img_out, 
    const Matrix_f* m_r,
    const Matrix_f* m_g,
    const Matrix_f* m_b
)
{
    uint32_t num_elts;
    uint32_t elt;
    Pixel_f* img_pxls;
    float*   m_r_elts;
    float*   m_g_elts;
    float*   m_b_elts;

    if (m_r->num_rows != m_g->num_rows || m_g->num_rows != m_b->num_rows ||
        m_r->num_cols != m_g->num_cols || m_g->num_cols != m_b->num_cols)
    {
        free_image_f(*img_out); *img_out = NULL;
        return JWSC_EARG("Inconsistent image channel sizes");
    }

    create_image_f(img_out, m_r->num_rows, m_r->num_cols);

    img_pxls = *((*img_out)->pxls);
    m_r_elts = *(m_r->elts);
    m_g_elts = *(m_g->elts);
    m_b_elts = *(m_b->elts);

    num_elts = m_r->num_rows*m_r->num_cols;

    for (elt = 0; elt < num_elts; elt++)
    {
        img_pxls[ elt ].r = m_r_elts[ elt ];
        img_pxls[ elt ].g = m_g_elts[ elt ];
        img_pxls[ elt ].b = m_b_elts[ elt ];
    }

    return NULL;
}

/*@}*/




/** 
 * @name create_image_from_matrix
 *
 * Creates a grayscale image from a matrix.
 * @{
 */

/**
 * @param  img_out  Result parameter. If @em *img_out is NULL, an image is
 *                  allocated; otherwise its space is re-used.
 * @param  m        Matrix to copy into the red, green and blue samples of 
 *                  the image.
 */
void create_image_from_matrix_f
(
    Image_f**   img_out, 
    const Matrix_f* m
)
{
    create_image_from_matrices_f(img_out, m, m, m);
}

/*@}*/




/** 
 * @name create_matrices_from_image
 *
 * Creates three matrices from an image.
 * @{
 */

/**
 * @param  m_r_out   Result parameter. Red sample matrix. If @em *m_r_out is 
 *                   NULL, an image is allocated; otherwise its space is 
 *                   re-used.
 * @param  m_g_out   Result parameter. Green sample matrix. If @em *m_g_out is 
 *                   NULL, an image is allocated; otherwise its space is 
 *                   re-used.
 * @param  m_b_out   Result parameter. Blue sample matrix. If @em *m_b_out is 
 *                   NULL, an image is allocated; otherwise its space is 
 *                   re-used.
 * @param  img       Image to copy into the matrices.
 */
void create_matrices_from_image_f
(
    Matrix_f**     m_r_out, 
    Matrix_f**     m_g_out, 
    Matrix_f**     m_b_out, 
    const Image_f* img
)
{
    uint32_t num_elts;
    uint32_t elt;
    Pixel_f* img_pxls;
    float*   m_r_elts;
    float*   m_g_elts;
    float*   m_b_elts;

    create_matrix_f(m_r_out, img->num_rows, img->num_cols);
    create_matrix_f(m_g_out, img->num_rows, img->num_cols);
    create_matrix_f(m_b_out, img->num_rows, img->num_cols);

    m_r_elts = *((*m_r_out)->elts);
    m_g_elts = *((*m_g_out)->elts);
    m_b_elts = *((*m_b_out)->elts);
    img_pxls = *(img->pxls);

    num_elts = img->num_rows*img->num_cols;

    for (elt = 0; elt < num_elts; elt++)
    {
        m_r_elts[ elt ] = img_pxls[ elt ].r;
        m_g_elts[ elt ] = img_pxls[ elt ].g;
        m_b_elts[ elt ] = img_pxls[ elt ].b;
    }
}

/*@}*/




/** 
 * @name create_matrix_from_image
 *
 * Creates an averaged matrix from an image.
 * @{
 */

/**
 * The weights are normalized to sum to one.
 *
 * The best weights for a human visible grayscale conversion are
 *
 * @code
 *   r=0.3   g=0.59   b=0.11
 * @endcode
 *
 * @param  m_out     Result parameter. If @em *m_out is NULL, a matrix is
 *                   allocated; otherwise its space is re-used.
 * @param  img       Image to copy into the matrix.
 * @param  r_weight  Weight of the red sample.
 * @param  g_weight  Weight of the green sample.
 * @param  b_weight  Weight of the blue sample.
 *
 * @note Weights must be non-negative.
 */
void create_matrix_from_image_f
(
    Matrix_f**     m_out, 
    const Image_f* img,
    float          r_weight,
    float          g_weight,
    float          b_weight
)
{
    uint32_t num_elts;
    uint32_t elt;
    float    weight_sum;
    float*   m_elts;
    Pixel_f* img_pxls;

    assert(r_weight >= 0 && g_weight >= 0 && b_weight >= 0);

    if ((weight_sum = r_weight + g_weight + b_weight) != 1.0)
    {
        r_weight *= 1.0 / weight_sum;
        g_weight *= 1.0 / weight_sum;
        b_weight *= 1.0 / weight_sum;
    }

    create_matrix_f(m_out, img->num_rows, img->num_cols);

    m_elts   = *((*m_out)->elts);
    img_pxls = *(img->pxls);

    num_elts = img->num_rows*img->num_cols;

    for (elt = 0; elt < num_elts; elt++)
    {
        m_elts[ elt ] = r_weight*img_pxls[ elt ].r +
                        g_weight*img_pxls[ elt ].g +
                        b_weight*img_pxls[ elt ].b;
    }
}

/*@}*/




/** 
 * @name copy_image
 *
 * Copies an image.
 * @{
 */

/**
 * @param  img_out  Result parameter. If @em *img_out is NULL, an image is
 *                  allocated; otherwise its space is re-used.
 * @param  img_in   Image to copy.
 */
void copy_image_f(Image_f** img_out, const Image_f* img_in)
{
    uint32_t num_pxls;
    uint32_t pxl;
    Pixel_f* img_out_pxls;
    Pixel_f* img_in_pxls;

    create_image_f(img_out, img_in->num_rows, img_in->num_cols);

    img_out_pxls = *((*img_out)->pxls);
    img_in_pxls  = *(img_in->pxls);

    num_pxls = img_in->num_rows*img_in->num_cols;

    for (pxl = 0; pxl < num_pxls; pxl++ )
    {
        img_out_pxls[ pxl ] = img_in_pxls[ pxl ];
    }
}

/*@}*/




/** 
 * @name free_image
 *
 * Frees an image.
 * @{
 */

/** @param img Image to free. */
void free_image_f(Image_f* img)
{
    if (img == NULL) return;

    free(*(img->pxls));
    free(img->pxls);
    free(img);
}

/*@}*/
