/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/** 
 * @file 
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Declarations for an image and associated functions.
 *
 * The image holds single precision floating point pixel elements.
 *
 * Although the image is indexed by row and column, the pixels are
 * allocated as a 1D array in row-major order. So iterating over the 
 * pixels can be done inside a single loop.
 */


#ifndef IMAGE_H
#define IMAGE_H


#include <jwsc/config.h>

#include <stdlib.h>
#include <inttypes.h>

#include "jwsc/base/error.h"
#include "jwsc/matrix/matrix.h"


#ifdef __cplusplus
namespace jwsc {
extern "C" {
#endif


/** 
 * @brief A 3-sample single precision floating point picture element. 
 *
 * Sample values range [0,1].
 */
typedef struct Pixel_f
{
    /** @brief Red sample. */
    float r;

    /** @brief Green sample. */
    float g;

    /** @brief Blue sample. */
    float b;
} 
Pixel_f;




/** @brief An image of Pixel_f. */ 
typedef struct Image_f
{
    /** @brief Number of rows. */
    uint32_t num_rows;

    /** @brief Number of columns. */
    uint32_t num_cols;

    /** @brief Pixels of the image. */
    Pixel_f** pxls;
} 
Image_f;




/** 
 * @name create_image
 *
 * Creates an image.
 * @{
 */

/** @brief Creates a single precision floating point image. */
void create_image_f
(
    Image_f** img_out, 
    uint32_t  num_rows, 
    uint32_t  num_cols
);

/*@}*/




/** 
 * @name create_init_image
 *
 * Creates and initializes an image.
 * @{
 */

/** @brief Creates and initializes a single precision floating point image. */
void create_init_image_f
(
    Image_f** img_out, 
    uint32_t  num_rows, 
    uint32_t  num_cols,
    Pixel_f   val
);

/*@}*/




/** 
 * @name create_zero_image
 *
 * Creates an image and initializes it with zeros.
 * @{
 */

/** 
 * @brief Creates a single precision floating point image and initializes it
 * with zeros.
 */
void create_zero_image_f
(
    Image_f** img_out, 
    uint32_t  num_rows, 
    uint32_t  num_cols
);

/*@}*/




/** 
 * @name create_random_image
 *
 * Creates an image and initializes it with random values.
 * @{
 */

/** 
 * @brief Creates a single precision floating point image and initializes it
 * with random values.
 */
void create_random_image_f
(
    Image_f** img_out, 
    uint32_t  num_rows, 
    uint32_t  num_cols,
    Pixel_f   min,
    Pixel_f   max
);

/*@}*/




/** 
 * @name create_image_from_matrices
 *
 * Creates an image from a three matrices.
 * @{
 */

/** 
 * @brief Creates a single precision floating point image from three matrices.
 */
Error* create_image_from_matrices_f
(
    Image_f**   img_out, 
    const Matrix_f* m_r,
    const Matrix_f* m_g,
    const Matrix_f* m_b
);

/*@}*/




/** 
 * @name create_image_from_matrix
 *
 * Creates a grayscale image from a matrix.
 * @{
 */

/** 
 * @brief Creates a single precision floating point grayscale image from a 
 * matrix.
 */
void create_image_from_matrix_f
(
    Image_f**       img_out, 
    const Matrix_f* m
);

/*@}*/




/** 
 * @name create_matrices_from_image
 *
 * Creates three matrices from an image.
 * @{
 */

/**
 * @brief Creates three single precision floating point matrices from an
 * image.
 */
void create_matrices_from_image_f
(
    Matrix_f**     m_r_out, 
    Matrix_f**     m_g_out, 
    Matrix_f**     m_b_out, 
    const Image_f* img
);

/*@}*/




/** 
 * @name create_matrix_from_image
 *
 * Creates an averaged matrix from an image.
 * @{
 */

/**
 * @brief Creates an averaged single precision floating point matrix from an
 * image.
 */
void create_matrix_from_image_f
(
    Matrix_f**     m_out, 
    const Image_f* img,
    float          r_weight,
    float          g_weight,
    float          b_weight
);

/*@}*/




/** 
 * @name copy_image
 *
 * Copies an image.
 * @{
 */

/** @brief Copies a single precision floating point image. */
void copy_image_f(Image_f** img_out, const Image_f* img_in);

/*@}*/




/** 
 * @name free_image
 *
 * Frees an image.
 * @{
 */

/** @brief Frees a single precision floatin point image. */
void free_image_f(Image_f* img);

/*@}*/


#ifdef __cplusplus
}
}
#endif


#endif
