/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/** 
 * @file 
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Definitions for functions to read and write images.
 */


#include <jwsc/config.h>

#include <stdlib.h>
#include <stdio.h>
#include <inttypes.h>
#include <assert.h>
#include <string.h>

#if defined JWSC_HAVE_MAGICK_MAGICKCORE_H
#include <magick/MagickCore.h>
#elif defined JWSC_HAVE_MAGICK_API_H
#include <magick/api.h>
#endif

#include "jwsc/base/error.h"
#include "jwsc/base/fname.h"
#include "jwsc/image/image.h"
#include "jwsc/image/jiff.h"
#include "jwsc/image/image_io.h"


/** 
 * @name read_image
 *
 * Reads an image from a file.
 * @{
 */

/**
 * The type of image to read is based on the file name suffix.
 *
 * @param  img_out  Result parameter. If @em *img_out is NULL, an image is
 *                  allocated; otherwise its space is re-used.
 * @param  fname    Name of the image file to read.
 *
 * @note File names are assumed to be of length <= 256 characters.
 *
 * @return On success, NULL is returned. On error, an Error is returned and @em
 * *img_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Could not read an image from @em fname.
 */
Error* read_image_f(Image_f** img_out, const char* fname)
{
    Error* e = NULL;

    if (fname_has_suffix(fname, "jiff"))
    {
        e = read_image_as_jiff_f(img_out, fname);
    }
    else
    {
#ifdef JWSC_HAVE_MAGICK
        e = read_image_with_magick_f(img_out, fname);
#else
        e = JWSC_EIO("Unknown image format");
        free_image_f(*img_out); *img_out = NULL;
#endif
    }

    return e;
}

/*@}*/




/** 
 * @name read_image_with_magick
 *
 * Reads an image from a file.
 * @{
 */

/**
 * @param  img_out  Result parameter. If @em *img_out is NULL, an image is
 *                  allocated; otherwise its space is re-used.
 * @param  fname    Name of the image file to read.
 *
 * @return On success, NULL is returned. On error, an Error is returned and @em
 * *img_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Could not read an image from @em fname.
 */
#ifdef JWSC_HAVE_MAGICK
Error* read_image_with_magick_f(Image_f** img_out, const char* fname)
{
    uint32_t num_rows, num_cols;

    ExceptionInfo magick_exception;
    ImageInfo*    magick_info = NULL;
    Image*        magick_img = NULL;
    Error*        e = NULL;
    
    GetExceptionInfo(&magick_exception);
    magick_info = CloneImageInfo(NULL);
    strncpy(magick_info->filename, fname, MaxTextExtent);

    if ((magick_img = ReadImage(magick_info, &magick_exception)) == NULL)
    {
        e = JWSC_EIO(magick_exception.reason);
        goto cleanup;
    }

    num_rows = (uint32_t)magick_img->rows;
    num_cols = (uint32_t)magick_img->columns;

    create_image_f(img_out, num_rows, num_cols);

    if (ExportImagePixels(magick_img, 0, 0, num_cols, num_rows, "RGB", 
                FloatPixel, *((*img_out)->pxls), &magick_exception) == 0)
    {
        e = JWSC_EIO(magick_exception.reason);
        goto cleanup;
    }

cleanup:
    if (magick_img) DestroyImage(magick_img);
    if (magick_info) DestroyImageInfo(magick_info);
    if (e) {free_image_f(*img_out); *img_out = NULL;}
    return e;
}
#endif

/*@}*/




/** 
 * @name write_image
 *
 * Writes an image to a file.
 * @{
 */

/**
 * @param  img    Image to write.
 * @param  fname  Name of the file to write the image to.
 *
 * @return On success, NULL is returned. On error, an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Could not write an image to @em fname.
 */
Error* write_image_f(const Image_f* img, const char* fname)
{
    JIFF_attributes attrs;
    Error* e = NULL;

    if (fname_has_suffix(fname, "jiff"))
    {
        init_jiff_attributes(&attrs);
        e = write_image_as_jiff_f(img, fname, &attrs);
    }
    else
    {
#ifdef JWSC_HAVE_MAGICK
        e = write_image_with_magick_f(img, fname);
#else
        e = JWSC_EIO("Unknown image format");
#endif
    }

    return e;
}

/*@}*/




/** 
 * @name write_image_with_magick
 *
 * Writes an image to a file.
 * @{
 */

/**
 * @param  img    Image to write.
 * @param  fname  Name of the file to write the image to.
 *
 * @return On success, NULL is returned. On error, an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Could not write an image to @em fname.
 */
#ifdef JWSC_HAVE_MAGICK
Error* write_image_with_magick_f(const Image_f* img, const char* fname)
{
    ExceptionInfo magick_exception;
    ImageInfo*    magick_info = NULL;
    Image*        magick_img = NULL;
    Error*        e = NULL;
    
    GetExceptionInfo(&magick_exception);
    magick_info = CloneImageInfo(NULL);

    if ((magick_img = ConstituteImage(img->num_cols, img->num_rows, "RGB", 
                    FloatPixel, *(img->pxls), &magick_exception)) == NULL)
    {
        e = JWSC_EIO(magick_exception.reason);
        goto cleanup;
    }

    strncpy(magick_img->filename, fname, MaxTextExtent);

    if (WriteImage(magick_info, magick_img) == 0)
    {
        e = JWSC_EIO(magick_img->exception.reason);
        goto cleanup;
    }

cleanup:
    if (magick_img) DestroyImage(magick_img);
    if (magick_info) DestroyImageInfo(magick_info);
    return e;
}
#endif

/*@}*/
