/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/** 
 * @file 
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Declarations for image utility functions.
 */


#ifndef IMAGE_UTIL_H
#define IMAGE_UTIL_H


#include <jwsc/config.h>

#include <stdlib.h>

#include "jwsc/base/error.h"
#include "jwsc/matrix/matrix.h"
#include "jwsc/image/image.h"


#ifdef __cplusplus
namespace jwsc {
extern "C" {
#endif


/** 
 * @name down_sample_image
 *
 * Down samples an image.
 * @{
 */

/** @brief Down samples a single precision image. */
Error* down_sample_image_f
(
    Image_f**      img_out, 
    const Image_f* img_in, 
    float          row_factor,
    float          col_factor
);

/*@}*/




/** 
 * @name extend_image
 *
 * Extends an image into a padding region.
 * @{
 */

/** @brief Extends a single precision image into a padding region. */
void extend_image_f
(
    Image_f**      img_out, 
    const Image_f* img_in, 
    uint32_t       row_padding,
    uint32_t       col_padding
);

/*@}*/




/** 
 * @name crop_image
 *
 * Crops an image.
 * @{
 */

/** @brief Crops a single precision image into a padding region. */
Error* crop_image_f
(
    Image_f**      img_out, 
    const Image_f* img_in, 
    uint32_t       row,
    uint32_t       col,
    uint32_t       num_rows,
    uint32_t       num_cols
);

/*@}*/




/** 
 * @name threshold_image
 *
 * Thresholds a image.
 * @{
 */

/** @brief Thresholds a single precision image. */
void threshold_image_f
(
    Matrix_f**     mat_out, 
    const Image_f* img_in,
    float          t
);

/*@}*/




/** 
 * @name threshold_matrix
 *
 * Thresholds a matrix.
 * @{
 */

/** @brief Thresholds a single precision matrix. */
void threshold_matrix_f
(
    Matrix_f**      mat_out, 
    const Matrix_f* mat_in,
    float           t
);

/*@}*/




/** 
 * @name dynamically_threshold_image
 *
 * Dynamically thresholds an image.
 * @{
 */

/** @brief Dynamically thresholds a single precision image. */
void dynamically_threshold_image_f
(
    Matrix_f**     mat_out, 
    const Image_f* img_in
);

/*@}*/




/** 
 * @name dynamically_threshold_matrix
 *
 * Dynamically thresholds a matrix.
 * @{
 */

/** @brief Dynamically thresholds a single precision matrix. */
void dynamically_threshold_matrix_f
(
    Matrix_f**      mat_out, 
    const Matrix_f* mat_in
);

/*@}*/


#ifdef __cplusplus
}
}
#endif


#endif
