/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/** 
 * @file 
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief JWS image file format.
 *
 * Specifies an integer or floating point image with grayscale or RGB
 * channels.
 */


#ifndef JIFF_H
#define JIFF_H


#include <jwsc/config.h>

#include <stdlib.h>

#include "jwsc/base/error.h"
#include "jwsc/image/image.h"


#ifdef __cplusplus
namespace jwsc {
extern "C" {
#endif


/** @brief Colorspace for a JIFF image. */
typedef enum
{
    JIFF_GRAY,     /*!< Single-channel, grayscaled image. */
    JIFF_RGB       /*!< Three-channel, RGB image. */
} 
JIFF_colorspace;


/** @brief Depth for colorspace channels. */
typedef enum 
{
    JIFF_8_BIT,  /*!< 8-bit channels. */
    JIFF_16_BIT, /*!< 16-bit channels. */
    JIFF_32_BIT  /*!< 32-bit channels. */
}
JIFF_depth;


/** @brief Channel type. */
typedef enum 
{
    JIFF_INTEGER,  /*!< Integer channels. */
    JIFF_FLOAT     /*!< Floating point channels. */
}
JIFF_type;


/** @brief Attributes for a JIFF image. */
typedef struct
{
    /** @brief Colorspace used by the JIFF image. */
    JIFF_colorspace  colorspace;

    /** @brief Depth of each channel in the colorspace (in bits). */
    JIFF_depth  depth;

    /** @brief Data-type of the channels in the colorspace. */
    JIFF_type  type;

    /** @brief Red channel averaging weight for grayscale conversion. */
    float red_w;

    /** @brief Green channel averaging weight for grayscale conversion. */
    float green_w;

    /** @brief Blue channel averaging weight for grayscale conversion. */
    float blue_w;
}
JIFF_attributes;


/** @brief Initializes a JIFF_attributes with defaults values. */
void init_jiff_attributes(JIFF_attributes* attrs);


/** @brief Reads an Image_f from a JIFF image file. */
Error* read_image_as_jiff_f(Image_f** img_out, const char* fname);


/** @brief Writes an Image_f to a JIFF image file. */
Error* write_image_as_jiff_f
(
    const Image_f*         img, 
    const char*            fname, 
    const JIFF_attributes* attrs
);


#ifdef __cplusplus
}
}
#endif


#endif
