/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/** 
 * @file 
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Declarations for an image block and associated functions.
 *
 * The image holds single precision floating point pixel elements.
 *
 * Although the image is indexed by image, row, and column, the pixels are
 * allocated as a 1D array in row-major order. So iterating over the 
 * pixels can be done inside a single loop.
 */


#ifndef IMGBLOCK_H
#define IMGBLOCK_H


#include <jwsc/config.h>

#include <stdlib.h>
#include <inttypes.h>

#include "jwsc/base/error.h"
#include "jwsc/matblock/matblock.h"
#include "jwsc/image/image.h"


#ifdef __cplusplus
namespace jwsc {
extern "C" {
#endif


/** @brief An image block of Pixel_f. */ 
typedef struct Imgblock_f
{
    /** @brief Number of images. */
    uint32_t num_imgs;

    /** @brief Number of rows. */
    uint32_t num_rows;

    /** @brief Number of columns. */
    uint32_t num_cols;

    /** @brief Pixels of the image block. */
    Pixel_f*** pxls;
} 
Imgblock_f;


/** @brief Orientation of an image in an imgblock. */
typedef enum
{
    IMGBLOCK_IMG_IMAGE,
    IMGBLOCK_ROW_IMAGE,
    IMGBLOCK_COL_IMAGE
}
Imgblock_image;


/** 
 * @name create_imgblock
 *
 * Creates an image block.
 * @{
 */

/** @brief Creates a single precision floating point image block. */
void create_imgblock_f
(
    Imgblock_f** img_out, 
    uint32_t     num_imgs, 
    uint32_t     num_rows, 
    uint32_t     num_cols
);

/*@}*/




/** 
 * @name create_init_imgblock
 *
 * Creates and initializes an image block.
 * @{
 */

/** 
 * @brief Creates and initializes a single precision floating point image
 * block.
 */
void create_init_imgblock_f
(
    Imgblock_f** img_out, 
    uint32_t     num_imgs, 
    uint32_t     num_rows, 
    uint32_t     num_cols,
    Pixel_f      val
);

/*@}*/




/** 
 * @name create_zero_imgblock
 *
 * Creates an image block and initializes it with zeros.
 * @{
 */

/** 
 * @brief Creates a single precision floating point image block and initializes 
 * it with zeros.
 */
void create_zero_imgblock_f
(
    Imgblock_f** img_out, 
    uint32_t     num_imgs, 
    uint32_t     num_rows, 
    uint32_t     num_cols
);

/*@}*/




/** 
 * @name create_random_imgblock
 *
 * Creates an image block and initializes it with random values.
 * @{
 */

/** 
 * @brief Creates a single precision floating point image block and initializes 
 * it with random values.
 */
void create_random_imgblock_f
(
    Imgblock_f** img_out, 
    uint32_t     num_imgs, 
    uint32_t     num_rows, 
    uint32_t     num_cols,
    Pixel_f      min,
    Pixel_f      max
);

/*@}*/




/** 
 * @name create_imgblock_from_matblocks
 *
 * Creates an image block from a three matrix blocks.
 * @{
 */

/** 
 * @brief Creates a single precision floating point image block from three 
 * matrix blocks.
 */
Error* create_imgblock_from_matblocks_f
(
    Imgblock_f**      img_out, 
    const Matblock_f* m_r,
    const Matblock_f* m_g,
    const Matblock_f* m_b
);

/*@}*/




/** 
 * @name create_imgblock_from_matblock
 *
 * Creates a grayscale image block from a matrix block.
 * @{
 */

/** 
 * @brief Creates a single precision floating point grayscale image block from 
 * a matrix block.
 */
void create_imgblock_from_matblock_f
(
    Imgblock_f**      img_out, 
    const Matblock_f* m
);

/*@}*/




/** 
 * @name create_matblocks_from_imgblock
 *
 * Creates three matrix blocks from an image block.
 * @{
 */

/**
 * @brief Creates three single precision floating point matrix blocks from an
 * image block.
 */
void create_matblocks_from_imgblock_f
(
    Matblock_f**      m_r_out, 
    Matblock_f**      m_g_out, 
    Matblock_f**      m_b_out, 
    const Imgblock_f* img
);

/*@}*/




/** 
 * @name create_matblock_from_imgblock
 *
 * Creates an averaged matrix block from an image block.
 * @{
 */

/**
 * @brief Creates an averaged single precision floating point matrix block from 
 * an image block.
 */
void create_matblock_from_imgblock_f
(
    Matblock_f**      m_out, 
    const Imgblock_f* img,
    float             r_weight,
    float             g_weight,
    float             b_weight
);

/*@}*/




/** 
 * @name copy_imgblock
 *
 * Copies an image block.
 * @{
 */

/** @brief Copies a single precision floating point image block. */
void copy_imgblock_f(Imgblock_f** img_out, const Imgblock_f* img_in);

/*@}*/




/**
 * @name copy_image_into_matblock
 *
 * Copies an image into a matblock.
 * @{
 */

/** @brief Copies a single precision floating point image into an image block.*/
Error* copy_image_into_imgblock_f
(
    Imgblock_f**      ib_out,
    const Imgblock_f* ib_in,
    const Image_f*    i,
    Imgblock_image    orient,
    uint32_t          index
);

/*@}*/




/**
 * @name copy_image_from_imgblock
 *
 * Copies an image from an imgblock.
 * @{
 */

/** @brief Copies a single precision floating point image from and image block.
 */
Error* copy_image_from_imgblock_f
(
    Image_f**         i_out,
    const Imgblock_f* ib,
    Imgblock_image    orient,
    uint32_t          index
);

/*@}*/




/** 
 * @name free_imgblock
 *
 * Frees an image block.
 * @{
 */

/** @brief Frees a single precision floatin point image block. */
void free_imgblock_f(Imgblock_f* img);

/*@}*/


#ifdef __cplusplus
}
}
#endif


#endif
