/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/** 
 * @file 
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Definitions for functions to read and write image blocks.
 */


#include <jwsc/config.h>

#include <stdlib.h>
#include <stdio.h>
#include <inttypes.h>
#include <assert.h>
#include <string.h>

#include "jwsc/base/error.h"
#include "jwsc/image/image.h"
#include "jwsc/image/image_io.h"
#include "jwsc/image/jiff.h"
#include "jwsc/imgblock/imgblock.h"
#include "jwsc/imgblock/imgblock_io.h"


/** 
 * @name read_imgblock_numbered
 *
 * Reads an image block from a numbered sequence of files.
 * @{
 */

/** 
 * An example for @em fname_fmt is
 *
 * @code image-%04d.tiff @endcode
 *
 * @param  imgblk_out  Result parameter. If @em *imgblk_out is NULL, an image 
 *                     block is allocated; otherwise its space is re-used.
 * @param  fname_fmt   Format for the file name with one printf style integer
 *                     conversion for the file number.
 * @param  num_imgs    Number of images to read into the image block.
 *
 * @return On success, NULL is returned. On error, an Error is returned and @em
 * *imgblk_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Could not read one of the images.
 * - @link Error_types::ERROR_INV_ARG ERROR_INV_ARG @endlink 
 *   The images to read do not have consistent images.
 *
 * @note Dimensions of all the files to read must be > 0.
 */
Error* read_imgblock_numbered_f
(
    Imgblock_f** imgblk_out, 
    const char*  fname_fmt,
    uint32_t     num_imgs
)
{
    char fname[256] = {0};

    uint32_t num_rows, num_cols;
    uint32_t i;

    Image_f* img = NULL;
    Error*   e;

    num_rows = 0;
    num_cols = 0;

    for (i = 0; i < num_imgs; i++)
    {
        snprintf(fname, 255, fname_fmt, i+1);

        if ((e = read_image_f(&img, fname)) != NULL)
        {
            free_imgblock_f(*imgblk_out); *imgblk_out = NULL;
            return e;
        }

        if (num_rows == 0 && num_cols == 0)
        {
            num_rows = img->num_rows;
            num_cols = img->num_cols;
            create_imgblock_f(imgblk_out, num_imgs, num_rows, num_cols);
        }

        assert(copy_image_into_imgblock_f(imgblk_out, *imgblk_out, img, 
                    IMGBLOCK_IMG_IMAGE, i) == NULL);
    }

    free_image_f(img);

    return NULL;
}

/*@}*/




/** 
 * @name read_imgblock_named
 *
 * Reads an image block from a named sequence of files.
 * @{
 */

/** 
 * @param  imgblk_out  Result parameter. If @em *imgblk_out is NULL, an image 
 *                     block is allocated; otherwise its space is re-used.
 * @param  fnames      Array of file names to read. Must contain at least 
 *                     @em num_imgs elts.
 * @param  num_imgs    Number of images to read into the image block.
 *
 * @return On success, NULL is returned. On error, an Error is returned and @em
 * *imgblk_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Could not read one of the images.
 * - @link Error_types::ERROR_INV_ARG ERROR_INV_ARG @endlink 
 *   The images to read do not have consistent images.
 *
 * @note Dimensions of all the files to read must be > 0.
 */
Error* read_imgblock_named_f
(
    Imgblock_f**   imgblk_out, 
    const char**   fnames,
    uint32_t       num_imgs
)
{
    uint32_t num_rows, num_cols;
    uint32_t i;

    Image_f* img = NULL;
    Error*   e;

    num_rows = 0;
    num_cols = 0;

    for (i = 0; i < num_imgs; i++)
    {
        if ((e = read_image_f(&img, fnames[ i ])) != NULL)
        {
            free_imgblock_f(*imgblk_out); *imgblk_out = NULL;
            return e;
        }

        if (num_rows == 0 && num_cols == 0)
        {
            num_rows = img->num_rows;
            num_cols = img->num_cols;
            create_imgblock_f(imgblk_out, num_imgs, num_rows, num_cols);
        }

        assert(copy_image_into_imgblock_f(imgblk_out, *imgblk_out, img, 
                    IMGBLOCK_IMG_IMAGE, i) == NULL);
    }

    free_image_f(img);

    return NULL;
}

/*@}*/




/** 
 * @name write_imgblock_numbered
 *
 * Writes an image block to a numbered sequence of files.
 * @{
 */

/** 
 * An example for @em fname_fmt is
 *
 * @code image-%04d.tiff @endcode
 *
 * @param  imgblk     Image block to write.
 * @param  fname_fmt  Format for the file name with one printf style integer
 *                    conversion for the file number.
 *
 * @return On success, NULL is returned. On error, an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Could not write an image.
 */
Error* write_imgblock_numbered_f
(
    const Imgblock_f* imgblk,
    const char*       fname_fmt
)
{
    char fname[256] = {0};

    uint32_t i;

    Image_f* img = NULL;
    Error*   e;

    for (i = 0; i < imgblk->num_imgs; i++)
    {
        assert(copy_image_from_imgblock_f(&img, imgblk, IMGBLOCK_IMG_IMAGE, i) 
                == NULL);

        snprintf(fname, 255, fname_fmt, i+1);

        if ((e = write_image_f(img, fname)) != NULL)
        {
            free_image_f(img);
            return e;
        }
    }

    free_image_f(img);

    return NULL;
}

/*@}*/




/** 
 * @name write_imgblock_named
 *
 * Writes an image block to a named sequence of files.
 * @{
 */

/** 
 * @param  imgblk  Image block to write.
 * @param  fnames  File names to write the images in the block to.
 *
 * @return On success, NULL is returned. On error, an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Could not write an image.
 */
Error* write_imgblock_named_f
(
    const Imgblock_f* imgblk,
    const char**      fnames
)
{
    uint32_t i;

    Image_f* img = NULL;
    Error*   e;

    for (i = 0; i < imgblk->num_imgs; i++)
    {
        assert(copy_image_from_imgblock_f(&img, imgblk, IMGBLOCK_IMG_IMAGE, i) 
                == NULL);

        if ((e = write_image_f(img, fnames[ i ])) != NULL)
        {
            free_image_f(img);
            return e;
        }
    }

    free_image_f(img);

    return NULL;
}

/*@}*/




/** 
 * @name write_imgblock_as_numbered_jiff
 *
 * Writes an image block to a numbered sequence of JIFF files.
 * @{
 */

/** 
 * An example for @em fname_fmt is
 *
 * @code image-%04d.jiff @endcode
 *
 * @param  imgblk     Image block to write.
 * @param  fname_fmt  Format for the file name with one printf style integer
 *                    conversion for the file number.
 * @param  attrs   JIFF file attributes.
 *
 * @return On success, NULL is returned. On error, an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Could not write an image.
 */
Error* write_imgblock_as_numbered_jiff_f
(
    const Imgblock_f*      imgblk,
    const char*            fname_fmt,
    const JIFF_attributes* attrs
)
{
    char fname[256] = {0};

    uint32_t i;

    Image_f* img = NULL;
    Error*   e;

    for (i = 0; i < imgblk->num_imgs; i++)
    {
        assert(copy_image_from_imgblock_f(&img, imgblk, IMGBLOCK_IMG_IMAGE, i) 
                == NULL);

        snprintf(fname, 255, fname_fmt, i+1);

        if ((e = write_image_as_jiff_f(img, fname, attrs)) != NULL)
        {
            free_image_f(img);
            return e;
        }
    }

    free_image_f(img);

    return NULL;
}

/*@}*/




/** 
 * @name write_imgblock_as_named_jiff
 *
 * Writes an image block to a named sequence of JIFF files.
 * @{
 */

/** 
 * @param  imgblk  Image block to write.
 * @param  fnames  File names to write the images in the block to.
 * @param  attrs   JIFF file attributes.
 *
 * @return On success, NULL is returned. On error, an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Could not write an image.
 */
Error* write_imgblock_as_named_jiff_f
(
    const Imgblock_f*      imgblk,
    const char**           fnames,
    const JIFF_attributes* attrs
)
{
    uint32_t i;

    Image_f* img = NULL;
    Error*   e;

    for (i = 0; i < imgblk->num_imgs; i++)
    {
        assert(copy_image_from_imgblock_f(&img, imgblk, IMGBLOCK_IMG_IMAGE, i) 
                == NULL);

        if ((e = write_image_as_jiff_f(img, fnames[ i ], attrs)) != NULL)
        {
            free_image_f(img);
            return e;
        }
    }

    free_image_f(img);

    return NULL;
}

/*@}*/
