/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/** 
 * @file 
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Declarations for functions to read and write image blocks.
 */


#ifndef IMGBLOCK_IO_H
#define IMGBLOCK_IO_H


#include <jwsc/config.h>

#include <stdlib.h>

#include "jwsc/base/error.h"
#include "jwsc/image/jiff.h"
#include "jwsc/imgblock/imgblock.h"


#ifdef __cplusplus
namespace jwsc {
extern "C" {
#endif


/** 
 * @name read_imgblock_numbered
 *
 * Reads an image block from a numbered sequence of files.
 * @{
 */

/** 
 * @brief Reads a single precision floating point image block from a numbered
 * sequence of files.
 */
Error* read_imgblock_numbered_f
(
    Imgblock_f** img_out, 
    const char*  fname_fmt,
    uint32_t     num_imgs
);

/*@}*/




/** 
 * @name read_imgblock_named
 *
 * Reads an image block from a named sequence of files.
 * @{
 */

/** 
 * @brief Reads a single precision floating point image block from a named
 * sequence of files.
 */
Error* read_imgblock_named_f
(
    Imgblock_f**   img_out, 
    const char**   fnames,
    uint32_t       num_imgs
);

/*@}*/




/** 
 * @name write_imgblock_numbered
 *
 * Writes an image block to a numbered sequence of files.
 * @{
 */

/** 
 * @brief Write a single precision floating point image block to a numbered
 * sequence of files.
 */
Error* write_imgblock_numbered_f
(
    const Imgblock_f* img,
    const char*       fname_fmt
);

/*@}*/




/** 
 * @name write_imgblock_named
 *
 * Writes an image block to a named sequence of files.
 * @{
 */

/** 
 * @brief Write a single precision floating point image block to a named
 * sequence of files.
 */
Error* write_imgblock_named_f
(
    const Imgblock_f* img,
    const char**      fnames
);

/*@}*/




/** 
 * @name write_imgblock_as_numbered_jiff
 *
 * Writes an image block to a numbered sequence of JIFF files.
 * @{
 */

/** 
 * @brief Write a single precision floating point image block to a numbered
 * sequence of JIFF files.
 */
Error* write_imgblock_as_numbered_jiff_f
(
    const Imgblock_f*      img,
    const char*            fname_fmt,
    const JIFF_attributes* attrs
);

/*@}*/




/** 
 * @name write_imgblock_as_named_jiff
 *
 * Writes an image block to a named sequence of JIFF files.
 * @{
 */

/** 
 * @brief Write a single precision floating point image block to a named
 * sequence of JIFF files.
 */
Error* write_imgblock_as_named_jiff_f
(
    const Imgblock_f*      img,
    const char**           fnames,
    const JIFF_attributes* attrs
);

/*@}*/


#ifdef __cplusplus
}
}
#endif


#endif
