/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Declarations for the image block surface detector.
 *
 * The surface detection algorithm is a variation of the standard Canny 
 * algorithm: the image block is convolved with the gradiant of Gaussian, then 
 * non-maximal suppression and hysterisis are applied to the resulting gradient 
 * map.
 */


#ifndef SURFACE_H 
#define SURFACE_H 


#include <jwsc/config.h>

#include <stdlib.h>
#include <inttypes.h>

#include "jwsc/base/error.h"
#include "jwsc/vector/vector.h"
#include "jwsc/imgblock/imgblock.h"
#include "jwsc/matblock/matblock.h"


#ifdef __cplusplus
namespace jwsc {
extern "C" {
#endif


/**
 * @brief Surface point in an image block or matrix block (single-sampled 
 * image).
 *
 * Currently includes the matrix, row and column of the surface point and the 
 * gradient information at the point. I'm not sure if the gradient info is 
 * really necessary, but it could be useful in some cases, perhaps to know the 
 * gradient magnitude at the point.
 */
typedef struct
{
    /** @brief Column the point lies in. */
    uint32_t col;

    /** @brief Row the point lies in. */
    uint32_t row;

    /** @brief Matrix (image) the point lies in. */
    uint32_t mat;

    /** @brief Rate of change in brightness along the columns at the point. */
    float dcol;

    /** @brief Rate of change in brightness along the rows at the point. */
    float drow;

    /** @brief Rate of change in brightness along the matrix at the point. */
    float dmat;

    /** @brief Intensity of the surface point, in the range [0,1]. */
    float intensity;
}
Surface_point_f;




/**
 * @brief A patch of surface on a plane defined by the position and
 * orientation of a Surface_point_f.
 *
 * The patches are parallelograms.
 */
typedef struct
{
    /**
     * @brief Four points composing the parallelogram for the patch.
     */
    Vector_f* pts[4];

    /**
     * @brief Normal vector for the plane the patch lies on.
     *
     * From the gradient vector of the Surface_point_f.
     */
    Vector_f* normal;

    /**
     * @brief Center (or anchor) point for the patch.
     *
     * From the location of the Surface_point_f.
     */
    Vector_f* center;

    /** @brief Ratio of reflected to incendent light on the patch. */
    float albedo;
}
Surface_patch_f;


/** 
 * @name detect_imgblock_surface_points
 *
 * Detects surface points in each sample of an image block.
 * @{
 */

/** 
 * @brief Detects surface points in each sample of a single precision image 
 * block.
 */
Error* detect_imgblock_surface_points_f
(
    Surface_point_f** r_pts_out,
    uint32_t*         num_r_pts_out,
    Surface_point_f** g_pts_out,
    uint32_t*         num_g_pts_out,
    Surface_point_f** b_pts_out,
    uint32_t*         num_b_pts_out,
    const Imgblock_f* img,
    float             sigma, 
    float             grad_img_weight,
    float             grad_row_weight,
    float             grad_col_weight,
    float             begin_thresh, 
    float             end_thresh
);

/*@}*/




/** 
 * @name detect_matblock_surface_points
 *
 * Detects surface points in a matrix block (single-sample image block).
 * @{
 */

/** 
 * @brief Detects surface points in a single precision matrix block 
 * (single-sample image block).
 */
Error* detect_matblock_surface_points_f
(
    Surface_point_f** pts_out,
    uint32_t*         num_pts_out,
    const Matblock_f* m,
    float             sigma, 
    float             grad_mat_weight,
    float             grad_row_weight,
    float             grad_col_weight,
    float             begin_thresh, 
    float             end_thresh
);

/*@}*/




/** 
 * @name sample_surface_points
 *
 * Randomly samples surface points from an array of them.
 * @{
 */

/**
 * @brief Randomly samples single precision surface points from an array of 
 * them.
 */
void sample_surface_points_f
(
    Surface_point_f**      pts_out, 
    uint32_t*              num_pts_out,
    const Surface_point_f* pts_in, 
    uint32_t               num_pts_in,
    float                  p
);

/*@}*/




/** 
 * @name color_surface_points
 *
 * Colors surface points in an image block.
 * @{
 */

/** @brief Colors single precision surface points in an image block. */
void color_surface_points_f
(
    Imgblock_f**           img_out,
    const Imgblock_f*      img_in,
    const Surface_point_f* pts, 
    uint32_t               num_pts,
    const Pixel_f*         pxl
);

/*@}*/




/** 
 * @name read_surface_points
 *
 * Reads an array of surface points in ASCII format from a file.
 * @{
 */

/**
 * @brief Reads an array of single precision surface points in ASCII format 
 * from a file.
 */
Error* read_surface_points_f
(
    Surface_point_f** pts_out,
    uint32_t*         num_pts_out,
    const char*       fname
);

/*@}*/




/** 
 * @name write_surface_points
 *
 * Writes an array of surface points in ASCII format to a file.
 * @{
 */

/**
 * @brief Writes an array of single precision surface pionts in ASCII format to 
 * a file.
 */
Error* write_surface_points_f
(
    const Surface_point_f* pts,
    uint32_t               num_pts,
    const char*            fname
);

/*@}*/




/** 
 * @name create_surface_patch
 *
 * Creates a Surface_patch from a Surface_point.
 * @{
 */

/** @brief Creates a Surface_patch_f from a Surface_point_f. */
void create_surface_patch_f
(
    Surface_patch_f**      patch_out,
    const Surface_point_f* point,
    float                  size,
    float                  x_scale,
    float                  y_scale,
    float                  z_scale
);

/*@}*/




/** 
 * @name free_surface_patch
 *
 * Frees a Surface_patch from a Surface_point.
 * @{
 */

/** @brief Frees a Surface_patch_f. */
void free_surface_patch_f(Surface_patch_f* patch);

/*@}*/


#ifdef __cplusplus
}
}
#endif


#endif 
