/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Declarations for a matrix block and associated functions.
 *
 * The matrix block holds integers and floating point real and complex numbers.
 *
 * Although the matrix block is indexed by matrix, row, and column, the 
 * elements are allocated as a 1D array in matrix,row,col order 
 * (that would be matrix-major order). So iterating over the elements can be 
 * done inside a single loop.
 */


#ifndef MATBLOCK_H 
#define MATBLOCK_H 


#include <jwsc/config.h>

#include <stdlib.h>
#include <inttypes.h>

#include "jwsc/base/error.h"
#include "jwsc/math/complex.h"
#include "jwsc/matrix/matrix.h"


#ifdef __cplusplus
namespace jwsc {
extern "C" {
#endif


/** @brief Matrix block of signed 8-bit values. */
typedef struct
{
    /** @brief Number of matrices. */
    uint32_t num_mats;

    /** @brief Number of rows. */
    uint32_t num_rows;

    /** @brief Number of columns. */
    uint32_t num_cols;

    /** @brief Elements of the matrix block. */
    int8_t*** elts;
}
Matblock_i8;

/** @brief Matrix block of signed 16-bit values. */
typedef struct
{
    /** @brief Number of matrices. */
    uint32_t num_mats;

    /** @brief Number of rows. */
    uint32_t num_rows;

    /** @brief Number of columns. */
    uint32_t num_cols;

    /** @brief Elements of the matrix block. */
    int16_t*** elts;
}
Matblock_i16;

/** @brief Matrix block of signed 32-bit values. */
typedef struct
{
    /** @brief Number of matrices. */
    uint32_t num_mats;

    /** @brief Number of rows. */
    uint32_t num_rows;

    /** @brief Number of columns. */
    uint32_t num_cols;

    /** @brief Elements of the matrix block. */
    int32_t*** elts;
}
Matblock_i32;

/** @brief Matrix block of signed 64-bit values. */
typedef struct
{
    /** @brief Number of matrices. */
    uint32_t num_mats;

    /** @brief Number of rows. */
    uint32_t num_rows;

    /** @brief Number of columns. */
    uint32_t num_cols;

    /** @brief Elements of the matrix block. */
    int64_t*** elts;
}
Matblock_i64;


/** @brief Matrix block of unsigned 8-bit values. */
typedef struct
{
    /** @brief Number of matrices. */
    int32_t num_mats;

    /** @brief Number of rows. */
    int32_t num_rows;

    /** @brief Number of columns. */
    int32_t num_cols;

    /** @brief Elements of the matrix block. */
    uint8_t*** elts;
}
Matblock_u8;

/** @brief Matrix block of unsigned 16-bit values. */
typedef struct
{
    /** @brief Number of matrices. */
    int32_t num_mats;

    /** @brief Number of rows. */
    uint32_t num_rows;

    /** @brief Number of columns. */
    uint32_t num_cols;

    /** @brief Elements of the matrix block. */
    uint16_t*** elts;
}
Matblock_u16;

/** @brief Matrix block of unsigned 32-bit values. */
typedef struct
{
    /** @brief Number of matrices. */
    int32_t num_mats;

    /** @brief Number of rows. */
    uint32_t num_rows;

    /** @brief Number of columns. */
    uint32_t num_cols;

    /** @brief Elements of the matrix block. */
    uint32_t*** elts;
}
Matblock_u32;

/** @brief Matrix block of unsigned 64-bit values. */
typedef struct
{
    /** @brief Number of matrices. */
    int32_t num_mats;

    /** @brief Number of rows. */
    uint32_t num_rows;

    /** @brief Number of columns. */
    uint32_t num_cols;

    /** @brief Elements of the matrix block. */
    uint64_t*** elts;
}
Matblock_u64;

/** @brief Matrix block of single precision floating point values. */
typedef struct
{
    /** @brief Number of matrices. */
    uint32_t num_mats;

    /** @brief Number of rows. */
    uint32_t num_rows;

    /** @brief Number of columns. */
    uint32_t num_cols;

    /** @brief Elements of the matrix block. */
    float*** elts;
}
Matblock_f;

/** @brief Matrix block of double precision floating point values. */
typedef struct
{
    /** @brief Number of matrices. */
    uint32_t num_mats;

    /** @brief Number of rows. */
    uint32_t num_rows;

    /** @brief Number of columns. */
    uint32_t num_cols;

    /** @brief Elements of the matrix block. */
    double*** elts;
}
Matblock_d;

/** @brief Matrix block of single precision complex floating point values. */
typedef struct
{
    /** @brief Number of matrices. */
    uint32_t num_mats;

    /** @brief Number of rows. */
    uint32_t num_rows;

    /** @brief Number of columns. */
    uint32_t num_cols;

    /** @brief Elements of the matrix block. */
    Complex_f*** elts;
}
Matblock_cf;

/** @brief Matrix block of double precision complex floating point values. */
typedef struct
{
    /** @brief Number of matrices. */
    uint32_t num_mats;

    /** @brief Number of rows. */
    uint32_t num_rows;

    /** @brief Number of columns. */
    uint32_t num_cols;

    /** @brief Elements of the matrix block. */
    Complex_d*** elts;
}
Matblock_cd;


/** @brief Orientation of a matrix in a matblock. */
typedef enum
{
    MATBLOCK_MAT_MATRIX,
    MATBLOCK_ROW_MATRIX,
    MATBLOCK_COL_MATRIX
}
Matblock_matrix;


/**
 * @name create_matblock
 *
 * Creates a matrix block of integer or floating point values.
 * @{
 */

/** @brief 8-bit unsigned integer. */
void create_matblock_u8
(
    Matblock_u8** m_out, 
    uint32_t      num_mats, 
    uint32_t      num_rows, 
    uint32_t      num_cols
);

/** @brief 32-bit unsigned integer. */
void create_matblock_u32
(
    Matblock_u32** m_out, 
    uint32_t       num_mats, 
    uint32_t       num_rows, 
    uint32_t       num_cols
);

/** @brief 32-bit signed integer. */
void create_matblock_i32
(
    Matblock_i32** m_out, 
    uint32_t       num_mats, 
    uint32_t       num_rows, 
    uint32_t       num_cols
);

/** @brief 64-bit signed integer. */
void create_matblock_i64
(
    Matblock_i64** m_out, 
    uint32_t       num_mats, 
    uint32_t       num_rows, 
    uint32_t       num_cols
);

/** @brief Single precision floating point. */
void create_matblock_f
(
    Matblock_f** m_out, 
    uint32_t     num_mats, 
    uint32_t     num_rows, 
    uint32_t     num_cols
);

/** @brief Double precision floating point. */
void create_matblock_d
(
    Matblock_d** m_out, 
    uint32_t     num_mats, 
    uint32_t     num_rows, 
    uint32_t     num_cols
);

/** @brief Single precision complex floating point. */
void create_matblock_cf
(
    Matblock_cf** m_out, 
    uint32_t      num_mats, 
    uint32_t      num_rows, 
    uint32_t      num_cols
);

/** @brief Double precision complexfloating point. */
void create_matblock_cd
(
    Matblock_cd** m_out, 
    uint32_t      num_mats, 
    uint32_t      num_rows, 
    uint32_t      num_cols
);

/*@}*/




/**
 * @name create_init_matblock
 *
 * Creates and initializes a matrix block of integer or floating point values.
 * @{
 */

/** @brief 8-bit unsigned integer. */
void create_init_matblock_u8
(
    Matblock_u8** m_out, 
    uint32_t      num_mats, 
    uint32_t      num_rows, 
    uint32_t      num_cols,
    uint8_t       val
);

/** @brief 32-bit unsigned integer. */
void create_init_matblock_u32
(
    Matblock_u32** m_out, 
    uint32_t       num_mats, 
    uint32_t       num_rows, 
    uint32_t       num_cols,
    uint32_t       val
);

/** @brief 32-bit signed integer. */
void create_init_matblock_i32
(
    Matblock_i32** m_out, 
    uint32_t       num_mats, 
    uint32_t       num_rows, 
    uint32_t       num_cols,
    int32_t        val
);

/** @brief 64-bit signed integer. */
void create_init_matblock_i64
(
    Matblock_i64** m_out, 
    uint32_t       num_mats, 
    uint32_t       num_rows, 
    uint32_t       num_cols,
    int64_t        val
);

/** @brief Single precision floating point. */
void create_init_matblock_f
(
    Matblock_f** m_out, 
    uint32_t     num_mats, 
    uint32_t     num_rows, 
    uint32_t     num_cols,
    float        val
);

/** @brief Double precision floating point. */
void create_init_matblock_d
(
    Matblock_d** m_out, 
    uint32_t     num_mats, 
    uint32_t     num_rows, 
    uint32_t     num_cols,
    double       val
);

/** @brief Single precision complex floating point. */
void create_init_matblock_cf
(
    Matblock_cf** m_out, 
    uint32_t      num_mats, 
    uint32_t      num_rows, 
    uint32_t      num_cols,
    Complex_f     val
);

/** @brief Double precision complexfloating point. */
void create_init_matblock_cd
(
    Matblock_cd** m_out, 
    uint32_t      num_mats, 
    uint32_t      num_rows, 
    uint32_t      num_cols,
    Complex_d     val
);

/*@}*/




/**
 * @name create_zero_matblock
 *
 * Creates a matrix block of integer or floating point values and initializes 
 * it with zeros.
 * @{
 */

/** @brief 8-bit unsigned integer. */
void create_zero_matblock_u8
(
    Matblock_u8** m_out, 
    uint32_t      num_mats, 
    uint32_t      num_rows, 
    uint32_t      num_cols
);

/** @brief 32-bit unsigned integer. */
void create_zero_matblock_u32
(
    Matblock_u32** m_out, 
    uint32_t       num_mats, 
    uint32_t       num_rows, 
    uint32_t       num_cols
);

/** @brief 32-bit signed integer. */
void create_zero_matblock_i32
(
    Matblock_i32** m_out, 
    uint32_t       num_mats, 
    uint32_t       num_rows, 
    uint32_t       num_cols
);

/** @brief 64-bit signed integer. */
void create_zero_matblock_i64
(
    Matblock_i64** m_out, 
    uint32_t       num_mats, 
    uint32_t       num_rows, 
    uint32_t       num_cols
);

/** @brief Single precision floating point. */
void create_zero_matblock_f
(
    Matblock_f** m_out, 
    uint32_t     num_mats, 
    uint32_t     num_rows, 
    uint32_t     num_cols
);

/** @brief Double precision floating point. */
void create_zero_matblock_d
(
    Matblock_d** m_out, 
    uint32_t     num_mats, 
    uint32_t     num_rows, 
    uint32_t     num_cols
);

/** @brief Single precision complex floating point. */
void create_zero_matblock_cf
(
    Matblock_cf** m_out, 
    uint32_t      num_mats, 
    uint32_t      num_rows, 
    uint32_t      num_cols
);

/** @brief Double precision complexfloating point. */
void create_zero_matblock_cd
(
    Matblock_cd** m_out, 
    uint32_t      num_mats, 
    uint32_t      num_rows, 
    uint32_t      num_cols
);

/*@}*/




/**
 * @name create_random_matblock
 *
 * Creates a matrix block of integer or floating point values and initializes it
 * with random values.
 * @{
 */

/** @brief 32-bit unsigned integer. */
void create_random_matblock_u32
(
    Matblock_u32** m_out, 
    uint32_t       num_mats, 
    uint32_t       num_rows, 
    uint32_t       num_cols,
    uint32_t       min,
    uint32_t       max
);

/** @brief 32-bit signed integer. */
void create_random_matblock_i32
(
    Matblock_i32** m_out, 
    uint32_t       num_mats, 
    uint32_t       num_rows, 
    uint32_t       num_cols,
    int32_t        min,
    int32_t        max
);

/** @brief 64-bit signed integer. */
void create_random_matblock_i64
(
    Matblock_i64** m_out, 
    uint32_t       num_mats, 
    uint32_t       num_rows, 
    uint32_t       num_cols,
    int64_t        min,
    int64_t        max
);

/** @brief Single precision floating point. */
void create_random_matblock_f
(
    Matblock_f** m_out, 
    uint32_t     num_mats, 
    uint32_t     num_rows, 
    uint32_t     num_cols,
    float        min,
    float        max
);

/** @brief Double precision floating point. */
void create_random_matblock_d
(
    Matblock_d** m_out, 
    uint32_t     num_mats, 
    uint32_t     num_rows, 
    uint32_t     num_cols,
    double       min,
    double       max
);

/** @brief Single precision complex floating point. */
void create_random_matblock_cf
(
    Matblock_cf** m_out, 
    uint32_t      num_mats, 
    uint32_t      num_rows, 
    uint32_t      num_cols,
    Complex_f     min,
    Complex_f     max
);

/** @brief Double precision complexfloating point. */
void create_random_matblock_cd
(
    Matblock_cd** m_out, 
    uint32_t      num_mats, 
    uint32_t      num_rows, 
    uint32_t      num_cols,
    Complex_d     min,
    Complex_d     max
);

/*@}*/




/**
 * @name copy_matblock
 *
 * Copies a matrix block into another.
 * @{
 */

/** @brief 32-bit unsigned integer. */
Error* copy_matblock_u32(Matblock_u32** m_out, const Matblock_u32* m_in);

/** @brief 32-bit signed integer. */
Error* copy_matblock_i32(Matblock_i32** m_out, const Matblock_i32* m_in);

/** @brief 64-bit signed integer. */
Error* copy_matblock_i64(Matblock_i64** m_out, const Matblock_i64* m_in);

/** @brief Single precision floating point. */
Error* copy_matblock_f(Matblock_f** m_out, const Matblock_f* m_in);

/** @brief Double precision floating point. */
Error* copy_matblock_d(Matblock_d** m_out, const Matblock_d* m_in);

/** @brief Single precision complex floating point. */
Error* copy_matblock_cf(Matblock_cf** m_out, const Matblock_cf* m_in);

/** @brief Double precision complex floating point. */
Error* copy_matblock_cd(Matblock_cd** m_out, const Matblock_cd* m_in);

/*@}*/




/**
 * @name copy_matblock_block
 *
 * Copies a block of a matrix block.
 * @{
 */

/** @brief Copies a block of a 32-bit unsigned integer matrix block. */
Error* copy_matblock_block_u32
(
    Matblock_u32**      m_out, 
    const Matblock_u32* m_in,
    uint32_t            mat_offset,
    uint32_t            row_offset,
    uint32_t            col_offset,
    uint32_t            num_mats,
    uint32_t            num_rows,
    uint32_t            num_cols
);

/** @brief Copies a block of a 32-bit signed integer matrix block. */
Error* copy_matblock_block_i32
(
    Matblock_i32**      m_out, 
    const Matblock_i32* m_in,
    uint32_t            mat_offset,
    uint32_t            row_offset,
    uint32_t            col_offset,
    uint32_t            num_mats,
    uint32_t            num_rows,
    uint32_t            num_cols
);

/** @brief Copies a block of a 64-bit signed integer matrix block. */
Error* copy_matblock_block_i64
(
    Matblock_i64**      m_out, 
    const Matblock_i64* m_in,
    uint32_t            mat_offset,
    uint32_t            row_offset,
    uint32_t            col_offset,
    uint32_t            num_mats,
    uint32_t            num_rows,
    uint32_t            num_cols
);

/** 
 * @brief Copies a block of a single precision floating point matrix block. */
Error* copy_matblock_block_f
(
    Matblock_f**      m_out, 
    const Matblock_f* m_in,
    uint32_t          mat_offset,
    uint32_t          row_offset,
    uint32_t          col_offset,
    uint32_t          num_mats,
    uint32_t          num_rows,
    uint32_t          num_cols
);

/** 
 * @brief Copies a block of a double precision floating point matrix block. */
Error* copy_matblock_block_d
(
    Matblock_d**      m_out, 
    const Matblock_d* m_in,
    uint32_t          mat_offset,
    uint32_t          row_offset,
    uint32_t          col_offset,
    uint32_t          num_mats,
    uint32_t          num_rows,
    uint32_t          num_cols
);

/** 
 * @brief Copies a block of a single precision complex floating point matrix 
 * block.
 */
Error* copy_matblock_block_cf
(
    Matblock_cf**      m_out, 
    const Matblock_cf* m_in,
    uint32_t           mat_offset,
    uint32_t           row_offset,
    uint32_t           col_offset,
    uint32_t           num_mats,
    uint32_t           num_rows,
    uint32_t           num_cols
);

/** 
 * @brief Copies a block of a double precision complex floating point matrix 
 * block.
 */
Error* copy_matblock_block_cd
(
    Matblock_cd**      m_out, 
    const Matblock_cd* m_in,
    uint32_t           mat_offset,
    uint32_t           row_offset,
    uint32_t           col_offset,
    uint32_t           num_mats,
    uint32_t           num_rows,
    uint32_t           num_cols
);

/*@}*/




/**
 * @name copy_matblock_into_matblock
 *
 * Copies a block of a matrix block into another.
 * @{
 */

/** 
 * @brief Copies a block of a 32-bit unsigned integer matrix block into 
 * another.
 */
Error* copy_matblock_block_into_matblock_u32
(
    Matblock_u32*       m_1, 
    uint32_t            mat_offset_1,
    uint32_t            row_offset_1,
    uint32_t            col_offset_1,
    const Matblock_u32* m_2,
    uint32_t            mat_offset_2,
    uint32_t            row_offset_2,
    uint32_t            col_offset_2,
    uint32_t            num_mats,
    uint32_t            num_rows,
    uint32_t            num_cols
);

/** 
 * @brief Copies a block of a 32-bit signed integer matrix block into 
 * another.
 */
Error* copy_matblock_block_into_matblock_i32
(
    Matblock_i32*       m_1, 
    uint32_t            mat_offset_1,
    uint32_t            row_offset_1,
    uint32_t            col_offset_1,
    const Matblock_i32* m_2,
    uint32_t            mat_offset_2,
    uint32_t            row_offset_2,
    uint32_t            col_offset_2,
    uint32_t            num_mats,
    uint32_t            num_rows,
    uint32_t            num_cols
);

/** 
 * @brief Copies a block of a 64-bit signed integer matrix block into 
 * another.
 */
Error* copy_matblock_block_into_matblock_i64
(
    Matblock_i64*       m_1, 
    uint32_t            mat_offset_1,
    uint32_t            row_offset_1,
    uint32_t            col_offset_1,
    const Matblock_i64* m_2,
    uint32_t            mat_offset_2,
    uint32_t            row_offset_2,
    uint32_t            col_offset_2,
    uint32_t            num_mats,
    uint32_t            num_rows,
    uint32_t            num_cols
);

/** 
 * @brief Copies a block of a single precision floating point matrix block into 
 * another.
 */
Error* copy_matblock_block_into_matblock_f
(
    Matblock_f*       m_1, 
    uint32_t          mat_offset_1,
    uint32_t          row_offset_1,
    uint32_t          col_offset_1,
    const Matblock_f* m_2,
    uint32_t          mat_offset_2,
    uint32_t          row_offset_2,
    uint32_t          col_offset_2,
    uint32_t          num_mats,
    uint32_t          num_rows,
    uint32_t          num_cols
);

/** 
 * @brief Copies a block of a double precision floating point matrix block into 
 * another.
 */
Error* copy_matblock_block_into_matblock_d
(
    Matblock_d*       m_1, 
    uint32_t          mat_offset_1,
    uint32_t          row_offset_1,
    uint32_t          col_offset_1,
    const Matblock_d* m_2,
    uint32_t          mat_offset_2,
    uint32_t          row_offset_2,
    uint32_t          col_offset_2,
    uint32_t          num_mats,
    uint32_t          num_rows,
    uint32_t          num_cols
);

/** 
 * @brief Copies a block of a single precision complex floating point matrix 
 * block into another.
 */
Error* copy_matblock_block_into_matblock_cf
(
    Matblock_cf*       m_1, 
    uint32_t           mat_offset_1,
    uint32_t           row_offset_1,
    uint32_t           col_offset_1,
    const Matblock_cf* m_2,
    uint32_t           mat_offset_2,
    uint32_t           row_offset_2,
    uint32_t           col_offset_2,
    uint32_t           num_mats,
    uint32_t           num_rows,
    uint32_t           num_cols
);

/** 
 * @brief Copies a block of a double precision complex floating point matrix
 * block into another.
 */
Error* copy_matblock_block_into_matblock_cd
(
    Matblock_cd*       m_1, 
    uint32_t           mat_offset_1,
    uint32_t           row_offset_1,
    uint32_t           col_offset_1,
    const Matblock_cd* m_2,
    uint32_t           mat_offset_2,
    uint32_t           row_offset_2,
    uint32_t           col_offset_2,
    uint32_t           num_mats,
    uint32_t           num_rows,
    uint32_t           num_cols
);

/*@}*/




/**
 * @name copy_matrix_into_matblock
 *
 * Copies a matrix into a matblock.
 * @{
 */

/** @brief Copies a 32-bit unsigned integer matrix into a matrix block. */
Error* copy_matrix_into_matblock_u32
(
    Matblock_u32**      mb_out,
    const Matblock_u32* mb_in,
    const Matrix_u32*   m,
    Matblock_matrix     orient,
    uint32_t            index
);

/** @brief Copies a 32-bit signed integer matrix into a matrix block. */
Error* copy_matrix_into_matblock_i32
(
    Matblock_i32**      mb_out,
    const Matblock_i32* mb_in,
    const Matrix_i32*   m,
    Matblock_matrix     orient,
    uint32_t            index
);

/** @brief Copies a 32-bit signed integer matrix into a matrix block. */
Error* copy_matrix_into_matblock_i64
(
    Matblock_i64**      mb_out,
    const Matblock_i64* mb_in,
    const Matrix_i64*   m,
    Matblock_matrix     orient,
    uint32_t            index
);

/** @brief Copies a 32-bit signed integer matrix into a matrix block. */
Error* copy_matrix_into_matblock_f
(
    Matblock_f**      mb_out,
    const Matblock_f* mb_in,
    const Matrix_f*   m,
    Matblock_matrix   orient,
    uint32_t          index
);

/** @brief Copies a 32-bit signed integer matrix into a matrix block. */
Error* copy_matrix_into_matblock_d
(
    Matblock_d**      mb_out,
    const Matblock_d* mb_in,
    const Matrix_d*   m,
    Matblock_matrix   orient,
    uint32_t          index
);

/** @brief Copies a 32-bit signed integer matrix into a matrix block. */
Error* copy_matrix_into_matblock_cf
(
    Matblock_cf**      mb_out,
    const Matblock_cf* mb_in,
    const Matrix_cf*   m,
    Matblock_matrix    orient,
    uint32_t           index
);

/** @brief Copies a 32-bit signed integer matrix into a matrix block. */
Error* copy_matrix_into_matblock_cd
(
    Matblock_cd**      mb_out,
    const Matblock_cd* mb_in,
    const Matrix_cd*   m,
    Matblock_matrix    orient,
    uint32_t           index
);

/*@}*/




/**
 * @name copy_matrix_from_matblock
 *
 * Copies a matrix from a matblock.
 * @{
 */

/** @brief Copies a 32-bit unsigned integer matrix from a matrix block. */
Error* copy_matrix_from_matblock_u32
(
    Matrix_u32**        m_out,
    const Matblock_u32* mb,
    Matblock_matrix     orient,
    uint32_t            index
);

/** @brief Copies a 32-bit signed integer matrix from a matrix block. */
Error* copy_matrix_from_matblock_i32
(
    Matrix_i32**        m_out,
    const Matblock_i32* mb,
    Matblock_matrix     orient,
    uint32_t            index
);

/** @brief Copies a 32-bit signed integer matrix from a matrix block. */
Error* copy_matrix_from_matblock_i64
(
    Matrix_i64**        m_out,
    const Matblock_i64* mb,
    Matblock_matrix     orient,
    uint32_t            index
);

/** @brief Copies a 32-bit signed integer matrix from a matrix block. */
Error* copy_matrix_from_matblock_f
(
    Matrix_f**        m_out,
    const Matblock_f* mb,
    Matblock_matrix   orient,
    uint32_t          index
);

/** @brief Copies a 32-bit signed integer matrix from a matrix block. */
Error* copy_matrix_from_matblock_d
(
    Matrix_d**        m_out,
    const Matblock_d* mb,
    Matblock_matrix   orient,
    uint32_t          index
);

/** @brief Copies a 32-bit signed integer matrix from a matrix block. */
Error* copy_matrix_from_matblock_cf
(
    Matrix_cf**        m_out,
    const Matblock_cf* mb,
    Matblock_matrix    orient,
    uint32_t           index
);

/** @brief Copies a 32-bit signed integer matrix from a matrix block. */
Error* copy_matrix_from_matblock_cd
(
    Matrix_cd**        m_out,
    const Matblock_cd* mb,
    Matblock_matrix    orient,
    uint32_t           index
);

/*@}*/




/**
 * @name free_matblock
 *
 * Frees a matrix block.
 * @{
 */

/** @brief Frees a 8-bit unsigned integer matrix block. */
void free_matblock_u8(Matblock_u8* m);

/** @brief Frees a 32-bit unsigned integer matrix block. */
void free_matblock_u32(Matblock_u32* m);

/** @brief Frees a 32-bit signed integer matrix block. */
void free_matblock_i32(Matblock_i32* m);

/** @brief Frees a 64-bit signed integer matrix block. */
void free_matblock_i64(Matblock_i64* m);

/** @brief Frees a single precision floating point matrix block. */
void free_matblock_f(Matblock_f* m);

/** @brief Frees a double precision floating point matrix block. */
void free_matblock_d(Matblock_d* m);

/** @brief Frees a single precision complex floating point matrix block. */
void free_matblock_cf(Matblock_cf* m);

/** @brief Frees a double precision complex floating point matrix block. */
void free_matblock_cd(Matblock_cd* m);

/*@}*/


#ifdef __cplusplus
}
}
#endif


#endif
