/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Declarations for matrix block FFT functions.
 *
 * If FFTW3 was built to use SIMD instrinctions, the real to complex and 
 * complex to real dft routines are accelerated. For the double precision
 * routines, however, only SSE2 ISA extensions are supported.
 */


#include <jwsc/config.h>

#include <stdlib.h>
#include <inttypes.h>
#include <assert.h>
#include <math.h>
#ifdef JWSC_HAVE_FFTW3_H
#include <fftw3.h>
#endif

#include "jwsc/matblock/matblock.h"
#include "jwsc/matblock/matblock_math.h"
#include "jwsc/matblock/matblock_fft.h"


/**
 * @name real_to_complex_matblock_dft
 *
 * Computes a complex discrete Fourier transform from a real matrix block.
 * @{
 */

/**
 * Takes advantage of the symmetry propery of the dft on real data: the number
 * of columns in @em m_out is only half of @em m_in. More precisely, 
 *
 * @code (*m_out)->num_cols == (m_in->num_cols / 2) + 1 @endcode
 *
 * Where integer division rounds down.
 *
 * @param  m_out  Result parameter. If @em *m_out is NULL, a matrix block is 
 *                allocated; otherwise its space is re-used.
 * @param  m_in   Real matrix block to convert.
 *
 * @note Use expand_symmetric_matblock_dft_d() to expand the symmetric columns 
 * in @em *m_out to the same dimensions as @em m_in.
 */
#ifdef JWSC_HAVE_FFTW3F
void real_to_complex_matblock_dft_f
(
    Matblock_cf**     m_out,
    const Matblock_f* m_in
)
{
    Matblock_f* s = NULL;

    real_to_complex_matblock_dft_scratch_f(m_out, m_in, &s);

    free_matblock_f(s);
}
#endif

/**
 * Takes advantage of the symmetry propery of the dft on real data: the number
 * of columns in @em m_out is only half of @em m_in. More precisely, 
 *
 * @code (*m_out)->num_cols == (m_in->num_cols / 2) + 1 @endcode
 *
 * Where integer division rounds down.
 *
 * @param  m_out  Result parameter. If @em *m_out is NULL, a matrix block is 
 *                allocated; otherwise its space is re-used.
 * @param  m_in   Real matrix block to convert.
 *
 * @note Use expand_symmetric_matblock_dft_d() to expand the symmetric columns 
 * in @em *m_out to the same dimensions as @em m_in.
 */
#ifdef JWSC_HAVE_FFTW3
void real_to_complex_matblock_dft_d
(
    Matblock_cd**     m_out,
    const Matblock_d* m_in
)
{
    Matblock_d* s = NULL;

    real_to_complex_matblock_dft_scratch_d(m_out, m_in, &s);

    free_matblock_d(s);
}
#endif

/*@}*/




/**
 * @name real_to_complex_matblock_dft_scratch
 *
 * Computes a complex discrete Fourier transform from a real matrix block
 * using a scratch buffer.
 * @{
 */

/**
 * Takes advantage of the symmetry propery of the dft on real data: the number
 * of columns in @em m_out is only half of @em m_in. More precisely, 
 *
 * @code (*m_out)->num_cols == (m_in->num_cols / 2) + 1 @endcode
 *
 * Where integer division rounds down.
 *
 * @param  m_out  Result parameter. If @em *m_out is NULL, a matrix block is 
 *                allocated; otherwise its space is re-used.
 * @param  m_in   Real matrix block to convert.
 * @param  s      Scratch Matblock_f.
 *
 * @note Use expand_symmetric_matblock_dft_d() to expand the symmetric columns 
 * in @em *m_out to the same dimensions as @em m_in.
 */
#ifdef JWSC_HAVE_FFTW3F
void real_to_complex_matblock_dft_scratch_f
(
    Matblock_cf**     m_out,
    const Matblock_f* m_in,
    Matblock_f**      s
)
{
    uint32_t    num_mats, num_rows, num_cols;
    fftwf_plan  plan;

    num_mats = m_in->num_mats;
    num_rows = m_in->num_rows;
    num_cols = m_in->num_cols;

    create_matblock_cf(m_out, num_mats, num_rows, (num_cols/2)+1);

    /* For fftw planning, input array does not need to be initialized, hence we
     * allow fftw to destroy the input to (possibly) get better efficiency. */
    create_matblock_f(s, num_mats, num_rows, num_cols);

    assert((plan = fftwf_plan_dft_r2c_3d(num_mats, num_rows, num_cols,
                    **((*s)->elts), (fftwf_complex*)**((*m_out)->elts), 
                    FFTW_MEASURE | FFTW_DESTROY_INPUT)) != NULL);

    /* Copy the input, now that planning is over, and execute. */
    copy_matblock_f(s, m_in);
    fftwf_execute(plan);

    fftwf_destroy_plan(plan);
}
#endif

/**
 * Takes advantage of the symmetry propery of the dft on real data: the number
 * of columns in @em m_out is only half of @em m_in. More precisely, 
 *
 * @code (*m_out)->num_cols == (m_in->num_cols / 2) + 1 @endcode
 *
 * Where integer division rounds down.
 *
 * @param  m_out  Result parameter. If @em *m_out is NULL, a matrix block is 
 *                allocated; otherwise its space is re-used.
 * @param  m_in   Real matrix block to convert.
 * @param  s      Scratch Matblock_d.
 *
 * @note Use expand_symmetric_matblock_dft_d() to expand the symmetric columns 
 * in @em *m_out to the same dimensions as @em m_in.
 */
#ifdef JWSC_HAVE_FFTW3
void real_to_complex_matblock_dft_scratch_d
(
    Matblock_cd**     m_out,
    const Matblock_d* m_in,
    Matblock_d**      s
)
{
    uint32_t    num_mats, num_rows, num_cols;
    fftw_plan   plan;

    num_mats = m_in->num_mats;
    num_rows = m_in->num_rows;
    num_cols = m_in->num_cols;

    create_matblock_cd(m_out, num_mats, num_rows, (num_cols/2)+1);

    /* For fftw planning, input array does not need to be initialized, hence we
     * allow fftw to destroy the input to (possibly) get better efficiency. */
    create_matblock_d(s, num_mats, num_rows, num_cols);

    assert((plan = fftw_plan_dft_r2c_3d(num_mats, num_rows, num_cols,
                    **((*s)->elts), (fftw_complex*)**((*m_out)->elts), 
                    FFTW_MEASURE | FFTW_DESTROY_INPUT)) != NULL);

    /* Copy the input, now that planning is over, and execute. */
    copy_matblock_d(s, m_in);
    fftw_execute(plan);

    fftw_destroy_plan(plan);
}
#endif

/*@}*/




/**
 * @name complex_to_real_matblock_dft
 *
 * Computes a real matrix block from a complex discrete Fourier transform.
 * @{
 */

/** 
 * Takes advantage of the symmetry propery of the dft on real data: the number
 * of columns in @em m_in is only half of @em m_out. More precisely,
 *
 * @code 
 * if (ncols_even)
 *     (*m_out)->num_cols == (m_in->num_cols-1) * 2
 * else
 *     (*m_out)->num_cols == (m_in->num_cols-1) * 2 + 1
 * @endcode
 *
 * See real_to_complex_dft_f() for information on how @em m_in came to be this
 * way.
 *
 * @param  m_out       Result parameter. If @em *m_out is NULL, a matrix block
 *                     is allocated; otherwise its space is re-used.
 * @param  m_in        Complex matrix block to convert.
 * @param  ncols_even  Boolean parameter: 0 if the number of columns in
 *                     the expanded symmetric @em m_in is odd; otherwise even.
 *
 * @note If @em m_in was already expanded to its symmetry, use 
 * shrink_symmetric_matblock_dft_f() to shrink it back before calling this
 * function.
 *
 * @note The FFT operations are unnormalized. If doing a real to complex to 
 * real dft operation, use scale_matblock_dft_f() to undo the effects of this.
 */ 
#ifdef JWSC_HAVE_FFTW3F
void complex_to_real_matblock_dft_f
(
    Matblock_f**       m_out,
    const Matblock_cf* m_in,
    uint8_t            ncols_even
)
{
    Matblock_cf* s = NULL;

    complex_to_real_matblock_dft_scratch_f(m_out, m_in, ncols_even, &s);

    free_matblock_cf(s);
}
#endif

/** 
 * Takes advantage of the symmetry propery of the dft on real data: the number
 * of columns in @em m_in is only half of @em m_out. More precisely,
 *
 * @code 
 * if (ncols_even)
 *     (*m_out)->num_cols == (m_in->num_cols-1) * 2
 * else
 *     (*m_out)->num_cols == (m_in->num_cols-1) * 2 + 1
 * @endcode
 *
 * See real_to_complex_dft_f() for information on how @em m_in came to be this
 * way.
 *
 * @param  m_out       Result parameter. If @em *m_out is NULL, a matrix block 
 *                     is allocated; otherwise its space is re-used.
 * @param  m_in        Complex matrix block to convert.
 * @param  ncols_even  Boolean parameter: 0 if the number of columns in
 *                     the expanded symmetric @em m_in is odd; otherwise even.
 *
 * @note If @em m_in was already expanded to its symmetry, use 
 * shrink_symmetric_matblock_dft_d() to shrink it back before calling this
 * function.
 *
 * @note The FFT operations are unnormalized. If doing a real to complex to 
 * real dft operation, use scale_matblock_dft_d() to undo the effects of this.
 */ 
#ifdef JWSC_HAVE_FFTW3
void complex_to_real_matblock_dft_d
(
    Matblock_d**       m_out,
    const Matblock_cd* m_in,
    uint8_t            ncols_even
)
{
    Matblock_cd* s = NULL;;

    complex_to_real_matblock_dft_scratch_d(m_out, m_in, ncols_even, &s);

    free_matblock_cd(s);
}
#endif

/*@}*/




/**
 * @name complex_to_real_matblock_dft
 *
 * Computes a real matrix block from a complex discrete Fourier transform
 * using a scratch buffer.
 * @{
 */

/** 
 * Takes advantage of the symmetry propery of the dft on real data: the number
 * of columns in @em m_in is only half of @em m_out. More precisely,
 *
 * @code 
 * if (ncols_even)
 *     (*m_out)->num_cols == (m_in->num_cols-1) * 2
 * else
 *     (*m_out)->num_cols == (m_in->num_cols-1) * 2 + 1
 * @endcode
 *
 * See real_to_complex_dft_f() for information on how @em m_in came to be this
 * way.
 *
 * @param  m_out       Result parameter. If @em *m_out is NULL, a matrix block
 *                     is allocated; otherwise its space is re-used.
 * @param  m_in        Complex matrix block to convert.
 * @param  ncols_even  Boolean parameter: 0 if the number of columns in
 *                     the expanded symmetric @em m_in is odd; otherwise even.
 * @param  s           Scratch Matblock_cf.
 *
 * @note If @em m_in was already expanded to its symmetry, use 
 * shrink_symmetric_matblock_dft_f() to shrink it back before calling this
 * function.
 *
 * @note The FFT operations are unnormalized. If doing a real to complex to 
 * real dft operation, use scale_matblock_dft_f() to undo the effects of this.
 */ 
#ifdef JWSC_HAVE_FFTW3F
void complex_to_real_matblock_dft_scratch_f
(
    Matblock_f**       m_out,
    const Matblock_cf* m_in,
    uint8_t            ncols_even,
    Matblock_cf**      s
)
{
    uint32_t     num_mats, num_rows, num_cols;
    fftwf_plan   plan;

    num_mats = m_in->num_mats;
    num_rows = m_in->num_rows;
    num_cols = (ncols_even) ? (m_in->num_cols-1)*2 : (m_in->num_cols-1)*2 + 1;

    create_matblock_f(m_out, num_mats, num_rows, num_cols);

    /* For fftw planning, input array does not need to be initialized, hence we
     * allow fftw to destroy the input to (possibly) get better efficiency. */
    create_matblock_cf(s, m_in->num_mats, m_in->num_rows, m_in->num_cols);

    assert((plan = fftwf_plan_dft_c2r_3d(num_mats, num_rows, num_cols,
                    (fftwf_complex*)**((*s)->elts), **((*m_out)->elts), 
                    FFTW_MEASURE | FFTW_DESTROY_INPUT)) != NULL);

    /* Copy the input, now that planning is over, and execute. */
    copy_matblock_cf(s, m_in);
    fftwf_execute(plan);

    fftwf_destroy_plan(plan);
}
#endif

/** 
 * Takes advantage of the symmetry propery of the dft on real data: the number
 * of columns in @em m_in is only half of @em m_out. More precisely,
 *
 * @code 
 * if (ncols_even)
 *     (*m_out)->num_cols == (m_in->num_cols-1) * 2
 * else
 *     (*m_out)->num_cols == (m_in->num_cols-1) * 2 + 1
 * @endcode
 *
 * See real_to_complex_dft_f() for information on how @em m_in came to be this
 * way.
 *
 * @param  m_out       Result parameter. If @em *m_out is NULL, a matrix block 
 *                     is allocated; otherwise its space is re-used.
 * @param  m_in        Complex matrix block to convert.
 * @param  ncols_even  Boolean parameter: 0 if the number of columns in
 *                     the expanded symmetric @em m_in is odd; otherwise even.
 * @param  s           Scratch Matblock_cd.
 *
 * @note If @em m_in was already expanded to its symmetry, use 
 * shrink_symmetric_matblock_dft_d() to shrink it back before calling this
 * function.
 *
 * @note The FFT operations are unnormalized. If doing a real to complex to 
 * real dft operation, use scale_matblock_dft_d() to undo the effects of this.
 */ 
#ifdef JWSC_HAVE_FFTW3
void complex_to_real_matblock_dft_scratch_d
(
    Matblock_d**       m_out,
    const Matblock_cd* m_in,
    uint8_t            ncols_even,
    Matblock_cd**      s
)
{
    uint32_t     num_mats, num_rows, num_cols;
    fftw_plan    plan;

    num_mats = m_in->num_mats;
    num_rows = m_in->num_rows;
    num_cols = (ncols_even) ? (m_in->num_cols-1)*2 : (m_in->num_cols-1)*2 + 1;

    create_matblock_d(m_out, num_mats, num_rows, num_cols);

    /* For fftw planning, input array does not need to be initialized, hence we
     * allow fftw to destroy the input to (possibly) get better efficiency. */
    create_matblock_cd(s, m_in->num_mats, m_in->num_rows, m_in->num_cols);

    assert((plan = fftw_plan_dft_c2r_3d(num_mats, num_rows, num_cols,
                    (fftw_complex*)**((*s)->elts), **((*m_out)->elts), 
                    FFTW_MEASURE | FFTW_DESTROY_INPUT)) != NULL);

    /* Copy the input, now that planning is over, and execute. */
    copy_matblock_cd(s, m_in);
    fftw_execute(plan);

    fftw_destroy_plan(plan);
}
#endif

/*@}*/




/**
 * @name expand_symmetric_matblock_dft
 *
 * Expands the symmetric columns of a dft.
 * @{
 */

/**
 * @param  m_out       Result parameter. If @em *m_out is NULL, a matrix block
 *                     is allocated; otherwise its space is re-used.
 * @param  m_in        Complex matrix block to expand.
 * @param  ncols_even  Boolean parameter: 0 if the number of columns to
 *                     expand @em m_in to is odd; otherwise even.
 *
 * @note If @em *m_out == @em m_in, then @em m_in is overwritten.
 */
void expand_symmetric_matblock_dft_f
(
    Matblock_cf**      m_out, 
    const Matblock_cf* m_in, 
    uint8_t            ncols_even
)
{
    uint32_t     num_mats, num_rows, num_cols;
    uint32_t     mat, row, col;
    Matblock_cf* m;

    num_mats = m_in->num_mats;
    num_rows = m_in->num_rows;
    num_cols = m_in->num_cols;

    m = (*m_out == m_in) ? NULL : *m_out;
    if (ncols_even)
    {
        create_matblock_cf(&m, num_mats, num_rows, (m_in->num_cols-1)*2);
    }
    else
    {
        create_matblock_cf(&m, num_mats, num_rows, (m_in->num_cols-1)*2 + 1);
    }

    assert(copy_matblock_block_into_matblock_cf(m, 0, 0, 0, m_in, 0, 0, 0, 
                m_in->num_mats, m_in->num_rows, m_in->num_cols) == NULL);

    for (mat = 0; mat < num_mats; mat++)
    {
        for (row = 0; row < num_rows; row++)
        {
            for (col = m->num_cols-1; col >= m_in->num_cols; col--)
            {
                m->elts[mat][row][col] = 
                    m_in->elts[mat][row][m->num_cols-col-1];
            }
        }
    }

    if (*m_out == m_in)
    {
        copy_matblock_cf(m_out, m);
        free_matblock_cf(m);
    }
    else
    {
        *m_out = m;
    }
}

/**
 * @param  m_out       Result parameter. If @em *m_out is NULL, a matrix block
 *                     is allocated; otherwise its space is re-used.
 * @param  m_in        Complex matrix block to expand.
 * @param  ncols_even  Boolean parameter: 0 if the number of columns to
 *                     expand @em m_in to is odd; otherwise even.
 *
 * @note If @em *m_out == @em m_in, then @em m_in is overwritten.
 */
void expand_symmetric_matblock_dft_d
(
    Matblock_cd**      m_out, 
    const Matblock_cd* m_in, 
    uint8_t            ncols_even
)
{
    uint32_t     num_mats, num_rows, num_cols;
    uint32_t     mat, row, col;
    Matblock_cd* m;

    num_mats = m_in->num_mats;
    num_rows = m_in->num_rows;
    num_cols = m_in->num_cols;

    m = (*m_out == m_in) ? NULL : *m_out;
    if (ncols_even)
    {
        create_matblock_cd(&m, num_mats, num_rows, (m_in->num_cols-1)*2);
    }
    else
    {
        create_matblock_cd(&m, num_mats, num_rows, (m_in->num_cols-1)*2 + 1);
    }

    assert(copy_matblock_block_into_matblock_cd(m, 0, 0, 0, m_in, 0, 0, 0, 
                m_in->num_mats, m_in->num_rows, m_in->num_cols) == NULL);

    for (mat = 0; mat < num_mats; mat++)
    {
        for (row = 0; row < num_rows; row++)
        {
            for (col = m->num_cols-1; col >= m_in->num_cols; col--)
            {
                m->elts[mat][row][col] = 
                    m_in->elts[mat][row][m->num_cols-col-1];
            }
        }
    }

    if (*m_out == m_in)
    {
        copy_matblock_cd(m_out, m);
        free_matblock_cd(m);
    }
    else
    {
        *m_out = m;
    }
}

/*@}*/




/**
 * @name shrink_symmetric_matblock_dft
 *
 * Shrinks the symmetric columns of a dft.
 * @{
 */

/**
 * @param  m_out  Result parameter. If @em *m_out is NULL, a matrix block is 
 *                allocated; otherwise its space is re-used.
 * @param  m_in   Complex matrix block to shrink.
 *
 * @note If @em *m_out == @em m_in, then @em m_in is overwritten.
 */
void shrink_symmetric_matblock_dft_f
(
    Matblock_cf**      m_out, 
    const Matblock_cf* m_in
)
{
    Matblock_cf* m;

    m = (*m_out == m_in) ? NULL : *m_out;
    create_matblock_cf(&m, m_in->num_mats, m_in->num_rows, 
            (m_in->num_cols/2)+1);

    assert(copy_matblock_block_into_matblock_cf(m, 0, 0, 0, m_in, 0, 0, 0, 
                m->num_mats, m->num_rows, m->num_cols) == NULL);

    if (*m_out == m_in)
    {
        copy_matblock_cf(m_out, m);
        free_matblock_cf(m);
    }
    else
    {
        *m_out = m;
    }
}

/**
 * @param  m_out  Result parameter. If @em *m_out is NULL, a matrix block is 
 *                allocated; otherwise its space is re-used.
 * @param  m_in   Complex matrix block to expand.
 *
 * @note If @em *m_out == @em m_in, then @em m_in is overwritten.
 */
void shrink_symmetric_matblock_dft_d
(
    Matblock_cd**      m_out, 
    const Matblock_cd* m_in
)
{
    Matblock_cd* m;

    m = (*m_out == m_in) ? NULL : *m_out;
    create_matblock_cd(&m, m_in->num_mats, m_in->num_rows, 
            (m_in->num_cols/2)+1);

    assert(copy_matblock_block_into_matblock_cd(m, 0, 0, 0, m_in, 0, 0, 0, 
                m->num_mats, m->num_rows, m->num_cols) == NULL);

    if (*m_out == m_in)
    {
        copy_matblock_cd(m_out, m);
        free_matblock_cd(m);
    }
    else
    {
        *m_out = m;
    }
}

/*@}*/




/**
 * @name scale_matblock_dft
 *
 * Scales the result of a real to complex to real dft operation.
 * @{
 */

/**
 * The FFT calculations are unscaled, so doing a real to complex to real 
 * operation results in an unscaled real matrix block. This function applies 
 * the proper scaling to undo this effect.
 *
 * @param  m_out  Result parameter. If @em *m_out is NULL, a matrix block is 
 *                allocated; otherwise its space is re-used.
 * @param  m_in   Matrix block to scale.
 *
 * @note If @em *m_out == @em m_in, then @em m_in is overwritten.
 */
void scale_matblock_dft_f(Matblock_f** m_out, const Matblock_f* m_in)
{
    float scale;

    scale = 1.0f / (float)(m_in->num_mats*m_in->num_rows*m_in->num_cols);
    multiply_matblock_by_scalar_f(m_out, m_in, scale);
}

/**
 * The FFT calculations are unscaled, so doing a real to complex to real 
 * operation results in an unscaled real matrix block. This function applies 
 * the proper scaling to undo this effect.
 *
 * @param  m_out  Result parameter. If @em *m_out is NULL, a matrix block is 
 *                allocated; otherwise its space is re-used.
 * @param  m_in   Matrix block to scale.
 *
 * @note If @em *m_out == @em m_in, then @em m_in is overwritten.
 */
void scale_matblock_dft_d(Matblock_d** m_out, const Matblock_d* m_in)
{
    float scale;

    scale = 1.0 / (double)(m_in->num_mats*m_in->num_rows*m_in->num_cols);
    multiply_matblock_by_scalar_d(m_out, m_in, scale);
}

/*@}*/




/**
 * @name shift_matblock_f
 *
 * Shifts a matrix block so its diagonal quadrants are swapped.
 * @{
 */

/**
 * @param  m_out  Result parameter. If @em *m_out is NULL, a matrix block
 *                is allocated; otherwise its space is re-used.
 * @param  m_in   Matrix block to shift.
 *
 * @note If @em *m_out == @em m_in, then @em m_in is overwritten.
 *
 * @note Use unshift_matblock_f() to undo the shift.
 */
void shift_matblock_f
(
    Matblock_f**       m_out,
    const Matblock_f*  m_in
)
{
    uint32_t num_mats, num_rows, num_cols;
    uint32_t mat, row, col;
    uint32_t half_mats, half_rows, half_cols;

    Matblock_f* m = NULL;

    num_mats  = m_in->num_mats;
    num_rows  = m_in->num_rows;
    num_cols  = m_in->num_cols;
    half_mats = num_mats / 2;
    half_rows = num_rows / 2;
    half_cols = num_cols / 2;

    m = (*m_out == m_in) ? NULL : *m_out;
    create_matblock_f(&m, num_mats, num_rows, num_cols);

    for (mat = 0; mat < num_mats; mat++)
    {
        if (mat < half_mats + (num_mats % 2))
        {
            for (row = 0; row < num_rows; row++)
            {
                if (row < half_rows + (num_rows % 2))
                {
                    for (col = 0; col < num_cols; col++)
                    {
                        if (col < half_cols + (num_cols % 2))
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat + half_mats ]
                                [ row + half_rows ]
                                [ col + half_cols ];
                        }
                        else
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat + half_mats ]
                                [ row + half_rows ]
                                [ col - half_cols - (num_cols % 2) ];
                        }
                    }
                }
                else
                {
                    for (col = 0; col < num_cols; col++)
                    {
                        if (col < half_cols + (num_cols % 2))
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat + half_mats ]
                                [ row - half_rows - (num_rows % 2) ]
                                [ col + half_cols ];
                        }
                        else
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat + half_mats ]
                                [ row - half_rows - (num_rows % 2) ]
                                [ col - half_cols - (num_cols % 2) ];
                        }
                    }
                }
            }
        }
        else
        {
            for (row = 0; row < num_rows; row++)
            {
                if (row < half_rows + (num_rows % 2))
                {
                    for (col = 0; col < num_cols; col++)
                    {
                        if (col < half_cols + (num_cols % 2))
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat - half_mats - (num_mats % 2) ]
                                [ row + half_rows ]
                                [ col + half_cols ];
                        }
                        else
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat - half_mats - (num_mats % 2) ]
                                [ row + half_rows ]
                                [ col - half_cols - (num_cols % 2) ];
                        }
                    }
                }
                else
                {
                    for (col = 0; col < num_cols; col++)
                    {
                        if (col < half_cols + (num_cols % 2))
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat - half_mats - (num_mats % 2) ]
                                [ row - half_rows - (num_rows % 2) ]
                                [ col + half_cols ];
                        }
                        else
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat - half_mats - (num_mats % 2) ]
                                [ row - half_rows - (num_rows % 2) ]
                                [ col - half_cols - (num_cols % 2) ];
                        }
                    }
                }
            }
        }
    }

    if (*m_out == m_in)
    {
        copy_matblock_f(m_out, m);
        free_matblock_f(m);
    }
    else
    {
        *m_out = m;
    }
}

/**
 * @param  m_out  Result parameter. If @em *m_out is NULL, a matrix block
 *                is allocated; otherwise its space is re-used.
 * @param  m_in   Matrix block to shift.
 *
 * @note If @em *m_out == @em m_in, then @em m_in is overwritten.
 *
 * @note Use unshift_matblock_d() to undo the shift.
 */
void shift_matblock_d
(
    Matblock_d**       m_out,
    const Matblock_d*  m_in
)
{
    uint32_t num_mats, num_rows, num_cols;
    uint32_t mat, row, col;
    uint32_t half_mats, half_rows, half_cols;

    Matblock_d* m = NULL;

    num_mats  = m_in->num_mats;
    num_rows  = m_in->num_rows;
    num_cols  = m_in->num_cols;
    half_mats = num_mats / 2;
    half_rows = num_rows / 2;
    half_cols = num_cols / 2;

    m = (*m_out == m_in) ? NULL : *m_out;
    create_matblock_d(&m, num_mats, num_rows, num_cols);

    for (mat = 0; mat < num_mats; mat++)
    {
        if (mat < half_mats + (num_mats % 2))
        {
            for (row = 0; row < num_rows; row++)
            {
                if (row < half_rows + (num_rows % 2))
                {
                    for (col = 0; col < num_cols; col++)
                    {
                        if (col < half_cols + (num_cols % 2))
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat + half_mats ]
                                [ row + half_rows ]
                                [ col + half_cols ];
                        }
                        else
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat + half_mats ]
                                [ row + half_rows ]
                                [ col - half_cols - (num_cols % 2) ];
                        }
                    }
                }
                else
                {
                    for (col = 0; col < num_cols; col++)
                    {
                        if (col < half_cols + (num_cols % 2))
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat + half_mats ]
                                [ row - half_rows - (num_rows % 2) ]
                                [ col + half_cols ];
                        }
                        else
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat + half_mats ]
                                [ row - half_rows - (num_rows % 2) ]
                                [ col - half_cols - (num_cols % 2) ];
                        }
                    }
                }
            }
        }
        else
        {
            for (row = 0; row < num_rows; row++)
            {
                if (row < half_rows + (num_rows % 2))
                {
                    for (col = 0; col < num_cols; col++)
                    {
                        if (col < half_cols + (num_cols % 2))
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat - half_mats - (num_mats % 2) ]
                                [ row + half_rows ]
                                [ col + half_cols ];
                        }
                        else
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat - half_mats - (num_mats % 2) ]
                                [ row + half_rows ]
                                [ col - half_cols - (num_cols % 2) ];
                        }
                    }
                }
                else
                {
                    for (col = 0; col < num_cols; col++)
                    {
                        if (col < half_cols + (num_cols % 2))
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat - half_mats - (num_mats % 2) ]
                                [ row - half_rows - (num_rows % 2) ]
                                [ col + half_cols ];
                        }
                        else
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat - half_mats - (num_mats % 2) ]
                                [ row - half_rows - (num_rows % 2) ]
                                [ col - half_cols - (num_cols % 2) ];
                        }
                    }
                }
            }
        }
    }

    if (*m_out == m_in)
    {
        copy_matblock_d(m_out, m);
        free_matblock_d(m);
    }
    else
    {
        *m_out = m;
    }
}

/**
 * @param  m_out  Result parameter. If @em *m_out is NULL, a matrix block
 *                is allocated; otherwise its space is re-used.
 * @param  m_in   Matrix block to shift.
 *
 * @note If @em *m_out == @em m_in, then @em m_in is overwritten.
 *
 * @note Use unshift_matblock_cf() to undo the shift.
 */
void shift_matblock_cf
(
    Matblock_cf**       m_out,
    const Matblock_cf*  m_in
)
{
    uint32_t num_mats, num_rows, num_cols;
    uint32_t mat, row, col;
    uint32_t half_mats, half_rows, half_cols;

    Matblock_cf* m = NULL;

    num_mats  = m_in->num_mats;
    num_rows  = m_in->num_rows;
    num_cols  = m_in->num_cols;
    half_mats = num_mats / 2;
    half_rows = num_rows / 2;
    half_cols = num_cols / 2;

    m = (*m_out == m_in) ? NULL : *m_out;
    create_matblock_cf(&m, num_mats, num_rows, num_cols);

    for (mat = 0; mat < num_mats; mat++)
    {
        if (mat < half_mats + (num_mats % 2))
        {
            for (row = 0; row < num_rows; row++)
            {
                if (row < half_rows + (num_rows % 2))
                {
                    for (col = 0; col < num_cols; col++)
                    {
                        if (col < half_cols + (num_cols % 2))
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat + half_mats ]
                                [ row + half_rows ]
                                [ col + half_cols ];
                        }
                        else
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat + half_mats ]
                                [ row + half_rows ]
                                [ col - half_cols - (num_cols % 2) ];
                        }
                    }
                }
                else
                {
                    for (col = 0; col < num_cols; col++)
                    {
                        if (col < half_cols + (num_cols % 2))
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat + half_mats ]
                                [ row - half_rows - (num_rows % 2) ]
                                [ col + half_cols ];
                        }
                        else
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat + half_mats ]
                                [ row - half_rows - (num_rows % 2) ]
                                [ col - half_cols - (num_cols % 2) ];
                        }
                    }
                }
            }
        }
        else
        {
            for (row = 0; row < num_rows; row++)
            {
                if (row < half_rows + (num_rows % 2))
                {
                    for (col = 0; col < num_cols; col++)
                    {
                        if (col < half_cols + (num_cols % 2))
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat - half_mats - (num_mats % 2) ]
                                [ row + half_rows ]
                                [ col + half_cols ];
                        }
                        else
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat - half_mats - (num_mats % 2) ]
                                [ row + half_rows ]
                                [ col - half_cols - (num_cols % 2) ];
                        }
                    }
                }
                else
                {
                    for (col = 0; col < num_cols; col++)
                    {
                        if (col < half_cols + (num_cols % 2))
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat - half_mats - (num_mats % 2) ]
                                [ row - half_rows - (num_rows % 2) ]
                                [ col + half_cols ];
                        }
                        else
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat - half_mats - (num_mats % 2) ]
                                [ row - half_rows - (num_rows % 2) ]
                                [ col - half_cols - (num_cols % 2) ];
                        }
                    }
                }
            }
        }
    }

    if (*m_out == m_in)
    {
        copy_matblock_cf(m_out, m);
        free_matblock_cf(m);
    }
    else
    {
        *m_out = m;
    }
}

/**
 * @param  m_out  Result parameter. If @em *m_out is NULL, a matrix block
 *                is allocated; otherwise its space is re-used.
 * @param  m_in   Matrix block to shift.
 *
 * @note If @em *m_out == @em m_in, then @em m_in is overwritten.
 *
 * @note Use unshift_matblock_cd() to undo the shift.
 */
void shift_matblock_cd
(
    Matblock_cd**       m_out,
    const Matblock_cd*  m_in
)
{
    uint32_t num_mats, num_rows, num_cols;
    uint32_t mat, row, col;
    uint32_t half_mats, half_rows, half_cols;

    Matblock_cd* m = NULL;

    num_mats  = m_in->num_mats;
    num_rows  = m_in->num_rows;
    num_cols  = m_in->num_cols;
    half_mats = num_mats / 2;
    half_rows = num_rows / 2;
    half_cols = num_cols / 2;

    m = (*m_out == m_in) ? NULL : *m_out;
    create_matblock_cd(&m, num_mats, num_rows, num_cols);

    for (mat = 0; mat < num_mats; mat++)
    {
        if (mat < half_mats + (num_mats % 2))
        {
            for (row = 0; row < num_rows; row++)
            {
                if (row < half_rows + (num_rows % 2))
                {
                    for (col = 0; col < num_cols; col++)
                    {
                        if (col < half_cols + (num_cols % 2))
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat + half_mats ]
                                [ row + half_rows ]
                                [ col + half_cols ];
                        }
                        else
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat + half_mats ]
                                [ row + half_rows ]
                                [ col - half_cols - (num_cols % 2) ];
                        }
                    }
                }
                else
                {
                    for (col = 0; col < num_cols; col++)
                    {
                        if (col < half_cols + (num_cols % 2))
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat + half_mats ]
                                [ row - half_rows - (num_rows % 2) ]
                                [ col + half_cols ];
                        }
                        else
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat + half_mats ]
                                [ row - half_rows - (num_rows % 2) ]
                                [ col - half_cols - (num_cols % 2) ];
                        }
                    }
                }
            }
        }
        else
        {
            for (row = 0; row < num_rows; row++)
            {
                if (row < half_rows + (num_rows % 2))
                {
                    for (col = 0; col < num_cols; col++)
                    {
                        if (col < half_cols + (num_cols % 2))
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat - half_mats - (num_mats % 2) ]
                                [ row + half_rows ]
                                [ col + half_cols ];
                        }
                        else
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat - half_mats - (num_mats % 2) ]
                                [ row + half_rows ]
                                [ col - half_cols - (num_cols % 2) ];
                        }
                    }
                }
                else
                {
                    for (col = 0; col < num_cols; col++)
                    {
                        if (col < half_cols + (num_cols % 2))
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat - half_mats - (num_mats % 2) ]
                                [ row - half_rows - (num_rows % 2) ]
                                [ col + half_cols ];
                        }
                        else
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat - half_mats - (num_mats % 2) ]
                                [ row - half_rows - (num_rows % 2) ]
                                [ col - half_cols - (num_cols % 2) ];
                        }
                    }
                }
            }
        }
    }

    if (*m_out == m_in)
    {
        copy_matblock_cd(m_out, m);
        free_matblock_cd(m);
    }
    else
    {
        *m_out = m;
    }
}

/*@}*/




/**
 * @name unshift_matblock
 *
 * Unshifts a matrix block that had its diagonal quadrants swapped.
 * @{
 */

/**
 * @param  m_out  Result parameter. If @em *m_out is NULL, a matrix block
 *                is allocated; otherwise its space is re-used.
 * @param  m_in   Matrix block to unshift.
 *
 * @note If @em *m_out == @em m_in, then @em m_in is overwritten.
 *
 * @note Undoes the effects of shift_matblock_f().
 */
void unshift_matblock_f
(
    Matblock_f**       m_out,
    const Matblock_f*  m_in
)
{
    uint32_t num_mats, num_rows, num_cols;
    uint32_t mat, row, col;
    uint32_t half_mats, half_rows, half_cols;

    Matblock_f* m = NULL;

    num_mats  = m_in->num_mats;
    num_rows  = m_in->num_rows;
    num_cols  = m_in->num_cols;
    half_mats = num_mats / 2;
    half_rows = num_rows / 2;
    half_cols = num_cols / 2;

    m = (*m_out == m_in) ? NULL : *m_out;
    create_matblock_f(&m, num_mats, num_rows, num_cols);

    for (mat = 0; mat < num_mats; mat++)
    {
        if (mat < half_mats)
        {
            for (row = 0; row < num_rows; row++)
            {
                if (row < half_rows)
                {
                    for (col = 0; col < num_cols; col++)
                    {
                        if (col < half_cols)
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat + half_mats + (num_mats % 2) ]
                                [ row + half_rows + (num_rows % 2) ]
                                [ col + half_cols + (num_cols % 2) ];
                        }
                        else
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat + half_mats + (num_mats % 2) ]
                                [ row + half_rows + (num_rows % 2) ]
                                [ col - half_cols ];
                        }
                    }
                }
                else
                {
                    for (col = 0; col < num_cols; col++)
                    {
                        if (col < half_cols)
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat + half_mats + (num_mats % 2) ]
                                [ row - half_rows ]
                                [ col + half_cols + (num_cols % 2) ];
                        }
                        else
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat + half_mats + (num_mats % 2) ]
                                [ row - half_rows]
                                [ col - half_cols];
                        }
                    }
                }
            }
        }
        else
        {
            for (row = 0; row < num_rows; row++)
            {
                if (row < half_rows)
                {
                    for (col = 0; col < num_cols; col++)
                    {
                        if (col < half_cols)
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat - half_mats ]
                                [ row + half_rows + (num_rows % 2) ]
                                [ col + half_cols + (num_cols % 2) ];
                        }
                        else
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat - half_mats ]
                                [ row + half_rows + (num_rows % 2) ]
                                [ col - half_cols ];
                        }
                    }
                }
                else
                {
                    for (col = 0; col < num_cols; col++)
                    {
                        if (col < half_cols)
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat - half_mats ]
                                [ row - half_rows ]
                                [ col + half_cols + (num_cols % 2) ];
                        }
                        else
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat - half_mats ]
                                [ row - half_rows]
                                [ col - half_cols];
                        }
                    }
                }
            }
        }
    }

    if (*m_out == m_in)
    {
        copy_matblock_f(m_out, m);
        free_matblock_f(m);
    }
    else
    {
        *m_out = m;
    }
}

/**
 * @param  m_out  Result parameter. If @em *m_out is NULL, a matrix block
 *                is allocated; otherwise its space is re-used.
 * @param  m_in   Matrix block to unshift.
 *
 * @note If @em *m_out == @em m_in, then @em m_in is overwritten.
 *
 * @note Undoes the effects of shift_matblock_d().
 */
void unshift_matblock_d
(
    Matblock_d**       m_out,
    const Matblock_d*  m_in
)
{
    uint32_t num_mats, num_rows, num_cols;
    uint32_t mat, row, col;
    uint32_t half_mats, half_rows, half_cols;

    Matblock_d* m = NULL;

    num_mats  = m_in->num_mats;
    num_rows  = m_in->num_rows;
    num_cols  = m_in->num_cols;
    half_mats = num_mats / 2;
    half_rows = num_rows / 2;
    half_cols = num_cols / 2;

    m = (*m_out == m_in) ? NULL : *m_out;
    create_matblock_d(&m, num_mats, num_rows, num_cols);

    for (mat = 0; mat < num_mats; mat++)
    {
        if (mat < half_mats)
        {
            for (row = 0; row < num_rows; row++)
            {
                if (row < half_rows)
                {
                    for (col = 0; col < num_cols; col++)
                    {
                        if (col < half_cols)
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat + half_mats + (num_mats % 2) ]
                                [ row + half_rows + (num_rows % 2) ]
                                [ col + half_cols + (num_cols % 2) ];
                        }
                        else
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat + half_mats + (num_mats % 2) ]
                                [ row + half_rows + (num_rows % 2) ]
                                [ col - half_cols ];
                        }
                    }
                }
                else
                {
                    for (col = 0; col < num_cols; col++)
                    {
                        if (col < half_cols)
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat + half_mats + (num_mats % 2) ]
                                [ row - half_rows ]
                                [ col + half_cols + (num_cols % 2) ];
                        }
                        else
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat + half_mats + (num_mats % 2) ]
                                [ row - half_rows]
                                [ col - half_cols];
                        }
                    }
                }
            }
        }
        else
        {
            for (row = 0; row < num_rows; row++)
            {
                if (row < half_rows)
                {
                    for (col = 0; col < num_cols; col++)
                    {
                        if (col < half_cols)
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat - half_mats ]
                                [ row + half_rows + (num_rows % 2) ]
                                [ col + half_cols + (num_cols % 2) ];
                        }
                        else
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat - half_mats ]
                                [ row + half_rows + (num_rows % 2) ]
                                [ col - half_cols ];
                        }
                    }
                }
                else
                {
                    for (col = 0; col < num_cols; col++)
                    {
                        if (col < half_cols)
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat - half_mats ]
                                [ row - half_rows ]
                                [ col + half_cols + (num_cols % 2) ];
                        }
                        else
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat - half_mats ]
                                [ row - half_rows]
                                [ col - half_cols];
                        }
                    }
                }
            }
        }
    }

    if (*m_out == m_in)
    {
        copy_matblock_d(m_out, m);
        free_matblock_d(m);
    }
    else
    {
        *m_out = m;
    }
}

/**
 * @param  m_out  Result parameter. If @em *m_out is NULL, a matrix block
 *                is allocated; otherwise its space is re-used.
 * @param  m_in   Matrix block to unshift.
 *
 * @note If @em *m_out == @em m_in, then @em m_in is overwritten.
 *
 * @note Undoes the effects of shift_matblock_cf().
 */
void unshift_matblock_cf
(
    Matblock_cf**       m_out,
    const Matblock_cf*  m_in
)
{
    uint32_t num_mats, num_rows, num_cols;
    uint32_t mat, row, col;
    uint32_t half_mats, half_rows, half_cols;

    Matblock_cf* m = NULL;

    num_mats  = m_in->num_mats;
    num_rows  = m_in->num_rows;
    num_cols  = m_in->num_cols;
    half_mats = num_mats / 2;
    half_rows = num_rows / 2;
    half_cols = num_cols / 2;

    m = (*m_out == m_in) ? NULL : *m_out;
    create_matblock_cf(&m, num_mats, num_rows, num_cols);

    for (mat = 0; mat < num_mats; mat++)
    {
        if (mat < half_mats)
        {
            for (row = 0; row < num_rows; row++)
            {
                if (row < half_rows)
                {
                    for (col = 0; col < num_cols; col++)
                    {
                        if (col < half_cols)
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat + half_mats + (num_mats % 2) ]
                                [ row + half_rows + (num_rows % 2) ]
                                [ col + half_cols + (num_cols % 2) ];
                        }
                        else
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat + half_mats + (num_mats % 2) ]
                                [ row + half_rows + (num_rows % 2) ]
                                [ col - half_cols ];
                        }
                    }
                }
                else
                {
                    for (col = 0; col < num_cols; col++)
                    {
                        if (col < half_cols)
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat + half_mats + (num_mats % 2) ]
                                [ row - half_rows ]
                                [ col + half_cols + (num_cols % 2) ];
                        }
                        else
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat + half_mats + (num_mats % 2) ]
                                [ row - half_rows]
                                [ col - half_cols];
                        }
                    }
                }
            }
        }
        else
        {
            for (row = 0; row < num_rows; row++)
            {
                if (row < half_rows)
                {
                    for (col = 0; col < num_cols; col++)
                    {
                        if (col < half_cols)
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat - half_cols ]
                                [ row + half_rows + (num_rows % 2) ]
                                [ col + half_cols + (num_cols % 2) ];
                        }
                        else
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat - half_cols ]
                                [ row + half_rows + (num_rows % 2) ]
                                [ col - half_cols ];
                        }
                    }
                }
                else
                {
                    for (col = 0; col < num_cols; col++)
                    {
                        if (col < half_cols)
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat - half_cols ]
                                [ row - half_rows ]
                                [ col + half_cols + (num_cols % 2) ];
                        }
                        else
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat - half_cols ]
                                [ row - half_rows]
                                [ col - half_cols];
                        }
                    }
                }
            }
        }
    }

    if (*m_out == m_in)
    {
        copy_matblock_cf(m_out, m);
        free_matblock_cf(m);
    }
    else
    {
        *m_out = m;
    }
}

/**
 * @param  m_out  Result parameter. If @em *m_out is NULL, a matrix block
 *                is allocated; otherwise its space is re-used.
 * @param  m_in   Matrix block to unshift.
 *
 * @note If @em *m_out == @em m_in, then @em m_in is overwritten.
 *
 * @note Undoes the effects of shift_matblock_cd().
 */
void unshift_matblock_cd
(
    Matblock_cd**       m_out,
    const Matblock_cd*  m_in
)
{
    uint32_t num_mats, num_rows, num_cols;
    uint32_t mat, row, col;
    uint32_t half_mats, half_rows, half_cols;

    Matblock_cd* m = NULL;

    num_mats  = m_in->num_mats;
    num_rows  = m_in->num_rows;
    num_cols  = m_in->num_cols;
    half_mats = num_mats / 2;
    half_rows = num_rows / 2;
    half_cols = num_cols / 2;

    m = (*m_out == m_in) ? NULL : *m_out;
    create_matblock_cd(&m, num_mats, num_rows, num_cols);

    for (mat = 0; mat < num_mats; mat++)
    {
        if (mat < half_mats)
        {
            for (row = 0; row < num_rows; row++)
            {
                if (row < half_rows)
                {
                    for (col = 0; col < num_cols; col++)
                    {
                        if (col < half_cols)
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat + half_mats + (num_mats % 2) ]
                                [ row + half_rows + (num_rows % 2) ]
                                [ col + half_cols + (num_cols % 2) ];
                        }
                        else
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat + half_mats + (num_mats % 2) ]
                                [ row + half_rows + (num_rows % 2) ]
                                [ col - half_cols ];
                        }
                    }
                }
                else
                {
                    for (col = 0; col < num_cols; col++)
                    {
                        if (col < half_cols)
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat + half_mats + (num_mats % 2) ]
                                [ row - half_rows ]
                                [ col + half_cols + (num_cols % 2) ];
                        }
                        else
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat + half_mats + (num_mats % 2) ]
                                [ row - half_rows]
                                [ col - half_cols];
                        }
                    }
                }
            }
        }
        else
        {
            for (row = 0; row < num_rows; row++)
            {
                if (row < half_rows)
                {
                    for (col = 0; col < num_cols; col++)
                    {
                        if (col < half_cols)
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat - half_cols ]
                                [ row + half_rows + (num_rows % 2) ]
                                [ col + half_cols + (num_cols % 2) ];
                        }
                        else
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat - half_cols ]
                                [ row + half_rows + (num_rows % 2) ]
                                [ col - half_cols ];
                        }
                    }
                }
                else
                {
                    for (col = 0; col < num_cols; col++)
                    {
                        if (col < half_cols)
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat - half_cols ]
                                [ row - half_rows ]
                                [ col + half_cols + (num_cols % 2) ];
                        }
                        else
                        {
                            m->elts[ mat ][ row ][ col ] = m_in->elts
                                [ mat - half_cols ]
                                [ row - half_rows]
                                [ col - half_cols];
                        }
                    }
                }
            }
        }
    }

    if (*m_out == m_in)
    {
        copy_matblock_cd(m_out, m);
        free_matblock_cd(m);
    }
    else
    {
        *m_out = m;
    }
}

/*@}*/
