/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Declarations for I/O operations on a matblock.
 */


#ifndef MATBLOCK_IO_H 
#define MATBLOCK_IO_H 


#include <jwsc/config.h>

#include <stdlib.h>
#include <stdio.h>
#include <inttypes.h>

#include "jwsc/base/error.h"
#include "jwsc/math/complex.h"
#include "jwsc/matblock/matblock.h"


#ifdef __cplusplus
namespace jwsc {
extern "C" {
#endif


/**
 * @name read_matblock_with_header
 *
 * Reads a matblock with a header from a file.
 * @{
 */

/** @brief Reads a 32-bit unsigned integer matblock. */
Error* read_matblock_with_header_u32(Matblock_u32** m_out, const char* fname);

/** @brief Reads a 32-bit signed integer matblock. */
Error* read_matblock_with_header_i32(Matblock_i32** m_out, const char* fname);

/** @brief Reads a 64-bit signed integer matblock. */
Error* read_matblock_with_header_i64(Matblock_i64** m_out, const char* fname);

/** @brief Reads a single precision floating point matblock. */
Error* read_matblock_with_header_f(Matblock_f** m_out, const char* fname);

/** @brief Reads a double precision floating point matblock. */
Error* read_matblock_with_header_d(Matblock_d** m_out, const char* fname);

/** @brief Reads a single precision complex floating point matblock. */
Error* read_matblock_with_header_cf(Matblock_cf** m_out, const char* fname);

/** @brief Reads a double precision complex floating point matblock. */
Error* read_matblock_with_header_cd(Matblock_cd** m_out, const char* fname);

/*@}*/




/**
 * @name read_matblock_with_header_fp
 *
 * Reads a matblock with a header from an opened file.
 * @{
 */

/** @brief Reads a 32-bit unsigned integer matblock. */
Error* read_matblock_with_header_fp_u32(Matblock_u32** m_out, FILE* fp);

/** @brief Reads a 32-bit signed integer matblock. */
Error* read_matblock_with_header_fp_i32(Matblock_i32** m_out, FILE* fp);

/** @brief Reads a 64-bit signed integer matblock. */
Error* read_matblock_with_header_fp_i64(Matblock_i64** m_out, FILE* fp);

/** @brief Reads a single precision floating point matblock. */
Error* read_matblock_with_header_fp_f(Matblock_f** m_out, FILE* fp);

/** @brief Reads a double precision floating point matblock. */
Error* read_matblock_with_header_fp_d(Matblock_d** m_out, FILE* fp);

/** @brief Reads a single precision complex floating point matblock. */
Error* read_matblock_with_header_fp_cf(Matblock_cf** m_out, FILE* fp);

/** @brief Reads a double precision complex floating point matblock. */
Error* read_matblock_with_header_fp_cd(Matblock_cd** m_out, FILE* fp);

/*@}*/




/**
 * @name write_matblock_with_header
 *
 * Writes a matblock with a header to a file.
 * @{
 */

/** @brief Writes a 32-bit unsigned integer matblock. */
Error* write_matblock_with_header_u32(const Matblock_u32* m, const char* fname);

/** @brief Writes a 32-bit signed integer matblock. */
Error* write_matblock_with_header_i32(const Matblock_i32* m, const char* fname);

/** @brief Writes a 64-bit signed integer matblock. */
Error* write_matblock_with_header_i64(const Matblock_i64* m, const char* fname);

/** @brief Writes a single precision floating point matblock. */
Error* write_matblock_with_header_f(const Matblock_f* m, const char* fname);

/** @brief Writes a double precision floating point matblock. */
Error* write_matblock_with_header_d(const Matblock_d* m, const char* fname);

/** @brief Writes a single precision complex floating point matblock. */
Error* write_matblock_with_header_cf(const Matblock_cf* m, const char* fname);

/** @brief Writes a double precision complex floating point matblock. */
Error* write_matblock_with_header_cd(const Matblock_cd* m, const char* fname);

/*@}*/


/**
 * @name write_matblock_with_header_fp
 *
 * Writes a matblock with a header to an opened file.
 * @{
 */

/** @brief Writes a 32-bit unsigned integer matblock. */
Error* write_matblock_with_header_fp_u32(const Matblock_u32* m, FILE* fp);

/** @brief Writes a 32-bit signed integer matblock. */
Error* write_matblock_with_header_fp_i32(const Matblock_i32* m, FILE* fp);

/** @brief Writes a 64-bit signed integer matblock. */
Error* write_matblock_with_header_fp_i64(const Matblock_i64* m, FILE* fp);

/** @brief Writes a single precision floating point matblock. */
Error* write_matblock_with_header_fp_f(const Matblock_f* m, FILE* fp);

/** @brief Writes a double precision floating point matblock. */
Error* write_matblock_with_header_fp_d(const Matblock_d* m, FILE* fp);

/** @brief Writes a single precision complex floating point matblock. */
Error* write_matblock_with_header_fp_cf(const Matblock_cf* m, FILE* fp);

/** @brief Writes a double precision complex floating point matblock. */
Error* write_matblock_with_header_fp_cd(const Matblock_cd* m, FILE* fp);

/*@}*/


#ifdef __cplusplus
}
}
#endif


#endif
