/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Declarations for mathematical matrix block operations.
 *
 * Many of the functions are accelerated if linked against an optimized BLAS
 * library.
 */


#ifndef MATBLOCK_MATH_H 
#define MATBLOCK_MATH_H 


#include <jwsc/config.h>

#include <stdlib.h>
#include <inttypes.h>

#include "jwsc/base/error.h"
#include "jwsc/math/complex.h"
#include "jwsc/matblock/matblock.h"


#ifdef __cplusplus
namespace jwsc {
extern "C" {
#endif


/**
 * @name add_scalar_to_matblock
 *
 * Adds a scalar to a matrix block.
 * @{
 */

/** @brief Adds a scalar to a single precision matrix block. */
void add_scalar_to_matblock_f
(
    Matblock_f**      m_out, 
    const Matblock_f* m_in, 
    float             a
);

/** @brief Adds a scalar to a double precision matrix block. */
void add_scalar_to_matblock_d
(
    Matblock_d**      m_out, 
    const Matblock_d* m_in, 
    double            a
);

/** @brief Adds a scalar to a single precision complex matrix block. */
void add_scalar_to_matblock_cf
(
    Matblock_cf**      m_out, 
    const Matblock_cf* m_in, 
    Complex_f          z
);

/** @brief Adds a scalar to a double precision complex matrix block. */
void add_scalar_to_matblock_cd
(
    Matblock_cd**      m_out, 
    const Matblock_cd* m_in, 
    Complex_d          z
);

/*@}*/




/** 
 * @name multiply_matblock_by_scalar
 *
 * Multiplies a matrix block by a scalar.
 * @{
 */

/** @brief Multiplies a single precision matrix block by a scalar. */
void multiply_matblock_by_scalar_f
(
    Matblock_f**      m_out, 
    const Matblock_f* m_in, 
    float             a
);

/** @brief Multiplies a double precision matrix block by a scalar. */
void multiply_matblock_by_scalar_d
(
    Matblock_d**      m_out, 
    const Matblock_d* m_in, 
    double            a
);

/** @brief Multiplies a single precision complex matrix block by a scalar. */
void multiply_matblock_by_scalar_cf
(
    Matblock_cf**      m_out, 
    const Matblock_cf* m_in, 
    Complex_f          z
);

/** @brief Multiplies a double precision complex matrix block by a scalar. */
void multiply_matblock_by_scalar_cd
(
    Matblock_cd**      m_out, 
    const Matblock_cd* m_in, 
    Complex_d          z
);

/*@}*/




/** 
 * @name multiply_matblocks_elt_wise
 *
 * Multiplies a two matrix blocks element by element.
 * @{
 */

/** @brief Multiples two single precision matrix blocks element by element. */
Error* multiply_matblocks_elt_wise_f
(
    Matblock_f**      m_out, 
    const Matblock_f* m_1, 
    const Matblock_f* m_2
);

/** @brief Multiples two double precision matrix blocks element by element. */
Error* multiply_matblocks_elt_wise_d
(
    Matblock_d**      m_out, 
    const Matblock_d* m_1, 
    const Matblock_d* m_2
);

/** 
 * @brief Multiples two single precision complex matrix blocks element by 
 * element.
 */
Error* multiply_matblocks_elt_wise_cf
(
    Matblock_cf**      m_out, 
    const Matblock_cf* m_1, 
    const Matblock_cf* m_2
);

/** 
 * @brief Multiples two double precision complex matrix blocks element by 
 * element.
 */
Error* multiply_matblocks_elt_wise_cd
(
    Matblock_cd**      m_out, 
    const Matblock_cd* m_1, 
    const Matblock_cd* m_2
);

/*@}*/




/** 
 * @name add_matblocks
 *
 * Adds two matrix blocks element-wise.
 * @{
 */

/** @brief Adds two single precision matrix blocks element-wise. */
Error* add_matblocks_f
(
    Matblock_f**      m_out, 
    const Matblock_f* m_1, 
    const Matblock_f* m_2
);

/** @brief Adds two double precision matrix blocks element-wise. */
Error* add_matblocks_d
(
    Matblock_d**      m_out, 
    const Matblock_d* m_1, 
    const Matblock_d* m_2
);

/*@}*/




/** 
 * @name subtract_matblocks
 *
 * Subtracts two matrix blocks element-wise.
 * @{
 */

/** @brief Subtracts two single precision matrix blocks element-wise. */
Error* subtract_matblocks_f
(
    Matblock_f**      m_out, 
    const Matblock_f* m_1, 
    const Matblock_f* m_2
);

/** @brief Subtracts two double precision matrix blocks element-wise. */
Error* subtract_matblocks_d
(
    Matblock_d**      m_out, 
    const Matblock_d* m_1, 
    const Matblock_d* m_2
);

/*@}*/




/** 
 * @name calc_matblock_asum
 *
 * Calculates the absolute sum of a real matrix block.
 * @{
 */

/** 
 * @brief Calculates the absolute value sum of a single precision matrix block.
 */
void calc_matblock_asum_f(float* sum_out, const Matblock_f* m);

/** 
 * @brief Calculates the absolute value sum of a double precision matrix block.
 */
void calc_matblock_asum_d(double* sum_out, const Matblock_d* m);

/*@}*/




/** 
 * @name normalize_matblock_sum
 *
 * Normalizes a real matrix block by its sum.
 * @{
 */

/** @brief Normalizes a single precision matrix block by its sum. */
void normalize_matblock_sum_f(Matblock_f** m_out, const Matblock_f* m_in);

/** @brief Normalizes a double precision matrix block by its sum. */
void normalize_matblock_sum_d(Matblock_d** m_out, const Matblock_d* m_in);

/*@}*/


/** 
 * @name abs_matblock
 *
 * Takes the absolute value of all elements of a matrix.
 * @{
 */

/** @brief Takes the absolute value of all elements of a single precision matrix. */
void abs_matblock_f(Matblock_f** m_out, const Matblock_f* m_in);

/** @brief Takes the absolute value of all elements of a double precision matrix. */
void abs_matblock_d(Matblock_d** m_out, const Matblock_d* m_in);


/*@}*/


/** 
 * @name threshold_matblock
 *
 * Thresholds a matblock by setting all values below the threshold to zero.
 * @{
 */

/** @brief Thresholds a single precision matrix block by setting all values below the threshold to zero. */
void threshold_matblock_f(Matblock_f** m_out, const Matblock_f* m_in, float threshold);

/** @brief Thresholds a double precision matrix block by setting all values below the threshold to zero. */
void threshold_matblock_d(Matblock_d** m_out, const Matblock_d* m_in, double threshold);

/*@}*/

#ifdef __cplusplus
}
}
#endif


#endif
