/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Declarations for a C interface to the functions from BLAS levels 
 * 1, 2, and 3.
 *
 * Wraps levels 1, 2, and 3 of the BLAS library (standard) from netlib. Not all 
 * functions are wrapped, just the ones I find useful.
 *
 * Manipulates the transpose parameters so that all matrices can be passed
 * as row-major ordered.
 */


#ifndef BLAS_H
#define BLAS_H


#include <jwsc/config.h>

#include <stdlib.h>

#include "jwsc/math/complex.h"


#ifdef __cplusplus
namespace jwsc {
extern "C" {
#endif


/* ===============  LEVEL 1 PROTOTYPES ==============*/


/**
 * @name [sdcz]swap 
 *
 * Swaps the elements of two vectors.
 * @{
 */

/** @brief Single precision. */
void blas_sswap(int n, float* x, int inc_x, float* y, int inc_y);

/** @brief Double precision. */
void blas_dswap(int n, double* x, int inc_x, double* y, int inc_y);

/** @brief Complex single precision. */
void blas_cswap(int n, Complex_f* x, int inc_x, Complex_f* y, int inc_y);

/** @brief Complex double precision. */
void blas_zswap(int n, Complex_d* x, int inc_x, Complex_d* y, int inc_y);

/*@}*/




/**
 * @name [sdcz]scal 
 *
 * Multiplies a vector by a scalar.
 * @{
 */

/** @brief Single precision. */
void blas_sscal(int n, float a, float* x, int inc_x);

/** @brief Double precision vector by a scalar. */
void blas_dscal(int n, double a, double* x, int inc_x);

/** @brief Complex single precision. */
void blas_cscal(int n, Complex_f a, Complex_f* x, int inc_x);

/** @brief Complex double precision. */
void blas_zscal(int n, Complex_d a, Complex_d* x, int inc_x);

/*@}*/




/**
 * @name [sdcz]copy 
 *
 * Copies a vector into another.
 * @{
 */

/** @brief Single precision. */
void blas_scopy(int n, const float* x, int inc_x, float* y, int inc_y);

/** @brief Double precision. */
void blas_dcopy(int n, const double* x, int inc_x, double* y, int inc_y);

/** @brief Complex single precision. */
void blas_ccopy(int n, const Complex_f* x, int inc_x, Complex_f* y, int inc_y);

/** @brief Complex double precision. */
void blas_zcopy(int n, const Complex_d* x, int inc_x, Complex_d* y, int inc_y);

/*@}*/




/**
 * @name [sdcz]axpy 
 *
 * Constant times a vector plus a vector.
 * @{
 */

/** @brief Single precision. */
void blas_saxpy
(
    int          n, 
    float        a, 
    const float* x, 
    int          inc_x, 
    float*       y, 
    int          inc_y
);

/** @brief Double precision. */
void blas_daxpy
(
    int           n, 
    double        a, 
    const double* x, 
    int           inc_x, 
    double*       y, 
    int           inc_y
);

/** @brief Complex single precision. */
void blas_caxpy
(
    int              n, 
    Complex_f        a, 
    const Complex_f* x, 
    int              inc_x, 
    Complex_f*       y, 
    int              inc_y
);

/** @brief Complex double precision. */
void blas_zaxpy
(
    int              n, 
    Complex_d        a, 
    const Complex_d* x, 
    int              inc_x, 
    Complex_d*       y, 
    int              inc_y
);

/*@}*/




/**
 * @name [sd]dot 
 *
 * Vector dot (inner) product.
 * @{
 */

/** @brief Single precision. */
float blas_sdot(int n, const float* x, int inc_x, const float* y, int inc_y);

/** @brief Double precision. */
double blas_ddot(int n, const double* x, int inc_x, const double* y, int inc_y);

/*@}*/




/**
 * @name [sd]nrm2 
 *
 * Vector magnitude (Euclidean norm).
 * @{
 */

/** @brief Single precision. */
float blas_snrm2(int n, const float* x, int inc_x);

/** @brief Double precision. */
double blas_dnrm2(int n, const double* x, int inc_x);

/*@}*/




/**
 * @name [sd]asum
 *
 * Sum of the magnitudes of the elements of a vector.
 * @{
 */

/** @brief Single precision. */
float blas_sasum(int n, const float* x, int inc_x);

/** @brief Double precision. */
double blas_dasum(int n, const double* x, int inc_x);

/*@}*/




/**
 * @name i[sdcz]amax
 *
 * Maximum absolute value in a vector.
 * @{
 */

/** @brief Single precision. */
int blas_isamax(int n, const float* x, int inc_x);

/** @brief Double precision. */
int blas_idamax(int n, const double* x, int inc_x);

/** @brief Complex single precision. */
int blas_icamax(int n, const Complex_f* x, int inc_x);

/** @brief Complex double precision. */
int blas_izamax(int n, const Complex_d* x, int inc_x);

/*@}*/




/* ===============  LEVEL 2 PROTOTYPES ==============*/


/**
 * @name [sdcz]gemv 
 *
 * General matrix-vector multiplication.
 * @{
 */

/** @brief Single precision. */
void blas_sgemv
(
    char         trans, 
    int          m, 
    int          n, 
    float        alpha, 
    const float* A, 
    int          lda,
    const float* x,
    int          inc_x,
    float        beta,
    float*       y,
    int          inc_y
);

/** @brief Double precision. */
void blas_dgemv
(
    char          trans, 
    int           m, 
    int           n, 
    double        alpha, 
    const double* A, 
    int           lda,
    const double* x,
    int           inc_x,
    double        beta,
    double*       y,
    int           inc_y
);

/** @brief Complex single precision. */
void blas_cgemv
(
    char             trans, 
    int              m, 
    int              n, 
    Complex_f        alpha, 
    const Complex_f* A, 
    int              lda,
    const Complex_f* x,
    int              inc_x,
    Complex_f        beta,
    Complex_f*       y,
    int              inc_y
);

/** @brief Complex double precision. */
void blas_zgemv
(
    char             trans, 
    int              m, 
    int              n, 
    Complex_d        alpha, 
    const Complex_d* A, 
    int              lda,
    const Complex_d* x,
    int              inc_x,
    Complex_d        beta,
    Complex_d*       y,
    int              inc_y
);

/*@}*/





/* ===============  LEVEL 3 PROTOTYPES ==============*/


/**
 * @name [sdcz]gemm 
 *
 * General matrix multiplication.
 * @{
 */

/** @brief Single precision. */
void blas_sgemm
(
    char         trans_a, 
    char         trans_b, 
    int          m, 
    int          n, 
    int          k, 
    float        alpha, 
    const float* A, 
    int          lda,
    const float* B,
    int          ldb,
    float        beta,
    float*       C,
    int          ldc
);

/** @brief Double precision. */
void blas_dgemm
(
    char          trans_a, 
    char          trans_b, 
    int           m, 
    int           n, 
    int           k, 
    double        alpha, 
    const double* A, 
    int           lda,
    const double* B,
    int           ldb,
    double        beta,
    double*       C,
    int           ldc
);

/** @brief Complex single precision. */
void blas_cgemm
(
    char             trans_a, 
    char             trans_b, 
    int              m, 
    int              n, 
    int              k, 
    Complex_f        alpha, 
    const Complex_f* A, 
    int              lda,
    const Complex_f* B,
    int              ldb,
    Complex_f        beta,
    Complex_f*       C,
    int              ldc
);

/** @brief Complex double precision. */
void blas_zgemm
(
    char             trans_a, 
    char             trans_b, 
    int              m, 
    int              n, 
    int              k, 
    Complex_d        alpha, 
    const Complex_d* A, 
    int              lda,
    const Complex_d* B,
    int              ldb,
    Complex_d        beta,
    Complex_d*       C,
    int              ldc
);

/*@}*/


#ifdef __cplusplus
}
}
#endif


#endif
