/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file 
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Definitions for functions from LAPACK.
 *
 * Wraps the LAPACK library from netlib.
 *
 * Input matrices are expected to be in row-major order; they are transposed
 * before and after calling the wrapped Fortran routine.
 */


#include <jwsc/config.h>

#include <stdlib.h>
#include <assert.h>

#include "jwsc/math/lapack.h"

#ifdef JWSC_HAVE_DMALLOC
#include <dmalloc.h>
#endif


extern float  sgetrf_(int*,int*,float*,int*,int*,int*);
extern double dgetrf_(int*,int*,double*,int*,int*,int*);

extern float  sgetri_(int*,float*,int*,int*,float*,int*,int*);
extern double dgetri_(int*,double*,int*,int*,double*,int*,int*);

extern float sgesdd_(char*,int*,int*,float*,int*,float*,float*,int*,float*,int*,float*,int*,int*,int*);
extern double dgesdd_(char*,int*,int*,double*,int*,double*,double*,int*,double*,int*,double*,int*,int*,int*);

extern float ssyev_(char*,char*,int*,float*,int*,float*,float*,int*,int*);
extern double dsyev_(char*,char*,int*,double*,int*,double*,double*,int*,int*);


/** @brief Double precision matrix transpose. */
static void trans_d(double** A_out, const double* A, int M, int N)
{
    int m, n;

    assert(*A_out != A);

    if (*A_out == NULL)
    {
        assert(*A_out = malloc(M*N*sizeof(double)));
    }

    for (m = 0; m < M; m++)
    {
        for (n = 0; n < N; n++)
        {
            (*A_out)[ n*M + m ] = A[ m*N + n ];
        }
    }
}


/** @brief Single precision matrix transpose. */
static void trans_f(float** A_out, const float* A, int M, int N)
{
    int i, j;

    assert(*A_out != A);

    if (*A_out == NULL)
    {
        assert(*A_out = malloc(M*N*sizeof(float)));
    }

    for (i = 0; i < M; i++)
    {
        for (j = 0; j < N; j++)
        {
            (*A_out)[ j*M + i ] = A[ i*N + j ];
        }
    }
}


/** 
 * The factorization has the form
 *
 * @code
 *     A = P * L * U
 * @endcode
 *
 *  where P is a permutation matrix, L is lower triangular with unit
 *  diagonal elements (lower trapezoidal if M > N), and U is upper
 *  triangular (upper trapezoidal if M < N).
 *
 * @param  M      Number of rows in @em A.
 * @param  N      Number of columns in @em A.
 * @param  A      On entry, the M-by-N matrix to be factored.
 *                On exit, the factors L and U from the factorization
 *                A = P*L*U; the unit diagonal elements of L are not stored.
 * @param  lda    Number of rows in @em A. Referred to as the leading 
 *                dimension of @em A.
 * @param  ipiv   The pivot indices; for 1 <= i <= min(M,N), row i of the
 *                matrix was interchanged with row IPIV(i).
 * @param  info   = 0  successful exit;
 *                < 0  if INFO = -i, the i-th argument had an illegal value;
 *                > 0  if INFO = i, U(i,i) is exactly zero. The factorization
 *                  has been completed, but the factor U is exactly
 *                  singular, and division by zero will occur if it is used
 *                  to solve a system of equations.
 */
void lapack_sgetrf
(
    int     M, 
    int     N, 
    float*  A, 
    int     lda,
    int*    ipiv,
    int*    info
)
{
    float*  A_t = NULL;

    trans_f(&A_t, A, M, N);
    sgetrf_(&N, &M, A_t, &lda, ipiv, info);
    trans_f(&A, A_t, N, M);

    free(A_t);
}


/** 
 * The factorization has the form
 *
 * @code
 *     A = P * L * U
 * @endcode
 *
 *  where P is a permutation matrix, L is lower triangular with unit
 *  diagonal elements (lower trapezoidal if M > N), and U is upper
 *  triangular (upper trapezoidal if M < N).
 *
 * @param  M      Number of rows in @em A.
 * @param  N      Number of columns in @em A.
 * @param  A      On entry, the M-by-N matrix to be factored.
 *                On exit, the factors L and U from the factorization
 *                A = P*L*U; the unit diagonal elements of L are not stored.
 * @param  lda    Number of rows in @em A. Referred to as the leading 
 *                dimension of @em A.
 * @param  ipiv   The pivot indices; for 1 <= i <= min(M,N), row i of the
 *                matrix was interchanged with row IPIV(i).
 * @param  info   = 0  successful exit;
 *                < 0  if INFO = -i, the i-th argument had an illegal value;
 *                > 0  if INFO = i, U(i,i) is exactly zero. The factorization
 *                  has been completed, but the factor U is exactly
 *                  singular, and division by zero will occur if it is used
 *                  to solve a system of equations.
 */
void lapack_dgetrf
(
    int     M, 
    int     N, 
    double* A, 
    int     lda,
    int*    ipiv,
    int*    info
)
{
    double* A_t = NULL;

    trans_d(&A_t, A, M, N);
    dgetrf_(&M, &N, A_t, &lda, ipiv, info);
    trans_d(&A, A_t, N, M);

    free(A_t);
}


/** 
 * This method inverts U and then computes inv(A) by solving the system
 *
 * @code
 *     inv(A)*L = inv(U) for inv(A).
 * @endcode
 *
 * @param  N      The order of the matrix A. N >= 0.
 * @param  A      On entry, the factors L and U from the factorization
 *                A = P*L*U as computed by SGETRF.
 *                On exit, if INFO = 0, the inverse of the original matrix A.
 * @param  lda    Number of rows in @em A. Referred to as the leading 
 *                dimension of @em A.
 * @param  ipiv   The pivot indices; for 1 <= i <= N, row i of the
 *                matrix was interchanged with row IPIV(i).
 * @param  work   On exit, if INFO=0, then WORK(1) returns the optimal LWORK.
 * @param  lwork  The dimension of the array WORK.  LWORK >= max(1,N).
 *                For optimal performance LWORK >= N*NB, where NB is
 *                the optimal blocksize returned by ILAENV.
 *                If LWORK = -1, then a workspace query is assumed; the routine
 *                only calculates the optimal size of the WORK array, returns
 *                this value as the first entry of the WORK array, and no error
 *                message related to LWORK is issued by XERBLA.
 * @param  info   = 0  successful exit;
 *                < 0  if INFO = -i, the i-th argument had an illegal value;
 *                > 0  if INFO = i, U(i,i) is exactly zero, the matrix is
 *                  singular and its inverse could not be computed.
 */
void lapack_sgetri
(
    int     N, 
    float*  A, 
    int     lda,
    int*    ipiv,
    float*  work, 
    int     lwork,
    int*    info
)
{
    float*  A_t = NULL;

    trans_f(&A_t, A, N, N);
    sgetri_(&N, A_t, &lda, ipiv, work, &lwork, info);
    trans_f(&A, A_t, N, N);

    free(A_t);
}


/** 
 * This method inverts U and then computes inv(A) by solving the system
 *
 * @code
 *     inv(A)*L = inv(U) for inv(A).
 * @endcode
 *
 * @param  N      The order of the matrix A. N >= 0.
 * @param  A      On entry, the factors L and U from the factorization
 *                A = P*L*U as computed by DGETRF.
 *                On exit, if INFO = 0, the inverse of the original matrix A.
 * @param  lda    Number of rows in @em A. Referred to as the leading 
 *                dimension of @em A.
 * @param  ipiv   The pivot indices; for 1 <= i <= N, row i of the
 *                matrix was interchanged with row IPIV(i).
 * @param  work   On exit, if INFO=0, then WORK(1) returns the optimal LWORK.
 * @param  lwork  The dimension of the array WORK.  LWORK >= max(1,N).
 *                For optimal performance LWORK >= N*NB, where NB is
 *                the optimal blocksize returned by ILAENV.
 *                If LWORK = -1, then a workspace query is assumed; the routine
 *                only calculates the optimal size of the WORK array, returns
 *                this value as the first entry of the WORK array, and no error
 *                message related to LWORK is issued by XERBLA.
 * @param  info   = 0  successful exit;
 *                < 0  if INFO = -i, the i-th argument had an illegal value;
 *                > 0  if INFO = i, U(i,i) is exactly zero, the matrix is
 *                  singular and its inverse could not be computed.
 */
void lapack_dgetri
(
    int     N, 
    double* A, 
    int     lda,
    int*    ipiv,
    double* work, 
    int     lwork,
    int*    info
)
{
    double* A_t = NULL;

    trans_d(&A_t, A, N, N);
    dgetri_(&N, A_t, &lda, ipiv, work, &lwork, info);
    trans_d(&A, A_t, N, N);

    free(A_t);
}


/** 
 * SGESDD computes the singular value decomposition (SVD) of a real M-by-N
 * matrix A, optionally computing the left and right singular vectors.  If
 * singular vectors are desired, it uses a divide-and-conquer algorithm.
 *
 *  The SVD is written
 *
 * @code
 *    A = U * SIGMA * transpose(V)
 * @endcode
 *
 * where SIGMA is an M-by-N matrix which is zero except for its min(m,n)
 * diagonal elements, U is an M-by-M orthogonal matrix, and V is an N-by-N
 * orthogonal matrix.  The diagonal elements of SIGMA are the singular values
 * of A; they are real and non-negative, and are returned in descending order.
 * The first min(m,n) columns of U and V are the left and right singular
 * vectors of A.
 *
 * Note that the routine returns VT = V**T, not V.
 *
 *
 * @param  JOBZ    (input) CHARACTER*1
 *                 Specifies options for computing all or part of the matrix U:
 *                 = 'A':  all M columns of U and all N rows of V**T are
 *                         returned in the arrays U and VT;
 *                 = 'S':  the first min(M,N) columns of U and the first
 *                         min(M,N) rows of V**T are returned in the arrays U
 *                         and VT;
 *                 = 'O':  If M >= N, the first N columns of U are overwritten
 *                         on the array A and all rows of V**T are returned in
 *                         the array VT;
 *                         otherwise, all columns of U are returned in the
 *                         array U and the first M rows of V**T are overwritten
 *                         in the array A;
 *                 = 'N':  no columns of U or rows of V**T are computed.
 *
 * @param  M       (input) INTEGER
 *                 The number of rows of the input matrix A.  M >= 0.
 *
 * @param  N       (input) INTEGER
 *                 The number of columns of the input matrix A.  N >= 0.
 *
 * @param  A       (input/output) REAL array, dimension (LDA,N)
 *                 On entry, the M-by-N matrix A.
 *                 On exit,
 *                 if JOBZ = 'O',  A is overwritten with the first N columns
 *                                 of U (the left singular vectors, stored
 *                                 columnwise) if M >= N;
 *                                 A is overwritten with the first M rows
 *                                 of V**T (the right singular vectors, stored
 *                                 rowwise) otherwise.
 *                 if JOBZ .ne. 'O', the contents of A are destroyed.
 *
 * @param  LDA     (input) INTEGER
 *                 The leading dimension of the array A.  LDA >= max(1,M).
 *
 * @param  S       (output) REAL array, dimension (min(M,N))
 *                 The singular values of A, sorted so that S(i) >= S(i+1).
 *
 * @param  U       (output) REAL array, dimension (LDU,UCOL)
 *                 UCOL = M if JOBZ = 'A' or JOBZ = 'O' and M < N;
 *                 UCOL = min(M,N) if JOBZ = 'S'.
 *                 If JOBZ = 'A' or JOBZ = 'O' and M < N, U contains the M-by-M
 *                 orthogonal matrix U;
 *                 if JOBZ = 'S', U contains the first min(M,N) columns of U
 *                 (the left singular vectors, stored columnwise);
 *                 if JOBZ = 'O' and M >= N, or JOBZ = 'N', U is not referenced.
 *
 * @param  LDU     (input) INTEGER
 *                 The leading dimension of the array U.  LDU >= 1; if
 *                 JOBZ = 'S' or 'A' or JOBZ = 'O' and M < N, LDU >= M.
 *
 * @param  VT      (output) REAL array, dimension (LDVT,N)
 *                 If JOBZ = 'A' or JOBZ = 'O' and M >= N, VT contains the
 *                 N-by-N orthogonal matrix V**T;
 *                 if JOBZ = 'S', VT contains the first min(M,N) rows of
 *                 V**T (the right singular vectors, stored rowwise);
 *                 if JOBZ = 'O' and M < N, or JOBZ = 'N', VT is not referenced.
 *
 * @param  LDVT    (input) INTEGER
 *                 The leading dimension of the array VT.  LDVT >= 1; if
 *                 JOBZ = 'A' or JOBZ = 'O' and M >= N, LDVT >= N;
 *                 if JOBZ = 'S', LDVT >= min(M,N).
 *
 * @param  WORK    (workspace/output) REAL array, dimension (MAX(1,LWORK))
 *                 On exit, if INFO = 0, WORK(1) returns the optimal LWORK;
 *
 * @param  LWORK   (input) INTEGER
 *                 The dimension of the array WORK. LWORK >= 1.
 *                 If JOBZ = 'N',
 *                   LWORK >= 3*min(M,N) + max(max(M,N),6*min(M,N)).
 *                 If JOBZ = 'O',
 *                   LWORK >= 3*min(M,N)*min(M,N) + 
 *                            max(max(M,N),5*min(M,N)*min(M,N)+4*min(M,N)).
 *                 If JOBZ = 'S' or 'A'
 *                   LWORK >= 3*min(M,N)*min(M,N) +
 *                            max(max(M,N),4*min(M,N)*min(M,N)+4*min(M,N)).
 *                 For good performance, LWORK should generally be larger.
 *                 If LWORK = -1 but other input arguments are legal, WORK(1)
 *                 returns the optimal LWORK.
 *
 * @param  IWORK   (workspace) INTEGER array, dimension (8*min(M,N))
 * @param  INFO    (output) INTEGER
 *                 = 0:  successful exit.
 *                 < 0:  if INFO = -i, the i-th argument had an illegal value.
 *                 > 0:  SBDSDC did not converge, updating process failed.
 */
void lapack_sgesdd
(
    char   JOBZ,
    int    M,
    int    N, 
    float* A, 
    int    LDA,
    float* S,
    float* U,
    int    LDU,
    float* VT,
    int    LDVT,
    float* WORK,
    int    LWORK,
    int*   IWORK,
    int*   INFO
)
{
    int    min_MN;
    float* A_t  = NULL;
    float* U_t  = NULL;
    float* VT_t = NULL;

    min_MN = (M <= N) ? M : N;

    if (JOBZ == 'A')
    {
        trans_f(&U_t, U, LDU, M);
        trans_f(&VT_t, VT, LDVT, N);
    }
    else if (JOBZ == 'S')
    {
        trans_f(&U_t, U, LDU, min_MN);
        trans_f(&VT_t, VT, LDVT, min_MN);
    }
    else if (JOBZ == 'O')
    {
        if (M < N) 
            trans_f(&U_t, U, LDU, M);
        else
            trans_f(&VT_t, VT, LDVT, N);
    }
    else if (JOBZ == 'N')
    {
        ;
    }
    else
    {
        *INFO = -1;
        return;
    }

    trans_f(&A_t, A, M, N);
    sgesdd_(&JOBZ, &M, &N, A_t, &LDA, S, U_t, &LDU, VT_t, &LDVT, WORK,
            &LWORK, IWORK, INFO);
    trans_f(&A, A_t, N, M);

    if (JOBZ == 'A')
    {
        trans_f(&U, U_t, M, LDU);
        trans_f(&VT, VT_t, N, LDVT);
    }
    else if (JOBZ == 'S')
    {
        trans_f(&U, U_t, min_MN, LDU);
        trans_f(&VT, VT_t, min_MN, LDVT);
    }
    else if (JOBZ == 'O')
    {
        if (M < N) 
            trans_f(&U, U_t, M, LDU);
        else
            trans_f(&VT, VT_t, N, LDVT);
    }

    free(A_t);
    free(U_t);
    free(VT_t);
}


/** 
 * DGESDD computes the singular value decomposition (SVD) of a real M-by-N
 * matrix A, optionally computing the left and right singular vectors.  If
 * singular vectors are desired, it uses a divide-and-conquer algorithm.
 *
 *  The SVD is written
 *
 * @code
 *    A = U * SIGMA * transpose(V)
 * @endcode
 *
 * where SIGMA is an M-by-N matrix which is zero except for its min(m,n)
 * diagonal elements, U is an M-by-M orthogonal matrix, and V is an N-by-N
 * orthogonal matrix.  The diagonal elements of SIGMA are the singular values
 * of A; they are real and non-negative, and are returned in descending order.
 * The first min(m,n) columns of U and V are the left and right singular
 * vectors of A.
 *
 * Note that the routine returns VT = V**T, not V.
 *
 *
 * @param  JOBZ    (input) CHARACTER*1
 *                 Specifies options for computing all or part of the matrix U:
 *                 = 'A':  all M columns of U and all N rows of V**T are
 *                         returned in the arrays U and VT;
 *                 = 'S':  the first min(M,N) columns of U and the first
 *                         min(M,N) rows of V**T are returned in the arrays U
 *                         and VT;
 *                 = 'O':  If M >= N, the first N columns of U are overwritten
 *                         on the array A and all rows of V**T are returned in
 *                         the array VT;
 *                         otherwise, all columns of U are returned in the
 *                         array U and the first M rows of V**T are overwritten
 *                         in the array A;
 *                 = 'N':  no columns of U or rows of V**T are computed.
 *
 * @param  M       (input) INTEGER
 *                 The number of rows of the input matrix A.  M >= 0.
 *
 * @param  N       (input) INTEGER
 *                 The number of columns of the input matrix A.  N >= 0.
 *
 * @param  A       (input/output) REAL array, dimension (LDA,N)
 *                 On entry, the M-by-N matrix A.
 *                 On exit,
 *                 if JOBZ = 'O',  A is overwritten with the first N columns
 *                                 of U (the left singular vectors, stored
 *                                 columnwise) if M >= N;
 *                                 A is overwritten with the first M rows
 *                                 of V**T (the right singular vectors, stored
 *                                 rowwise) otherwise.
 *                 if JOBZ .ne. 'O', the contents of A are destroyed.
 *
 * @param  LDA     (input) INTEGER
 *                 The leading dimension of the array A.  LDA >= max(1,M).
 *
 * @param  S       (output) REAL array, dimension (min(M,N))
 *                 The singular values of A, sorted so that S(i) >= S(i+1).
 *
 * @param  U       (output) REAL array, dimension (LDU,UCOL)
 *                 UCOL = M if JOBZ = 'A' or JOBZ = 'O' and M < N;
 *                 UCOL = min(M,N) if JOBZ = 'S'.
 *                 If JOBZ = 'A' or JOBZ = 'O' and M < N, U contains the M-by-M
 *                 orthogonal matrix U;
 *                 if JOBZ = 'S', U contains the first min(M,N) columns of U
 *                 (the left singular vectors, stored columnwise);
 *                 if JOBZ = 'O' and M >= N, or JOBZ = 'N', U is not referenced.
 *
 * @param  LDU     (input) INTEGER
 *                 The leading dimension of the array U.  LDU >= 1; if
 *                 JOBZ = 'S' or 'A' or JOBZ = 'O' and M < N, LDU >= M.
 *
 * @param  VT      (output) REAL array, dimension (LDVT,N)
 *                 If JOBZ = 'A' or JOBZ = 'O' and M >= N, VT contains the
 *                 N-by-N orthogonal matrix V**T;
 *                 if JOBZ = 'S', VT contains the first min(M,N) rows of
 *                 V**T (the right singular vectors, stored rowwise);
 *                 if JOBZ = 'O' and M < N, or JOBZ = 'N', VT is not referenced.
 *
 * @param  LDVT    (input) INTEGER
 *                 The leading dimension of the array VT.  LDVT >= 1; if
 *                 JOBZ = 'A' or JOBZ = 'O' and M >= N, LDVT >= N;
 *                 if JOBZ = 'S', LDVT >= min(M,N).
 *
 * @param  WORK    (workspace/output) REAL array, dimension (MAX(1,LWORK))
 *                 On exit, if INFO = 0, WORK(1) returns the optimal LWORK;
 *
 * @param  LWORK   (input) INTEGER
 *                 The dimension of the array WORK. LWORK >= 1.
 *                 If JOBZ = 'N',
 *                   LWORK >= 3*min(M,N) + max(max(M,N),6*min(M,N)).
 *                 If JOBZ = 'O',
 *                   LWORK >= 3*min(M,N)*min(M,N) + 
 *                            max(max(M,N),5*min(M,N)*min(M,N)+4*min(M,N)).
 *                 If JOBZ = 'S' or 'A'
 *                   LWORK >= 3*min(M,N)*min(M,N) +
 *                            max(max(M,N),4*min(M,N)*min(M,N)+4*min(M,N)).
 *                 For good performance, LWORK should generally be larger.
 *                 If LWORK = -1 but other input arguments are legal, WORK(1)
 *                 returns the optimal LWORK.
 *
 * @param  IWORK   (workspace) INTEGER array, dimension (8*min(M,N))
 * @param  INFO    (output) INTEGER
 *                 = 0:  successful exit.
 *                 < 0:  if INFO = -i, the i-th argument had an illegal value.
 *                 > 0:  SBDSDC did not converge, updating process failed.
 */
void lapack_dgesdd
(
    char    JOBZ,
    int     M,
    int     N, 
    double* A, 
    int     LDA,
    double* S,
    double* U,
    int     LDU,
    double* VT,
    int     LDVT,
    double* WORK,
    int     LWORK,
    int*    IWORK,
    int*    INFO
)
{
    int     min_MN;
    double* A_t  = NULL;
    double* U_t  = NULL;
    double* VT_t = NULL;

    min_MN = (M <= N) ? M : N;

    if (JOBZ == 'A')
    {
        trans_d(&U_t, U, LDU, M);
        trans_d(&VT_t, VT, LDVT, N);
    }
    else if (JOBZ == 'S')
    {
        trans_d(&U_t, U, LDU, min_MN);
        trans_d(&VT_t, VT, LDVT, min_MN);
    }
    else if (JOBZ == 'O')
    {
        if (M < N) 
            trans_d(&U_t, U, LDU, M);
        else
            trans_d(&VT_t, VT, LDVT, N);
    }
    else if (JOBZ == 'N')
    {
        ;
    }
    else
    {
        *INFO = -1;
        return;
    }

    trans_d(&A_t, A, M, N);
    dgesdd_(&JOBZ, &M, &N, A_t, &LDA, S, U_t, &LDU, VT_t, &LDVT, WORK,
            &LWORK, IWORK, INFO);
    trans_d(&A, A_t, N, M);

    if (JOBZ == 'A')
    {
        trans_d(&U, U_t, M, LDU);
        trans_d(&VT, VT_t, N, LDVT);
    }
    else if (JOBZ == 'S')
    {
        trans_d(&U, U_t, min_MN, LDU);
        trans_d(&VT, VT_t, min_MN, LDVT);
    }
    else if (JOBZ == 'O')
    {
        if (M < N) 
            trans_d(&U, U_t, M, LDU);
        else
            trans_d(&VT, VT_t, N, LDVT);
    }

    free(A_t);
    free(U_t);
    free(VT_t);
}


/**
 *  SSYEV computes all eigenvalues and, optionally, eigenvectors of a
 *  real symmetric matrix A.
 *
 *  @param  JOBZ    (input) CHARACTER*1
 *                  = 'N':  Compute eigenvalues only;
 *                  = 'V':  Compute eigenvalues and eigenvectors.
 *
 *  @param  UPLO    (input) CHARACTER*1
 *                  = 'U':  Upper triangle of A is stored;
 *                  = 'L':  Lower triangle of A is stored.
 *
 *  @param  N       (input) INTEGER
 *                  The order of the matrix A.  N >= 0.
 *
 *  @param  A       (input/output) SINGLE PRECISION array, dimension (LDA, N)
 *                  On entry, the symmetric matrix A.  If UPLO = 'U', the
 *                  leading N-by-N upper triangular part of A contains the
 *                  upper triangular part of the matrix A.  If UPLO = 'L',
 *                  the leading N-by-N lower triangular part of A contains
 *                  the lower triangular part of the matrix A.
 *                  On exit, if JOBZ = 'V', then if INFO = 0, A contains the
 *                  orthonormal eigenvectors of the matrix A.
 *                  If JOBZ = 'N', then on exit the lower triangle (if UPLO='L')
 *                  or the upper triangle (if UPLO='U') of A, including the
 *                  diagonal, is destroyed.
 *
 *  @param  LDA     (input) INTEGER
 *                  The leading dimension of the array A.  LDA >= max(1,N).
 *
 *  @param  W       (output) SINGLE PRECISION array, dimension (N)
 *                  If INFO = 0, the eigenvalues in ascending order.
 *
 *  @param  WORK    (workspace/output) SINGLE PRECISION array, dimension
 *                  (MAX(1,LWORK)) On exit, if INFO = 0, WORK(1) returns the
 *                  optimal LWORK.
 *
 *  @param  LWORK   (input) INTEGER
 *                  The length of the array WORK.  LWORK >= max(1,3*N-1).
 *                  For optimal efficiency, LWORK >= (NB+2)*N,
 *                  where NB is the blocksize for DSYTRD returned by ILAENV.
 *
 *                  If LWORK = -1, then a workspace query is assumed; the
 *                  routine only calculates the optimal size of the WORK
 *                  array, returns this value as the first entry of the WORK
 *                  array, and no error message related to LWORK is issued by
 *                  XERBLA.
 *
 *  @param  INFO    (output) INTEGER
 *                  = 0:  successful exit
 *                  < 0:  if INFO = -i, the i-th argument had an illegal value
 *                  > 0:  if INFO = i, the algorithm failed to converge; i
 *                        off-diagonal elements of an intermediate tridiagonal
 *                        form did not converge to zero.
 */
void lapack_ssyev
(
    char    JOBZ,
    char    UPLO,
    int     N, 
    float*  A, 
    int     LDA,
    float*  W,
    float*  WORK,
    int     LWORK,
    int*    INFO
)
{
    float* A_t = NULL;

    if (JOBZ != 'N' && JOBZ != 'V')
    {
        *INFO = -1;
        return;
    }

    trans_f(&A_t, A, N, N);
    ssyev_(&JOBZ, &UPLO, &N, A_t, &LDA, W, WORK, &LWORK, INFO);
    trans_f(&A, A_t, N, N);

    free(A_t);
}


/**
 *  DSYEV computes all eigenvalues and, optionally, eigenvectors of a
 *  real symmetric matrix A.
 *
 *  @param  JOBZ    (input) CHARACTER*1
 *                  = 'N':  Compute eigenvalues only;
 *                  = 'V':  Compute eigenvalues and eigenvectors.
 *
 *  @param  UPLO    (input) CHARACTER*1
 *                  = 'U':  Upper triangle of A is stored;
 *                  = 'L':  Lower triangle of A is stored.
 *
 *  @param  N       (input) INTEGER
 *                  The order of the matrix A.  N >= 0.
 *
 *  @param  A       (input/output) DOUBLE PRECISION array, dimension (LDA, N)
 *                  On entry, the symmetric matrix A.  If UPLO = 'U', the
 *                  leading N-by-N upper triangular part of A contains the
 *                  upper triangular part of the matrix A.  If UPLO = 'L',
 *                  the leading N-by-N lower triangular part of A contains
 *                  the lower triangular part of the matrix A.
 *                  On exit, if JOBZ = 'V', then if INFO = 0, A contains the
 *                  orthonormal eigenvectors of the matrix A.
 *                  If JOBZ = 'N', then on exit the lower triangle (if UPLO='L')
 *                  or the upper triangle (if UPLO='U') of A, including the
 *                  diagonal, is destroyed.
 *
 *  @param  LDA     (input) INTEGER
 *                  The leading dimension of the array A.  LDA >= max(1,N).
 *
 *  @param  W       (output) DOUBLE PRECISION array, dimension (N)
 *                  If INFO = 0, the eigenvalues in ascending order.
 *
 *  @param  WORK    (workspace/output) DOUBLE PRECISION array, dimension
 *                  (MAX(1,LWORK)) On exit, if INFO = 0, WORK(1) returns the
 *                  optimal LWORK.
 *
 *  @param  LWORK   (input) INTEGER
 *                  The length of the array WORK.  LWORK >= max(1,3*N-1).
 *                  For optimal efficiency, LWORK >= (NB+2)*N,
 *                  where NB is the blocksize for DSYTRD returned by ILAENV.
 *
 *                  If LWORK = -1, then a workspace query is assumed; the
 *                  routine only calculates the optimal size of the WORK
 *                  array, returns this value as the first entry of the WORK
 *                  array, and no error message related to LWORK is issued by
 *                  XERBLA.
 *
 *  @param  INFO    (output) INTEGER
 *                  = 0:  successful exit
 *                  < 0:  if INFO = -i, the i-th argument had an illegal value
 *                  > 0:  if INFO = i, the algorithm failed to converge; i
 *                        off-diagonal elements of an intermediate tridiagonal
 *                        form did not converge to zero.
 */
void lapack_dsyev
(
    char    JOBZ,
    char    UPLO,
    int     N, 
    double* A, 
    int     LDA,
    double* W,
    double* WORK,
    int     LWORK,
    int*    INFO
)
{
    double* A_t = NULL;

    if (JOBZ != 'N' && JOBZ != 'V')
    {
        *INFO = -1;
        return;
    }

    trans_d(&A_t, A, N, N);
    dsyev_(&JOBZ, &UPLO, &N, A_t, &LDA, W, WORK, &LWORK, INFO);
    trans_d(&A, A_t, N, N);

    free(A_t);
}
