/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file 
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Declarations for functions from LAPACK.
 *
 * Wraps the LAPACK library from netlib.
 *
 * Input matrices are expected to be in row-major order; they are transposed
 * before and after calling the wrapped Fortran routine.
 */


#ifndef LAPACK_H
#define LAPACK_H


#include <jwsc/config.h>

#include <stdlib.h>


#ifdef __cplusplus
namespace jwsc {
extern "C" {
#endif


/** 
 * @brief Computes an LU factorization of a general M-by-N matrix A
 * using partial pivoting with row interchanges.
 */
void lapack_sgetrf
(
    int     M, 
    int     N, 
    float*  A, 
    int     lda,
    int*    ipiv,
    int*    info
);


/** 
 * @brief Computes an LU factorization of a general M-by-N matrix A
 * using partial pivoting with row interchanges.
 */
void lapack_dgetrf
(
    int     M, 
    int     N, 
    double* A, 
    int     lda,
    int*    ipiv,
    int*    info
);


/** 
 * @brief computes the inverse of a matrix using the LU factorization
 * computed by SGETRF.
 */
void lapack_sgetri
(
    int     N, 
    float*  A, 
    int     lda,
    int*    ipiv,
    float*  work, 
    int     lwork,
    int*    info
);


/** 
 * @brief computes the inverse of a matrix using the LU factorization
 * computed by DGETRF.
 */
void lapack_dgetri
(
    int     N, 
    double* A, 
    int     lda,
    int*    ipiv,
    double* work, 
    int     lwork,
    int*    info
);


/** 
 * @brief computes the SVD of a matrix using a divide and conquer algorithm
 * implemented by SGESDD.
 */
void lapack_sgesdd
(
    char    JOBZ,
    int     M,
    int     N, 
    float*  A, 
    int     LDA,
    float*  S,
    float*  U,
    int     LDU,
    float*  VT,
    int     LDVT,
    float*  WORK,
    int     LWORK,
    int*    IWORK,
    int*    INFO
);


/** 
 * @brief computes the SVD of a matrix using a divide and conquer algorithm
 * implemented by DGESDD.
 */
void lapack_dgesdd
(
    char    JOBZ,
    int     M,
    int     N, 
    double* A, 
    int     LDA,
    double* S,
    double* U,
    int     LDU,
    double* VT,
    int     LDVT,
    double* WORK,
    int     LWORK,
    int*    IWORK,
    int*    INFO
);


/** 
 * @brief Computes all eigenvalues and, optionally, eigenvectors of a real
 * symmetric matrix A.
 */
void lapack_ssyev
(
    char    JOBZ,
    char    UPLO,
    int     N, 
    float*  A, 
    int     LDA,
    float*  W,
    float*  WORK,
    int     LWORK,
    int*    INFO
);


/** 
 * @brief Computes all eigenvalues and, optionally, eigenvectors of a real
 * symmetric matrix A.
 */
void lapack_dsyev
(
    char    JOBZ,
    char    UPLO,
    int     N, 
    double* A, 
    int     LDA,
    double* W,
    double* WORK,
    int     LWORK,
    int*    INFO
);


#ifdef __cplusplus
}
}
#endif


#endif
