/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Declarations for a matrix and associated functions.
 *
 * The matrix holds integers and floating point real and complex numbers.
 *
 * Although the matrix is indexed by row and column, the elements are
 * allocated as a 1D array in row-major order. So iterating over the 
 * elements can be done inside a single loop.
 */


#ifndef MATRIX_H 
#define MATRIX_H 


#include <jwsc/config.h>

#include <stdlib.h>
#include <inttypes.h>

#include "jwsc/base/error.h"
#include "jwsc/math/complex.h"
#include "jwsc/vector/vector.h"


#ifdef __cplusplus
namespace jwsc {
extern "C" {
#endif


/** @brief Matrix of signed 8-bit values. */
typedef struct
{
    /** @brief Number of rows. */
    uint32_t num_rows;

    /** @brief Number of columns. */
    uint32_t num_cols;

    /** @brief Elements of the matrix. */
    int8_t** elts;
}
Matrix_i8;

/** @brief Matrix of signed 16-bit values. */
typedef struct
{
    /** @brief Number of rows. */
    uint32_t num_rows;

    /** @brief Number of columns. */
    uint32_t num_cols;

    /** @brief Elements of the matrix. */
    int16_t** elts;
}
Matrix_i16;

/** @brief Matrix of signed 32-bit values. */
typedef struct
{
    /** @brief Number of rows. */
    uint32_t num_rows;

    /** @brief Number of columns. */
    uint32_t num_cols;

    /** @brief Elements of the matrix. */
    int32_t** elts;
}
Matrix_i32;

/** @brief Matrix of signed 64-bit values. */
typedef struct
{
    /** @brief Number of rows. */
    uint32_t num_rows;

    /** @brief Number of columns. */
    uint32_t num_cols;

    /** @brief Elements of the matrix. */
    int64_t** elts;
}
Matrix_i64;


/** @brief Matrix of unsigned 8-bit values. */
typedef struct
{
    /** @brief Number of rows. */
    int32_t num_rows;

    /** @brief Number of columns. */
    int32_t num_cols;

    /** @brief Elements of the matrix. */
    uint8_t** elts;
}
Matrix_u8;

/** @brief Matrix of unsigned 16-bit values. */
typedef struct
{
    /** @brief Number of rows. */
    uint32_t num_rows;

    /** @brief Number of columns. */
    uint32_t num_cols;

    /** @brief Elements of the matrix. */
    uint16_t** elts;
}
Matrix_u16;

/** @brief Matrix of unsigned 32-bit values. */
typedef struct
{
    /** @brief Number of rows. */
    uint32_t num_rows;

    /** @brief Number of columns. */
    uint32_t num_cols;

    /** @brief Elements of the matrix. */
    uint32_t** elts;
}
Matrix_u32;

/** @brief Matrix of unsigned 64-bit values. */
typedef struct
{
    /** @brief Number of rows. */
    uint32_t num_rows;

    /** @brief Number of columns. */
    uint32_t num_cols;

    /** @brief Elements of the matrix. */
    uint64_t** elts;
}
Matrix_u64;

/** @brief Matrix of single precision floating point values. */
typedef struct
{
    /** @brief Number of rows. */
    uint32_t num_rows;

    /** @brief Number of columns. */
    uint32_t num_cols;

    /** @brief Elements of the matrix. */
    float** elts;
}
Matrix_f;

/** @brief Matrix of double precision floating point values. */
typedef struct
{
    /** @brief Number of rows. */
    uint32_t num_rows;

    /** @brief Number of columns. */
    uint32_t num_cols;

    /** @brief Elements of the matrix. */
    double** elts;
}
Matrix_d;

/** @brief Matrix of single precision complex floating point values. */
typedef struct
{
    /** @brief Number of rows. */
    uint32_t num_rows;

    /** @brief Number of columns. */
    uint32_t num_cols;

    /** @brief Elements of the matrix. */
    Complex_f** elts;
}
Matrix_cf;

/** @brief Matrix of double precision complex floating point values. */
typedef struct
{
    /** @brief Number of rows. */
    uint32_t num_rows;

    /** @brief Number of columns. */
    uint32_t num_cols;

    /** @brief Elements of the matrix. */
    Complex_d** elts;
}
Matrix_cd;


/** @brief Orientation of a vector in a matrix. */
typedef enum
{
    MATRIX_ROW_VECTOR,
    MATRIX_COL_VECTOR
}
Matrix_vector;




/**
 * @name create_matrix
 *
 * Creates a matrix of integer or floating point values.
 * @{
 */

/** @brief 32-bit unsigned integer. */
void create_matrix_u32
(
    Matrix_u32** m_out, 
    uint32_t     num_rows, 
    uint32_t     num_cols
);

/** @brief 32-bit signed integer. */
void create_matrix_i32
(
    Matrix_i32** m_out, 
    uint32_t     num_rows, 
    uint32_t     num_cols
);

/** @brief 64-bit signed integer. */
void create_matrix_i64
(
    Matrix_i64** m_out, 
    uint32_t     num_rows, 
    uint32_t     num_cols
);

/** @brief Single precision floating point. */
void create_matrix_f
(
    Matrix_f** m_out, 
    uint32_t   num_rows, 
    uint32_t   num_cols
);

/** @brief Double precision floating point. */
void create_matrix_d
(
    Matrix_d** m_out, 
    uint32_t   num_rows, 
    uint32_t   num_cols
);

/** @brief Single precision complex floating point. */
void create_matrix_cf
(
    Matrix_cf** m_out, 
    uint32_t    num_rows, 
    uint32_t    num_cols
);

/** @brief Double precision complexfloating point. */
void create_matrix_cd
(
    Matrix_cd** m_out, 
    uint32_t    num_rows, 
    uint32_t    num_cols
);

/*@}*/




/**
 * @name create_init_matrix
 *
 * Creates and initializes a matrix of integer or floating point values.
 * @{
 */

/** @brief 32-bit unsigned integer. */
void create_init_matrix_u32
(
    Matrix_u32** m_out, 
    uint32_t     num_rows, 
    uint32_t     num_cols,
    uint32_t     val
);

/** @brief 32-bit signed integer. */
void create_init_matrix_i32
(
    Matrix_i32** m_out, 
    uint32_t     num_rows, 
    uint32_t     num_cols,
    int32_t      val
);

/** @brief 64-bit signed integer. */
void create_init_matrix_i64
(
    Matrix_i64** m_out, 
    uint32_t     num_rows, 
    uint32_t     num_cols,
    int64_t      val
);

/** @brief Single precision floating point. */
void create_init_matrix_f
(
    Matrix_f** m_out, 
    uint32_t   num_rows, 
    uint32_t   num_cols,
    float      val
);

/** @brief Double precision floating point. */
void create_init_matrix_d
(
    Matrix_d** m_out, 
    uint32_t   num_rows, 
    uint32_t   num_cols,
    double     val
);

/** @brief Single precision complex floating point. */
void create_init_matrix_cf
(
    Matrix_cf** m_out, 
    uint32_t    num_rows, 
    uint32_t    num_cols,
    Complex_f   val
);

/** @brief Double precision complexfloating point. */
void create_init_matrix_cd
(
    Matrix_cd** m_out, 
    uint32_t    num_rows, 
    uint32_t    num_cols,
    Complex_d   val
);

/*@}*/




/**
 * @name create_identity_matrix
 *
 * Creates a matrix of integer or floating point values and initializes it with 
 * zeros and a diagonal of ones.
 * @{
 */

/** @brief 32-bit unsigned integer. */
void create_identity_matrix_u32(Matrix_u32** m_out, uint32_t size);

/** @brief 32-bit signed integer. */
void create_identity_matrix_i32(Matrix_i32** m_out, uint32_t size);

/** @brief 64-bit signed integer. */
void create_identity_matrix_i64(Matrix_i64** m_out, uint32_t size);

/** @brief Single precision floating point. */
void create_identity_matrix_f(Matrix_f** m_out, uint32_t size);

/** @brief Double precision floating point. */
void create_identity_matrix_d(Matrix_d** m_out, uint32_t size);

/*@}*/




/**
 * @name create_zero_matrix
 *
 * Creates a matrix of integer or floating point values and initializes it with 
 * zeros.
 * @{
 */

/** @brief 32-bit unsigned integer. */
void create_zero_matrix_u32
(
    Matrix_u32** m_out, 
    uint32_t     num_rows, 
    uint32_t     num_cols
);

/** @brief 32-bit signed integer. */
void create_zero_matrix_i32
(
    Matrix_i32** m_out, 
    uint32_t     num_rows, 
    uint32_t     num_cols
);

/** @brief 64-bit signed integer. */
void create_zero_matrix_i64
(
    Matrix_i64** m_out, 
    uint32_t     num_rows, 
    uint32_t     num_cols
);

/** @brief Single precision floating point. */
void create_zero_matrix_f
(
    Matrix_f** m_out, 
    uint32_t   num_rows, 
    uint32_t   num_cols
);

/** @brief Double precision floating point. */
void create_zero_matrix_d
(
    Matrix_d** m_out, 
    uint32_t   num_rows, 
    uint32_t   num_cols
);

/** @brief Single precision complex floating point. */
void create_zero_matrix_cf
(
    Matrix_cf** m_out, 
    uint32_t    num_rows, 
    uint32_t    num_cols
);

/** @brief Double precision complexfloating point. */
void create_zero_matrix_cd
(
    Matrix_cd** m_out, 
    uint32_t    num_rows, 
    uint32_t    num_cols
);

/*@}*/




/**
 * @name create_diagonal_matrix
 *
 * Creates a square matrix of integer or floating point values and initializes
 * its diagonal using values from a vector.
 * @{
 */

/** @brief 32-bit unsigned integer. */
void create_diagnonal_matrix_u32
(
    Matrix_u32**      m_out, 
    const Vector_u32* v
);

/** @brief 32-bit signed integer. */
void create_diagonal_matrix_i32
(
    Matrix_i32**      m_out, 
    const Vector_i32* v
);

/** @brief 64-bit signed integer. */
void create_diagonal_matrix_i64
(
    Matrix_i64**      m_out, 
    const Vector_i64* v
);

/** @brief Single precision floating point. */
void create_diagonal_matrix_f
(
    Matrix_f**      m_out, 
    const Vector_f* v
);

/** @brief Double precision floating point. */
void create_diagonal_matrix_d
(
    Matrix_d**      m_out, 
    const Vector_d* v
);

/** @brief Single precision complex floating point. */
void create_diagonal_matrix_cf
(
    Matrix_cf**      m_out, 
    const Vector_cf* v
);

/** @brief Double precision complexfloating point. */
void create_diagonal_matrix_cd
(
    Matrix_cd**      m_out, 
    const Vector_cd* v
);

/*@}*/




/**
 * @name create_random_matrix
 *
 * Creates a matrix of integer or floating point values and initializes it
 * with random values.
 * @{
 */

/** @brief 32-bit unsigned integer. */
void create_random_matrix_u32
(
    Matrix_u32** m_out, 
    uint32_t     num_rows, 
    uint32_t     num_cols,
    uint32_t     min,
    uint32_t     max
);

/** @brief 32-bit signed integer. */
void create_random_matrix_i32
(
    Matrix_i32** m_out, 
    uint32_t     num_rows, 
    uint32_t     num_cols,
    int32_t      min,
    int32_t      max
);

/** @brief 64-bit signed integer. */
void create_random_matrix_i64
(
    Matrix_i64** m_out, 
    uint32_t     num_rows, 
    uint32_t     num_cols,
    int64_t      min,
    int64_t      max
);

/** @brief Single precision floating point. */
void create_random_matrix_f
(
    Matrix_f** m_out, 
    uint32_t   num_rows, 
    uint32_t   num_cols,
    float      min,
    float      max
);

/** @brief Double precision floating point. */
void create_random_matrix_d
(
    Matrix_d** m_out, 
    uint32_t   num_rows, 
    uint32_t   num_cols,
    double     min,
    double     max
);

/** @brief Single precision complex floating point. */
void create_random_matrix_cf
(
    Matrix_cf** m_out, 
    uint32_t    num_rows, 
    uint32_t    num_cols,
    Complex_f   min,
    Complex_f   max
);

/** @brief Double precision complexfloating point. */
void create_random_matrix_cd
(
    Matrix_cd** m_out, 
    uint32_t    num_rows, 
    uint32_t    num_cols,
    Complex_d   min,
    Complex_d   max
);

/*@}*/




/**
 * @name copy_matrix
 *
 * Copies a matrix into another.
 * @{
 */

/** @brief 32-bit unsigned integer. */
Error* copy_matrix_u32(Matrix_u32** m_out, const Matrix_u32* m_in);

/** @brief 32-bit signed integer. */
Error* copy_matrix_i32(Matrix_i32** m_out, const Matrix_i32* m_in);

/** @brief 64-bit signed integer. */
Error* copy_matrix_i64(Matrix_i64** m_out, const Matrix_i64* m_in);

/** @brief Single precision floating point. */
Error* copy_matrix_f(Matrix_f** m_out, const Matrix_f* m_in);

/** @brief Double precision floating point. */
Error* copy_matrix_d(Matrix_d** m_out, const Matrix_d* m_in);

/** @brief Single precision complex floating point. */
Error* copy_matrix_cf(Matrix_cf** m_out, const Matrix_cf* m_in);

/** @brief Double precision complex floating point. */
Error* copy_matrix_cd(Matrix_cd** m_out, const Matrix_cd* m_in);

/*@}*/




/**
 * @name copy_matrix_block
 *
 * Copies a block of a matrix.
 * @{
 */

/** @brief Copies a block of a 32-bit unsigned integer matrix. */
Error* copy_matrix_block_u32
(
    Matrix_u32**      m_out, 
    const Matrix_u32* m_in,
    uint32_t          row_offset,
    uint32_t          col_offset,
    uint32_t          num_rows,
    uint32_t          num_cols
);

/** @brief Copies a block of a 32-bit signed integer matrix. */
Error* copy_matrix_block_i32
(
    Matrix_i32**      m_out, 
    const Matrix_i32* m_in,
    uint32_t          row_offset,
    uint32_t          col_offset,
    uint32_t          num_rows,
    uint32_t          num_cols
);

/** @brief Copies a block of a 64-bit signed integer matrix. */
Error* copy_matrix_block_i64
(
    Matrix_i64**      m_out, 
    const Matrix_i64* m_in,
    uint32_t          row_offset,
    uint32_t          col_offset,
    uint32_t          num_rows,
    uint32_t          num_cols
);

/** @brief Copies a block of a single precision floating point matrix. */
Error* copy_matrix_block_f
(
    Matrix_f**      m_out, 
    const Matrix_f* m_in,
    uint32_t        row_offset,
    uint32_t        col_offset,
    uint32_t        num_rows,
    uint32_t        num_cols
);

/** @brief Copies a block of a double precision floating point matrix. */
Error* copy_matrix_block_d
(
    Matrix_d**      m_out, 
    const Matrix_d* m_in,
    uint32_t        row_offset,
    uint32_t        col_offset,
    uint32_t        num_rows,
    uint32_t        num_cols
);

/** @brief Copies a block of a single precision complex floating point matrix.*/
Error* copy_matrix_block_cf
(
    Matrix_cf**      m_out, 
    const Matrix_cf* m_in,
    uint32_t         row_offset,
    uint32_t         col_offset,
    uint32_t         num_rows,
    uint32_t         num_cols
);

/** @brief Copies a block of a double precision complex floating point matrix.*/
Error* copy_matrix_block_cd
(
    Matrix_cd**      m_out, 
    const Matrix_cd* m_in,
    uint32_t         row_offset,
    uint32_t         col_offset,
    uint32_t         num_rows,
    uint32_t         num_cols
);

/*@}*/




/**
 * @name copy_matrix_block_into_matrix
 *
 * Copies a block of a matrix.
 * @{
 */

/** @brief Copies a block of a 32-bit unsigned integer matrix into another. */
Error* copy_matrix_block_into_matrix_u32
(
    Matrix_u32*       m_1, 
    uint32_t          row_offset_1,
    uint32_t          col_offset_1,
    const Matrix_u32* m_2,
    uint32_t          row_offset_2,
    uint32_t          col_offset_2,
    uint32_t          num_rows,
    uint32_t          num_cols
);

/** @brief Copies a block of a 32-bit signed integer matrix into another. */
Error* copy_matrix_block_into_matrix_i32
(
    Matrix_i32*       m_1, 
    uint32_t          row_offset_1,
    uint32_t          col_offset_1,
    const Matrix_i32* m_2,
    uint32_t          row_offset_2,
    uint32_t          col_offset_2,
    uint32_t          num_rows,
    uint32_t          num_cols
);

/** @brief Copies a block of a 64-bit signed integer matrix into another. */
Error* copy_matrix_block_into_matrix_i64
(
    Matrix_i64*       m_1, 
    uint32_t          row_offset_1,
    uint32_t          col_offset_1,
    const Matrix_i64* m_2,
    uint32_t          row_offset_2,
    uint32_t          col_offset_2,
    uint32_t          num_rows,
    uint32_t          num_cols
);

/** 
 * @brief Copies a block of a single precision floating point matrix into 
 * another.
 */
Error* copy_matrix_block_into_matrix_f
(
    Matrix_f*       m_1, 
    uint32_t        row_offset_1,
    uint32_t        col_offset_1,
    const Matrix_f* m_2,
    uint32_t        row_offset_2,
    uint32_t        col_offset_2,
    uint32_t        num_rows,
    uint32_t        num_cols
);

/** 
 * @brief Copies a block of a double precision floating point matrix into 
 * another.
 */
Error* copy_matrix_block_into_matrix_d
(
    Matrix_d*       m_1, 
    uint32_t        row_offset_1,
    uint32_t        col_offset_1,
    const Matrix_d* m_2,
    uint32_t        row_offset_2,
    uint32_t        col_offset_2,
    uint32_t        num_rows,
    uint32_t        num_cols
);

/** 
 * @brief Copies a block of a single precision complex floating point matrix 
 * into another.
 */
Error* copy_matrix_block_into_matrix_cf
(
    Matrix_cf*       m_1, 
    uint32_t         row_offset_1,
    uint32_t         col_offset_1,
    const Matrix_cf* m_2,
    uint32_t         row_offset_2,
    uint32_t         col_offset_2,
    uint32_t         num_rows,
    uint32_t         num_cols
);

/** 
 * @brief Copies a block of a double precision complex floating point matrix 
 * into another.
 */
Error* copy_matrix_block_into_matrix_cd
(
    Matrix_cd*       m_1, 
    uint32_t         row_offset_1,
    uint32_t         col_offset_1,
    const Matrix_cd* m_2,
    uint32_t         row_offset_2,
    uint32_t         col_offset_2,
    uint32_t         num_rows,
    uint32_t         num_cols
);

/*@}*/




/**
 * @name copy_vector_into_matrix
 *
 * Copies a vector into a matrix.
 * @{
 */

/** @brief Copies a 32-bit unsigned integer vector into a matrix. */
Error* copy_vector_into_matrix_u32
(
    Matrix_u32**      m_out,
    const Matrix_u32* m_in,
    const Vector_u32* v,
    Matrix_vector     orient,
    uint32_t          index
);

/** @brief Copies a 32-bit signed integer vector into a matrix. */
Error* copy_vector_into_matrix_i32
(
    Matrix_i32**      m_out,
    const Matrix_i32* m_in,
    const Vector_i32* v,
    Matrix_vector     orient,
    uint32_t          index
);

/** @brief Copies a 32-bit signed integer vector into a matrix. */
Error* copy_vector_into_matrix_i64
(
    Matrix_i64**      m_out,
    const Matrix_i64* m_in,
    const Vector_i64* v,
    Matrix_vector     orient,
    uint32_t          index
);

/** @brief Copies a 32-bit signed integer vector into a matrix. */
Error* copy_vector_into_matrix_f
(
    Matrix_f**      m_out,
    const Matrix_f* m_in,
    const Vector_f* v,
    Matrix_vector   orient,
    uint32_t        index
);

/** @brief Copies a 32-bit signed integer vector into a matrix. */
Error* copy_vector_into_matrix_d
(
    Matrix_d**      m_out,
    const Matrix_d* m_in,
    const Vector_d* v,
    Matrix_vector   orient,
    uint32_t        index
);

/** @brief Copies a 32-bit signed integer vector into a matrix. */
Error* copy_vector_into_matrix_cf
(
    Matrix_cf**      m_out,
    const Matrix_cf* m_in,
    const Vector_cf* v,
    Matrix_vector    orient,
    uint32_t         index
);

/** @brief Copies a 32-bit signed integer vector into a matrix. */
Error* copy_vector_into_matrix_cd
(
    Matrix_cd**      m_out,
    const Matrix_cd* m_in,
    const Vector_cd* v,
    Matrix_vector    orient,
    uint32_t         index
);

/*@}*/




/**
 * @name copy_vector_from_matrix
 *
 * Copies a vector from a matrix.
 * @{
 */

/** @brief Copies a 32-bit unsigned integer vector from a matrix. */
Error* copy_vector_from_matrix_u32
(
    Vector_u32**      v_out,
    const Matrix_u32* m,
    Matrix_vector     orient,
    uint32_t          index
);

/** @brief Copies a 32-bit signed integer vector from a matrix. */
Error* copy_vector_from_matrix_i32
(
    Vector_i32**      v_out,
    const Matrix_i32* m,
    Matrix_vector     orient,
    uint32_t          index
);

/** @brief Copies a 32-bit signed integer vector from a matrix. */
Error* copy_vector_from_matrix_i64
(
    Vector_i64**      v_out,
    const Matrix_i64* m,
    Matrix_vector     orient,
    uint32_t          index
);

/** @brief Copies a 32-bit signed integer vector from a matrix. */
Error* copy_vector_from_matrix_f
(
    Vector_f**      v_out,
    const Matrix_f* m,
    Matrix_vector   orient,
    uint32_t        index
);

/** @brief Copies a 32-bit signed integer vector from a matrix. */
Error* copy_vector_from_matrix_d
(
    Vector_d**      v_out,
    const Matrix_d* m,
    Matrix_vector   orient,
    uint32_t        index
);

/** @brief Copies a 32-bit signed integer vector from a matrix. */
Error* copy_vector_from_matrix_cf
(
    Vector_cf**      v_out,
    const Matrix_cf* m,
    Matrix_vector    orient,
    uint32_t         index
);

/** @brief Copies a 32-bit signed integer vector from a matrix. */
Error* copy_vector_from_matrix_cd
(
    Vector_cd**      v_out,
    const Matrix_cd* m,
    Matrix_vector    orient,
    uint32_t         index
);

/*@}*/




/**
 * @name free_matrix
 *
 * Frees a matrix.
 * @{
 */

/** @brief Frees a 32-bit unsigned integer matrix. */
void free_matrix_u32(Matrix_u32* m);

/** @brief Frees a 32-bit signed integer matrix. */
void free_matrix_i32(Matrix_i32* m);

/** @brief Frees a 64-bit signed integer matrix. */
void free_matrix_i64(Matrix_i64* m);

/** @brief Frees a single precision floating point matrix. */
void free_matrix_f(Matrix_f* m);

/** @brief Frees a double precision floating point matrix. */
void free_matrix_d(Matrix_d* m);

/** @brief Frees a single precision complex floating point matrix. */
void free_matrix_cf(Matrix_cf* m);

/** @brief Frees a double precision complex floating point matrix. */
void free_matrix_cd(Matrix_cd* m);

/*@}*/


#ifdef __cplusplus
}
}
#endif


#endif
