/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Declarations for matrix FFT functions.
 *
 * If FFTW3 was built to use SIMD instrinctions, the real to complex and 
 * complex to real dft routines are accelerated. For the double precision
 * routines, however, only SSE2 ISA extensions are supported.
 */


#ifndef MATRIX_FFT_H
#define MATRIX_FFT_H


#include <jwsc/config.h>

#include <stdlib.h>

#include "jwsc/matrix/matrix.h"


#ifdef __cplusplus
namespace jwsc {
extern "C" {
#endif


/**
 * @name real_to_complex_matrix_dft
 *
 * Computes a complex discrete Fourier transform from a real matrix.
 * @{
 */

/**
 * @brief Computes a complex discrete Fourier transform from a single
 * precision real matrix.
 */
#ifdef JWSC_HAVE_FFTW3F
void real_to_complex_matrix_dft_f
(
    Matrix_cf**     m_out,
    const Matrix_f* m_in
);
#endif

/**
 * @brief Computes a complex discrete Fourier transform from a double
 * precision real matrix.
 */
#ifdef JWSC_HAVE_FFTW3
void real_to_complex_matrix_dft_d
(
    Matrix_cd**     m_out,
    const Matrix_d* m_in
);
#endif

/*@}*/




/**
 * @name complex_to_real_matrix_dft
 *
 * Computes a real matrix from a complex discrete Fourier transform.
 * @{
 */

/** 
 * @brief Computes a single precision real matrix from a complex discrete 
 * Fourier transform
 */ 
#ifdef JWSC_HAVE_FFTW3F
void complex_to_real_matrix_dft_f
(
    Matrix_f**       m_out,
    const Matrix_cf* m_in,
    uint8_t          ncols_even
);
#endif

/** 
 * @brief Computes a double precision real matrix from a complex discrete 
 * Fourier transform
 */ 
#ifdef JWSC_HAVE_FFTW3
void complex_to_real_matrix_dft_d
(
    Matrix_d**       m_out,
    const Matrix_cd* m_in,
    uint8_t          ncols_even
);
#endif

/*@}*/




/**
 * @name expand_symmetric_matrix_dft
 *
 * Expands the symmetric columns of a dft.
 * @{
 */

/** @brief Expands the symmetric columns of a single precision dft. */
void expand_symmetric_matrix_dft_f
(
    Matrix_cf**      m_out, 
    const Matrix_cf* m_in, 
    uint8_t          ncols_even
);

/** @brief Expands the symmetric columns of a double precision dft. */
void expand_symmetric_matrix_dft_d
(
    Matrix_cd**      m_out, 
    const Matrix_cd* m_in, 
    uint8_t          ncols_even
);

/*@}*/




/**
 * @name shrink_symmetric_matrix_dft
 *
 * Shrinks the symmetric columns of a dft.
 * @{
 */

/** @brief Shrinks the symmetric columns of a single precision dft. */
Error* shrink_symmetric_matrix_dft_f(Matrix_cf** m_out, const Matrix_cf* m_in);

/**
 * @brief Shrinks the symmetric columns of a double precision dft.
 */
Error* shrink_symmetric_matrix_dft_d(Matrix_cd** m_out, const Matrix_cd* m_in);

/*@}*/




/**
 * @name scale_matrix_dft
 *
 * Scales the result of a real to complex to real dft operation.
 * @{
 */

/**
 * @brief Scales the single precision result of a real to complex to real dft 
 * operation.
 */
void scale_matrix_dft_f(Matrix_f** m_out, const Matrix_f* m_in);

/**
 * @brief Scales the double precision result of a real to complex to real dft 
 * operation.
 */
void scale_matrix_dft_d(Matrix_d** m_out, const Matrix_d* m_in);

/*@}*/




/**
 * @name shift_matrix
 *
 * Shifts a matrix so its diagonal quadrants are swapped.
 * @{
 */

/**
 * @brief Shifts a single precision matrix so its diagonal quadrants are 
 * swapped.
 */
void shift_matrix_f
(
    Matrix_f**       m_out,
    const Matrix_f*  m_in
);

/**
 * @brief Shifts a double precision matrix so its diagonal quadrants are 
 * swapped.
 */
void shift_matrix_d
(
    Matrix_d**       m_out,
    const Matrix_d*  m_in
);

/**
 * @brief Shifts a single precision complex matrix so its diagonal quadrants 
 * are swapped.
 */
void shift_matrix_cf
(
    Matrix_cf**       m_out,
    const Matrix_cf*  m_in
);

/**
 * @brief Shifts a double precision complex matrix so its diagonal quadrants 
 * are swapped.
 */
void shift_matrix_cd
(
    Matrix_cd**       m_out,
    const Matrix_cd*  m_in
);

/*@}*/




/**
 * @name unshift_matrix
 *
 * Unshifts a matrix that had its diagonal quadrants swapped.
 * @{
 */

/**
 * @brief Unshifts a single precision matrix that had its diagonal quadrants 
 * swapped.
 */
void unshift_matrix_f
(
    Matrix_f**       m_out,
    const Matrix_f*  m_in
);

/**
 * @brief Unshifts a double precision matrix that had its diagonal quadrants 
 * swapped.
 */
void unshift_matrix_d
(
    Matrix_d**       m_out,
    const Matrix_d*  m_in
);

/**
 * @brief Unshifts a single precision complex matrix that had its diagonal 
 * quadrants swapped.
 */
void unshift_matrix_cf
(
    Matrix_cf**       m_out,
    const Matrix_cf*  m_in
);

/**
 * @brief Unshifts a double precision complex matrix that had its diagonal 
 * quadrants swapped.
 */
void unshift_matrix_cd
(
    Matrix_cd**       m_out,
    const Matrix_cd*  m_in
);

/*@}*/


#ifdef __cplusplus
}
}
#endif


#endif
