/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Definitions for I/O operations on a Matrix container.
 */


#include <jwsc/config.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <inttypes.h>
#include <assert.h>
#include <ctype.h>

#include "jwsc/base/error.h"
#include "jwsc/base/file_io.h"
#include "jwsc/math/complex.h"
#include "jwsc/matrix/matrix.h"
#include "jwsc/matrix/matrix_io.h"

#ifdef JWSC_HAVE_DMALLOC
#include <dmalloc.h>
#endif


/** @brief Reads a line of a matrix from a file pointer. */
static int read_matrix_line(char** line, FILE* fp)
{
    static const uint32_t block_size = 1024;

    uint32_t num_elts;
    uint32_t blocks;

    char c;

    num_elts = 0;
    blocks   = 0;

    while ((c = fgetc(fp)) != EOF && c != '\n') 
    {
        if (num_elts % block_size == 0)
        {
            *line = realloc(*line, (++blocks)*block_size*sizeof(char));
            assert(*line);
        }

        (*line)[ num_elts++ ] = c;
    }

    if (num_elts % block_size == 0)
    {
        *line = realloc(*line, (blocks*block_size+1)*sizeof(char));
        assert(*line);
    }
    (*line)[ num_elts ] = '\0';

    return (c == EOF);
}


/** @brief Counts the number of space-separated matrix columns in a line. */
static uint32_t count_matrix_line_columns(const char* line)
{
    uint32_t count;
    uint32_t done;
    uint32_t in_column;

    const char* c;

    count     = 0;
    done      = 0;
    in_column = 0;

    c = line;

    while(!done)
    {
        if (*c == '\0')
        {
            done = 1;
        }
        else if (isspace(*c))
        {
            in_column = 0;
        }
        else if (!in_column)
        {
            count++;
            in_column = 1;
        }
        c++;
    }

    return count;
}


/** @brief Increments a pointer to a matrix line to the next column. */
static const char* advance_line_column_ptr(const char* col)
{
    while (*col != ' ' && *col != '\0') col++;

    while (*col == ' ' && *col != '\0') col++;

    return col;
}




/**
 * @name read_matrix
 *
 * Reads a matrix from a file.
 * @{
 */

/**
 * Format of the matrix in the file is expected to be sequential lines 
 * containing rows of the matrix. Blank lines and lines beginning with '#'  
 * prior to the first row of the matrix are skipped.  For example,
 *
 * @code
 * # Comment.
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m_out   Result parameter. If @em *m_out is NULL, a matrix is
 *                 allocated; otherwise its space is re-used.
 * @param  fname   Name of the file to read the matrix from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *m_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the matrix in the file is 
 *   not formatted properly.
 */
Error* read_matrix_u32(Matrix_u32** m_out, const char* fname)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "r")) == NULL)
    {
        free_matrix_u32(*m_out); *m_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    if ((e = read_matrix_fp_u32(m_out, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        free_matrix_u32(*m_out); *m_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the matrix in the file is expected to be sequential lines 
 * containing rows of the matrix. Blank lines and lines beginning with '#'  
 * prior to the first row of the matrix are skipped.  For example,
 *
 * @code
 * # Comment.
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m_out   Result parameter. If @em *m_out is NULL, a matrix is
 *                 allocated; otherwise its space is re-used.
 * @param  fname   Name of the file to read the matrix from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *m_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the matrix in the file is 
 *   not formatted properly.
 */
Error* read_matrix_i32(Matrix_i32** m_out, const char* fname)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "r")) == NULL)
    {
        free_matrix_i32(*m_out); *m_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    if ((e = read_matrix_fp_i32(m_out, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        free_matrix_i32(*m_out); *m_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the matrix in the file is expected to be sequential lines 
 * containing rows of the matrix. Blank lines and lines beginning with '#'  
 * prior to the first row of the matrix are skipped.  For example,
 *
 * @code
 * # Comment.
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m_out   Result parameter. If @em *m_out is NULL, a matrix is
 *                 allocated; otherwise its space is re-used.
 * @param  fname   Name of the file to read the matrix from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *m_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the matrix in the file is 
 *   not formatted properly.
 */
Error* read_matrix_i64(Matrix_i64** m_out, const char* fname)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "r")) == NULL)
    {
        free_matrix_i64(*m_out); *m_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    if ((e = read_matrix_fp_i64(m_out, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        free_matrix_i64(*m_out); *m_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the matrix in the file is expected to be sequential lines 
 * containing rows of the matrix. Blank lines and lines beginning with '#'  
 * prior to the first row of the matrix are skipped.  For example,
 *
 * @code
 * # Comment.
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m_out   Result parameter. If @em *m_out is NULL, a matrix is
 *                 allocated; otherwise its space is re-used.
 * @param  fname   Name of the file to read the matrix from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *m_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the matrix in the file is 
 *   not formatted properly.
 */
Error* read_matrix_f(Matrix_f** m_out, const char* fname)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "r")) == NULL)
    {
        free_matrix_f(*m_out); *m_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    if ((e = read_matrix_fp_f(m_out, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        free_matrix_f(*m_out); *m_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the matrix in the file is expected to be sequential lines 
 * containing rows of the matrix. Blank lines and lines beginning with '#'  
 * prior to the first row of the matrix are skipped.  For example,
 *
 * @code
 * # Comment.
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m_out   Result parameter. If @em *m_out is NULL, a matrix is
 *                 allocated; otherwise its space is re-used.
 * @param  fname   Name of the file to read the matrix from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *m_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the matrix in the file is 
 *   not formatted properly.
 */
Error* read_matrix_d(Matrix_d** m_out, const char* fname)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "r")) == NULL)
    {
        free_matrix_d(*m_out); *m_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    if ((e = read_matrix_fp_d(m_out, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        free_matrix_d(*m_out); *m_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the matrix in the file is expected to be sequential lines 
 * containing rows of the matrix. Blank lines and lines beginning with '#'  
 * prior to the first row of the matrix are skipped.  For example,
 *
 * @code
 * # Comment.
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m_out   Result parameter. If @em *m_out is NULL, a matrix is
 *                 allocated; otherwise its space is re-used.
 * @param  fname   Name of the file to read the matrix from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *m_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the matrix in the file is 
 *   not formatted properly.
 */
Error* read_matrix_cf(Matrix_cf** m_out, const char* fname)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "r")) == NULL)
    {
        free_matrix_cf(*m_out); *m_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    if ((e = read_matrix_fp_cf(m_out, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        free_matrix_cf(*m_out); *m_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the matrix in the file is expected to be sequential lines 
 * containing rows of the matrix. Blank lines and lines beginning with '#'  
 * prior to the first row of the matrix are skipped.  For example,
 *
 * @code
 * # Comment.
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m_out   Result parameter. If @em *m_out is NULL, a matrix is
 *                 allocated; otherwise its space is re-used.
 * @param  fname   Name of the file to read the matrix from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *m_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the matrix in the file is 
 *   not formatted properly.
 */
Error* read_matrix_cd(Matrix_cd** m_out, const char* fname)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "r")) == NULL)
    {
        free_matrix_cd(*m_out); *m_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    if ((e = read_matrix_fp_cd(m_out, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        free_matrix_cd(*m_out); *m_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/*@}*/




/**
 * @name read_matrix_fp
 *
 * Reads a matrix from an opened file.
 * @{
 */

/**
 * Format of the matrix in the file is expected to be sequential lines 
 * containing rows of the matrix. Blank lines and lines beginning with '#'  
 * prior to the first row of the matrix are skipped.  For example,
 *
 * @code
 * # Comment.
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m_out   Result parameter. If @em *m_out is NULL, a matrix is
 *                 allocated; otherwise its space is re-used.
 * @param  fp      File to read the matrix from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *m_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the matrix in the file is 
 *   not formatted properly.
 */
Error* read_matrix_fp_u32(Matrix_u32** m_out, FILE* fp)
{
    static const uint32_t block_size = 1024;

    uint32_t num_elts, num_rows, num_cols;
    uint32_t elt, col;
    uint32_t blocks;
    uint32_t done;

    char* line = NULL;
    const char* line_col;

    uint32_t  val;
    uint32_t* elts   = NULL;
    uint32_t* m_elts = NULL;

    if (skip_fp_spaces_and_comments(fp) == 0)
    {
        free_matrix_u32(*m_out); *m_out = NULL;
        return JWSC_EIO("Nothing to read");
    }

    num_elts = 0;
    num_cols = 0;
    num_rows = 0;
    blocks   = 0;
    done     = 0;

    while (!done)
    {
        read_matrix_line(&line, fp);

        if (num_cols == 0)
        {
            if ((num_cols = count_matrix_line_columns(line)) == 0)
            {
                free(line);
                free_matrix_u32(*m_out); *m_out = NULL;
                return JWSC_EIO("Matrix to read not formatted properly");
            }
        }
        else
        {
            if (count_matrix_line_columns(line) != num_cols)
            {
                free(line);
                free(elts);
                free_matrix_u32(*m_out); *m_out = NULL;
                return JWSC_EIO("Matrix to read not formatted properly");
            }
        }

        line_col = line;

        for (col = 0; col < num_cols; col++)
        {
            if (sscanf(line_col, "%u", &val) != 1)
            {
                free(line);
                free(elts);
                free_matrix_u32(*m_out); *m_out = NULL;
                return JWSC_EIO("Matrix to read not formatted properly");
            }

            line_col = advance_line_column_ptr(line_col);

            if (num_elts % block_size == 0)
            {
                elts = realloc(elts, (++blocks)*block_size*
                        sizeof(uint32_t));
                assert(elts);
            }

            elts[ num_elts++ ] = val;
        }

        num_rows++;

        done = !skip_fp_spaces_and_comments(fp);
    }

    create_matrix_u32(m_out, num_rows, num_cols);
    m_elts = *((*m_out)->elts);

    for (elt = 0; elt < num_elts; elt++)
    {
        m_elts[ elt ] = elts[ elt ];
    }

    free(line);
    free(elts);

    return NULL;
}

/**
 * Format of the matrix in the file is expected to be sequential lines 
 * containing rows of the matrix. Blank lines and lines beginning with '#'  
 * prior to the first row of the matrix are skipped.  For example,
 *
 * @code
 * # Comment.
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m_out   Result parameter. If @em *m_out is NULL, a matrix is
 *                 allocated; otherwise its space is re-used.
 * @param  fp      File to read the matrix from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *m_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the matrix in the file is 
 *   not formatted properly.
 */
Error* read_matrix_fp_i32(Matrix_i32** m_out, FILE* fp)
{
    static const uint32_t block_size = 1024;

    uint32_t num_elts, num_rows, num_cols;
    uint32_t elt, col;
    uint32_t blocks;
    uint32_t done;

    char* line = NULL;
    const char* line_col;

    int32_t  val;
    int32_t* elts   = NULL;
    int32_t* m_elts = NULL;

    if (skip_fp_spaces_and_comments(fp) == 0)
    {
        free_matrix_i32(*m_out); *m_out = NULL;
        return JWSC_EIO("Nothing to read");
    }

    num_elts = 0;
    num_cols = 0;
    num_rows = 0;
    blocks   = 0;
    done     = 0;

    while (!done)
    {
        read_matrix_line(&line, fp);

        if (num_cols == 0)
        {
            if ((num_cols = count_matrix_line_columns(line)) == 0)
            {
                free(line);
                free_matrix_i32(*m_out); *m_out = NULL;
                return JWSC_EIO("Matrix to read not formatted properly");
            }
        }
        else
        {
            if (count_matrix_line_columns(line) != num_cols)
            {
                free(line);
                free(elts);
                free_matrix_i32(*m_out); *m_out = NULL;
                return JWSC_EIO("Matrix to read not formatted properly");
            }
        }

        line_col = line;

        for (col = 0; col < num_cols; col++)
        {
            if (sscanf(line_col, "%d", &val) != 1)
            {
                free(line);
                free(elts);
                free_matrix_i32(*m_out); *m_out = NULL;
                return JWSC_EIO("Matrix to read not formatted properly");
            }

            line_col = advance_line_column_ptr(line_col);

            if (num_elts % block_size == 0)
            {
                elts = realloc(elts, (++blocks)*block_size*sizeof(int32_t));
                assert(elts);
            }

            elts[ num_elts++ ] = val;
        }

        num_rows++;

        done = !skip_fp_spaces_and_comments(fp);
    }

    create_matrix_i32(m_out, num_rows, num_cols);
    m_elts = *((*m_out)->elts);

    for (elt = 0; elt < num_elts; elt++)
    {
        m_elts[ elt ] = elts[ elt ];
    }

    free(line);
    free(elts);

    return NULL;
}

/**
 * Format of the matrix in the file is expected to be sequential lines 
 * containing rows of the matrix. Blank lines and lines beginning with '#'  
 * prior to the first row of the matrix are skipped.  For example,
 *
 * @code
 * # Comment.
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m_out   Result parameter. If @em *m_out is NULL, a matrix is
 *                 allocated; otherwise its space is re-used.
 * @param  fp      File to read the matrix from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *m_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the matrix in the file is 
 *   not formatted properly.
 */
Error* read_matrix_fp_i64(Matrix_i64** m_out, FILE* fp)
{
    static const uint32_t block_size = 1024;

    uint32_t num_elts, num_rows, num_cols;
    uint32_t elt, col;
    uint32_t blocks;
    uint32_t done;

    char* line = NULL;
    const char* line_col;

    int64_t  val;
    int64_t* elts   = NULL;
    int64_t* m_elts = NULL;

    if (skip_fp_spaces_and_comments(fp) == 0)
    {
        free_matrix_i64(*m_out); *m_out = NULL;
        return JWSC_EIO("Nothing to read");
    }

    num_elts = 0;
    num_cols = 0;
    num_rows = 0;
    blocks   = 0;
    done     = 0;

    while (!done)
    {
        read_matrix_line(&line, fp);

        if (num_cols == 0)
        {
            if ((num_cols = count_matrix_line_columns(line)) == 0)
            {
                free(line);
                free_matrix_i64(*m_out); *m_out = NULL;
                return JWSC_EIO("Matrix to read not formatted properly");
            }
        }
        else
        {
            if (count_matrix_line_columns(line) != num_cols)
            {
                free(line);
                free(elts);
                free_matrix_i64(*m_out); *m_out = NULL;
                return JWSC_EIO("Matrix to read not formatted properly");
            }
        }

        line_col = line;

        for (col = 0; col < num_cols; col++)
        {
            if (sscanf(line_col, "%" SCNd64, &val) != 1)
            {
                free(line);
                free(elts);
                free_matrix_i64(*m_out); *m_out = NULL;
                return JWSC_EIO("Matrix to read not formatted properly");
            }

            line_col = advance_line_column_ptr(line_col);

            if (num_elts % block_size == 0)
            {
                elts = realloc(elts, (++blocks)*block_size*sizeof(int64_t));
                assert(elts);
            }

            elts[ num_elts++ ] = val;
        }

        num_rows++;

        done = !skip_fp_spaces_and_comments(fp);
    }

    create_matrix_i64(m_out, num_rows, num_cols);
    m_elts = *((*m_out)->elts);

    for (elt = 0; elt < num_elts; elt++)
    {
        m_elts[ elt ] = elts[ elt ];
    }

    free(line);
    free(elts);

    return NULL;
}

/**
 * Format of the matrix in the file is expected to be sequential lines 
 * containing rows of the matrix. Blank lines and lines beginning with '#'  
 * prior to the first row of the matrix are skipped.  For example,
 *
 * @code
 * # Comment.
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m_out   Result parameter. If @em *m_out is NULL, a matrix is
 *                 allocated; otherwise its space is re-used.
 * @param  fp      File to read the matrix from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *m_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the matrix in the file is 
 *   not formatted properly.
 */
Error* read_matrix_fp_f(Matrix_f** m_out, FILE* fp)
{
    static const uint32_t block_size = 1024;

    uint32_t num_elts, num_rows, num_cols;
    uint32_t elt, col;
    uint32_t blocks;
    uint32_t done;

    char* line = NULL;
    const char* line_col;

    float  val;
    float* elts   = NULL;
    float* m_elts = NULL;

    if (skip_fp_spaces_and_comments(fp) == 0)
    {
        free_matrix_f(*m_out); *m_out = NULL;
        return JWSC_EIO("Nothing to read");
    }

    num_elts = 0;
    num_cols = 0;
    num_rows = 0;
    blocks   = 0;
    done     = 0;

    while (!done)
    {
        read_matrix_line(&line, fp);

        if (num_cols == 0)
        {
            if ((num_cols = count_matrix_line_columns(line)) == 0)
            {
                free(line);
                free_matrix_f(*m_out); *m_out = NULL;
                return JWSC_EIO("Matrix to read not formatted properly");
            }
        }
        else
        {
            if (count_matrix_line_columns(line) != num_cols)
            {
                free(line);
                free(elts);
                free_matrix_f(*m_out); *m_out = NULL;
                return JWSC_EIO("Matrix to read not formatted properly");
            }
        }

        line_col = line;

        for (col = 0; col < num_cols; col++)
        {
            if (sscanf(line_col, "%f", &val) != 1)
            {
                free(line);
                free(elts);
                free_matrix_f(*m_out); *m_out = NULL;
                return JWSC_EIO("Matrix to read not formatted properly");
            }

            line_col = advance_line_column_ptr(line_col);

            if (num_elts % block_size == 0)
            {
                elts = realloc(elts, (++blocks)*block_size*sizeof(float));
                assert(elts);
            }

            elts[ num_elts++ ] = val;
        }

        num_rows++;

        done = !skip_fp_spaces_and_comments(fp);
    }

    create_matrix_f(m_out, num_rows, num_cols);
    m_elts = *((*m_out)->elts);

    for (elt = 0; elt < num_elts; elt++)
    {
        m_elts[ elt ] = elts[ elt ];
    }

    free(line);
    free(elts);

    return NULL;
}

/**
 * Format of the matrix in the file is expected to be sequential lines 
 * containing rows of the matrix. Blank lines and lines beginning with '#'  
 * prior to the first row of the matrix are skipped.  For example,
 *
 * @code
 * # Comment.
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m_out   Result parameter. If @em *m_out is NULL, a matrix is
 *                 allocated; otherwise its space is re-used.
 * @param  fp      File to read the matrix from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *m_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the matrix in the file is 
 *   not formatted properly.
 */
Error* read_matrix_fp_d(Matrix_d** m_out, FILE* fp)
{
    static const uint32_t block_size = 1024;

    uint32_t num_elts, num_rows, num_cols;
    uint32_t elt, col;
    uint32_t blocks;
    uint32_t done;

    char* line = NULL;
    const char* line_col;

    double  val;
    double* elts   = NULL;
    double* m_elts = NULL;

    if (skip_fp_spaces_and_comments(fp) == 0)
    {
        free_matrix_d(*m_out); *m_out = NULL;
        return JWSC_EIO("Nothing to read");
    }

    num_elts = 0;
    num_cols = 0;
    num_rows = 0;
    blocks   = 0;
    done     = 0;

    while (!done)
    {
        read_matrix_line(&line, fp);

        if (num_cols == 0)
        {
            if ((num_cols = count_matrix_line_columns(line)) == 0)
            {
                free(line);
                free_matrix_d(*m_out); *m_out = NULL;
                return JWSC_EIO("Matrix to read not formatted properly");
            }
        }
        else
        {
            if (count_matrix_line_columns(line) != num_cols)
            {
                free(line);
                free(elts);
                free_matrix_d(*m_out); *m_out = NULL;
                return JWSC_EIO("Matrix to read not formatted properly");
            }
        }

        line_col = line;

        for (col = 0; col < num_cols; col++)
        {
            if (sscanf(line_col, "%lf", &val) != 1)
            {
                free(line);
                free(elts);
                free_matrix_d(*m_out); *m_out = NULL;
                return JWSC_EIO("Matrix to read not formatted properly");
            }

            line_col = advance_line_column_ptr(line_col);

            if (num_elts % block_size == 0)
            {
                elts = realloc(elts, (++blocks)*block_size*sizeof(double));
                assert(elts);
            }

            elts[ num_elts++ ] = val;
        }

        num_rows++;

        done = !skip_fp_spaces_and_comments(fp);
    }

    create_matrix_d(m_out, num_rows, num_cols);
    m_elts = *((*m_out)->elts);

    for (elt = 0; elt < num_elts; elt++)
    {
        m_elts[ elt ] = elts[ elt ];
    }

    free(line);
    free(elts);

    return NULL;
}

/**
 * Format of the matrix in the file is expected to be sequential lines 
 * containing rows of the matrix. Blank lines and lines beginning with '#'  
 * prior to the first row of the matrix are skipped.  For example,
 *
 * @code
 * # Comment.
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m_out   Result parameter. If @em *m_out is NULL, a matrix is
 *                 allocated; otherwise its space is re-used.
 * @param  fp      File to read the matrix from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *m_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the matrix in the file is 
 *   not formatted properly.
 */
Error* read_matrix_fp_cf(Matrix_cf** m_out, FILE* fp)
{
    static const uint32_t block_size = 1024;

    uint32_t num_elts, num_rows, num_cols;
    uint32_t elt, col;
    uint32_t blocks;
    uint32_t done;

    char* line = NULL;
    const char* line_col;

    Complex_f  val;
    Complex_f* elts   = NULL;
    Complex_f* m_elts = NULL;

    if (skip_fp_spaces_and_comments(fp) == 0)
    {
        free_matrix_cf(*m_out); *m_out = NULL;
        return JWSC_EIO("Nothing to read");
    }

    num_elts = 0;
    num_cols = 0;
    num_rows = 0;
    blocks   = 0;
    done     = 0;

    while (!done)
    {
        read_matrix_line(&line, fp);

        if (num_cols == 0)
        {
            if ((num_cols = count_matrix_line_columns(line)) == 0)
            {
                free(line);
                free_matrix_cf(*m_out); *m_out = NULL;
                return JWSC_EIO("Matrix to read not formatted properly");
            }
            num_cols /= 2;
        }
        else
        {
            if (count_matrix_line_columns(line) != num_cols*2)
            {
                free(line);
                free(elts);
                free_matrix_cf(*m_out); *m_out = NULL;
                return JWSC_EIO("Matrix to read not formatted properly");
            }
        }

        line_col = line;

        for (col = 0; col < num_cols; col++)
        {
            if (sscanf(line_col, "%f %f", &(val.r), &(val.i)) != 2)
            {
                free(line);
                free(elts);
                free_matrix_cf(*m_out); *m_out = NULL;
                return JWSC_EIO("Matrix to read not formatted properly");
            }

            line_col = advance_line_column_ptr(line_col);
            line_col = advance_line_column_ptr(line_col);

            if (num_elts % block_size == 0)
            {
                elts = realloc(elts, (++blocks)*block_size*
                        sizeof(Complex_f));
                assert(elts);
            }

            elts[ num_elts++ ] = val;
        }

        num_rows++;

        done = !skip_fp_spaces_and_comments(fp);
    }

    create_matrix_cf(m_out, num_rows, num_cols);
    m_elts = *((*m_out)->elts);

    for (elt = 0; elt < num_elts; elt++)
    {
        m_elts[ elt ] = elts[ elt ];
    }

    free(line);
    free(elts);

    return NULL;
}

/**
 * Format of the matrix in the file is expected to be sequential lines 
 * containing rows of the matrix. Blank lines and lines beginning with '#'  
 * prior to the first row of the matrix are skipped.  For example,
 *
 * @code
 * # Comment.
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m_out   Result parameter. If @em *m_out is NULL, a matrix is
 *                 allocated; otherwise its space is re-used.
 * @param  fp      File to read the matrix from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *m_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the matrix in the file is 
 *   not formatted properly.
 */
Error* read_matrix_fp_cd(Matrix_cd** m_out, FILE* fp)
{
    static const uint32_t block_size = 1024;

    uint32_t num_elts, num_rows, num_cols;
    uint32_t elt, col;
    uint32_t blocks;
    uint32_t done;

    char* line = NULL;
    const char* line_col;

    Complex_d  val;
    Complex_d* elts   = NULL;
    Complex_d* m_elts = NULL;

    if (skip_fp_spaces_and_comments(fp) == 0)
    {
        free_matrix_cd(*m_out); *m_out = NULL;
        return JWSC_EIO("Nothing to read");
    }

    num_elts = 0;
    num_cols = 0;
    num_rows = 0;
    blocks   = 0;
    done     = 0;

    while (!done)
    {
        read_matrix_line(&line, fp);

        if (num_cols == 0)
        {
            if ((num_cols = count_matrix_line_columns(line)) == 0)
            {
                free(line);
                free_matrix_cd(*m_out); *m_out = NULL;
                return JWSC_EIO("Matrix to read not formatted properly");
            }
            num_cols /= 2;
        }
        else
        {
            if (count_matrix_line_columns(line) != num_cols*2)
            {
                free(line);
                free(elts);
                free_matrix_cd(*m_out); *m_out = NULL;
                return JWSC_EIO("Matrix to read not formatted properly");
            }
        }

        line_col = line;

        for (col = 0; col < num_cols; col++)
        {
            if (sscanf(line_col, "%lf %lf", &(val.r), &(val.i)) != 2)
            {
                free(line);
                free(elts);
                free_matrix_cd(*m_out); *m_out = NULL;
                return JWSC_EIO("Matrix to read not formatted properly");
            }

            line_col = advance_line_column_ptr(line_col);
            line_col = advance_line_column_ptr(line_col);

            if (num_elts % block_size == 0)
            {
                elts = realloc(elts, (++blocks)*block_size*
                        sizeof(Complex_d));
                assert(elts);
            }

            elts[ num_elts++ ] = val;
        }

        num_rows++;

        done = !skip_fp_spaces_and_comments(fp);
    }

    create_matrix_cd(m_out, num_rows, num_cols);
    m_elts = *((*m_out)->elts);

    for (elt = 0; elt < num_elts; elt++)
    {
        m_elts[ elt ] = elts[ elt ];
    }

    free(line);
    free(elts);

    return NULL;
}

/*@}*/




/**
 * @name write_matrix
 *
 * Writes a matrix to a file.
 * @{
 */

/**
 * Format of the matrix written is sequential lines containing rows of the 
 * Matrix.
 *
 * @code
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m      Matrix to write.
 * @param  fname  Name of the file to write the matrix to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_matrix_u32(const Matrix_u32* m, const char* fname)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return JWSC_EIO(strerror(errno));
    }

    if ((e = write_matrix_fp_u32(m, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the matrix written is sequential lines containing rows of the 
 * Matrix.
 *
 * @code
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m      Matrix to write.
 * @param  fname  Name of the file to write the matrix to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_matrix_i32(const Matrix_i32* m, const char* fname)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return JWSC_EIO(strerror(errno));
    }

    if ((e = write_matrix_fp_i32(m, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the matrix written is sequential lines containing rows of the 
 * Matrix.
 *
 * @code
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m      Matrix to write.
 * @param  fname  Name of the file to write the matrix to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_matrix_i64(const Matrix_i64* m, const char* fname)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return JWSC_EIO(strerror(errno));
    }

    if ((e = write_matrix_fp_i64(m, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the matrix written is sequential lines containing rows of the 
 * Matrix.
 *
 * @code
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m      Matrix to write.
 * @param  fname  Name of the file to write the matrix to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_matrix_f(const Matrix_f* m, const char* fname)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return JWSC_EIO(strerror(errno));
    }

    if ((e = write_matrix_fp_f(m, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the matrix written is sequential lines containing rows of the 
 * Matrix.
 *
 * @code
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m      Matrix to write.
 * @param  fname  Name of the file to write the matrix to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_matrix_d(const Matrix_d* m, const char* fname)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return JWSC_EIO(strerror(errno));
    }

    if ((e = write_matrix_fp_d(m, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the matrix written is sequential lines containing rows of the 
 * Matrix.
 *
 * @code
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m      Matrix to write.
 * @param  fname  Name of the file to write the matrix to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_matrix_cf(const Matrix_cf* m, const char* fname)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return JWSC_EIO(strerror(errno));
    }

    if ((e = write_matrix_fp_cf(m, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the matrix written is sequential lines containing rows of the 
 * Matrix.
 *
 * @code
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m      Matrix to write.
 * @param  fname  Name of the file to write the matrix to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_matrix_cd(const Matrix_cd* m, const char* fname)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return JWSC_EIO(strerror(errno));
    }

    if ((e = write_matrix_fp_cd(m, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/*@}*/




/**
 * @name write_matrix_fp
 *
 * Writes a matrix to an opened file.
 * @{
 */

/**
 * Format of the matrix written is sequential lines containing rows of the 
 * matrix.
 *
 * @code
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m   Matrix to write.
 * @param  fp  File to write the matrix to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_matrix_fp_u32(const Matrix_u32* m, FILE* fp)
{
    uint32_t num_rows, num_cols;
    uint32_t row, col;

    num_rows = m->num_rows;
    num_cols = m->num_cols;
    for (row = 0; row < num_rows; row++)
    {
        for (col = 0; col < num_cols; col++)
        {
            fprintf(fp, "%u ", m->elts[ row ][ col ]);
        }
        fprintf(fp, "\n");
    }

    return NULL;
}

/**
 * Format of the matrix written is sequential lines containing rows of the 
 * matrix.
 *
 * @code
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m   Matrix to write.
 * @param  fp  File to write the matrix to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_matrix_fp_i32(const Matrix_i32* m, FILE* fp)
{
    uint32_t num_rows, num_cols;
    uint32_t row, col;

    num_rows = m->num_rows;
    num_cols = m->num_cols;
    for (row = 0; row < num_rows; row++)
    {
        for (col = 0; col < num_cols; col++)
        {
            fprintf(fp, "%d ", m->elts[ row ][ col ]);
        }
        fprintf(fp, "\n");
    }

    return NULL;
}

/**
 * Format of the matrix written is sequential lines containing rows of the 
 * matrix.
 *
 * @code
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m   Matrix to write.
 * @param  fp  File to write the matrix to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_matrix_fp_i64(const Matrix_i64* m, FILE* fp)
{
    uint32_t num_rows, num_cols;
    uint32_t row, col;

    num_rows = m->num_rows;
    num_cols = m->num_cols;
    for (row = 0; row < num_rows; row++)
    {
        for (col = 0; col < num_cols; col++)
        {
            fprintf(fp, "%" PRId64 " ", m->elts[ row ][ col ]);
        }
        fprintf(fp, "\n");
    }

    return NULL;
}

/**
 * Format of the matrix written is sequential lines containing rows of the 
 * matrix.
 *
 * @code
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m   Matrix to write.
 * @param  fp  File to write the matrix to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_matrix_fp_f(const Matrix_f* m, FILE* fp)
{
    uint32_t num_rows, num_cols;
    uint32_t row, col;

    num_rows = m->num_rows;
    num_cols = m->num_cols;
    for (row = 0; row < num_rows; row++)
    {
        for (col = 0; col < num_cols; col++)
        {
            fprintf(fp, "%f ", m->elts[ row ][ col ]);
        }
        fprintf(fp, "\n");
    }

    return NULL;
}

/**
 * Format of the matrix written is sequential lines containing rows of the 
 * matrix.
 *
 * @code
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m   Matrix to write.
 * @param  fp  File to write the matrix to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_matrix_fp_d(const Matrix_d* m, FILE* fp)
{
    uint32_t num_rows, num_cols;
    uint32_t row, col;

    num_rows = m->num_rows;
    num_cols = m->num_cols;
    for (row = 0; row < num_rows; row++)
    {
        for (col = 0; col < num_cols; col++)
        {
            fprintf(fp, "%.8e ", m->elts[ row ][ col ]);
        }
        fprintf(fp, "\n");
    }

    return NULL;
}

/**
 * Format of the matrix written is sequential lines containing rows of the 
 * matrix.
 *
 * @code
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m   Matrix to write.
 * @param  fp  File to write the matrix to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_matrix_fp_cf(const Matrix_cf* m, FILE* fp)
{
    uint32_t num_rows, num_cols;
    uint32_t row, col;

    num_rows = m->num_rows;
    num_cols = m->num_cols;
    for (row = 0; row < num_rows; row++)
    {
        for (col = 0; col < num_cols; col++)
        {
            fprintf(fp, "%f %f ", m->elts[ row ][ col ].r, 
                                  m->elts[ row ][ col ].i);
        }
        fprintf(fp, "\n");
    }

    return NULL;
}

/**
 * Format of the matrix written is sequential lines containing rows of the 
 * matrix.
 *
 * @code
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m   Matrix to write.
 * @param  fp  File to write the matrix to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_matrix_fp_cd(const Matrix_cd* m, FILE* fp)
{
    uint32_t num_rows, num_cols;
    uint32_t row, col;

    num_rows = m->num_rows;
    num_cols = m->num_cols;
    for (row = 0; row < num_rows; row++)
    {
        for (col = 0; col < num_cols; col++)
        {
            fprintf(fp, "%lf %lf ", m->elts[ row ][ col ].r,
                                    m->elts[ row ][ col ].i);
        }
        fprintf(fp, "\n");
    }

    return NULL;
}

/*@}*/




/**
 * @name read_matrix_with_header
 *
 * Reads a matrix with a header from a file.
 * @{
 */

/**
 * Format of the matrix in the file is expected to a line containing the 
 * number of rows and columns followed by sequential lines containing rows of 
 * the matrix. Blank lines and lines beginning with '#' prior to the first row 
 * of the matrix are skipped.  For example,
 *
 * @code
 * # Comment
 * num_rows=3 num_cols=3
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m_out   Result parameter. If @em *m_out is NULL, a matrix is
 *                 allocated; otherwise its space is re-used.
 * @param  fname   Name of the file to read the matrix from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *m_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the matrix in the file is 
 *   not formatted properly.
 */
Error* read_matrix_with_header_u32(Matrix_u32** m_out, const char* fname)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "r")) == NULL)
    {
        free_matrix_u32(*m_out); *m_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    if ((e = read_matrix_with_header_fp_u32(m_out, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        free_matrix_u32(*m_out); *m_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the matrix in the file is expected to a line containing the 
 * number of rows and columns followed by sequential lines containing rows of 
 * the matrix. Blank lines and lines beginning with '#' prior to the first row 
 * of the matrix are skipped.  For example,
 *
 * @code
 * # Comment
 * num_rows=3 num_cols=3
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m_out   Result parameter. If @em *m_out is NULL, a matrix is
 *                 allocated; otherwise its space is re-used.
 * @param  fname   Name of the file to read the matrix from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *m_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the matrix in the file is 
 *   not formatted properly.
 */
Error* read_matrix_with_header_i32(Matrix_i32** m_out, const char* fname)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "r")) == NULL)
    {
        free_matrix_i32(*m_out); *m_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    if ((e = read_matrix_with_header_fp_i32(m_out, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        free_matrix_i32(*m_out); *m_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the matrix in the file is expected to a line containing the 
 * number of rows and columns followed by sequential lines containing rows of 
 * the matrix. Blank lines and lines beginning with '#' prior to the first row 
 * of the matrix are skipped.  For example,
 *
 * @code
 * # Comment
 * num_rows=3 num_cols=3
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m_out   Result parameter. If @em *m_out is NULL, a matrix is
 *                 allocated; otherwise its space is re-used.
 * @param  fname   Name of the file to read the matrix from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *m_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the matrix in the file is 
 *   not formatted properly.
 */
Error* read_matrix_with_header_i64(Matrix_i64** m_out, const char* fname)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "r")) == NULL)
    {
        free_matrix_i64(*m_out); *m_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    if ((e = read_matrix_with_header_fp_i64(m_out, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        free_matrix_i64(*m_out); *m_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the matrix in the file is expected to a line containing the 
 * number of rows and columns followed by sequential lines containing rows of 
 * the matrix. Blank lines and lines beginning with '#' prior to the first row 
 * of the matrix are skipped.  For example,
 *
 * @code
 * # Comment
 * num_rows=3 num_cols=3
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m_out   Result parameter. If @em *m_out is NULL, a matrix is
 *                 allocated; otherwise its space is re-used.
 * @param  fname   Name of the file to read the matrix from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *m_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the matrix in the file is 
 *   not formatted properly.
 */
Error* read_matrix_with_header_f(Matrix_f** m_out, const char* fname)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "r")) == NULL)
    {
        free_matrix_f(*m_out); *m_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    if ((e = read_matrix_with_header_fp_f(m_out, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        free_matrix_f(*m_out); *m_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the matrix in the file is expected to a line containing the 
 * number of rows and columns followed by sequential lines containing rows of 
 * the matrix. Blank lines and lines beginning with '#' prior to the first row 
 * of the matrix are skipped.  For example,
 *
 * @code
 * # Comment
 * num_rows=3 num_cols=3
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m_out   Result parameter. If @em *m_out is NULL, a matrix is
 *                 allocated; otherwise its space is re-used.
 * @param  fname   Name of the file to read the matrix from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *m_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the matrix in the file is 
 *   not formatted properly.
 */
Error* read_matrix_with_header_d(Matrix_d** m_out, const char* fname)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "r")) == NULL)
    {
        free_matrix_d(*m_out); *m_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    if ((e = read_matrix_with_header_fp_d(m_out, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        free_matrix_d(*m_out); *m_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the matrix in the file is expected to a line containing the 
 * number of rows and columns followed by sequential lines containing rows of 
 * the matrix. Blank lines and lines beginning with '#' prior to the first row 
 * of the matrix are skipped.  For example,
 *
 * @code
 * # Comment
 * num_rows=3 num_cols=3
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m_out   Result parameter. If @em *m_out is NULL, a matrix is
 *                 allocated; otherwise its space is re-used.
 * @param  fname   Name of the file to read the matrix from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *m_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the matrix in the file is 
 *   not formatted properly.
 */
Error* read_matrix_with_header_cf(Matrix_cf** m_out, const char* fname)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "r")) == NULL)
    {
        free_matrix_cf(*m_out); *m_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    if ((e = read_matrix_with_header_fp_cf(m_out, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        free_matrix_cf(*m_out); *m_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the matrix in the file is expected to a line containing the 
 * number of rows and columns followed by sequential lines containing rows of 
 * the matrix. Blank lines and lines beginning with '#' prior to the first row 
 * of the matrix are skipped.  For example,
 *
 * @code
 * # Comment
 * num_rows=3 num_cols=3
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m_out   Result parameter. If @em *m_out is NULL, a matrix is
 *                 allocated; otherwise its space is re-used.
 * @param  fname   Name of the file to read the matrix from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *m_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the matrix in the file is 
 *   not formatted properly.
 */
Error* read_matrix_with_header_cd(Matrix_cd** m_out, const char* fname)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "r")) == NULL)
    {
        free_matrix_cd(*m_out); *m_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    if ((e = read_matrix_with_header_fp_cd(m_out, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        free_matrix_cd(*m_out); *m_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/*@}*/




/**
 * @name read_matrix_with_header_fp
 *
 * Reads a matrix with a header from an opened file.
 * @{
 */

/**
 * Format of the matrix in the file is expected to a line containing the 
 * number of rows and columns followed by sequential lines containing rows of 
 * the matrix. Blank lines and lines beginning with '#' prior to the first row 
 * of the matrix are skipped.  For example,
 *
 * @code
 * # Comment
 * num_rows=3 num_cols=3
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m_out   Result parameter. If @em *m_out is NULL, a matrix is
 *                 allocated; otherwise its space is re-used.
 * @param  fp      File to read the matrix from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *m_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the matrix in the file is 
 *   not formatted properly.
 */
Error* read_matrix_with_header_fp_u32(Matrix_u32** m_out, FILE* fp)
{
    uint32_t    num_rows, num_cols;
    uint32_t    row, col;
    Matrix_u32* m;

    if (skip_fp_spaces_and_comments(fp) == 0)
    {
        free_matrix_u32(*m_out); *m_out = NULL;
        return JWSC_EIO("No matrix to read in file");
    }

    if (fscanf(fp, "num_rows=%u num_cols=%u\n", &num_rows, &num_cols) != 2)
    {
        free_matrix_u32(*m_out); *m_out = NULL;
        return JWSC_EIO("File to read matrix from not formatted properly");
    }

    create_matrix_u32(m_out, num_rows, num_cols);
    m = *m_out;

    for (row = 0; row < num_rows; row++)
    {
        for (col = 0; col < num_cols; col++)
        {
            if (fscanf(fp, "%u", &(m->elts[ row ][ col ])) != 1)
            {
                free_matrix_u32(*m_out); *m_out = NULL;
                return JWSC_EIO("File to read matrix from not formatted properly");
            }
        }
    }

    return NULL;
}

/**
 * Format of the matrix in the file is expected to a line containing the 
 * number of rows and columns followed by sequential lines containing rows of 
 * the matrix. Blank lines and lines beginning with '#' prior to the first row 
 * of the matrix are skipped.  For example,
 *
 * @code
 * # Comment
 * num_rows=3 num_cols=3
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m_out   Result parameter. If @em *m_out is NULL, a matrix is
 *                 allocated; otherwise its space is re-used.
 * @param  fp      File to read the matrix from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *m_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the matrix in the file is 
 *   not formatted properly.
 */
Error* read_matrix_with_header_fp_i32(Matrix_i32** m_out, FILE* fp)
{
    uint32_t    num_rows, num_cols;
    uint32_t    row, col;
    Matrix_i32* m;

    if (skip_fp_spaces_and_comments(fp) == 0)
    {
        free_matrix_i32(*m_out); *m_out = NULL;
        return JWSC_EIO("No matrix to read in file");
    }

    if (fscanf(fp, "num_rows=%u num_cols=%u\n", &num_rows, &num_cols) != 2)
    {
        free_matrix_i32(*m_out); *m_out = NULL;
        return JWSC_EIO("File to read matrix from not formatted properly");
    }

    create_matrix_i32(m_out, num_rows, num_cols);
    m = *m_out;

    for (row = 0; row < num_rows; row++)
    {
        for (col = 0; col < num_cols; col++)
        {
            if (fscanf(fp, "%d", &(m->elts[ row ][ col ])) != 1)
            {
                free_matrix_i32(*m_out); *m_out = NULL;
                return JWSC_EIO("File to read matrix from not formatted properly");
            }
        }
    }

    return NULL;
}

/**
 * Format of the matrix in the file is expected to a line containing the 
 * number of rows and columns followed by sequential lines containing rows of 
 * the matrix. Blank lines and lines beginning with '#' prior to the first row 
 * of the matrix are skipped.  For example,
 *
 * @code
 * # Comment
 * num_rows=3 num_cols=3
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m_out   Result parameter. If @em *m_out is NULL, a matrix is
 *                 allocated; otherwise its space is re-used.
 * @param  fp      File to read the matrix from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *m_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the matrix in the file is 
 *   not formatted properly.
 */
Error* read_matrix_with_header_fp_i64(Matrix_i64** m_out, FILE* fp)
{
    uint32_t    num_rows, num_cols;
    uint32_t    row, col;
    Matrix_i64* m;

    if (skip_fp_spaces_and_comments(fp) == 0)
    {
        free_matrix_i64(*m_out); *m_out = NULL;
        return JWSC_EIO("No matrix to read in file");
    }

    if (fscanf(fp, "num_rows=%u num_cols=%u\n", &num_rows, &num_cols) != 2)
    {
        free_matrix_i64(*m_out); *m_out = NULL;
        return JWSC_EIO("File to read matrix from not formatted properly");
    }

    create_matrix_i64(m_out, num_rows, num_cols);
    m = *m_out;

    for (row = 0; row < num_rows; row++)
    {
        for (col = 0; col < num_cols; col++)
        {
            if (fscanf(fp, "%" SCNd64, &(m->elts[ row ][ col ])) != 1)
            {
                free_matrix_i64(*m_out); *m_out = NULL;
                return JWSC_EIO("File to read matrix from not formatted properly");
            }
        }
    }

    return NULL;
}

/**
 * Format of the matrix in the file is expected to a line containing the 
 * number of rows and columns followed by sequential lines containing rows of 
 * the matrix. Blank lines and lines beginning with '#' prior to the first row 
 * of the matrix are skipped.  For example,
 *
 * @code
 * # Comment
 * num_rows=3 num_cols=3
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m_out   Result parameter. If @em *m_out is NULL, a matrix is
 *                 allocated; otherwise its space is re-used.
 * @param  fp      File to read the matrix from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *m_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the matrix in the file is 
 *   not formatted properly.
 */
Error* read_matrix_with_header_fp_f(Matrix_f** m_out, FILE* fp)
{
    uint32_t  num_rows, num_cols;
    uint32_t  row, col;
    Matrix_f* m;

    if (skip_fp_spaces_and_comments(fp) == 0)
    {
        free_matrix_f(*m_out); *m_out = NULL;
        return JWSC_EIO("No matrix to read in file");
    }

    if (fscanf(fp, "num_rows=%u num_cols=%u\n", &num_rows, &num_cols) != 2)
    {
        free_matrix_f(*m_out); *m_out = NULL;
        return JWSC_EIO("File to read matrix from not formatted properly");
    }

    create_matrix_f(m_out, num_rows, num_cols);
    m = *m_out;

    for (row = 0; row < num_rows; row++)
    {
        for (col = 0; col < num_cols; col++)
        {
            if (fscanf(fp, "%f", &(m->elts[ row ][ col ])) != 1)
            {
                free_matrix_f(*m_out); *m_out = NULL;
                return JWSC_EIO("File to read matrix from not formatted properly");
            }
        }
    }

    return NULL;
}

/**
 * Format of the matrix in the file is expected to a line containing the 
 * number of rows and columns followed by sequential lines containing rows of 
 * the matrix. Blank lines and lines beginning with '#' prior to the first row 
 * of the matrix are skipped.  For example,
 *
 * @code
 * # Comment
 * num_rows=3 num_cols=3
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m_out   Result parameter. If @em *m_out is NULL, a matrix is
 *                 allocated; otherwise its space is re-used.
 * @param  fp      File to read the matrix from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *m_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the matrix in the file is 
 *   not formatted properly.
 */
Error* read_matrix_with_header_fp_d(Matrix_d** m_out, FILE* fp)
{
    uint32_t  num_rows, num_cols;
    uint32_t  row, col;
    Matrix_d* m;

    if (skip_fp_spaces_and_comments(fp) == 0)
    {
        free_matrix_d(*m_out); *m_out = NULL;
        return JWSC_EIO("No matrix to read in file");
    }

    if (fscanf(fp, "num_rows=%u num_cols=%u\n", &num_rows, &num_cols) != 2)
    {
        free_matrix_d(*m_out); *m_out = NULL;
        return JWSC_EIO("File to read matrix from not formatted properly");
    }

    create_matrix_d(m_out, num_rows, num_cols);
    m = *m_out;

    for (row = 0; row < num_rows; row++)
    {
        for (col = 0; col < num_cols; col++)
        {
            if (fscanf(fp, "%lf", &(m->elts[ row ][ col ])) != 1)
            {
                free_matrix_d(*m_out); *m_out = NULL;
                return JWSC_EIO("File to read matrix from not formatted properly");
            }
        }
    }

    return NULL;
}

/**
 * Format of the matrix in the file is expected to a line containing the 
 * number of rows and columns followed by sequential lines containing rows of 
 * the matrix. Blank lines and lines beginning with '#' prior to the first row 
 * of the matrix are skipped.  For example,
 *
 * @code
 * # Comment
 * num_rows=3 num_cols=3
 * 1 2 3 1 2 3
 * 4 5 6 4 5 6
 * 7 8 9 7 8 9
 * @endcode
 *
 * @param  m_out   Result parameter. If @em *m_out is NULL, a matrix is
 *                 allocated; otherwise its space is re-used.
 * @param  fp      File to read the matrix from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *m_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the matrix in the file is 
 *   not formatted properly.
 */
Error* read_matrix_with_header_fp_cf(Matrix_cf** m_out, FILE* fp)
{
    uint32_t   num_rows, num_cols;
    uint32_t   row, col;
    Matrix_cf* m;

    if (skip_fp_spaces_and_comments(fp) == 0)
    {
        free_matrix_cf(*m_out); *m_out = NULL;
        return JWSC_EIO("No matrix to read in file");
    }

    if (fscanf(fp, "num_rows=%u num_cols=%u\n", &num_rows, &num_cols) != 2)
    {
        free_matrix_cf(*m_out); *m_out = NULL;
        return JWSC_EIO("File to read matrix from not formatted properly");
    }

    create_matrix_cf(m_out, num_rows, num_cols);
    m = *m_out;

    for (row = 0; row < num_rows; row++)
    {
        for (col = 0; col < num_cols; col++)
        {
            if (fscanf(fp, "%f %f", &(m->elts[ row ][ col ].r), 
                                    &(m->elts[ row ][ col ].i)) != 2)
            {
                free_matrix_cf(*m_out); *m_out = NULL;
                return JWSC_EIO("File to read matrix from not formatted properly");
            }
        }
    }

    return NULL;
}

/**
 * Format of the matrix in the file is expected to a line containing the 
 * number of rows and columns followed by sequential lines containing rows of 
 * the matrix. Blank lines and lines beginning with '#' prior to the first row 
 * of the matrix are skipped.  For example,
 *
 * @code
 * # Comment
 * num_rows=3 num_cols=3
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m_out   Result parameter. If @em *m_out is NULL, a matrix is
 *                 allocated; otherwise its space is re-used.
 * @param  fp      File to read the matrix from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *m_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the matrix in the file is 
 *   not formatted properly.
 */
Error* read_matrix_with_header_fp_cd(Matrix_cd** m_out, FILE* fp)
{
    uint32_t   num_rows, num_cols;
    uint32_t   row, col;
    Matrix_cd* m;

    if (skip_fp_spaces_and_comments(fp) == 0)
    {
        free_matrix_cd(*m_out); *m_out = NULL;
        return JWSC_EIO("No matrix to read in file");
    }

    if (fscanf(fp, "num_rows=%u num_cols=%u\n", &num_rows, &num_cols) != 2)
    {
        free_matrix_cd(*m_out); *m_out = NULL;
        return JWSC_EIO("File to read matrix from not formatted properly");
    }

    create_matrix_cd(m_out, num_rows, num_cols);
    m = *m_out;

    for (row = 0; row < num_rows; row++)
    {
        for (col = 0; col < num_cols; col++)
        {
            if (fscanf(fp, "%lf %lf", &(m->elts[ row ][ col ].r),
                                      &(m->elts[ row ][ col ].i)) == 2)
            {
                free_matrix_cd(*m_out); *m_out = NULL;
                return JWSC_EIO("File to read matrix from not formatted properly");
            }
        }
    }

    return NULL;
}

/*@}*/




/**
 * @name write_matrix
 *
 * Writes a matrix with a header to a file.
 * @{
 */

/**
 * Format of the matrix written is a line containing the number of elements 
 * followed by sequential lines containg rows of the matrix. For example,
 *
 * @code
 * num_rows=3 num_cols=3
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m      Matrix to write.
 * @param  fname  Name of the file to read the matrix from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_matrix_with_header_u32(const Matrix_u32* m, const char* fname)
{
    FILE* fp;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return JWSC_EIO(strerror(errno));
    }

    write_matrix_with_header_fp_u32(m, fp);

    if (fclose(fp) != 0)
    {
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the matrix written is a line containing the number of elements 
 * followed by sequential lines containg rows of the matrix. For example,
 *
 * @code
 * num_rows=3 num_cols=3
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m      Matrix to write.
 * @param  fname  Name of the file to read the matrix from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_matrix_with_header_i32(const Matrix_i32* m, const char* fname)
{
    FILE* fp;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return JWSC_EIO(strerror(errno));
    }

    write_matrix_with_header_fp_i32(m, fp);

    if (fclose(fp) != 0)
    {
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the matrix written is a line containing the number of elements 
 * followed by sequential lines containg rows of the matrix. For example,
 *
 * @code
 * num_rows=3 num_cols=3
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m      Matrix to write.
 * @param  fname  Name of the file to read the matrix from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_matrix_with_header_i64(const Matrix_i64* m, const char* fname)
{
    FILE* fp;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return JWSC_EIO(strerror(errno));
    }

    write_matrix_with_header_fp_i64(m, fp);

    if (fclose(fp) != 0)
    {
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the matrix written is a line containing the number of elements 
 * followed by sequential lines containg rows of the matrix. For example,
 *
 * @code
 * num_rows=3 num_cols=3
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m      Matrix to write.
 * @param  fname  Name of the file to read the matrix from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_matrix_with_header_f(const Matrix_f* m, const char* fname)
{
    FILE* fp;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return JWSC_EIO(strerror(errno));
    }

    write_matrix_with_header_fp_f(m, fp);

    if (fclose(fp) != 0)
    {
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the matrix written is a line containing the number of elements 
 * followed by sequential lines containg rows of the matrix. For example,
 *
 * @code
 * num_rows=3 num_cols=3
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m      Matrix to write.
 * @param  fname  Name of the file to read the matrix from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_matrix_with_header_d(const Matrix_d* m, const char* fname)
{
    FILE* fp;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return JWSC_EIO(strerror(errno));
    }

    write_matrix_with_header_fp_d(m, fp);

    if (fclose(fp) != 0)
    {
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the matrix written is a line containing the number of elements 
 * followed by sequential lines containg rows of the matrix. For example,
 *
 * @code
 * num_rows=3 num_cols=3
 * 1 2 3 1 2 3
 * 4 5 6 4 5 6
 * 7 8 9 7 8 9
 * @endcode
 *
 * @param  m      Matrix to write.
 * @param  fname  Name of the file to read the matrix from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_matrix_with_header_cf(const Matrix_cf* m, const char* fname)
{
    FILE* fp;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return JWSC_EIO(strerror(errno));
    }

    write_matrix_with_header_fp_cf(m, fp);

    if (fclose(fp) != 0)
    {
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the matrix written is a line containing the number of elements 
 * followed by sequential lines containg rows of the matrix. For example,
 *
 * @code
 * num_rows=3 num_cols=3
 * 1 2 3 1 2 3 
 * 4 5 6 4 5 6
 * 7 8 9 7 8 9
 * @endcode
 *
 * @param  m      Matrix to write.
 * @param  fname  Name of the file to read the matrix from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_matrix_with_header_cd(const Matrix_cd* m, const char* fname)
{
    FILE* fp;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return JWSC_EIO(strerror(errno));
    }

    write_matrix_with_header_fp_cd(m, fp);

    if (fclose(fp) != 0)
    {
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/*@}*/




/**
 * @name write_matrix_with_header_fp
 *
 * Writes a matrix with a header to an opened file.
 * @{
 */

/**
 * Format of the matrix written is a line containing the number of elements 
 * followed by sequential lines containg rows of the matrix. For example,
 *
 * @code
 * num_rows=3 num_cols=3
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m   Matrix to write.
 * @param  fp  File to write the matrix to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_matrix_with_header_fp_u32(const Matrix_u32* m, FILE* fp)
{
    uint32_t num_rows, num_cols;
    uint32_t row, col;

    num_rows = m->num_rows;
    num_cols = m->num_cols;
    fprintf(fp, "num_rows=%u num_cols=%u\n", num_rows, num_cols);

    for (row = 0; row < num_rows; row++)
    {
        for (col = 0; col < num_cols; col++)
        {
            fprintf(fp, "%u ", m->elts[ row ][ col ]);
        }
        fprintf(fp, "\n");
    }

    return NULL;
}

/**
 * Format of the matrix written is a line containing the number of elements 
 * followed by sequential lines containg rows of the matrix. For example,
 *
 * @code
 * num_rows=3 num_cols=3
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m   Matrix to write.
 * @param  fp  File to write the matrix to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_matrix_with_header_fp_i32(const Matrix_i32* m, FILE* fp)
{
    uint32_t num_rows, num_cols;
    uint32_t row, col;

    num_rows = m->num_rows;
    num_cols = m->num_cols;
    fprintf(fp, "num_rows=%u num_cols=%u\n", num_rows, num_cols);

    for (row = 0; row < num_rows; row++)
    {
        for (col = 0; col < num_cols; col++)
        {
            fprintf(fp, "%d ", m->elts[ row ][ col ]);
        }
        fprintf(fp, "\n");
    }

    return NULL;
}

/**
 * Format of the matrix written is a line containing the number of elements 
 * followed by sequential lines containg rows of the matrix. For example,
 *
 * @code
 * num_rows=3 num_cols=3
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m   Matrix to write.
 * @param  fp  File to write the matrix to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_matrix_with_header_fp_i64(const Matrix_i64* m, FILE* fp)
{
    uint32_t num_rows, num_cols;
    uint32_t row, col;

    num_rows = m->num_rows;
    num_cols = m->num_cols;
    fprintf(fp, "num_rows=%u num_cols=%u\n", num_rows, num_cols);

    for (row = 0; row < num_rows; row++)
    {
        for (col = 0; col < num_cols; col++)
        {
            fprintf(fp, "%" PRId64 " ", m->elts[ row ][ col ]);
        }
        fprintf(fp, "\n");
    }

    return NULL;
}

/**
 * Format of the matrix written is a line containing the number of elements 
 * followed by sequential lines containg rows of the matrix. For example,
 *
 * @code
 * num_rows=3 num_cols=3
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m   Matrix to write.
 * @param  fp  File to write the matrix to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_matrix_with_header_fp_f(const Matrix_f* m, FILE* fp)
{
    uint32_t num_rows, num_cols;
    uint32_t row, col;

    num_rows = m->num_rows;
    num_cols = m->num_cols;
    fprintf(fp, "num_rows=%u num_cols=%u\n", num_rows, num_cols);

    for (row = 0; row < num_rows; row++)
    {
        for (col = 0; col < num_cols; col++)
        {
            fprintf(fp, "%f ", m->elts[ row ][ col ]);
        }
        fprintf(fp, "\n");
    }

    return NULL;
}

/**
 * Format of the matrix written is a line containing the number of elements 
 * followed by sequential lines containg rows of the matrix. For example,
 *
 * @code
 * num_rows=3 num_cols=3
 * 1 2 3 
 * 4 5 6 
 * 7 8 9
 * @endcode
 *
 * @param  m   Matrix to write.
 * @param  fp  File to write the matrix to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_matrix_with_header_fp_d(const Matrix_d* m, FILE* fp)
{
    uint32_t num_rows, num_cols;
    uint32_t row, col;

    num_rows = m->num_rows;
    num_cols = m->num_cols;
    fprintf(fp, "num_rows=%u num_cols=%u\n", num_rows, num_cols);

    for (row = 0; row < num_rows; row++)
    {
        for (col = 0; col < num_cols; col++)
        {
            fprintf(fp, "%.8e ", m->elts[ row ][ col ]);
        }
        fprintf(fp, "\n");
    }

    return NULL;
}

/**
 * Format of the matrix written is a line containing the number of elements 
 * followed by sequential lines containg rows of the matrix. For example,
 *
 * @code
 * num_rows=3 num_cols=3
 * 1 2 3 1 2 3
 * 4 5 6 4 5 6
 * 7 8 9 7 8 9
 * @endcode
 *
 * @param  m   Matrix to write.
 * @param  fp  File to write the matrix to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_matrix_with_header_fp_cf(const Matrix_cf* m, FILE* fp)
{
    uint32_t num_rows, num_cols;
    uint32_t row, col;

    num_rows = m->num_rows;
    num_cols = m->num_cols;
    fprintf(fp, "num_rows=%u num_cols=%u\n", num_rows, num_cols);

    for (row = 0; row < num_rows; row++)
    {
        for (col = 0; col < num_cols; col++)
        {
            fprintf(fp, "%f %f ", m->elts[ row ][ col ].r,
                                  m->elts[ row ][ col ].i);
        }
        fprintf(fp, "\n");
    }

    return NULL;
}

/**
 * Format of the matrix written is a line containing the number of elements 
 * followed by sequential lines containg rows of the matrix. For example,
 *
 * @code
 * num_rows=3 num_cols=3
 * 1 2 3 1 2 3
 * 4 5 6 4 5 6
 * 7 8 9 7 8 9
 * @endcode
 *
 * @param  m   Matrix to write.
 * @param  fp  File to write the matrix to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_matrix_with_header_fp_cd(const Matrix_cd* m, FILE* fp)
{
    uint32_t num_rows, num_cols;
    uint32_t row, col;

    num_rows = m->num_rows;
    num_cols = m->num_cols;
    fprintf(fp, "num_rows=%u num_cols=%u\n", num_rows, num_cols);

    for (row = 0; row < num_rows; row++)
    {
        for (col = 0; col < num_cols; col++)
        {
            fprintf(fp, "%lf %lf ", m->elts[ row ][ col ].r,
                                    m->elts[ row ][ col ].i);
        }
        fprintf(fp, "\n");
    }

    return NULL;
}

/*@}*/
