/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Declarations for mathematical matrix operations.
 *
 * Many of the functions are accelerated if linked against an optimized BLAS
 * library.
 */


#ifndef MATRIX_MATH_H 
#define MATRIX_MATH_H 


#include <jwsc/config.h>

#include <stdlib.h>
#include <inttypes.h>

#include "jwsc/base/error.h"
#include "jwsc/math/complex.h"
#include "jwsc/vector/vector.h"
#include "jwsc/matrix/matrix.h"


#ifdef __cplusplus
namespace jwsc {
extern "C" {
#endif


/**
 * @name add_scalar_to_matrix
 *
 * Adds a scalar to a matrix.
 * @{
 */

/** @brief Adds a scalar to a single precision matrix. */
void add_scalar_to_matrix_f(Matrix_f** m_out, const Matrix_f* m_in, float a);

/** @brief Adds a scalar to a double precision matrix. */
void add_scalar_to_matrix_d(Matrix_d** m_out, const Matrix_d* m_in, double a);

/** @brief Adds a scalar to a single precision complex matrix. */
void add_scalar_to_matrix_cf
(
    Matrix_cf**      m_out, 
    const Matrix_cf* m_in, 
    Complex_f        z
);

/** @brief Adds a scalar to a double precision complex matrix. */
void add_scalar_to_matrix_cd
(
    Matrix_cd**      m_out, 
    const Matrix_cd* m_in, 
    Complex_d        z
);

/*@}*/




/** 
 * @name multiply_matrix_by_scalar
 *
 * Multiplies a matrix by a scalar.
 * @{
 */

/** @brief Multiplies a single precision matrix by a scalar. */
void multiply_matrix_by_scalar_f
(
    Matrix_f**      m_out, 
    const Matrix_f* m_in, 
    float           a
);

/** @brief Multiplies a double precision matrix by a scalar. */
void multiply_matrix_by_scalar_d
(
    Matrix_d**      m_out, 
    const Matrix_d* m_in, 
    double          a
);

/** @brief Multiplies a single precision complex matrix by a scalar. */
void multiply_matrix_by_scalar_cf
(
    Matrix_cf**      m_out, 
    const Matrix_cf* m_in, 
    Complex_f        z
);

/** @brief Multiplies a double precision complex matrix by a scalar. */
void multiply_matrix_by_scalar_cd
(
    Matrix_cd**      m_out, 
    const Matrix_cd* m_in, 
    Complex_d        z
);

/*@}*/




/** 
 * @name multiply_matrices
 *
 * Multiplies a two matrices.
 * @{
 */

/** @brief Multiples two single precision matrices. */
Error* multiply_matrices_f
(
    Matrix_f**      m_out, 
    const Matrix_f* m_1, 
    const Matrix_f* m_2
);

/** @brief Multiples two double precision matrices. */
Error* multiply_matrices_d
(
    Matrix_d**      m_out, 
    const Matrix_d* m_1, 
    const Matrix_d* m_2
);

/** @brief Multiples two single precision complex matrices. */
Error* multiply_matrices_cf
(
    Matrix_cf**      m_out, 
    const Matrix_cf* m_1, 
    const Matrix_cf* m_2
);

/** @brief Multiples two double precision complex matrices. */
Error* multiply_matrices_cd
(
    Matrix_cd**      m_out, 
    const Matrix_cd* m_1, 
    const Matrix_cd* m_2
);

/*@}*/




/** 
 * @name multiply_matrices_elt_wise
 *
 * Multiplies a two matrices element by element.
 * @{
 */

/** @brief Multiples two single precision matrices element by element. */
Error* multiply_matrices_elt_wise_f
(
    Matrix_f**      m_out, 
    const Matrix_f* m_1, 
    const Matrix_f* m_2
);

/** @brief Multiples two double precision matrices element by element. */
Error* multiply_matrices_elt_wise_d
(
    Matrix_d**      m_out, 
    const Matrix_d* m_1, 
    const Matrix_d* m_2
);

/** 
 * @brief Multiples two single precision complex matrices element by element.
 */
Error* multiply_matrices_elt_wise_cf
(
    Matrix_cf**      m_out, 
    const Matrix_cf* m_1, 
    const Matrix_cf* m_2
);

/** 
 * @brief Multiples two double precision complex matrices element by element.
 */
Error* multiply_matrices_elt_wise_cd
(
    Matrix_cd**      m_out, 
    const Matrix_cd* m_1, 
    const Matrix_cd* m_2
);

/*@}*/




/** 
 * @name transpose_matrix
 *
 * Transposes a matrix.
 * @{
 */

/** @brief Transposes a single precision matrix. */
void transpose_matrix_f(Matrix_f** m_out, const Matrix_f* m_in);

/** @brief Transposes a double precision matrix. */
void transpose_matrix_d(Matrix_d** m_out, const Matrix_d* m_in);

/** @brief Transposes a single precision matrix. */
void transpose_matrix_cf(Matrix_cf** m_out, const Matrix_cf* m_in);

/** @brief Transposes a double precision complex matrix. */
void transpose_matrix_cd(Matrix_cd** m_out, const Matrix_cd* m_in);

/*@}*/




/** 
 * @name multiply_matrix_and_vector
 *
 * Multiplies a matrix and a vector.
 * @{
 */

/** @brief Multiplies a single precision matrix and vector. */
Error* multiply_matrix_and_vector_f
(
    Vector_f**      v_out, 
    const Matrix_f* m_in, 
    const Vector_f* v_in
);

/** @brief Multiplies a double precision matrix and vector. */
Error* multiply_matrix_and_vector_d
(
    Vector_d**      v_out, 
    const Matrix_d* m_in, 
    const Vector_d* v_in
);

/** @brief Multiplies a single precision complex matrix and vector. */
Error* multiply_matrix_and_vector_cf
(
    Vector_cf**      v_out, 
    const Matrix_cf* m_in, 
    const Vector_cf* v_in
);

/** @brief Multiplies a double precision complex matrix and vector. */
Error* multiply_matrix_and_vector_cd
(
    Vector_cd**      v_out, 
    const Matrix_cd* m_in, 
    const Vector_cd* v_in
);

/*@}*/




/** 
 * @name add_matrices
 *
 * Adds two matrices element-wise.
 * @{
 */

/** @brief Adds two single precision matrices element-wise. */
Error* add_matrices_f
(
    Matrix_f**      m_out, 
    const Matrix_f* m_1, 
    const Matrix_f* m_2
);

/** @brief Adds two double precision matrices element-wise. */
Error* add_matrices_d
(
    Matrix_d**      m_out, 
    const Matrix_d* m_1, 
    const Matrix_d* m_2
);

/** @brief Adds two single precision complex matrices element-wise. */
Error* add_matrices_cf
(
    Matrix_cf**      m_out, 
    const Matrix_cf* m_1, 
    const Matrix_cf* m_2
);

/** @brief Adds two double precision complex matrices element-wise. */
Error* add_matrices_cd
(
    Matrix_cd**      m_out, 
    const Matrix_cd* m_1, 
    const Matrix_cd* m_2
);

/*@}*/




/** 
 * @name subtract_matrices
 *
 * Subtracts two matrices element-wise.
 * @{
 */

/** @brief Subtracts two single precision matrices element-wise. */
Error* subtract_matrices_f
(
    Matrix_f**      m_out, 
    const Matrix_f* m_1, 
    const Matrix_f* m_2
);

/** @brief Subtracts two double precision matrices element-wise. */
Error* subtract_matrices_d
(
    Matrix_d**      m_out, 
    const Matrix_d* m_1, 
    const Matrix_d* m_2
);

/** @brief Subtracts two single precision complex matrices element-wise. */
Error* subtract_matrices_cf
(
    Matrix_cf**      m_out, 
    const Matrix_cf* m_1, 
    const Matrix_cf* m_2
);

/** @brief Subtracts two double precision complex matrices element-wise. */
Error* subtract_matrices_cd
(
    Matrix_cd**      m_out, 
    const Matrix_cd* m_1, 
    const Matrix_cd* m_2
);

/*@}*/




/** 
 * @name calc_matrix_asum
 *
 * Calculates the absolute sum of a real matrix.
 * @{
 */

/** @brief Calculates the absolute value sum of a single precision matrix. */
void calc_matrix_asum_f(float* sum_out, const Matrix_f* m);

/** @brief Calculates the absolute value sum of a double precision matrix. */
void calc_matrix_asum_d(double* sum_out, const Matrix_d* m);

/*@}*/




/** 
 * @name normalize_matrix_sum
 *
 * Normalizes a real matrix by its sum.
 * @{
 */

/** @brief Normalizes a single precision matrix by its sum. */
void normalize_matrix_sum_f(Matrix_f** m_out, const Matrix_f* m_in);

/** @brief Normalizes a double precision matrix by its sum. */
void normalize_matrix_sum_d(Matrix_d** m_out, const Matrix_d* m_in);

/*@}*/




/** 
 * @name calc_frobenius_norm
 *
 * Calculates the Frobenius norm (Euclidean norm) of a matrix.
 * @{
 */

/** @brief Calculates the Frobenius norm of a single precision matrix. */
void calc_frobenius_norm_f(float* norm_out, const Matrix_f* m);

/** @brief Calculates the Frobenius norm of a double precision matrix. */
void calc_frobenius_norm_d(double* norm_out, const Matrix_d* m);

/*@}*/




/** 
 * @name calc_frobenius_norm_diff
 *
 * Calculates the Frobenius norm (Euclidean norm) of a difference of two
 * matrices.
 * @{
 */

/** 
 * @brief Calculates the Frobenius norm of a difference of two single
 * precision matrix.
 */
Error* calc_frobenius_norm_diff_f
(
    float*          norm_out, 
    const Matrix_f* m_1,
    const Matrix_f* m_2
);

/** 
 * @brief Calculates the Frobenius norm of a difference of two double
 * precision matrix.
 */
Error* calc_frobenius_norm_diff_d
(
    double*         norm_out, 
    const Matrix_d* m_1,
    const Matrix_d* m_2
);

/*@}*/




/** 
 * @name calc_frobenius_norm_abs_diff
 *
 * Calculates the Frobenius norm (Euclidean norm) of a difference of two
 * matrices, applying the absolute value to each matrix first.
 * @{
 */

/** 
 * @brief Calculates the Frobenius norm of a difference of two single
 * precision matrix, applying the absolute value to each matrix first.
 */
Error* calc_frobenius_norm_abs_diff_f
(
    float*          norm_out, 
    const Matrix_f* m_1,
    const Matrix_f* m_2
);

/** 
 * @brief Calculates the Frobenius norm of a difference of two double
 * precision matrix, applying the absolute value to each matrix first.
 */
Error* calc_frobenius_norm_abs_diff_d
(
    double*         norm_out, 
    const Matrix_d* m_1,
    const Matrix_d* m_2
);

/*@}*/




#if defined JWSC_HAVE_LAPACK
/** 
 * @name lu_decompose_matrix
 *
 * LU decomposes a matrix.
 * @{
 */

/** @brief LU decomposes a single precision matrix. */
Error* lu_decompose_matrix_f
(
    Matrix_f**      m_out, 
    Vector_i32**    pivot_out,
    const Matrix_f* m_in
);

/** @brief LU decomposes a double precision matrix. */
Error* lu_decompose_matrix_d
(
    Matrix_d**      m_out, 
    Vector_i32**    pivot_out,
    const Matrix_d* m_in
);

/*@}*/
#endif




#if defined JWSC_HAVE_LAPACK
/** 
 * @name invert_matrix
 *
 * Inverts a matrix.
 * @{
 */

/** @brief Inverts a single precision matrix. */
Error* invert_matrix_f(Matrix_f** m_out, const Matrix_f* m_in);

/** @brief Inverts a double precision matrix. */
Error* invert_matrix_d(Matrix_d** m_out, const Matrix_d* m_in);

/*@}*/
#endif




#if defined JWSC_HAVE_LAPACK
/** 
 * @name invert_lu_matrix
 *
 * Inverts an LU decomposed matrix.
 * @{
 */

/** @brief Inverts an LU decomposed single precision matrix. */
Error* invert_lu_matrix_f
(
    Matrix_f**        m_out, 
    const Matrix_f*   m_in, 
    const Vector_i32* pivot
);

/** @brief Inverts an LU decomposed double precision matrix. */
Error* invert_lu_matrix_d
(
    Matrix_d**        m_out, 
    const Matrix_d*   m_in, 
    const Vector_i32* pivot
);

/*@}*/
#endif




#if defined JWSC_HAVE_LAPACK
/** 
 * @name calc_matrix_determinant
 *
 * Calculates the determinant of a matrix.
 * @{
 */

/** 
 * @brief Calculates the determinant of a single precision 
 * matrix.
 */
Error* calc_matrix_determinant_f(float* det_out, const Matrix_f* m);

/** @brief Calculates the determinant of a double precision matrix. */
Error* calc_matrix_determinant_d(double* det_out, const Matrix_d* m);

/*@}*/
#endif




/** 
 * @name calc_lu_matrix_determinant
 *
 * Calculates the determinant of an LU decomposed matrix.
 * @{
 */

/** 
 * @brief Calculates the determinant of an LU decomposed single precision 
 * matrix.
 */
Error* calc_lu_matrix_determinant_f
(
    float*            det_out,
    const Matrix_f*   m, 
    const Vector_i32* pivot
);

/** 
 * @brief Calculates the determinant of an LU decomposed double precision 
 * matrix.
 */
Error* calc_lu_matrix_determinant_d
(
    double*           det_out,
    const Matrix_d*   m, 
    const Vector_i32* pivot
);

/*@}*/




#if defined JWSC_HAVE_LAPACK
/** 
 * @name solve_real_symmetric_matrix_eigenproblem
 *
 * Solves the eigenproblem for a real, symmetric matrix.
 * @{
 */

/** 
 * @brief Solves the eigenproblem for a single precision, real and symmetric 
 * matrix.
 */
Error* solve_real_symmetric_matrix_eigenproblem_f
(
    Matrix_f**      V_out,
    Vector_f**      D_out, 
    const Matrix_f* M
);

/**
 * @brief Solves the eigenproblem for a doulbe precision, real and symmetric 
 * matrix.
 */
Error* solve_real_symmetric_matrix_eigenproblem_d
(
    Matrix_d**      V_out,
    Vector_d**      D_out, 
    const Matrix_d* M
);

/*@}*/
#endif




#if defined JWSC_HAVE_LAPACK
/** 
 * @name singular_value_decompose_matrix
 *
 * Singular value decomposition of a matrix.
 * @{
 */

/** @brief Singular value decomposition of a single precision matrix. */
Error* singular_value_decompose_matrix_f
(
    Matrix_f**      U_out,
    Vector_f**      S_out, 
    Matrix_f**      Vt_out,
    const Matrix_f* M
);

/** @brief Singular value decomposition of a double precision matrix. */
Error* singular_value_decompose_matrix_d
(
    Matrix_d**      U_out,
    Vector_d**      S_out, 
    Matrix_d**      Vt_out,
    const Matrix_d* M
);

/*@}*/
#endif




#if defined JWSC_HAVE_LAPACK
/** 
 * @name pseudoinvert_matrix_no_svd
 *
 * Pseudoinverts a matrix without using singular value decomposition.
 * @{
 */

/** 
 * @brief Pseudoinverts a single precision matrix without using singular
 * value decomposition.
 */
Error* pseudoinvert_matrix_no_svd_f
(
    Matrix_f**      M_out,
    const Matrix_f* M_in
);

/** 
 * @brief Pseudoinverts a double precision matrix without using singular
 * value decomposition.
 */
Error* pseudoinvert_matrix_no_svd_d
(
    Matrix_d**      M_out,
    const Matrix_d* M_in
);

/*@}*/
#endif




#if defined JWSC_HAVE_LAPACK
/** 
 * @name pseudoinvert_matrix
 *
 * Pseudoinverts a matrix using singular value decomposition.
 * @{
 */

/** 
 * @brief Pseudoinverts a single precision matrix using singular value
 * decomposition.
 */
Error* pseudoinvert_matrix_f
(
    Matrix_f**      M_out,
    const Matrix_f* M_in
);

/** 
 * @brief Pseudoinverts a double precision matrix using singular value
 * decomposition.
 */
Error* pseudoinvert_matrix_d
(
    Matrix_d**      M_out,
    const Matrix_d* M_in
);

/*@}*/
#endif




/** 
 * @name create_euler_rotation_matrix
 *
 * Creates a standard Euler rotation matrix with angles phi, theta and psi.
 * @{
 */

/** @brief Creates a single precision Euler rotation matrix. */
void create_euler_rotation_matrix_f
(
    Matrix_f** m_out,
    float      phi,
    float      theta,
    float      psi
);

/** @brief Creates a double precision Euler rotation matrix. */
void create_euler_rotation_matrix_d
(
    Matrix_d** m_out,
    float      phi,
    float      theta,
    float      psi
);

/*@}*/




/** 
 * @name create_3d_rotation_matrix
 *
 * Creates a matrix for rotating points around an arbitrary vector.
 * @{
 */

/** 
 * @brief Creates a single precision matrix for rotating points around an
 * arbitrary vector.
 */
Error* create_3d_rotation_matrix_1_f
(
    Matrix_f**      m_out, 
    float           phi,
    const Vector_f* v
);

/** 
 * @brief Creates a double precision matrix for rotating points around an
 * arbitrary vector.
 */
Error* create_3d_rotation_matrix_1_d
(
    Matrix_d**      m_out, 
    double          phi,
    const Vector_d* v
);

/** 
 * @brief Creates a single precision matrix for rotating points around an
 * arbitrary vector.
 */
Error* create_3d_rotation_matrix_2_f
(
    Matrix_f** m_out, 
    float      phi,
    float      x,
    float      y,
    float      z
);

/** 
 * @brief Creates a double precision matrix for rotating points around an
 * arbitrary vector.
 */
Error* create_3d_rotation_matrix_2_d
(
    Matrix_d** m_out, 
    double     phi,
    double     x,
    double     y,
    double     z
);

/*@}*/




/** 
 * @name create_3d_homo_rotation_matrix
 *
 * Creates a homogeneous matrix for rotating points around an arbitrary vector.
 * @{
 */

/** 
 * @brief Creates a single precision homogeneous matrix for rotating points 
 * around an arbitrary vector.
 */
Error* create_3d_homo_rotation_matrix_1_f
(
    Matrix_f**      m_out, 
    float           phi,
    const Vector_f* v
);

/** 
 * @brief Creates a double precision homogeneous matrix for rotating points 
 * around an arbitrary vector.
 */
Error* create_3d_homo_rotation_matrix_1_d
(
    Matrix_d**      m_out, 
    double          phi,
    const Vector_d* v
);

/** 
 * @brief Creates a single precision homogeneous matrix for rotating points 
 * around an arbitrary vector.
 */
Error* create_3d_homo_rotation_matrix_2_f
(
    Matrix_f** m_out, 
    float      phi,
    float      x,
    float      y,
    float      z
);

/** 
 * @brief Creates a double precision homogeneous matrix for rotating points 
 * around an arbitrary vector.
 */
Error* create_3d_homo_rotation_matrix_2_d
(
    Matrix_d** m_out, 
    double     phi,
    double     x,
    double     y,
    double     z
);

/*@}*/




/** 
 * @name create_3d_scaling_matrix
 *
 * Creates a matrix for scaling points.
 * @{
 */

/** @brief Creates a single precision matrix for scaling points. */
Error* create_3d_scaling_matrix_1_f
(
    Matrix_f**      m_out, 
    const Vector_f* v
);

/** @brief Creates a double precision matrix for scaling points. */
Error* create_3d_scaling_matrix_1_d
(
    Matrix_d**      m_out, 
    const Vector_d* v
);

/** 
 * @brief Creates a single precision matrix for scaling points. */
Error* create_3d_scaling_matrix_2_f
(
    Matrix_f** m_out, 
    float      x,
    float      y,
    float      z
);

/** @brief Creates a double precision matrix for scaling points. */
Error* create_3d_scaling_matrix_2_d
(
    Matrix_d** m_out, 
    double     x,
    double     y,
    double     z
);

/*@}*/




/** 
 * @name create_3d_homo_scaling_matrix
 *
 * Creates a homogeneous matrix for scaling points.
 * @{
 */

/** @brief Creates a single precision homogeneous matrix for scaling points. */
Error* create_3d_homo_scaling_matrix_1_f
(
    Matrix_f**      m_out, 
    const Vector_f* v
);

/** @brief Creates a double precision homogeneous matrix for scaling points. */
Error* create_3d_homo_scaling_matrix_1_d
(
    Matrix_d**      m_out, 
    const Vector_d* v
);

/** @brief Creates a single precision homogeneous matrix for scaling points. */
Error* create_3d_homo_scaling_matrix_2_f
(
    Matrix_f** m_out, 
    float      x,
    float      y,
    float      z
);

/** @brief Creates a double precision homogeneous matrix for scaling points. */
Error* create_3d_homo_scaling_matrix_2_d
(
    Matrix_d** m_out, 
    double     x,
    double     y,
    double     z
);

/*@}*/




/** 
 * @name create_3d_homo_translation_matrix
 *
 * Creates a homogeneous matrix for translating points in R^3.
 * @{
 */

/** 
 * @brief Creates a single precision homogeneous matrix for translating points 
 * in R^3.
 */
Error* create_3d_homo_translation_matrix_1_f
(
    Matrix_f**      m_out, 
    const Vector_f* v
);

/** 
 * @brief Creates a double precision homogeneous matrix for translating points 
 * in R^3.
 */
Error* create_3d_homo_translation_matrix_1_d
(
    Matrix_d**      m_out, 
    const Vector_d* v
);

/** 
 * @brief Creates a single precision homogeneous matrix for translating points 
 * in R^3.
 */
Error* create_3d_homo_translation_matrix_2_f
(
    Matrix_f** m_out, 
    float      x,
    float      y,
    float      z
);

/** 
 * @brief Creates a double precision homogeneous matrix for translating points 
 * in R^3.
 */
Error* create_3d_homo_translation_matrix_2_d
(
    Matrix_d** m_out, 
    double     x,
    double     y,
    double     z
);

/*@}*/


#ifdef __cplusplus
}
}
#endif


#endif
