/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/** 
 * @file 
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Definitions for matrix utility functions.
 */


#include <jwsc/config.h>

#include <stdlib.h>
#include <inttypes.h>
#include <math.h>
#include <assert.h>

#include "jwsc/matrix/matrix.h"


/** 
 * @name extend_matrix
 *
 * Extends an matrix into a padding region.
 * @{
 */

/**
 * @param  mat_out      Result parameter. If @em *mat_out is NULL, a matrix
 *                      is allocated; otherwise its space is re-used.
 * @param  mat_in       Matrix to extend.
 * @param  row_padding  Number or rows for padding on top and bottom.
 * @param  col_padding  Number of columns for padding on left and right.
 *
 * @note If @em *mat_out == @em mat_in, then @em mat_in is overwritten.
 */
Error* extend_matrix_f
(
    Matrix_f**      mat_out, 
    const Matrix_f* mat_in, 
    uint32_t        row_padding,
    uint32_t        col_padding
)
{
    uint32_t num_rows, num_cols;
    uint32_t num_prows, num_pcols;
    uint32_t prow, pcol;

    Matrix_f* mat;

    num_rows = mat_in->num_rows;
    num_cols = mat_in->num_cols;

    num_prows = num_rows + 2*row_padding;
    num_pcols = num_cols + 2*col_padding;

    mat = (*mat_out == mat_in) ? NULL : *mat_out;
    create_matrix_f(&mat, num_prows, num_pcols);

    for (prow = 0; prow < (num_prows - row_padding); prow++)
    {
        if (prow < row_padding)
        {
            for (pcol = 0; pcol < col_padding; pcol++)
            {
                mat->elts[ prow ][ pcol ] 
                    = mat_in->elts[ 0 ][ 0 ];

                mat->elts[ prow ][ num_pcols-pcol-1 ] 
                    = mat_in->elts[ 0 ][ num_cols-1 ];

                mat->elts[ num_prows-prow-1 ][ pcol ] 
                    = mat_in->elts[ num_rows-1 ][ 0 ];

                mat->elts[ num_prows-prow-1 ][ num_pcols-pcol-1 ] 
                    = mat_in->elts[ num_rows-1 ][ num_cols-1 ];
            }
            for (pcol = col_padding; pcol < (num_pcols-col_padding); pcol++)
            {
                mat->elts[ prow ][ pcol ] 
                    = mat_in->elts[ 0 ][ pcol-col_padding ];
                mat->elts[ num_prows-prow-1 ][ pcol ] 
                    = mat_in->elts[ num_rows-1 ][ pcol-col_padding ];
            }
        }
        else
        {
            for (pcol = 0; pcol < col_padding; pcol++)
            {
                mat->elts[ prow ][ pcol ] 
                    = mat_in->elts[ prow-row_padding ][ 0 ];
                mat->elts[ prow ][ num_pcols-pcol-1 ] 
                    = mat_in->elts[ prow-row_padding ][ num_cols-1 ];
            }
            for (pcol = col_padding; pcol < (num_pcols-col_padding); pcol++)
            {
                mat->elts[ prow ][ pcol ] 
                    = mat_in->elts[ prow-row_padding][ pcol-col_padding ];
            }
        }
    }

    *mat_out = mat;

    return NULL;
}

/*@}*/
