/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Declarations for multivariate probability density functions.
 */


#ifndef MV_PDF_H
#define MV_PDF_H


#include <jwsc/config.h>

#include <stdlib.h>
#include <inttypes.h>

#include "jwsc/base/error.h"
#include "jwsc/vector/vector.h"
#include "jwsc/matrix/matrix.h"


#ifdef __cplusplus
namespace jwsc {
extern "C" {
#endif


#if defined JWSC_HAVE_LAPACK
/** 
 * @name mv_gaussian_pdf
 *
 * Computes the multivariate Gaussian pdf at a point.
 * @{
 */

/** 
 * @brief Computes the single precision multivariate Gaussian pdf at a point.
 */
Error* mv_gaussian_pdf_f
(
    float*          p_out,
    const Vector_f* mu, 
    const Matrix_f* sigma, 
    const Vector_f* x
);

/** 
 * @brief Computes the double precision multivariate Gaussian pdf at a point.
 */
Error* mv_gaussian_pdf_d
(
    double*         p_out,
    const Vector_d* mu, 
    const Matrix_d* sigma, 
    const Vector_d* x
);

/*@}*/
#endif


#if defined JWSC_HAVE_LAPACK
/** 
 * @name set_mv_gaussian_pdf
 *
 * Computes the multivariate Gaussian pdf at a set of points.
 * @{
 */

/** 
 * @brief Computes the single precision multivariate Gaussian pdf at a set of
 * points.
 */
Error* set_mv_gaussian_pdf_f
(
    Vector_f**      p_out,
    const Vector_f* mu, 
    const Matrix_f* sigma, 
    const Matrix_f* x
);

/** 
 * @brief Computes the double precision multivariate Gaussian pdf at a set of
 * points.
 */
Error* set_mv_gaussian_pdf_d
(
    Vector_d**      p_out,
    const Vector_d* mu, 
    const Matrix_d* sigma, 
    const Matrix_d* x
);

/*@}*/
#endif


#if defined JWSC_HAVE_LAPACK
/** 
 * @name log_mv_gaussian_pdf
 *
 * Computes the log of a multivariate Gaussian pdf at a point.
 * @{
 */

/** 
 * @brief Computes the single precision log of a multivariate Gaussian pdf at
 * a point.
 */
Error* log_mv_gaussian_pdf_f
(
    float*          p_out,
    const Vector_f* mu, 
    const Matrix_f* sigma, 
    const Vector_f* x
);

/** 
 * @brief Computes the double precision log of a multivariate Gaussian pdf at
 * a point.
 */
Error* log_mv_gaussian_pdf_d
(
    double*         p_out,
    const Vector_d* mu, 
    const Matrix_d* sigma, 
    const Vector_d* x
);

/*@}*/
#endif


/** 
 * @name sample_standard_mv_gaussian_pdf
 *
 * Draws a random sample from the standard multivariate Gaussian pdf.
 * @{
 */

/**
 * @brief Draws a single precision random sample from the standard
 * multivariate Gaussian pdf.
 */
void sample_standard_mv_gaussian_pdf_f
(
    Vector_f** x_out, 
    uint32_t   D
);

/**
 * @brief Draws a double precision random sample from the standard
 * multivariate Gaussian pdf.
 */
void sample_standard_mv_gaussian_pdf_d
(
    Vector_d** x_out, 
    uint32_t   D
);

/*@}*/


/** 
 * @name set_sample_standard_mv_gaussian_pdf
 *
 * Draws a set of random samples from the standard multivariate Gaussian pdf.
 * @{
 */

/**
 * @brief Draws a set of single precision random samples from the standard
 * multivariate Gaussian pdf.
 */
void set_sample_standard_mv_gaussian_pdf_f
(
    Matrix_f** x_out, 
    uint32_t   N,
    uint32_t   D
);

/**
 * @brief Draws a set of double precision random samples from the standard
 * multivariate Gaussian pdf.
 */
void set_sample_standard_mv_gaussian_pdf_d
(
    Matrix_d** x_out, 
    uint32_t   N,
    uint32_t   D
);

/*@}*/


#if defined JWSC_HAVE_LAPACK
/** 
 * @name sample_mv_gaussian_pdf
 *
 * Draws a random sample from the multivariate Gaussian pdf.
 * @{
 */

/**
 * @brief Draws a single precision random sample from the multivariate
 * Gaussian pdf.
 */
Error* sample_mv_gaussian_pdf_f
(
    Vector_f**      x_out,
    const Vector_f* mu, 
    const Matrix_f* sigma
);

/**
 * @brief Draws a double precision random sample from the multivariate
 * Gaussian pdf.
 */
Error* sample_mv_gaussian_pdf_d
(
    Vector_d**      x_out,
    const Vector_d* mu, 
    const Matrix_d* sigma
);

/*@}*/
#endif


#if defined JWSC_HAVE_LAPACK
/** 
 * @name set_sample_mv_gaussian_pdf
 *
 * Draws a set of random samples from the multivariate Gaussian pdf.
 * @{
 */

/**
 * @brief Draws a set of single precision random samples from the multivariate
 * Gaussian pdf.
 */
Error* set_sample_mv_gaussian_pdf_f
(
    Matrix_f**      x_out,
    uint32_t        N,
    const Vector_f* mu, 
    const Matrix_f* sigma
);

/**
 * @brief Draws a set of double precision random samples from the multivariate
 * Gaussian pdf.
 */
Error* set_sample_mv_gaussian_pdf_d
(
    Matrix_d**      x_out,
    uint32_t        N,
    const Vector_d* mu, 
    const Matrix_d* sigma
);

/*@}*/
#endif


#ifdef __cplusplus
}
}
#endif


#endif
