/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Multivariate probability mass functions.
 */


#include <jwsc/config.h>

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <math.h>
#include <inttypes.h>

#include "jwsc/vector/vector.h"
#include "jwsc/vector/vector_math.h"
#include "jwsc/matrix/matrix.h"
#include "jwsc/matrix/matrix_math.h"
#include "jwsc/prob/pmf.h"
#include "jwsc/prob/mv_pmf.h"


/** 
 * @name mv_bernoulli_pmf
 *
 * Computes the multivariate Bernoulli pmf at a point.
 * @{
 */

/**
 * @param  p_out  Result parameter.
 * @param  mu     Probabilities of success of the multivariate Bernoulli.
 * @param  x      Value to evaluate the pmf under.
 */
void mv_bernoulli_pmf_f
(
    float*            p_out,
    const Vector_f*   mu,
    const Vector_u32* x
)
{
    uint32_t d, D;
    float    p;

    assert(mu->num_elts == x->num_elts);

    D = mu->num_elts;
 
    p = 1.0;
    for (d = 0; d < D; d++)
    {
        p *= (x->elts[ d ]) ? mu->elts[ d ] : 1.0 - mu->elts[ d ];
    }

    *p_out = p;
}

/**
 * @param  p_out  Result parameter.
 * @param  mu     Probabilities of success of the multivariate Bernoulli.
 * @param  x      Value to evaluate the pmf under.
 */
void mv_bernoulli_pmf_d
(
    double*           p_out,
    const Vector_d*   mu, 
    const Vector_u32* x
)
{
    uint32_t d, D;
    double   p;

    assert(mu->num_elts == x->num_elts);

    D = mu->num_elts;
 
    p = 1.0;
    for (d = 0; d < D; d++)
    {
        p *= (x->elts[ d ]) ? mu->elts[ d ] : 1.0 - mu->elts[ d ];
    }

    *p_out = p;
}

/*@}*/


/** 
 * @name sample_mv_bernoulli_pmf
 *
 * Draws a random sample from the multivariate Bernoulli pmf.
 * @{
 */

/**
 * @param  x_out  Result paramter. Contains the sample.
 * @param  mu     Probabilities of success of the Bernoulli.
 *
 * @note Does not seed the rand() function. If this is desired, srand() must
 * be called prior to calling this function.
 */
void sample_mv_bernoulli_pmf_f
(
    Vector_u32**    x_out,
    const Vector_f* mu
)
{
    uint32_t    d, D;
    Vector_u32* x;

    D = mu->num_elts;

    create_vector_u32(x_out, D);
    x = *x_out;

    for (d = 0; d < D; d++)
    {
        x->elts[ d ] = sample_bernoulli_pmf_f(mu->elts[ d ]);
    }
}

/**
 * @param  x_out  Result paramter. Contains the sample.
 * @param  mu     Probabilities of success of the Bernoulli.
 *
 * @note Does not seed the rand() function. If this is desired, srand() must
 * be called prior to calling this function.
 */
void sample_mv_bernoulli_pmf_d
(
    Vector_u32**    x_out,
    const Vector_d* mu
)
{
    uint32_t    d, D;
    Vector_u32* x;

    D = mu->num_elts;

    create_vector_u32(x_out, D);
    x = *x_out;

    for (d = 0; d < D; d++)
    {
        x->elts[ d ] = sample_bernoulli_pmf_d(mu->elts[ d ]);
    }
}

/*@}*/


/** 
 * @name set_sample_mv_bernoulli_pmf
 *
 * Draws a set of random samples from the multivariate Bernoulli pmf.
 * @{
 */

/**
 * @param  x_out  Result paramter. Contains the sample.
 * @param  N      Number of samples in the set.
 * @param  mu     Probabilities of success of the Bernoulli.
 *
 * @note Does not seed the rand() function. If this is desired, srand() must
 * be called prior to calling this function.
 */
void set_sample_mv_bernoulli_pmf_f
(
    Matrix_u32**    x_out,
    uint32_t        N,
    const Vector_f* mu
)
{
    uint32_t    D;
    uint32_t    n;
    Vector_u32* x_n = NULL;

    D = mu->num_elts;

    create_matrix_u32(x_out, N, D);

    for (n = 0; n < N; n++)
    {
        sample_mv_bernoulli_pmf_f(&x_n, mu);
        copy_vector_into_matrix_u32(x_out, *x_out, x_n, MATRIX_ROW_VECTOR, n);
    }

    free_vector_u32(x_n);
}

/**
 * @param  x_out  Result paramter. Contains the sample.
 * @param  N      Number of samples in the set.
 * @param  mu     Probabilities of success of the Bernoulli.
 *
 * @note Does not seed the rand() function. If this is desired, srand() must
 * be called prior to calling this function.
 */
void set_sample_mv_bernoulli_pmf_d
(
    Matrix_u32**    x_out,
    uint32_t        N,
    const Vector_d* mu
)
{
    uint32_t    D;
    uint32_t    n;
    Vector_u32* x_n = NULL;

    D = mu->num_elts;

    create_matrix_u32(x_out, N, D);

    for (n = 0; n < N; n++)
    {
        sample_mv_bernoulli_pmf_d(&x_n, mu);
        copy_vector_into_matrix_u32(x_out, *x_out, x_n, MATRIX_ROW_VECTOR, n);
    }

    free_vector_u32(x_n);
}

/*@}*/
