/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Multivariate probability mass functions.
 */


#ifndef MV_PMF_H
#define MV_PMF_H


#include <jwsc/config.h>

#include <stdlib.h>
#include <inttypes.h>

#include "jwsc/vector/vector.h"
#include "jwsc/matrix/matrix.h"


#ifdef __cplusplus
namespace jwsc {
extern "C" {
#endif


/** 
 * @name mv_bernoulli_pmf
 *
 * Computes the multivariate Bernoulli pmf at a point.
 * @{
 */

/** 
 * @brief Computes the single precision multivariate Bernoulli pmf at a point.
 */
void mv_bernoulli_pmf_f
(
    float*            p_out,
    const Vector_f*   mu, 
    const Vector_u32* x
);

/** 
 * @brief Computes the double precision multivariate Bernoulli pmf at a point.
 */
void mv_bernoulli_pmf_d
(
    double*           p_out,
    const Vector_d*   mu, 
    const Vector_u32* x
);

/*@}*/


/** 
 * @name sample_mv_bernoulli_pmf
 *
 * Draws a random sample from the multivariate Bernoulli pmf.
 * @{
 */

/**
 * @brief Draws a single precision random sample from the multivariate
 * Bernoulli pmf.
 */
void sample_mv_bernoulli_pmf_f
(
    Vector_u32**    x_out,
    const Vector_f* mu
);

/**
 * @brief Draws a double precision random sample from the multivariate
 * Bernoulli pmf.
 */
void sample_mv_bernoulli_pmf_d
(
    Vector_u32**    x_out,
    const Vector_d* mu
);

/*@}*/


/** 
 * @name set_sample_mv_bernoulli_pmf
 *
 * Draws a set of random samples from the multivariate Bernoulli pmf.
 * @{
 */

/**
 * @brief Draws a set of single precision random samples from the multivariate
 * Bernoulli pmf.
 */
void set_sample_mv_bernoulli_pmf_f
(
    Matrix_u32**    x_out,
    uint32_t        N,
    const Vector_f* mu
);

/**
 * @brief Draws a set of double precision random samples from the multivariate
 * Bernoulli pmf.
 */
void set_sample_mv_bernoulli_pmf_d
(
    Matrix_u32**    x_out,
    uint32_t        N,
    const Vector_d* mu
);

/*@}*/


#ifdef __cplusplus
}
}
#endif


#endif
