/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Declarations for univariate probability density functions.
 */


#ifndef PDF_H
#define PDF_H


#include <jwsc/config.h>

#include <stdlib.h>


#ifdef __cplusplus
namespace jwsc {
extern "C" {
#endif


/** 
 * @name standard_gaussian_pdf
 *
 * Computes the univariate standard Gaussian pdf at a point.
 * @{
 */

/**
 * @brief Computes the single precision univariate standard Gaussian pdf at a
 * point.
 */
float standard_gaussian_pdf_f(float x);

/** 
 * @brief Computes the double precision univariate standard Gaussian pdf at a
 * point.
 */
double standard_gaussian_pdf_d(double x);

/*@}*/


/** 
 * @name gaussian_pdf
 *
 * Computes the univariate Gaussian pdf at a point.
 * @{
 */

/** @brief Computes the single precision univariate Gaussian pdf at a point. */
float gaussian_pdf_f(float mu, float sigma, float x);

/** @brief Computes the double precision univariate Gaussian pdf at a point. */
double gaussian_pdf_d(double mu, double sigma, double x);

/*@}*/


/** 
 * @name log_standard_gaussian_pdf
 *
 * Computes the log of the univariate standard Gaussian pdf at a point.
 * @{
 */

/** 
 * @brief Computes the single precision log of the univariate standard Gaussian 
 * pdf at a point.
 */
float log_standard_gaussian_pdf_f(float x);

/** 
 * @brief Computes the double precision log of the univariate standard
 * Gaussian pdf at a point.
 */
double log_standard_gaussian_pdf_d(double x);

/*@}*/


/** 
 * @name log_gaussian_pdf
 *
 * Computes the log of the univariate Gaussian pdf at a point.
 * @{
 */

/** 
 * @brief Computes the single precision log of the univariate Gaussian pdf at
 * a point.
 */
float log_gaussian_pdf_f(float mu, float sigma, float x);

/** 
 * @brief Computes the double precision log of the univariate Gaussian pdf at
 * a point.
 */
double log_gaussian_pdf_d(double mu, double sigma, double x);

/*@}*/


#if defined JWSC_HAVE_SLATEC || defined JWSC_HAVE_ERF
/** 
 * @name integrate_gaussian_pdf
 *
 * Integrates the Gaussian pdf over a finite interval using the error function.
 * @{
 */

/** 
 * @brief Integrates the single precision Gaussian pdf over a finite interval
 * using the error function.
 */
float integrate_gaussian_pdf_f
(
    float mu, 
    float sigma, 
    float a, 
    float b
);


/** 
 * @brief Integrates the double precision Gaussian pdf over a finite interval
 * using the error function.
 */
double integrate_gaussian_pdf_d
(
    double mu, 
    double sigma, 
    double a, 
    double b
);

/*@}*/
#endif


/** 
 * @name sample_standard_gaussian_pdf
 *
 * Draws a random sample from the standard Gaussian pdf.
 * @{
 */

/** 
 * @brief Draws a single precision random sample from the standard Gaussian
 * pdf.
 */
float sample_standard_gaussian_pdf_f(float a, float b);

/** 
 * @brief Draws a double precision random sample from the standard Gaussian
 * pdf.
 */
double sample_standard_gaussian_pdf_d(double a, double b);

/*@}*/


/** 
 * @name sample_gaussian_pdf
 *
 * Draws a random sample from the Gaussian pdf.
 * @{
 */

/** @brief Draws a single precision random sample from the Gaussian pdf. */
float sample_gaussian_pdf_f
(
    float mu, 
    float sigma, 
    float a, 
    float b
);

/** @brief Draws a double precision random sample from the Gaussian pdf. */
double sample_gaussian_pdf_d
(
    double mu, 
    double sigma, 
    double a, 
    double b
);

/*@}*/


#if defined JWSC_HAVE_TGAMMA
/** 
 * @name gamma_pdf
 *
 * Computes the univariate gamma pdf at a point.
 * @{
 */

/** @brief Computes the single precision univariate gamma pdf at a point. */
float gamma_pdf_f(float alpha, float beta, float x);

/** @brief Computes the double precision univariate gamma pdf at a point. */
double gamma_pdf_d(double alpha, double beta, double x);

/*@}*/
#endif


#if defined JWSC_HAVE_LGAMMA
/** 
 * @name log_gamma_pdf
 *
 * Computes the log of the univariate gamma pdf at a point.
 * @{
 */

/** 
 * @brief Computes the single precision log of the univariate gamma pdf at
 * a point.
 */
float log_gamma_pdf_f(float alpha, float beta, float x);

/** 
 * @brief Computes the double precision log of the univariate gamma pdf at
 * a point.
 */
double log_gamma_pdf_d(double alpha, double beta, double x);

/*@}*/
#endif


#if defined JWSC_HAVE_TGAMMA
/** 
 * @name sample_gamma_pdf
 *
 * Draws a random sample from the gamma pdf.
 * @{
 */

/** @brief Draws a single precision random sample from the gamma pdf. */
float sample_gamma_pdf_f
(
    float alpha, 
    float beta, 
    float a, 
    float b
);

/** @brief Draws a double precision random sample from the gamma pdf. */
double sample_gamma_pdf_d
(
    double alpha, 
    double beta, 
    double a, 
    double b
);

/*@}*/
#endif


/** 
 * @name sample_uniform_pdf
 *
 * Draws a random sample from the Uniform pdf.
 * @{
 */

/** @brief Draws a single precision random sample from the Uniform pdf. */
float sample_uniform_pdf_f(float a, float b);

/** @brief Draws a double precision random sample from the Uniform pdf. */
double sample_uniform_pdf_d(double a, double b);

/*@}*/


/** 
 * @name sample_sine_pdf
 *
 * Draws a random sample from the Sine pdf over the interval [0,PI/2].
 * @{
 */

/** 
 * @brief Draws a single precision random sample from the Sine pdf over the
 * interval [0,PI/2].
 */
float sample_sine_pdf_f();

/** 
 * @brief Draws a double precision random sample from the Sine pdf over the
 * interval [0.PI/2].
 */
double sample_sine_pdf_d();

/*@}*/


#ifdef __cplusplus
}
}
#endif


#endif
