/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Declarations for univariate probability mass functions.
 */


#ifndef PMF_H
#define PMF_H


#include <jwsc/config.h>

#include <stdlib.h>
#include <inttypes.h>


#ifdef __cplusplus
namespace jwsc {
extern "C" {
#endif


/** 
 * @name poission_pmf
 *
 * Computes the Poission pmf at a point.
 * @{
 */

/** @brief Computes the double precision Poisson pmf at a point. */
float poisson_pmf_f(float lambda, uint32_t n);

/** @brief Computes the double precision Poisson pmf at a point. */
double poisson_pmf_d(double lambda, uint32_t n);

/*@}*/


/** 
 * @name sample_poisson_pmf
 *
 * Draws a random sample from the Poisson pmf.
 * @{
 */

/** @brief Draws a single precision random sample from the Poisson pmf. */
uint32_t sample_poisson_pmf_f(float lambda, uint32_t a, uint32_t b);

/** @brief Draws a double precision random sample from the Poisson pmf. */
uint32_t sample_poisson_pmf_d(double lambda, uint32_t a, uint32_t b);

/*@}*/


/** 
 * @name bernoulli_pmf
 *
 * Computes the Bernoulli pmf at a point.
 * @{
 */

/** @brief Computes the single precision Bernoulli pmf at a point. */
float bernoulli_pmf_f(float p, uint32_t n);

/** @brief Computes the double precision Bernoulli pmf at a point. */
double bernoulli_pmf_d(double p, uint32_t n);

/*@}*/


/** 
 * @name sample_bernoulli_pmf
 *
 * Draws a random sample from the Bernoulli pmf.
 * @{
 */

/** @brief Draws a single precision random sample from the Bernoulli pmf. */
uint8_t sample_bernoulli_pmf_f(float p);

/** @brief Draws a double precision random sample from the Bernoulli pmf. */
uint8_t sample_bernoulli_pmf_d(double p);

/*@}*/


/** 
 * @name geometric_pmf
 *
 * Computes the Geometric pmf at a point.
 * @{
 */

/** @brief Computes the single precision Geometric pmf at a point. */
float geometric_pmf_f(float p, uint32_t n);

/** @brief Computes the double precision Geometric pmf at a point. */
double geometric_pmf_d(double p, uint32_t n);

/*@}*/


/** 
 * @name log_geometric_pmf
 *
 * Computes the log of the Geometric pmf at a point.
 * @{
 */

/** @brief Computes the single precision log Geometric pmf at a point. */
float log_geometric_pmf_f(float p, uint32_t n);

/** @brief Computes the double precision log Geometric pmf at a point. */
double log_geometric_pmf_d(double p, uint32_t n);

/*@}*/


/** 
 * @name sample_geometric_pmf
 *
 * Draws a random sample from the Geometric pmf.
 * @{
 */

/** @brief Draws a single precision random sample from the Geometric pmf. */
uint32_t sample_geometric_pmf_f(float p);

/** @brief Draws a double precision random sample from the Geometric pmf. */
uint32_t sample_geometric_pmf_d(double p);

/*@}*/


#ifdef __cplusplus
}
}
#endif


#endif
