/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Collection of functions for stochastic dynamics on an energy function 
 * to generate random samples.
 */


#ifndef DYNAMICS_H 
#define DYNAMICS_H 


#include <jwsc/config.h>

#include <stdlib.h>
#include <inttypes.h>

#include "jwsc/base/error.h"
#include "jwsc/vector/vector.h"
#include "jwsc/matrix/matrix.h"


#ifdef __cplusplus
namespace jwsc {
extern "C" {
#endif


/** 
 * @name stochastic_dynamics
 *
 * Leapfrog discretization of stochastic dynamics to sample from an energy
 * function.
 * @{
 */

/** @brief Double precision leapfrog stochastic dynamics. */
Error* stochastic_dynamics_1_d
(
    uint32_t iterations,
    double   delta_t,
    double   alpha,
    uint32_t kick,
    void*    params,
    void     (*get_pt_from_params)(Vector_d** p_out, const void* params),
    Error*   (*grad_energy_func)(Vector_d** grad_out, void* params),
    Error*   (*accept_sample)(const Vector_d* sample, const Vector_d* momenta, void* params)
);


/** @brief Double precision leapfrog stochastic dynamics. */
Error* stochastic_dynamics_2_d
(
    uint32_t        iterations,
    const Vector_d* delta_t,
    double          alpha,
    uint32_t        kick,
    void*           params,
    void            (*get_pt_from_params)(Vector_d** p_out, const void* params),
    Error*          (*grad_energy_func)(Vector_d** grad_out, void* params),
    Error*          (*accept_sample)(const Vector_d* sample, const Vector_d* momenta, void* params)
);

/*@}*/




/** 
 * @name hybrid_monte_carlo
 *
 * Stochastic dynamics combined with Metropolis-Hastings for quickly converging
 * and unbiased sampling.
 * @{
 */

/** @brief Double precision hybrid monte carlo. */
Error* hybrid_monte_carlo_1_d
(
    uint32_t iterations,
    uint32_t leap_itera,
    double   delta_t,
    double   alpha,
    uint32_t kick,
    void*    params,
    void     (*get_pt_from_params)(Vector_d** p_out, const void* params),
    Error*   (*set_params_from_pt)(const Vector_d* p, void* params),
    Error*   (*energy_func)(double* energy_out, void* params),
    Error*   (*grad_energy_func)(Vector_d** grad_out, void* params),
    Error*   (*accept_sample)(const Vector_d* sample, const Vector_d* momenta, void* params),
    Error*   (*reject_sample)(const Vector_d* sample, const Vector_d* momenta, void* params)
);

/** @brief Double precision hybrid monte carlo. */
Error* hybrid_monte_carlo_2_d
(
    uint32_t        iterations,
    uint32_t        leap_itera,
    const Vector_d* delta_t,
    double          alpha,
    uint32_t        kick,
    void*           params,
    void            (*get_pt_from_params)(Vector_d** p_out, const void* params),
    Error*          (*set_params_from_pt)(const Vector_d* p, void* params),
    Error*          (*energy_func)(double* energy_out, void* params),
    Error*          (*grad_energy_func)(Vector_d** grad_out, void* params),
    Error*          (*accept_sample)(const Vector_d* sample, const Vector_d* momenta, void* params),
    Error*          (*reject_sample)(const Vector_d* sample, const Vector_d* momenta, void* params)
);

/*@}*/




/** 
 * @name langevin
 *
 * Langevin dynamics for generating samples from a energy function.
 * @{
 */

/** @brief Single precision Langevin dynamics. */
Error* langevin_1_f
(
    uint32_t iterations,
    float    delta_t,
    void*    params,
    void     (*get_pt_from_params)(Vector_f** p_out, const void* params),
    Error*   (*grad_energy_func)(Vector_f** grad_out, void* params),
    Error*   (*accept_sample)(const Vector_f* sample, void* params)
);

/** @brief Single precision Langevin dynamics. */
Error* langevin_2_f
(
    uint32_t        iterations,
    const Vector_f* delta_t,
    void*           params,
    void            (*get_pt_from_params)(Vector_f** p_out, const void* params),
    Error*          (*grad_energy_func)(Vector_f** grad_out, void* params),
    Error*          (*accept_sample)(const Vector_f* sample, void* params)
);

/** @brief Double precision Langevin dynamics. */
Error* langevin_1_d
(
    uint32_t iterations,
    double   delta_t,
    void*    params,
    void     (*get_pt_from_params)(Vector_d** p_out, const void* params),
    Error*   (*grad_energy_func)(Vector_d** grad_out, void* params),
    Error*   (*accept_sample)(const Vector_d* sample, void* params)
);

/** @brief Double precision Langevin dynamics. */
Error* langevin_2_d
(
    uint32_t        iterations,
    const Vector_d* delta_t,
    void*           params,
    void            (*get_pt_from_params)(Vector_d** p_out, const void* params),
    Error*          (*grad_energy_func)(Vector_d** grad_out, void* params),
    Error*          (*accept_sample)(const Vector_d* sample, void* params)
);

/*@}*/




/** 
 * @name langevin_hyperdynamics
 *
 * Voter's hyperdynamics biasing for generating samples from an energy
 * function using Langevin dynamics.
 * @{
 */

/** @brief Single precision Voter hyperdynamics. */
Error* langevin_hyperdynamics_1_f
(
    uint32_t iterations,
    float    delta_t,
    float    gamma,
    float    eta,
    float    h,
    float    d,
    void*    params,
    void     (*get_pt_from_params)(Vector_f** p_out, const void* params),
    Error*   (*set_params_from_pt)(const Vector_f* p, void* params),
    Error*   (*energy_func)(float* energy_out, void* params),
    Error*   (*grad_energy_func)(Vector_f** grad_out, void* params),
    Error*   (*accept_sample)(const Vector_f* sample, void* params, float energy_bias)
);

/** @brief Single precision Voter hyperdynamics. */
Error* langevin_hyperdynamics_2_f
(
    uint32_t        iterations,
    const Vector_f* init_delta_t,
    float           gamma,
    float           eta,
    float           h,
    float           d,
    void*           params,
    void            (*get_pt_from_params)(Vector_f** p_out, const void* params),
    Error*          (*set_params_from_pt)(const Vector_f* p, void* params),
    Error*          (*energy_func)(float* energy_out, void* params),
    Error*          (*grad_energy_func)(Vector_f** grad_out, void* params),
    Error*          (*accept_sample)(const Vector_f* sample, void* params, float energy_bias)
);

/** @brief Double precision Voter hyperdynamics. */
Error* langevin_hyperdynamics_1_d
(
    uint32_t iterations,
    double   delta_t,
    double   gamma,
    double   eta,
    double   h,
    double   d,
    void*    params,
    void     (*get_pt_from_params)(Vector_d** p_out, const void* params),
    Error*   (*set_params_from_pt)(const Vector_d* p, void* params),
    Error*   (*energy_func)(double* energy_out, void* params),
    Error*   (*grad_energy_func)(Vector_d** grad_out, void* params),
    Error*   (*accept_sample)(const Vector_d* sample, void* params, double energy_bias)
);

/** @brief Double precision Voter hyperdynamics. */
Error* langevin_hyperdynamics_2_d
(
    uint32_t        iterations,
    const Vector_d* init_delta_t,
    double          gamma,
    double          eta,
    double          h,
    double          d,
    void*           params,
    void            (*get_pt_from_params)(Vector_d** p_out, const void* params),
    Error*          (*set_params_from_pt)(const Vector_d* p, void* params),
    Error*          (*energy_func)(double* energy_out, void* params),
    Error*          (*grad_energy_func)(Vector_d** grad_out, void* params),
    Error*          (*accept_sample)(const Vector_d* sample, void* params, double energy_bias)
);

/*@}*/


#ifdef __cplusplus
}
}
#endif


#endif
