/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Collection of functions for solving regression problems.
 */


#ifndef REGRESS_H 
#define REGRESS_H 


#include <jwsc/config.h>

#include <stdlib.h>
#include <inttypes.h>

#include "jwsc/base/error.h"
#include "jwsc/vector/vector.h"
#include "jwsc/matrix/matrix.h"


#ifdef __cplusplus
namespace jwsc {
extern "C" {
#endif


#if defined JWSC_HAVE_LAPACK
/** 
 * @name linear_least_squares_regress
 *
 * Linear regression using least squares.
 * @{
 */

/** @brief Single precision linear regression using least squares. */
Error* linear_least_squares_regress_f
(
    Vector_f**      w_out,
    const Matrix_f* x,
    const Vector_f* y
);

/** @brief Double precision linear regression using least squares. */
Error* linear_least_squares_regress_d
(
    Vector_d**      w_out,
    const Matrix_d* x,
    const Vector_d* y
);

/*@}*/




/** 
 * @name linear_least_squares_ransac_regress
 *
 * Linear regression using RANSAC and least squares.
 * @{
 */

/** @brief Single precision linear regression using RANSAC and least squares. */
Error* linear_least_squares_ransac_regress_f
(
    Vector_f**      w_out,
    const Matrix_f* x,
    const Vector_f* y,
    uint32_t        n,
    uint32_t        k,
    float           t,
    uint32_t        d
);

/** @brief Double precision linear regression using RANSAC and least squares. */
Error* linear_least_squares_ransac_regress_d
(
    Vector_d**      w_out,
    const Matrix_d* x,
    const Vector_d* y,
    uint32_t        n,
    uint32_t        k,
    double          t,
    uint32_t        d
);

/*@}*/
#endif


#ifdef __cplusplus
}
}
#endif


#endif
