/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Collection of functions for computing sample statistics.
 */


#ifndef STAT_H 
#define STAT_H 


#include <jwsc/config.h>

#include <stdlib.h>
#include <inttypes.h>

#include "jwsc/base/error.h"
#include "jwsc/vector/vector.h"
#include "jwsc/matrix/matrix.h"


#ifdef __cplusplus
namespace jwsc {
extern "C" {
#endif


/** 
 * @name sample_mean
 *
 * Computes the mean of a random sample.
 * @{
 */

/** @brief Computes the single precision sample mean of a random sample. */
void sample_mean_f(float* mean_out, const Vector_f* x);

/** @brief Computes the double precision sample mean of a random sample. */
void sample_mean_d(double* mean_out, const Vector_d* x);

/*@}*/


/** 
 * @name sample_variance
 *
 * Computes the variance of a random sample.
 * @{
 */

/** @brief Computes the single precision sample variance of a random sample. */
void sample_variance_f(float* var_out, const Vector_f* x);

/** @brief Computes the double precision sample variance of a random sample. */
void sample_variance_d(double* var_out, const Vector_d* x);

/*@}*/


/** 
 * @name sample_error
 *
 * Computes the standard error of a random sample mean.
 * @{
 */

/** 
 * @brief Computes the single precision standard error of a random sample
 * mean.
 */
void sample_error_f(float* err_out, const Vector_f* x);

/** 
 * @brief Computes the double precision standard error of a random sample
 * mean.
 */
void sample_error_d(double* err_out, const Vector_d* x);

/*@}*/


/** 
 * @name sample_stats
 *
 * Computes the mean, variance, and standard error of a random sample.
 * @{
 */

/** 
 * @brief Computes the single precision mean, variance, and standard error of
 * a random sample.
 */
void sample_stats_f
(
    float*          mean_out, 
    float*          var_out,
    float*          err_out, 
    const Vector_f* x
);

/** 
 * @brief Computes the double precision mean, variance, and standard error of
 * a random sample.
 */
void sample_stats_d
(
    double*         mean_out, 
    double*         var_out, 
    double*         err_out, 
    const Vector_d* x
);

/*@}*/


/** 
 * @name ind_mv_sample_stats
 *
 * Computes the mean, variance, and standard error of a set of independent
 * random variables in a sample.
 * @{
 */

/** 
 * @brief Computes the single precision mean, variance, and standard error of
 * a set of independent random variables in a sample.
 */
void ind_mv_sample_stats_f
(
    Vector_f**      means_out, 
    Vector_f**      vars_out,
    Vector_f**      errs_out, 
    const Matrix_f* x
);

/** 
 * @brief Computes the double precision mean, variance, and standard error of
 * a set of independent random variables in a sample.
 */
void ind_mv_sample_stats_d
(
    Vector_d**      means_out, 
    Vector_d**      vars_out, 
    Vector_d**      errs_out, 
    const Matrix_d* x
);

/*@}*/


/** 
 * @name mv_sample_mean
 *
 * Computes the mean of a multivariate random sample.
 * @{
 */

/** 
 * @brief Computes the single precision sample mean of a multivariate random
 * sample.
 */
void mv_sample_mean_f(Vector_f** mean_out, const Matrix_f* x);

/** 
 * @brief Computes the double precision sample mean of a multivariate random
 * sample.
 */
void mv_sample_mean_d(Vector_d** mean_out, const Matrix_d* x);

/*@}*/


/** 
 * @name mv_sample_variance
 *
 * Computes the variance of a multivariate random sample.
 * @{
 */

/** 
 * @brief Computes the single precision sample covariance of a multivariate
 * random sample.
 */
void mv_sample_covariance_f(Matrix_f** cov_out, const Matrix_f* x);

/** 
 * @brief Computes the double precision sample covariance of a multivariate
 * random sample.
 */
void mv_sample_covariance_d(Matrix_d** cov_out, const Matrix_d* x);

/*@}*/


/** 
 * @name mv_sample_stats
 *
 * Computes the mean, variance, and standard error of a multivariate random
 * sample.
 * @{
 */

/** 
 * @brief Computes the single precision mean, covariance, and standard error of
 * a multivariate random sample.
 */
void mv_sample_stats_f
(
    Vector_f**      mean_out, 
    Matrix_f**      cov_out,
    const Matrix_f* x
);

/** 
 * @brief Computes the double precision mean, covariance, and standard error of
 * a multivariate random sample.
 */
void mv_sample_stats_d
(
    Vector_d**      mean_out, 
    Matrix_d**      cov_out, 
    const Matrix_d* x
);

/*@}*/


#ifdef __cplusplus
}
}
#endif


#endif
