/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Declarations for a vector and associated functions.
 *
 * The vector holds integers and floating point real and complex numbers.
 *
 * @note Accelerated if linked against an optimized BLAS.
 *
 * Some of the functions are accelerated if linked against an optimized BLAS
 * library.
 */


#include <jwsc/config.h>

#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include <inttypes.h>

#include "jwsc/base/error.h"
#include "jwsc/math/complex.h"
#include "jwsc/math/blas.h"
#include "jwsc/vector/vector.h"

#ifdef JWSC_HAVE_DMALLOC
#include <dmalloc.h>
#endif


/**
 * @name create_vector
 *
 * Creates a vector of integer or floating point values.
 * @{
 */

/**
 * Does not initialize the vector.
 *
 * @param  v_out     Result parameter. If @em *v_out is NULL, a vector is
 *                   allocated; otherwise its space is re-used.
 * @param  num_elts  Number of elements in the vector.
 *
 * @note Number of elements must be > 0.
 */
void create_vector_u32(Vector_u32** v_out, uint32_t num_elts)
{
    Vector_u32* v;

    assert(num_elts > 0);

    if (*v_out == NULL)
    {
        assert(*v_out = malloc(sizeof(Vector_u32)));
        (*v_out)->num_elts = 0;
        (*v_out)->elts     = NULL;
    }
    v = *v_out;

    if (v->num_elts != num_elts)
    {
        v->elts = realloc(v->elts, num_elts*sizeof(uint32_t));
        assert(v->elts);
        v->num_elts = num_elts;
    }
}

/**
 * Does not initialize the vector.
 *
 * @param  v_out     Result parameter. If @em *v_out is NULL, a vector is
 *                   allocated; otherwise its space is re-used.
 * @param  num_elts  Number of elements in the vector.
 *
 * @note Number of elements must be > 0.
 */
void create_vector_i32(Vector_i32** v_out, uint32_t num_elts)
{
    Vector_i32* v;

    assert(num_elts > 0);

    if (*v_out == NULL)
    {
        assert(*v_out = malloc(sizeof(Vector_i32)));
        (*v_out)->num_elts = 0;
        (*v_out)->elts     = NULL;
    }
    v = *v_out;

    if (v->num_elts != num_elts)
    {
        v->elts = realloc(v->elts, num_elts*sizeof(int32_t));
        assert(v->elts);
        v->num_elts = num_elts;
    }
}

/**
 * Does not initialize the vector.
 *
 * @param  v_out     Result parameter. If @em *v_out is NULL, a vector is
 *                   allocated; otherwise its space is re-used.
 * @param  num_elts  Number of elements in the vector.
 *
 * @note Number of elements must be > 0.
 */
void create_vector_i64(Vector_i64** v_out, uint32_t num_elts)
{
    Vector_i64* v;

    assert(num_elts > 0);

    if (*v_out == NULL)
    {
        assert(*v_out = malloc(sizeof(Vector_i64)));
        (*v_out)->num_elts = 0;
        (*v_out)->elts     = NULL;
    }
    v = *v_out;

    if (v->num_elts != num_elts)
    {
        v->elts = realloc(v->elts, num_elts*sizeof(int64_t));
        assert(v->elts);
        v->num_elts = num_elts;
    }
}

/**
 * Does not initialize the vector.
 *
 * @param  v_out     Result parameter. If @em *v_out is NULL, a vector is
 *                   allocated; otherwise its space is re-used.
 * @param  num_elts  Number of elements in the vector.
 *
 * @note Number of elements must be > 0.
 */
void create_vector_f(Vector_f** v_out, uint32_t num_elts)
{
    Vector_f* v;

    assert(num_elts > 0);

    if (*v_out == NULL)
    {
        assert(*v_out = malloc(sizeof(Vector_f)));
        (*v_out)->num_elts = 0;
        (*v_out)->elts     = NULL;
    }
    v = *v_out;

    if (v->num_elts != num_elts)
    {
        v->elts = realloc(v->elts, num_elts*sizeof(float));
        assert(v->elts);
        v->num_elts = num_elts;
    }
}

/**
 * Does not initialize the vector.
 *
 * @param  v_out     Result parameter. If @em *v_out is NULL, a vector is
 *                   allocated; otherwise its space is re-used.
 * @param  num_elts  Number of elements in the vector.
 *
 * @note Number of elements must be > 0.
 */
void create_vector_d(Vector_d** v_out, uint32_t num_elts)
{
    Vector_d* v;

    assert(num_elts > 0);

    if (*v_out == NULL)
    {
        assert(*v_out = malloc(sizeof(Vector_d)));
        (*v_out)->num_elts = 0;
        (*v_out)->elts     = NULL;
    }
    v = *v_out;

    if (v->num_elts != num_elts)
    {
        v->elts = realloc(v->elts, num_elts*sizeof(double));
        assert(v->elts);
        v->num_elts = num_elts;
    }
}

/**
 * Does not initialize the vector.
 *
 * @param  v_out     Result parameter. If @em *v_out is NULL, a vector is
 *                   allocated; otherwise its space is re-used.
 * @param  num_elts  Number of elements in the vector.
 *
 * @note Number of elements must be > 0.
 */
void create_vector_cf(Vector_cf** v_out, uint32_t num_elts)
{
    Vector_cf* v;

    assert(num_elts > 0);

    if (*v_out == NULL)
    {
        assert(*v_out = malloc(sizeof(Vector_cf)));
        (*v_out)->num_elts = 0;
        (*v_out)->elts     = NULL;
    }
    v = *v_out;

    if (v->num_elts != num_elts)
    {
        v->elts = realloc(v->elts, num_elts*sizeof(Complex_f));
        assert(v->elts);
        v->num_elts = num_elts;
    }
}

/**
 * Does not initialize the vector.
 *
 * @param  v_out     Result parameter. If @em *v_out is NULL, a vector is
 *                   allocated; otherwise its space is re-used.
 * @param  num_elts  Number of elements in the vector.
 *
 * @note Number of elements must be > 0.
 */
void create_vector_cd(Vector_cd** v_out, uint32_t num_elts)
{
    Vector_cd* v;

    assert(num_elts > 0);

    if (*v_out == NULL)
    {
        assert(*v_out = malloc(sizeof(Vector_cd)));
        (*v_out)->num_elts = 0;
        (*v_out)->elts     = NULL;
    }
    v = *v_out;

    if (v->num_elts != num_elts)
    {
        v->elts = realloc(v->elts, num_elts*sizeof(Complex_d));
        assert(v->elts);
        v->num_elts = num_elts;
    }
}

/*@}*/




/**
 * @name create_init_vector
 *
 * Creates and initializes a vector of integer or floating point values.
 * @{
 */

/**
 * @param  v_out     Result parameter. If @em *v_out is NULL, a vector is
 *                   allocated; otherwise its space is re-used.
 * @param  num_elts  Number of elements in the vector.
 * @param  val       Value to initialize all elements of the vector with.
 *
 * @note Number of elements must be > 0.
 */
void create_init_vector_u32(Vector_u32** v_out, uint32_t num_elts, uint32_t val)
{
    uint32_t    elt;
    Vector_u32* v;

    create_vector_u32(v_out, num_elts);
    v = *v_out;

    for (elt = 0; elt < num_elts; elt++)
    {
        v->elts[ elt ] = val;
    }
}

/**
 * @param  v_out     Result parameter. If @em *v_out is NULL, a vector is
 *                   allocated; otherwise its space is re-used.
 * @param  num_elts  Number of elements in the vector.
 * @param  val       Value to initialize all elements of the vector with.
 *
 * @note Number of elements must be > 0.
 */
void create_init_vector_i32(Vector_i32** v_out, uint32_t num_elts, int32_t val)
{
    uint32_t    elt;
    Vector_i32* v;

    create_vector_i32(v_out, num_elts);
    v = *v_out;

    for (elt = 0; elt < num_elts; elt++)
    {
        v->elts[ elt ] = val;
    }
}

/**
 * @param  v_out     Result parameter. If @em *v_out is NULL, a vector is
 *                   allocated; otherwise its space is re-used.
 * @param  num_elts  Number of elements in the vector.
 * @param  val       Value to initialize all elements of the vector with.
 *
 * @note Number of elements must be > 0.
 */
void create_init_vector_i64(Vector_i64** v_out, uint32_t num_elts, int64_t val)
{
    uint32_t    elt;
    Vector_i64* v;

    create_vector_i64(v_out, num_elts);
    v = *v_out;

    for (elt = 0; elt < num_elts; elt++)
    {
        v->elts[ elt ] = val;
    }
}

/**
 * @param  v_out     Result parameter. If @em *v_out is NULL, a vector is
 *                   allocated; otherwise its space is re-used.
 * @param  num_elts  Number of elements in the vector.
 * @param  val       Value to initialize all elements of the vector with.
 *
 * @note Number of elements must be > 0.
 */
void create_init_vector_f(Vector_f** v_out, uint32_t num_elts, float val)
{
    uint32_t  elt;
    Vector_f* v;

    create_vector_f(v_out, num_elts);
    v = *v_out;

    for (elt = 0; elt < num_elts; elt++)
    {
        v->elts[ elt ] = val;
    }
}

/**
 * @param  v_out     Result parameter. If @em *v_out is NULL, a vector is
 *                   allocated; otherwise its space is re-used.
 * @param  num_elts  Number of elements in the vector.
 * @param  val       Value to initialize all elements of the vector with.
 *
 * @note Number of elements must be > 0.
 */
void create_init_vector_d(Vector_d** v_out, uint32_t num_elts, double val)
{
    uint32_t  elt;
    Vector_d* v;

    create_vector_d(v_out, num_elts);
    v = *v_out;

    for (elt = 0; elt < num_elts; elt++)
    {
        v->elts[ elt ] = val;
    }
}

/**
 * @param  v_out     Result parameter. If @em *v_out is NULL, a vector is
 *                   allocated; otherwise its space is re-used.
 * @param  num_elts  Number of elements in the vector.
 * @param  val       Value to initialize all elements of the vector with.
 *
 * @note Number of elements must be > 0.
 */
void create_init_vector_cf(Vector_cf** v_out, uint32_t num_elts, Complex_f val)
{
    uint32_t   elt;
    Vector_cf* v;

    create_vector_cf(v_out, num_elts);
    v = *v_out;

    for (elt = 0; elt < num_elts; elt++)
    {
        v->elts[ elt ] = val;
    }
}

/**
 * @param  v_out     Result parameter. If @em *v_out is NULL, a vector is
 *                   allocated; otherwise its space is re-used.
 * @param  num_elts  Number of elements in the vector.
 * @param  val       Value to initialize all elements of the vector with.
 *
 * @note Number of elements must be > 0.
 */
void create_init_vector_cd(Vector_cd** v_out, uint32_t num_elts, Complex_d val)
{
    uint32_t   elt;
    Vector_cd* v;

    create_vector_cd(v_out, num_elts);
    v = *v_out;

    for (elt = 0; elt < num_elts; elt++)
    {
        v->elts[ elt ] = val;
    }
}

/*@}*/




/**
 * @name create_zero_vector
 *
 * Creates a vector of integer or floating point values and initializes it with 
 * zeros.
 * @{
 */

/**
 * @param  v_out     Result parameter. If @em *v_out is NULL, a vector is
 *                   allocated; otherwise its space is re-used.
 * @param  num_elts  Number of elements in the vector.
 *
 * @note Number of elements must be > 0.
 */
void create_zero_vector_u32(Vector_u32** v_out, uint32_t num_elts)
{
    create_init_vector_u32(v_out, num_elts, 0);
}


/**
 * @param  v_out     Result parameter. If @em *v_out is NULL, a vector is
 *                   allocated; otherwise its space is re-used.
 * @param  num_elts  Number of elements in the vector.
 *
 * @note Number of elements must be > 0.
 */
void create_zero_vector_i32(Vector_i32** v_out, uint32_t num_elts)
{
    create_init_vector_i32(v_out, num_elts, 0);
}


/**
 * @param  v_out     Result parameter. If @em *v_out is NULL, a vector is
 *                   allocated; otherwise its space is re-used.
 * @param  num_elts  Number of elements in the vector.
 *
 * @note Number of elements must be > 0.
 */
void create_zero_vector_i64(Vector_i64** v_out, uint32_t num_elts)
{
    create_init_vector_i64(v_out, num_elts, 0);
}


/**
 * @param  v_out     Result parameter. If @em *v_out is NULL, a vector is
 *                   allocated; otherwise its space is re-used.
 * @param  num_elts  Number of elements in the vector.
 *
 * @note Number of elements must be > 0.
 */
void create_zero_vector_f(Vector_f** v_out, uint32_t num_elts)
{
    create_init_vector_f(v_out, num_elts, 0);
}


/**
 * @param  v_out     Result parameter. If @em *v_out is NULL, a vector is
 *                   allocated; otherwise its space is re-used.
 * @param  num_elts  Number of elements in the vector.
 *
 * @note Number of elements must be > 0.
 */
void create_zero_vector_d(Vector_d** v_out, uint32_t num_elts)
{
    create_init_vector_d(v_out, num_elts, 0);
}


/**
 * @param  v_out     Result parameter. If @em *v_out is NULL, a vector is
 *                   allocated; otherwise its space is re-used.
 * @param  num_elts  Number of elements in the vector.
 */
void create_zero_vector_cf(Vector_cf** v_out, uint32_t num_elts)
{
    Complex_f z = {0};
    create_init_vector_cf(v_out, num_elts, z);
}


/**
 * @param  v_out     Result parameter. If @em *v_out is NULL, a vector is
 *                   allocated; otherwise its space is re-used.
 * @param  num_elts  Number of elements in the vector.
 *
 * @note Number of elements must be > 0.
 */
void create_zero_vector_cd(Vector_cd** v_out, uint32_t num_elts)
{
    Complex_d z = {0};
    create_init_vector_cd(v_out, num_elts, z);
}

/*@}*/




/**
 * @name create_random_vector
 *
 * Creates a vector of integer or floating point values and initializes it
 * with random values.
 * @{
 */

/**
 * Calls rand() to generate the random numbers.
 *
 * @param  v_out     Result parameter. If @em *v_out is NULL, a vector is
 *                   allocated; otherwise its space is re-used.
 * @param  num_elts  Number of elements in the vector.
 * @param  min       Minimum random value.
 * @param  max       Maximum random value.
 *
 * @note Number of elements must be > 0.
 *
 * @warning Does not check @em min and @em max to see if they would cause
 * overflow.
 */
void create_random_vector_u32
(
    Vector_u32** v_out,
    uint32_t     num_elts,
    uint32_t     min,
    uint32_t     max
)
{
    uint32_t    elt;
    double      r;
    Vector_u32* v;

    create_vector_u32(v_out, num_elts);
    v = *v_out;

    for (elt = 0; elt < num_elts; elt++)
    {
        r = (double)rand() / (double)RAND_MAX;
        v->elts[ elt ] = min + (uint32_t)floor((max - min)*r + 0.5);
    }
}

/**
 * Calls rand() to generate the random numbers.
 *
 * @param  v_out     Result parameter. If @em *v_out is NULL, a vector is
 *                   allocated; otherwise its space is re-used.
 * @param  num_elts  Number of elements in the vector.
 * @param  min       Minimum random value.
 * @param  max       Maximum random value.
 *
 * @note Number of elements must be > 0.
 *
 * @warning Does not check @em min and @em max to see if they would cause
 * overflow.
 */
void create_random_vector_i32
(
    Vector_i32** v_out,
    uint32_t     num_elts,
    int32_t      min,
    int32_t      max
)
{
    uint32_t    elt;
    double      r;
    Vector_i32* v;

    create_vector_i32(v_out, num_elts);
    v = *v_out;

    for (elt = 0; elt < num_elts; elt++)
    {
        r = (double)rand() / (double)RAND_MAX;
        v->elts[ elt ] = min + (int32_t)floor((max - min)*r + 0.5);
    }
}

/**
 * Calls rand() to generate the random numbers.
 *
 * @param  v_out     Result parameter. If @em *v_out is NULL, a vector is
 *                   allocated; otherwise its space is re-used.
 * @param  num_elts  Number of elements in the vector.
 * @param  min       Minimum random value.
 * @param  max       Maximum random value.
 *
 * @note Number of elements must be > 0.
 *
 * @warning Does not check @em min and @em max to see if they would cause
 * overflow.
 */
void create_random_vector_i64
(
    Vector_i64** v_out,
    uint32_t     num_elts,
    int64_t      min,
    int64_t      max
)
{
    uint32_t    elt;
    double      r;
    Vector_i64* v;

    create_vector_i64(v_out, num_elts);
    v = *v_out;

    for (elt = 0; elt < num_elts; elt++)
    {
        r = (double)rand() / (double)RAND_MAX;
        v->elts[ elt ] = min + (int64_t)floor((max - min)*r + 0.5);
    }
}

/**
 * Calls rand() to generate the random numbers.
 *
 * @param  v_out     Result parameter. If @em *v_out is NULL, a vector is
 *                   allocated; otherwise its space is re-used.
 * @param  num_elts  Number of elements in the vector.
 * @param  min       Minimum random value.
 * @param  max       Maximum random value.
 *
 * @note Number of elements must be > 0.
 *
 * @warning Does not check @em min and @em max to see if they would cause
 * overflow.
 */
void create_random_vector_f
(
    Vector_f** v_out,
    uint32_t   num_elts,
    float      min,
    float      max
)
{
    uint32_t  elt;
    float     r;
    Vector_f* v;

    create_vector_f(v_out, num_elts);
    v = *v_out;

    for (elt = 0; elt < num_elts; elt++)
    {
        r = (float)rand() / (float)RAND_MAX;
        v->elts[ elt ] = min + (max - min)*r;
    }
}

/**
 * Calls rand() to generate the random numbers.
 *
 * @param  v_out     Result parameter. If @em *v_out is NULL, a vector is
 *                   allocated; otherwise its space is re-used.
 * @param  num_elts  Number of elements in the vector.
 * @param  min       Minimum random value.
 * @param  max       Maximum random value.
 *
 * @note Number of elements must be > 0.
 *
 * @warning Does not check @em min and @em max to see if they would cause
 * overflow.
 */
void create_random_vector_d
(
    Vector_d** v_out,
    uint32_t   num_elts,
    double     min,
    double     max
)
{
    uint32_t  elt;
    double    r;
    Vector_d* v;

    create_vector_d(v_out, num_elts);
    v = *v_out;

    for (elt = 0; elt < num_elts; elt++)
    {
        r = (double)rand() / (double)RAND_MAX;
        v->elts[ elt ] = min + (max - min)*r;
    }
}

/**
 * Calls rand() to generate the random numbers.
 *
 * @param  v_out     Result parameter. If @em *v_out is NULL, a vector is
 *                   allocated; otherwise its space is re-used.
 * @param  num_elts  Number of elements in the vector.
 * @param  min       Minimum random value.
 * @param  max       Maximum random value.
 *
 * @note Number of elements must be > 0.
 *
 * @warning Does not check @em min and @em max to see if they would cause
 * overflow.
 */
void create_random_vector_cf
(
    Vector_cf** v_out,
    uint32_t    num_elts,
    Complex_f   min,
    Complex_f   max
)
{
    uint32_t   elt;
    float      r;
    Vector_cf* v;

    create_vector_cf(v_out, num_elts);
    v = *v_out;

    for (elt = 0; elt < num_elts; elt++)
    {
        r = (float)rand() / (float)RAND_MAX;
        v->elts[ elt ].r = min.r + (max.r - min.r)*r;
        r = (float)rand() / (float)RAND_MAX;
        v->elts[ elt ].i = min.i + (max.i - min.i)*r;
    }
}

/**
 * Calls rand() to generate the random numbers.
 *
 * @param  v_out     Result parameter. If @em *v_out is NULL, a vector is
 *                   allocated; otherwise its space is re-used.
 * @param  num_elts  Number of elements in the vector.
 * @param  min       Minimum random value.
 * @param  max       Maximum random value.
 *
 * @note Number of elements must be > 0.
 *
 * @warning Does not check @em min and @em max to see if they would cause
 * overflow.
 */
void create_random_vector_cd
(
    Vector_cd** v_out,
    uint32_t    num_elts,
    Complex_d   min,
    Complex_d   max
)
{
    uint32_t   elt;
    double     r;
    Vector_cd* v;

    create_vector_cd(v_out, num_elts);
    v = *v_out;

    for (elt = 0; elt < num_elts; elt++)
    {
        r = (double)rand() / (double)RAND_MAX;
        v->elts[ elt ].r = min.r + (max.r - min.r)*r;
        r = (double)rand() / (double)RAND_MAX;
        v->elts[ elt ].i = min.i + (max.i - min.i)*r;
    }
}

/*@}*/




/**
 * @name copy_vector
 *
 * Copies a vector into another.
 * @{
 */

/**
 * @param  v_out  Result parameter. If @em *v_out is NULL, a vector is 
 *                allocated; otherwise its space is re-used.
 * @param  v_in   Vector to copy.
 *
 * @return On success, NULL is returned. On error, an Error is returned.
 * - @link Error_types::ERROR_INV_ARG ERROR_INV_ARG @endlink 
 *   Attempting to copy vector into itself.
 *
 * @note Does not support copying a vector into itself, i.e., 
 * @em *v_out == @em v_in.
 */
Error* copy_vector_u32(Vector_u32** v_out, const Vector_u32* v_in)
{
    uint32_t    num_elts, elt;
    Vector_u32* v;

    if (*v_out == v_in)
    {
        return JWSC_EARG("Copying a vector into itself is not supported");
    }

    num_elts = v_in->num_elts;
    create_vector_u32(v_out, num_elts);
    v = *v_out;

    for (elt = 0; elt < num_elts; elt++)
    {
        v->elts[ elt ] = v_in->elts[ elt ];
    }

    return NULL;
}

/**
 * @param  v_out  Result parameter. If @em *v_out is NULL, a vector is 
 *                allocated; otherwise its space is re-used.
 * @param  v_in   Vector to copy.
 *
 * @return On success, NULL is returned. On error, an Error is returned.
 * - @link Error_types::ERROR_INV_ARG ERROR_INV_ARG @endlink 
 *   Attempting to copy vector into itself.
 *
 * @note Does not support copying a vector into itself, i.e., 
 * @em *v_out == @em v_in.
 */
Error* copy_vector_i32(Vector_i32** v_out, const Vector_i32* v_in)
{
    uint32_t    num_elts, elt;
    Vector_i32* v;

    if (*v_out == v_in)
    {
        return JWSC_EARG("Copying a vector into itself is not supported");
    }

    num_elts = v_in->num_elts;
    create_vector_i32(v_out, num_elts);
    v = *v_out;

    for (elt = 0; elt < num_elts; elt++)
    {
        v->elts[ elt ] = v_in->elts[ elt ];
    }

    return NULL;
}

/**
 * @param  v_out  Result parameter. If @em *v_out is NULL, a vector is 
 *                allocated; otherwise its space is re-used.
 * @param  v_in   Vector to copy.
 *
 * @return On success, NULL is returned. On error, an Error is returned.
 * - @link Error_types::ERROR_INV_ARG ERROR_INV_ARG @endlink 
 *   Attempting to copy vector into itself.
 *
 * @note Does not support copying a vector into itself, i.e., 
 * @em *v_out == @em v_in.
 */
Error* copy_vector_i64(Vector_i64** v_out, const Vector_i64* v_in)
{
    uint32_t    num_elts, elt;
    Vector_i64* v;

    if (*v_out == v_in)
    {
        return JWSC_EARG("Copying a vector into itself is not supported");
    }

    num_elts = v_in->num_elts;
    create_vector_i64(v_out, num_elts);
    v = *v_out;

    for (elt = 0; elt < num_elts; elt++)
    {
        v->elts[ elt ] = v_in->elts[ elt ];
    }

    return NULL;
}

/**
 * @param  v_out  Result parameter. If @em *v_out is NULL, a vector is 
 *                allocated; otherwise its space is re-used.
 * @param  v_in   Vector to copy.
 *
 * @return On success, NULL is returned. On error, an Error is returned.
 * - @link Error_types::ERROR_INV_ARG ERROR_INV_ARG @endlink 
 *   Attempting to copy vector into itself.
 *
 * @note Does not support copying a vector into itself, i.e., 
 * @em *v_out == @em v_in.
 */
Error* copy_vector_f(Vector_f** v_out, const Vector_f* v_in)
{
    uint32_t  num_elts;
    Vector_f* v;
#if defined JWSC_HAVE_BLAS || defined JWSC_HAVE_ACCELERATE
#else
    uint32_t  elt;
#endif

    if (*v_out == v_in)
    {
        return JWSC_EARG("Copying a vector into itself is not supported");
    }

    num_elts = v_in->num_elts;
    create_vector_f(v_out, num_elts);
    v = *v_out;

#if defined JWSC_HAVE_BLAS || defined JWSC_HAVE_ACCELERATE
    blas_scopy(num_elts, v_in->elts, 1, v->elts, 1);
#else
    for (elt = 0; elt < num_elts; elt++)
    {
        v->elts[ elt ] = v_in->elts[ elt ];
    }
#endif

    return NULL;
}

/**
 * @param  v_out  Result parameter. If @em *v_out is NULL, a vector is 
 *                allocated; otherwise its space is re-used.
 * @param  v_in   Vector to copy.
 *
 * @return On success, NULL is returned. On error, an Error is returned.
 * - @link Error_types::ERROR_INV_ARG ERROR_INV_ARG @endlink 
 *   Attempting to copy vector into itself.
 *
 * @note Does not support copying a vector into itself, i.e., 
 * @em *v_out == @em v_in.
 */
Error* copy_vector_d(Vector_d** v_out, const Vector_d* v_in)
{
    uint32_t  num_elts;
    Vector_d* v;
#if defined JWSC_HAVE_BLAS || defined JWSC_HAVE_ACCELERATE
#else
    uint32_t  elt;
#endif

    if (*v_out == v_in)
    {
        return JWSC_EARG("Copying a vector into itself is not supported");
    }

    num_elts = v_in->num_elts;
    create_vector_d(v_out, num_elts);
    v = *v_out;

#if defined JWSC_HAVE_BLAS || defined JWSC_HAVE_ACCELERATE
    blas_dcopy(num_elts, v_in->elts, 1, v->elts, 1);
#else
    for (elt = 0; elt < num_elts; elt++)
    {
        v->elts[ elt ] = v_in->elts[ elt ];
    }
#endif

    return NULL;
}

/**
 * @param  v_out  Result parameter. If @em *v_out is NULL, a vector is 
 *                allocated; otherwise its space is re-used.
 * @param  v_in   Vector to copy.
 *
 * @return On success, NULL is returned. On error, an Error is returned.
 * - @link Error_types::ERROR_INV_ARG ERROR_INV_ARG @endlink 
 *   Attempting to copy vector into itself.
 *
 * @note Does not support copying a vector into itself, i.e., 
 * @em *v_out == @em v_in.
 */
Error* copy_vector_cf(Vector_cf** v_out, const Vector_cf* v_in)
{
    uint32_t  num_elts;
    Vector_cf* v;
#if defined JWSC_HAVE_BLAS || defined JWSC_HAVE_ACCELERATE
#else
    uint32_t  elt;
#endif

    if (*v_out == v_in)
    {
        return JWSC_EARG("Copying a vector into itself is not supported");
    }

    num_elts = v_in->num_elts;
    create_vector_cf(v_out, num_elts);
    v = *v_out;

#if defined JWSC_HAVE_BLAS || defined JWSC_HAVE_ACCELERATE
    blas_ccopy(num_elts, v_in->elts, 1, v->elts, 1);
#else
    for (elt = 0; elt < num_elts; elt++)
    {
        v->elts[ elt ] = v_in->elts[ elt ];
    }
#endif

    return NULL;
}

/**
 * @param  v_out  Result parameter. If @em *v_out is NULL, a vector is 
 *                allocated; otherwise its space is re-used.
 * @param  v_in   Vector to copy.
 *
 * @return On success, NULL is returned. On error, an Error is returned.
 * - @link Error_types::ERROR_INV_ARG ERROR_INV_ARG @endlink 
 *   Attempting to copy vector into itself.
 *
 * @note Does not support copying a vector into itself, i.e., 
 * @em *v_out == @em v_in.
 */
Error* copy_vector_cd(Vector_cd** v_out, const Vector_cd* v_in)
{
    uint32_t  num_elts;
    Vector_cd* v;
#if defined JWSC_HAVE_BLAS || defined JWSC_HAVE_ACCELERATE
#else
    uint32_t  elt;
#endif

    if (*v_out == v_in)
    {
        return JWSC_EARG("Copying a vector into itself is not supported");
    }

    num_elts = v_in->num_elts;
    create_vector_cd(v_out, num_elts);
    v = *v_out;

#if defined JWSC_HAVE_BLAS || defined JWSC_HAVE_ACCELERATE
    blas_zcopy(num_elts, v_in->elts, 1, v->elts, 1);
#else
    for (elt = 0; elt < num_elts; elt++)
    {
        v->elts[ elt ] = v_in->elts[ elt ];
    }
#endif

    return NULL;
}

/*@}*/




/**
 * @name copy_vector_section
 *
 * Copies a section of a vector into another.
 * @{
 */

/**
 * @param  v_out     Result parameter. If @em *v_out is NULL, a vector
 *                   is allocated; otherwise its space is re-used.
 * @param  v_in      vector to copy a section from.
 * @param  offset    Position of first element to begin copying from.
 * @param  num_elts  Number of elements to copy.
 *
 * @return On success, NULL is returned. On error, an Error is returned and 
 * @em *v_out is freed and set to NULL (if it is not @em v_in).
 * - @link Error_types::ERROR_INV_ARG ERROR_INV_ARG @endlink 
 *   Section to copy was outside the bounds of @em v_in.
 * - @link Error_types::ERROR_INV_ARG ERROR_INV_ARG @endlink 
 *   Attempting to copy vector into itself.
 *
 * @note Does not support copying a vector into itself, i.e., 
 * @em *m_out == @em m_in.
 */
Error* copy_vector_section_u32
(
    Vector_u32**      v_out, 
    const Vector_u32* v_in,
    uint32_t          offset,
    uint32_t          num_elts
)
{
    uint32_t    elt;
    Vector_u32* v;

    if (*v_out == v_in)
    {
        return JWSC_EARG("Copying a vector into itself is not supported");
    }

    if (v_in->num_elts < (offset+num_elts))
    {
        free_vector_u32(*v_out); v_out = NULL;
        return JWSC_EARG("Section to copy outside bounds of input vector");
    }

    create_vector_u32(v_out, num_elts);
    v = *v_out;

    for (elt = 0; elt < num_elts; elt++)
    {
        v->elts[ elt ] = v_in->elts[ elt+offset ];
    }

    return NULL;
}

/**
 * @param  v_out     Result parameter. If @em *v_out is NULL, a vector
 *                   is allocated; otherwise its space is re-used.
 * @param  v_in      vector to copy a section from.
 * @param  offset    Position of first element to begin copying from.
 * @param  num_elts  Number of elements to copy.
 *
 * @return On success, NULL is returned. On error, an Error is returned and 
 * @em *v_out is freed and set to NULL (if it is not @em v_in).
 * - @link Error_types::ERROR_INV_ARG ERROR_INV_ARG @endlink 
 *   Section to copy was outside the bounds of @em v_in.
 * - @link Error_types::ERROR_INV_ARG ERROR_INV_ARG @endlink 
 *   Attempting to copy vector into itself.
 *
 * @note Does not support copying a vector into itself, i.e., 
 * @em *m_out == @em m_in.
 */
Error* copy_vector_section_i32
(
    Vector_i32**      v_out, 
    const Vector_i32* v_in,
    uint32_t          offset,
    uint32_t          num_elts
)
{
    uint32_t    elt;
    Vector_i32* v;

    if (*v_out == v_in)
    {
        return JWSC_EARG("Copying a vector into itself is not supported");
    }

    if (v_in->num_elts < (offset+num_elts))
    {
        free_vector_i32(*v_out); v_out = NULL;
        return JWSC_EARG("Section to copy outside bounds of input vector");
    }

    create_vector_i32(v_out, num_elts);
    v = *v_out;

    for (elt = 0; elt < num_elts; elt++)
    {
        v->elts[ elt ] = v_in->elts[ elt+offset ];
    }

    return NULL;
}

/**
 * @param  v_out     Result parameter. If @em *v_out is NULL, a vector
 *                   is allocated; otherwise its space is re-used.
 * @param  v_in      vector to copy a section from.
 * @param  offset    Position of first element to begin copying from.
 * @param  num_elts  Number of elements to copy.
 *
 * @return On success, NULL is returned. On error, an Error is returned and 
 * @em *v_out is freed and set to NULL (if it is not @em v_in).
 * - @link Error_types::ERROR_INV_ARG ERROR_INV_ARG @endlink 
 *   Section to copy was outside the bounds of @em v_in.
 * - @link Error_types::ERROR_INV_ARG ERROR_INV_ARG @endlink 
 *   Attempting to copy vector into itself.
 *
 * @note Does not support copying a vector into itself, i.e., 
 * @em *m_out == @em m_in.
 */
Error* copy_vector_section_i64
(
    Vector_i64**      v_out, 
    const Vector_i64* v_in,
    uint32_t          offset,
    uint32_t          num_elts
)
{
    uint32_t    elt;
    Vector_i64* v;

    if (*v_out == v_in)
    {
        return JWSC_EARG("Copying a vector into itself is not supported");
    }

    if (v_in->num_elts < (offset+num_elts))
    {
        free_vector_i64(*v_out); v_out = NULL;
        return JWSC_EARG("Section to copy outside bounds of input vector");
    }

    create_vector_i64(v_out, num_elts);
    v = *v_out;

    for (elt = 0; elt < num_elts; elt++)
    {
        v->elts[ elt ] = v_in->elts[ elt+offset ];
    }

    return NULL;
}

/**
 * @param  v_out     Result parameter. If @em *v_out is NULL, a vector
 *                   is allocated; otherwise its space is re-used.
 * @param  v_in      vector to copy a section from.
 * @param  offset    Position of first element to begin copying from.
 * @param  num_elts  Number of elements to copy.
 *
 * @return On success, NULL is returned. On error, an Error is returned and 
 * @em *v_out is freed and set to NULL (if it is not @em v_in).
 * - @link Error_types::ERROR_INV_ARG ERROR_INV_ARG @endlink 
 *   Section to copy was outside the bounds of @em v_in.
 * - @link Error_types::ERROR_INV_ARG ERROR_INV_ARG @endlink 
 *   Attempting to copy vector into itself.
 *
 * @note Does not support copying a vector into itself, i.e., 
 * @em *m_out == @em m_in.
 */
Error* copy_vector_section_f
(
    Vector_f**      v_out, 
    const Vector_f* v_in,
    uint32_t        offset,
    uint32_t        num_elts
)
{
    Vector_f* v;
#if defined JWSC_HAVE_BLAS || defined JWSC_HAVE_ACCELERATE
#else
    uint32_t  elt;
#endif

    if (*v_out == v_in)
    {
        return JWSC_EARG("Copying a vector into itself is not supported");
    }

    if (v_in->num_elts < (offset+num_elts))
    {
        free_vector_f(*v_out); v_out = NULL;
        return JWSC_EARG("Section to copy outside bounds of input vector");
    }

    create_vector_f(v_out, num_elts);
    v = *v_out;

#if defined JWSC_HAVE_BLAS || defined JWSC_HAVE_ACCELERATE
    blas_scopy(num_elts, &(v_in->elts[ offset ]), 1, v->elts, 1);
#else
    for (elt = 0; elt < num_elts; elt++)
    {
        v->elts[ elt ] = v_in->elts[ elt+offset ];
    }
#endif

    return NULL;
}

/**
 * @param  v_out     Result parameter. If @em *v_out is NULL, a vector
 *                   is allocated; otherwise its space is re-used.
 * @param  v_in      vector to copy a section from.
 * @param  offset    Position of first element to begin copying from.
 * @param  num_elts  Number of elements to copy.
 *
 * @return On success, NULL is returned. On error, an Error is returned and 
 * @em *v_out is freed and set to NULL (if it is not @em v_in).
 * - @link Error_types::ERROR_INV_ARG ERROR_INV_ARG @endlink 
 *   Section to copy was outside the bounds of @em v_in.
 * - @link Error_types::ERROR_INV_ARG ERROR_INV_ARG @endlink 
 *   Attempting to copy vector into itself.
 *
 * @note Does not support copying a vector into itself, i.e., 
 * @em *m_out == @em m_in.
 */
Error* copy_vector_section_d
(
    Vector_d**      v_out, 
    const Vector_d* v_in,
    uint32_t        offset,
    uint32_t        num_elts
)
{
    Vector_d* v;
#if defined JWSC_HAVE_BLAS || defined JWSC_HAVE_ACCELERATE
#else
    uint32_t  elt;
#endif

    if (*v_out == v_in)
    {
        return JWSC_EARG("Copying a vector into itself is not supported");
    }

    if (v_in->num_elts < (offset+num_elts))
    {
        free_vector_d(*v_out); v_out = NULL;
        return JWSC_EARG("Section to copy outside bounds of input vector");
    }

    create_vector_d(v_out, num_elts);
    v = *v_out;

#if defined JWSC_HAVE_BLAS || defined JWSC_HAVE_ACCELERATE
    blas_dcopy(num_elts, &(v_in->elts[ offset ]), 1, v->elts, 1);
#else
    for (elt = 0; elt < num_elts; elt++)
    {
        v->elts[ elt ] = v_in->elts[ elt+offset ];
    }
#endif

    return NULL;
}

/**
 * @param  v_out     Result parameter. If @em *v_out is NULL, a vector
 *                   is allocated; otherwise its space is re-used.
 * @param  v_in      vector to copy a section from.
 * @param  offset    Position of first element to begin copying from.
 * @param  num_elts  Number of elements to copy.
 *
 * @return On success, NULL is returned. On error, an Error is returned and 
 * @em *v_out is freed and set to NULL (if it is not @em v_in).
 * - @link Error_types::ERROR_INV_ARG ERROR_INV_ARG @endlink 
 *   Section to copy was outside the bounds of @em v_in.
 * - @link Error_types::ERROR_INV_ARG ERROR_INV_ARG @endlink 
 *   Attempting to copy vector into itself.
 *
 * @note Does not support copying a vector into itself, i.e., 
 * @em *m_out == @em m_in.
 */
Error* copy_vector_section_cf
(
    Vector_cf**      v_out, 
    const Vector_cf* v_in,
    uint32_t         offset,
    uint32_t         num_elts
)
{
    Vector_cf* v;
#if defined JWSC_HAVE_BLAS || defined JWSC_HAVE_ACCELERATE
#else
    uint32_t  elt;
#endif

    if (*v_out == v_in)
    {
        return JWSC_EARG("Copying a vector into itself is not supported");
    }

    if (v_in->num_elts < (offset+num_elts))
    {
        free_vector_cf(*v_out); v_out = NULL;
        return JWSC_EARG("Section to copy outside bounds of input vector");
    }

    create_vector_cf(v_out, num_elts);
    v = *v_out;

#if defined JWSC_HAVE_BLAS || defined JWSC_HAVE_ACCELERATE
    blas_ccopy(num_elts, &(v_in->elts[ offset ]), 1, v->elts, 1);
#else
    for (elt = 0; elt < num_elts; elt++)
    {
        v->elts[ elt ] = v_in->elts[ elt+offset ];
    }
#endif

    return NULL;
}

/**
 * @param  v_out     Result parameter. If @em *v_out is NULL, a vector
 *                   is allocated; otherwise its space is re-used.
 * @param  v_in      vector to copy a section from.
 * @param  offset    Position of first element to begin copying from.
 * @param  num_elts  Number of elements to copy.
 *
 * @return On success, NULL is returned. On error, an Error is returned and 
 * @em *v_out is freed and set to NULL (if it is not @em v_in).
 * - @link Error_types::ERROR_INV_ARG ERROR_INV_ARG @endlink 
 *   Section to copy was outside the bounds of @em v_in.
 * - @link Error_types::ERROR_INV_ARG ERROR_INV_ARG @endlink 
 *   Attempting to copy vector into itself.
 *
 * @note Does not support copying a vector into itself, i.e., 
 * @em *v_out == @em v_in.
 */
Error* copy_vector_section_cd
(
    Vector_cd**      v_out, 
    const Vector_cd* v_in,
    uint32_t         offset,
    uint32_t         num_elts
)
{
    Vector_cd* v;
#if defined JWSC_HAVE_BLAS || defined JWSC_HAVE_ACCELERATE
#else
    uint32_t  elt;
#endif

    if (*v_out == v_in)
    {
        return JWSC_EARG("Copying a vector into itself is not supported");
    }

    if (v_in->num_elts < (offset+num_elts))
    {
        free_vector_cd(*v_out); v_out = NULL;
        return JWSC_EARG("Section to copy outside bounds of input vector");
    }

    create_vector_cd(v_out, num_elts);
    v = *v_out;

#if defined JWSC_HAVE_BLAS || defined JWSC_HAVE_ACCELERATE
    blas_zcopy(num_elts, &(v_in->elts[ offset ]), 1, v->elts, 1);
#else
    for (elt = 0; elt < num_elts; elt++)
    {
        v->elts[ elt ] = v_in->elts[ elt+offset ];
    }
#endif

    return NULL;
}

/*@}*/




/**
 * @name copy_vector_section_into_vector
 *
 * Copies a section of a vector.
 * @{
 */

/**
 * @param  v_1       Vector to copy the section into.
 * @param  offset_1  Position of first elt to copy into @em v_1.
 * @param  v_2       Vector to copy the section from.
 * @param  offset_2  Position of first elt to copy from @em v_2.
 * @param  num_elts  Number of elements in the section to copy.
 *
 * @return On success, NULL is returned. On error, an Error is returned.
 * - @link Error_types::ERROR_INV_ARG ERROR_INV_ARG @endlink 
 *   Section to copy was outside the bounds of @em v_1.
 * - @link Error_types::ERROR_INV_ARG ERROR_INV_ARG @endlink 
 *   Attempting to copy vector into itself.
 *
 * @note Does not support copying a vector into itself, i.e., 
 * @em v_1 == @em v_2.
 */
Error* copy_vector_section_into_vector_u32
(
    Vector_u32*       v_1, 
    uint32_t          offset_1,
    const Vector_u32* v_2,
    uint32_t          offset_2,
    uint32_t          num_elts
)
{
    uint32_t elt;

    if (v_1 == v_2)
    {
        return JWSC_EARG("Copying a vector into itself is not supported");
    }

    /* Test if the size of the section to copy is in the bounds of the input
     * vector. */
    if (v_2->num_elts < (offset_2+num_elts) ||
        v_1->num_elts < (offset_1+num_elts))
    {
        return JWSC_EARG("Section to copy outside bounds of input vector");
    }

    for (elt = 0; elt < num_elts; elt++)
    {
        v_1->elts[ elt+offset_1 ] = v_2->elts[ elt+offset_2 ];
    }

    return NULL;
}

/**
 * @param  v_1       Vector to copy the section into.
 * @param  offset_1  Position of first elt to copy into @em v_1.
 * @param  v_2       Vector to copy the section from.
 * @param  offset_2  Position of first elt to copy from @em v_2.
 * @param  num_elts  Number of elts to copy.
 *
 * @return On success, NULL is returned. On error, an Error is returned.
 * - @link Error_types::ERROR_INV_ARG ERROR_INV_ARG @endlink 
 *   Section to copy was outside the bounds of @em v_1.
 * - @link Error_types::ERROR_INV_ARG ERROR_INV_ARG @endlink 
 *   Attempting to copy vector into itself.
 *
 * @note Does not support copying a vector into itself, i.e., 
 * @em v_1 == @em v_2.
 */
Error* copy_vector_section_into_vector_i32
(
    Vector_i32*       v_1, 
    uint32_t          offset_1,
    const Vector_i32* v_2,
    uint32_t          offset_2,
    uint32_t          num_elts
)
{
    uint32_t elt;

    if (v_1 == v_2)
    {
        return JWSC_EARG("Copying a vector into itself is not supported");
    }

    /* Test if the size of the section to copy is in the bounds of the input
     * vector. */
    if (v_2->num_elts < (offset_2+num_elts) ||
        v_1->num_elts < (offset_1+num_elts))
    {
        return JWSC_EARG("Section to copy outside bounds of input vector");
    }

    for (elt = 0; elt < num_elts; elt++)
    {
        v_1->elts[ elt+offset_1 ] = v_2->elts[ elt+offset_2 ];
    }

    return NULL;
}

/**
 * @param  v_1       Vector to copy the section into.
 * @param  offset_1  Position of first elt to copy into @em v_1.
 * @param  v_2       Vector to copy the section from.
 * @param  offset_2  Position of first elt to copy from @em v_2.
 * @param  num_elts  Number of elts to copy.
 *
 * @return On success, NULL is returned. On error, an Error is returned.
 * - @link Error_types::ERROR_INV_ARG ERROR_INV_ARG @endlink 
 *   Section to copy was outside the bounds of @em v_1.
 * - @link Error_types::ERROR_INV_ARG ERROR_INV_ARG @endlink 
 *   Attempting to copy vector into itself.
 *
 * @note Does not support copying a vector into itself, i.e., 
 * @em v_1 == @em v_2.
 */
Error* copy_vector_section_into_vector_i64
(
    Vector_i64*       v_1, 
    uint32_t          offset_1,
    const Vector_i64* v_2,
    uint32_t          offset_2,
    uint32_t          num_elts
)
{
    uint32_t elt;

    if (v_1 == v_2)
    {
        return JWSC_EARG("Copying a vector into itself is not supported");
    }

    /* Test if the size of the section to copy is in the bounds of the input
     * vector. */
    if (v_2->num_elts < (offset_2+num_elts) ||
        v_1->num_elts < (offset_1+num_elts))
    {
        return JWSC_EARG("Section to copy outside bounds of input vector");
    }

    for (elt = 0; elt < num_elts; elt++)
    {
        v_1->elts[ elt+offset_1 ] = v_2->elts[ elt+offset_2 ];
    }

    return NULL;
}

/**
 * @param  v_1       Vector to copy the section into.
 * @param  offset_1  Position of first elt to copy into @em v_1.
 * @param  v_2       Vector to copy the section from.
 * @param  offset_2  Position of first elt to copy from @em v_2.
 * @param  num_elts  Number of elts to copy.
 *
 * @return On success, NULL is returned. On error, an Error is returned.
 * - @link Error_types::ERROR_INV_ARG ERROR_INV_ARG @endlink 
 *   Section to copy was outside the bounds of @em v_1.
 * - @link Error_types::ERROR_INV_ARG ERROR_INV_ARG @endlink 
 *   Attempting to copy vector into itself.
 *
 * @note Does not support copying a vector into itself, i.e., 
 * @em v_1 == @em v_2.
 */
Error* copy_vector_section_into_vector_f
(
    Vector_f*       v_1, 
    uint32_t        offset_1,
    const Vector_f* v_2,
    uint32_t        offset_2,
    uint32_t        num_elts
)
{
    uint32_t elt;

    if (v_1 == v_2)
    {
        return JWSC_EARG("Copying a vector into itself is not supported");
    }

    /* Test if the size of the section to copy is in the bounds of the input
     * vector. */
    if (v_2->num_elts < (offset_2+num_elts) ||
        v_1->num_elts < (offset_1+num_elts))
    {
        return JWSC_EARG("Section to copy outside bounds of input vector");
    }

    for (elt = 0; elt < num_elts; elt++)
    {
        v_1->elts[ elt+offset_1 ] = v_2->elts[ elt+offset_2 ];
    }

    return NULL;
}

/**
 * @param  v_1       Vector to copy the section into.
 * @param  offset_1  Position of first elt to copy into @em v_1.
 * @param  v_2       Vector to copy the section from.
 * @param  offset_2  Position of first elt to copy from @em v_2.
 * @param  num_elts  Number of elts to copy.
 *
 * @return On success, NULL is returned. On error, an Error is returned.
 * - @link Error_types::ERROR_INV_ARG ERROR_INV_ARG @endlink 
 *   Section to copy was outside the bounds of @em v_1.
 * - @link Error_types::ERROR_INV_ARG ERROR_INV_ARG @endlink 
 *   Attempting to copy vector into itself.
 *
 * @note Does not support copying a vector into itself, i.e., 
 * @em v_1 == @em v_2.
 */
Error* copy_vector_section_into_vector_d
(
    Vector_d*       v_1, 
    uint32_t        offset_1,
    const Vector_d* v_2,
    uint32_t        offset_2,
    uint32_t        num_elts
)
{
    uint32_t elt;

    if (v_1 == v_2)
    {
        return JWSC_EARG("Copying a vector into itself is not supported");
    }

    /* Test if the size of the section to copy is in the bounds of the input
     * vector. */
    if (v_2->num_elts < (offset_2+num_elts) ||
        v_1->num_elts < (offset_1+num_elts))
    {
        return JWSC_EARG("Section to copy outside bounds of input vector");
    }

    for (elt = 0; elt < num_elts; elt++)
    {
        v_1->elts[ elt+offset_1 ] = v_2->elts[ elt+offset_2 ];
    }

    return NULL;
}

/**
 * @param  v_1       Vector to copy the section into.
 * @param  offset_1  Position of first elt to copy into @em v_1.
 * @param  v_2       Vector to copy the section from.
 * @param  offset_2  Position of first elt to copy from @em v_2.
 * @param  num_elts  Number of elts to copy.
 *
 * @return On success, NULL is returned. On error, an Error is returned.
 * - @link Error_types::ERROR_INV_ARG ERROR_INV_ARG @endlink 
 *   Section to copy was outside the bounds of @em v_1.
 * - @link Error_types::ERROR_INV_ARG ERROR_INV_ARG @endlink 
 *   Attempting to copy vector into itself.
 *
 * @note Does not support copying a vector into itself, i.e., 
 * @em v_1 == @em v_2.
 */
Error* copy_vector_section_into_vector_cf
(
    Vector_cf*       v_1, 
    uint32_t         offset_1,
    const Vector_cf* v_2,
    uint32_t         offset_2,
    uint32_t         num_elts
)
{
    uint32_t elt;

    if (v_1 == v_2)
    {
        return JWSC_EARG("Copying a vector into itself is not supported");
    }

    /* Test if the size of the section to copy is in the bounds of the input
     * vector. */
    if (v_2->num_elts < (offset_2+num_elts) ||
        v_1->num_elts < (offset_1+num_elts))
    {
        return JWSC_EARG("Section to copy outside bounds of input vector");
    }

    for (elt = 0; elt < num_elts; elt++)
    {
        v_1->elts[ elt+offset_1 ] = v_2->elts[ elt+offset_2 ];
    }

    return NULL;
}

/**
 * @param  v_1       Vector to copy the section into.
 * @param  offset_1  Position of first elt to copy into @em v_1.
 * @param  v_2       Vector to copy the section from.
 * @param  offset_2  Position of first elt to copy from @em v_2.
 * @param  num_elts  Number of elts to copy.
 *
 * @return On success, NULL is returned. On error, an Error is returned.
 * - @link Error_types::ERROR_INV_ARG ERROR_INV_ARG @endlink 
 *   Section to copy was outside the bounds of @em v_1.
 * - @link Error_types::ERROR_INV_ARG ERROR_INV_ARG @endlink 
 *   Attempting to copy vector into itself.
 *
 * @note Dimensions must be > 0.
 *
 * @note Does not support copying a vector into itself, i.e., 
 * @em v_1 == @em v_2.
 */
Error* copy_vector_section_into_vector_cd
(
    Vector_cd*       v_1, 
    uint32_t         offset_1,
    const Vector_cd* v_2,
    uint32_t         offset_2,
    uint32_t         num_elts
)
{
    uint32_t elt;

    if (v_1 == v_2)
    {
        return JWSC_EARG("Copying a vector into itself is not supported");
    }

    /* Test if the size of the section to copy is in the bounds of the input
     * vector. */
    if (v_2->num_elts < (offset_2+num_elts) ||
        v_1->num_elts < (offset_1+num_elts))
    {
        return JWSC_EARG("Section to copy outside bounds of input vector");
    }

    for (elt = 0; elt < num_elts; elt++)
    {
        v_1->elts[ elt+offset_1 ] = v_2->elts[ elt+offset_2 ];
    }

    return NULL;
}

/*@}*/




/**
 * @name cat_vectors
 *
 * Concatenates two vectors.
 * @{
 */

/**
 * @param  v_out  Result parameter. If @em *v_out is NULL, a vector
 *                is allocated; otherwise its space is re-used.
 * @param  v_1    First vector in the concatenation.
 * @param  v_2    Second vector in the concatenation.
 *
 * @note If @em *v_out == @em v_in, v_in is overwritten.
 */
void cat_vectors_u32
(
    Vector_u32**      v_out, 
    const Vector_u32* v_1,
    const Vector_u32* v_2
)
{
    uint32_t N_1, N_2;
    Vector_u32* v;

    N_1 = v_1->num_elts;
    N_2 = v_2->num_elts;

    v = (*v_out == v_1 || *v_out == v_2) ? NULL : *v_out;
    create_vector_u32(&v, N_1+N_2);

    copy_vector_section_into_vector_u32(v, 0, v_1, 0, N_1); 
    copy_vector_section_into_vector_u32(v, N_1, v_2, 0, N_2); 

    if (*v_out == v_1 || *v_out == v_2)
    {
        copy_vector_u32(v_out, v);
    }
}

/**
 * @param  v_out  Result parameter. If @em *v_out is NULL, a vector
 *                is allocated; otherwise its space is re-used.
 * @param  v_1    First vector in the concatenation.
 * @param  v_2    Second vector in the concatenation.
 *
 * @note If @em *v_out == @em v_in, v_in is overwritten.
 */
void cat_vectors_i32
(
    Vector_i32**      v_out, 
    const Vector_i32* v_1,
    const Vector_i32* v_2
)
{
    uint32_t N_1, N_2;
    Vector_i32* v;

    N_1 = v_1->num_elts;
    N_2 = v_2->num_elts;

    v = (*v_out == v_1 || *v_out == v_2) ? NULL : *v_out;
    create_vector_i32(&v, N_1+N_2);

    copy_vector_section_into_vector_i32(v, 0, v_1, 0, N_1); 
    copy_vector_section_into_vector_i32(v, N_1, v_2, 0, N_2); 

    if (*v_out == v_1 || *v_out == v_2)
    {
        copy_vector_i32(v_out, v);
    }
}

/**
 * @param  v_out  Result parameter. If @em *v_out is NULL, a vector
 *                is allocated; otherwise its space is re-used.
 * @param  v_1    First vector in the concatenation.
 * @param  v_2    Second vector in the concatenation.
 *
 * @note If @em *v_out == @em v_in, v_in is overwritten.
 */
void cat_vectors_i64
(
    Vector_i64**      v_out, 
    const Vector_i64* v_1,
    const Vector_i64* v_2
)
{
    uint32_t N_1, N_2;
    Vector_i64* v;

    N_1 = v_1->num_elts;
    N_2 = v_2->num_elts;

    v = (*v_out == v_1 || *v_out == v_2) ? NULL : *v_out;
    create_vector_i64(&v, N_1+N_2);

    copy_vector_section_into_vector_i64(v, 0, v_1, 0, N_1); 
    copy_vector_section_into_vector_i64(v, N_1, v_2, 0, N_2); 

    if (*v_out == v_1 || *v_out == v_2)
    {
        copy_vector_i64(v_out, v);
    }
}

/**
 * @param  v_out  Result parameter. If @em *v_out is NULL, a vector
 *                is allocated; otherwise its space is re-used.
 * @param  v_1    First vector in the concatenation.
 * @param  v_2    Second vector in the concatenation.
 *
 * @note If @em *v_out == @em v_in, v_in is overwritten.
 */
void cat_vectors_f
(
    Vector_f**      v_out, 
    const Vector_f* v_1,
    const Vector_f* v_2
)
{
    uint32_t N_1, N_2;
    Vector_f* v;

    N_1 = v_1->num_elts;
    N_2 = v_2->num_elts;

    v = (*v_out == v_1 || *v_out == v_2) ? NULL : *v_out;
    create_vector_f(&v, N_1+N_2);

    copy_vector_section_into_vector_f(v, 0, v_1, 0, N_1); 
    copy_vector_section_into_vector_f(v, N_1, v_2, 0, N_2); 

    if (*v_out == v_1 || *v_out == v_2)
    {
        copy_vector_f(v_out, v);
    }
}

/**
 * @param  v_out  Result parameter. If @em *v_out is NULL, a vector
 *                is allocated; otherwise its space is re-used.
 * @param  v_1    First vector in the concatenation.
 * @param  v_2    Second vector in the concatenation.
 *
 * @note If @em *v_out == @em v_in, v_in is overwritten.
 */
void cat_vectors_d
(
    Vector_d**      v_out, 
    const Vector_d* v_1,
    const Vector_d* v_2
)
{
    uint32_t N_1, N_2;
    Vector_d* v;

    N_1 = v_1->num_elts;
    N_2 = v_2->num_elts;

    v = (*v_out == v_1 || *v_out == v_2) ? NULL : *v_out;
    create_vector_d(&v, N_1+N_2);

    copy_vector_section_into_vector_d(v, 0, v_1, 0, N_1); 
    copy_vector_section_into_vector_d(v, N_1, v_2, 0, N_2); 

    if (*v_out == v_1 || *v_out == v_2)
    {
        copy_vector_d(v_out, v);
    }
}

/**
 * @param  v_out  Result parameter. If @em *v_out is NULL, a vector
 *                is allocated; otherwise its space is re-used.
 * @param  v_1    First vector in the concatenation.
 * @param  v_2    Second vector in the concatenation.
 *
 * @note If @em *v_out == @em v_in, v_in is overwritten.
 */
void cat_vectors_cf
(
    Vector_cf**      v_out, 
    const Vector_cf* v_1,
    const Vector_cf* v_2
)
{
    uint32_t N_1, N_2;
    Vector_cf* v;

    N_1 = v_1->num_elts;
    N_2 = v_2->num_elts;

    v = (*v_out == v_1 || *v_out == v_2) ? NULL : *v_out;
    create_vector_cf(&v, N_1+N_2);

    copy_vector_section_into_vector_cf(v, 0, v_1, 0, N_1); 
    copy_vector_section_into_vector_cf(v, N_1, v_2, 0, N_2); 

    if (*v_out == v_1 || *v_out == v_2)
    {
        copy_vector_cf(v_out, v);
    }
}

/**
 * @param  v_out  Result parameter. If @em *v_out is NULL, a vector
 *                is allocated; otherwise its space is re-used.
 * @param  v_1    First vector in the concatenation.
 * @param  v_2    Second vector in the concatenation.
 *
 * @note If @em *v_out == @em v_in, v_in is overwritten.
 */
void cat_vectors_cd
(
    Vector_cd**      v_out, 
    const Vector_cd* v_1,
    const Vector_cd* v_2
)
{
    uint32_t N_1, N_2;
    Vector_cd* v;

    N_1 = v_1->num_elts;
    N_2 = v_2->num_elts;

    v = (*v_out == v_1 || *v_out == v_2) ? NULL : *v_out;
    create_vector_cd(&v, N_1+N_2);

    copy_vector_section_into_vector_cd(v, 0, v_1, 0, N_1); 
    copy_vector_section_into_vector_cd(v, N_1, v_2, 0, N_2); 

    if (*v_out == v_1 || *v_out == v_2)
    {
        copy_vector_cd(v_out, v);
    }
}

/*@}*/




/**
 * @name free_vector
 *
 * Frees a vector.
 * @{
 */

/** 
 * @param  v  Vector to free.
 */
void free_vector_u32(Vector_u32* v)
{
    if (v == NULL) return;

    free(v->elts);
    free(v);
}

/** 
 * @param  v  Vector to free.
 */
void free_vector_i32(Vector_i32* v)
{
    if (v == NULL) return;

    free(v->elts);
    free(v);
}

/** 
 * @param  v  Vector to free.
 */
void free_vector_i64(Vector_i64* v)
{
    if (v == NULL) return;

    free(v->elts);
    free(v);
}

/** 
 * @param  v  Vector to free.
 */
void free_vector_f(Vector_f* v)
{
    if (v == NULL) return;

    free(v->elts);
    free(v);
}

/** 
 * @param  v  Vector to free.
 */
void free_vector_d(Vector_d* v)
{
    if (v == NULL) return;

    free(v->elts);
    free(v);
}

/** 
 * @param  v  Vector to free.
 */
void free_vector_cf(Vector_cf* v)
{
    if (v == NULL) return;

    free(v->elts);
    free(v);
}

/** 
 * @param  v  Vector to free.
 */
void free_vector_cd(Vector_cd* v)
{
    if (v == NULL) return;

    free(v->elts);
    free(v);
}

/*@})*/
