/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Declarations for a vector and associated functions.
 *
 * The vector holds integers and floating point real and complex numbers.
 */


#ifndef VECTOR_H 
#define VECTOR_H 


#include <jwsc/config.h>

#include <stdlib.h>
#include <inttypes.h>

#include "jwsc/base/error.h"
#include "jwsc/math/complex.h"


#ifdef __cplusplus
namespace jwsc {
extern "C" {
#endif


/** @brief Vector of signed 8-bit values. */
typedef struct
{
    /** @brief Number of elements. */
    uint32_t num_elts;

    /** @brief Array of elements. */
    int8_t* elts;
}
Vector_i8;

/** @brief Vector of signed 16-bit values. */
typedef struct
{
    /** @brief Number of elements. */
    uint32_t num_elts;

    /** @brief Array of elements. */
    int16_t* elts;
}
Vector_i16;

/** @brief Vector of signed 32-bit values. */
typedef struct
{
    /** @brief Number of elements. */
    uint32_t num_elts;

    /** @brief Array of elements. */
    int32_t* elts;
}
Vector_i32;

/** @brief Vector of signed 64-bit values. */
typedef struct
{
    /** @brief Number of elements. */
    uint32_t num_elts;

    /** @brief Array of elements. */
    int64_t* elts;
}
Vector_i64;

/** @brief Vector of unsigned 8-bit values. */
typedef struct
{
    /** @brief Number of elements. */
    uint32_t num_elts;

    /** @brief Array of elements. */
    uint8_t* elts;
}
Vector_u8;

/** @brief Vector of unsigned 16-bit values. */
typedef struct
{
    /** @brief Number of elements. */
    uint32_t num_elts;

    /** @brief Array of elements. */
    uint16_t* elts;
}
Vector_u16;

/** @brief Vector of unsigned 32-bit values. */
typedef struct
{
    /** @brief Number of elements. */
    uint32_t num_elts;

    /** @brief Array of elements. */
    uint32_t* elts;
}
Vector_u32;

/** @brief Vector of unsigned 64-bit values. */
typedef struct
{
    /** @brief Number of elements. */
    uint32_t num_elts;

    /** @brief Array of elements. */
    uint64_t* elts;
}
Vector_u64;

/** @brief Vector of single precision floating point values. */
typedef struct
{
    /** @brief Number of elements. */
    uint32_t num_elts;

    /** @brief Array of elements. */
    float* elts;
}
Vector_f;

/** @brief Vector of double precision floating point values. */
typedef struct
{
    /** @brief Number of elements. */
    uint32_t num_elts;

    /** @brief Array of elements. */
    double* elts;
}
Vector_d;

/** @brief Vector of single precision complex floating point values. */
typedef struct
{
    /** @brief Number of elements. */
    uint32_t num_elts;

    /** @brief Array of elements. */
    Complex_f* elts;
}
Vector_cf;

/** @brief Vector of double precision complex floating point values. */
typedef struct
{
    /** @brief Number of elements. */
    uint32_t num_elts;

    /** @brief Array of elements. */
    Complex_d* elts;
}
Vector_cd;




/**
 * @name create_vector
 *
 * Creates a vector of integer or floating point values.
 * @{
 */

/** @brief 32-bit unsigned integer. */
void create_vector_u32(Vector_u32** v_out, uint32_t num_elts);

/** @brief 32-bit signed integer. */
void create_vector_i32(Vector_i32** v_out, uint32_t num_elts);

/** @brief 64-bit signed integer. */
void create_vector_i64(Vector_i64** v_out, uint32_t num_elts);

/** @brief Single precision floating point. */
void create_vector_f(Vector_f** v_out, uint32_t num_elts);

/** @brief Double precision floating point. */
void create_vector_d(Vector_d** v_out, uint32_t num_elts);

/** @brief Single precision complex floating point. */
void create_vector_cf(Vector_cf** v_out, uint32_t num_elts);

/** @brief Double precision complex floating point. */
void create_vector_cd(Vector_cd** v_out, uint32_t num_elts);

/*@}*/




/**
 * @name create_init_vector
 *
 * Creates and initializes a vector of integer or floating point values.
 * @{
 */

/** @brief 32-bit unsigned integer. */
void create_init_vector_u32(Vector_u32** v_out, uint32_t num_elts,uint32_t val);

/** @brief 32-bit signed integer. */
void create_init_vector_i32(Vector_i32** v_out, uint32_t num_elts, int32_t val);

/** @brief 64-bit signed integer. */
void create_init_vector_i64(Vector_i64** v_out, uint32_t num_elts, int64_t val);

/** @brief Single precision floating point. */
void create_init_vector_f(Vector_f** v_out, uint32_t num_elts, float val);

/** @brief Double precision floating point. */
void create_init_vector_d(Vector_d** v_out, uint32_t num_elts, double val);

/** @brief Single precision complex floating point. */
void create_init_vector_cf(Vector_cf** v_out, uint32_t num_elts, Complex_f val);

/** @brief Double precision complex floating point. */
void create_init_vector_cd(Vector_cd** v_out, uint32_t num_elts, Complex_d val);

/*@}*/




/**
 * @name create_zero_vector
 *
 * Creates a vector of integer or floating point values and initializes it with 
 * zeros.
 * @{
 */

/** @brief 32-bit unsigned integer. */
void create_zero_vector_u32(Vector_u32** v_out, uint32_t num_elts);

/** @brief 32-bit signed integer. */
void create_zero_vector_i32(Vector_i32** v_out, uint32_t num_elts);

/** @brief 64-bit signed integer. */
void create_zero_vector_i64(Vector_i64** v_out, uint32_t num_elts);

/** @brief Single precision floating point. */
void create_zero_vector_f(Vector_f** v_out, uint32_t num_elts);

/** @brief Double precision floating point. */
void create_zero_vector_d(Vector_d** v_out, uint32_t num_elts);

/** @brief Single precision complex floating point. */
void create_zero_vector_cf(Vector_cf** v_out, uint32_t num_elts);

/** @brief Double precision complex floating point. */
void create_zero_vector_cd(Vector_cd** v_out, uint32_t num_elts);

/*@}*/




/**
 * @name create_random_vector
 *
 * Creates a vector of integer or floating point values and initializes it
 * with random values.
 * @{
 */

/** @brief 32-bit unsigned integer. */
void create_random_vector_u32
(
    Vector_u32** v_out, 
    uint32_t     num_elts,
    uint32_t     min,
    uint32_t     max
);

/** @brief 32-bit signed integer. */
void create_random_vector_i32
(
    Vector_i32** v_out, 
    uint32_t     num_elts,
    int32_t      min,
    int32_t      max
);

/** @brief 64-bit signed integer. */
void create_random_vector_i64
(
    Vector_i64** v_out, 
    uint32_t     num_elts,
    int64_t      min,
    int64_t      max
);

/** @brief Single precision floating point. */
void create_random_vector_f
(
    Vector_f** v_out, 
    uint32_t   num_elts,
    float      min,
    float      max
);

/** @brief Double precision floating point. */
void create_random_vector_d
(
    Vector_d** v_out, 
    uint32_t   num_elts,
    double     min,
    double     max
);

/** @brief Single precision complex floating point. */
void create_random_vector_cf
(
    Vector_cf** v_out, 
    uint32_t    num_elts,
    Complex_f   min,
    Complex_f   max
);

/** @brief Double precision complex floating point. */
void create_random_vector_cd
(
    Vector_cd** v_out, 
    uint32_t    num_elts,
    Complex_d   min,
    Complex_d   max
);

/*@}*/




/**
 * @name copy_vector
 *
 * Copies a vector into another.
 * @{
 */

/** @brief 32-bit unsigned integer. */
Error* copy_vector_u32(Vector_u32** v_out, const Vector_u32* v_in);

/** @brief 32-bit signed integer. */
Error* copy_vector_i32(Vector_i32** v_out, const Vector_i32* v_in);

/** @brief 64-bit signed integer. */
Error* copy_vector_i64(Vector_i64** v_out, const Vector_i64* v_in);

/** @brief Single precision floating point. */
Error* copy_vector_f(Vector_f** v_out, const Vector_f* v_in);

/** @brief Double precision floating point. */
Error* copy_vector_d(Vector_d** v_out, const Vector_d* v_in);

/** @brief Single precision complex floating point. */
Error* copy_vector_cf(Vector_cf** v_out, const Vector_cf* v_in);

/** @brief Double precision complex floating point. */
Error* copy_vector_cd(Vector_cd** v_out, const Vector_cd* v_in);

/*@}*/




/**
 * @name copy_vector_section
 *
 * Copies a section of a vector into another.
 * @{
 */

/** @brief 32-bit unsigned integer. */
Error* copy_vector_section_u32
(
    Vector_u32**      v_out, 
    const Vector_u32* v_in,
    uint32_t          offset,
    uint32_t          num_elts
);

/** @brief 32-bit signed integer. */
Error* copy_vector_section_i32
(
    Vector_i32**      v_out, 
    const Vector_i32* v_in,
    uint32_t          offset,
    uint32_t          num_elts
);

/** @brief 64-bit signed integer. */
Error* copy_vector_section_i64
(
    Vector_i64**      v_out, 
    const Vector_i64* v_in,
    uint32_t          offset,
    uint32_t          num_elts
);

/** @brief Single precision floating point. */
Error* copy_vector_section_f
(
    Vector_f**      v_out, 
    const Vector_f* v_in,
    uint32_t        offset,
    uint32_t        num_elts
);

/** @brief Double precision floating point. */
Error* copy_vector_section_d
(
    Vector_d**      v_out, 
    const Vector_d* v_in,
    uint32_t        offset,
    uint32_t        num_elts
);

/** @brief Single precision complex floating point. */
Error* copy_vector_section_cf
(
    Vector_cf**      v_out, 
    const Vector_cf* v_in,
    uint32_t         offset,
    uint32_t         num_elts
);

/** @brief Double precision complex floating point. */
Error* copy_vector_section_cd
(
    Vector_cd**      v_out, 
    const Vector_cd* v_in,
    uint32_t         offset,
    uint32_t         num_elts
);

/*@}*/




/**
 * @name copy_vector_section_into_vector
 *
 * Copies a section of a vector.
 * @{
 */

/** @brief Copies a section of a 32-bit unsigned integer vector into another. */
Error* copy_vector_section_into_vector_u32
(
    Vector_u32*       v_1, 
    uint32_t          offset_1,
    const Vector_u32* v_2,
    uint32_t          offset_2,
    uint32_t          num_elts
);

/** @brief Copies a section of a 32-bit signed integer vector into another. */
Error* copy_vector_section_into_vector_i32
(
    Vector_i32*       v_1, 
    uint32_t          offset_1,
    const Vector_i32* v_2,
    uint32_t          offset_2,
    uint32_t          num_elts
);

/** @brief Copies a section of a 64-bit signed integer vector into another. */
Error* copy_vector_section_into_vector_i64
(
    Vector_i64*       v_1, 
    uint32_t          offset_1,
    const Vector_i64* v_2,
    uint32_t          offset_2,
    uint32_t          num_elts
);

/** 
 * @brief Copies a section of a single precision floating point vector into 
 * another.
 */
Error* copy_vector_section_into_vector_f
(
    Vector_f*       v_1, 
    uint32_t        offset_1,
    const Vector_f* v_2,
    uint32_t        offset_2,
    uint32_t        num_elts
);

/** 
 * @brief Copies a section of a double precision floating point vector into 
 * another.
 */
Error* copy_vector_section_into_vector_d
(
    Vector_d*       v_1, 
    uint32_t        offset_1,
    const Vector_d* v_2,
    uint32_t        offset_2,
    uint32_t        num_elts
);

/** 
 * @brief Copies a section of a single precision complex floating point vector 
 * into another.
 */
Error* copy_vector_section_into_vector_cf
(
    Vector_cf*       v_1, 
    uint32_t         offset_1,
    const Vector_cf* v_2,
    uint32_t         offset_2,
    uint32_t         num_elts
);

/** 
 * @brief Copies a section of a double precision complex floating point vector 
 * into another.
 */
Error* copy_vector_section_into_vector_cd
(
    Vector_cd*       v_1, 
    uint32_t         offset_1,
    const Vector_cd* v_2,
    uint32_t         offset_2,
    uint32_t         num_elts
);

/*@}*/




/**
 * @name cat_vectors
 *
 * Concatenates two vectors.
 * @{
 */

/** @brief Concatenates two 32-bit unsigned integer vectors. */
void cat_vectors_u32
(
    Vector_u32**      v_out, 
    const Vector_u32* v_1, 
    const Vector_u32* v_2
);

/** @brief Concatenates two 32-bit signed integer vectors. */
void cat_vectors_i32
(
    Vector_i32**      v_out, 
    const Vector_i32* v_1, 
    const Vector_i32* v_2
);

/** @brief Concatenates two 64-bit signed integer vectors. */
void cat_vectors_i64
(
    Vector_i64**      v_out, 
    const Vector_i64* v_1, 
    const Vector_i64* v_2
);

/** @brief Concatenates two single precision floating point vectors. */
void cat_vectors_f
(
    Vector_f**      v_out, 
    const Vector_f* v_1, 
    const Vector_f* v_2
);

/** @brief Concatenates two double precision floating point vectors. */
void cat_vectors_d
(
    Vector_d**      v_out, 
    const Vector_d* v_1, 
    const Vector_d* v_2
);

/** @brief Concatenates two single precision complex floating point vectors. */
void cat_vectors_cf
(
    Vector_cf**      v_out, 
    const Vector_cf* v_1, 
    const Vector_cf* v_2
);

/** @brief Concatenates two double precision complex floating point vectors. */
void cat_vectors_cd
(
    Vector_cd**      v_out, 
    const Vector_cd* v_1, 
    const Vector_cd* v_2
);

/*@}*/




/**
 * @name free_vector
 *
 * Frees a vector.
 * @{
 */

/** @brief 32-bit unsigned integer. */
void free_vector_u32(Vector_u32* v);

/** @brief 32-bit signed integer. */
void free_vector_i32(Vector_i32* v);

/** @brief 64-bit signed integer. */
void free_vector_i64(Vector_i64* v);

/** @brief Single precision floating point. */
void free_vector_f(Vector_f* v);

/** @brief Double precision floating point. */
void free_vector_d(Vector_d* v);

/** @brief Single precision complex floating point. */
void free_vector_cf(Vector_cf* v);

/** @brief Double precision complex floating point. */
void free_vector_cd(Vector_cd* v);

/*@}*/


#ifdef __cplusplus
}
}
#endif


#endif
