/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Definitions for I/O operations on a vector.
 */


#include <jwsc/config.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <inttypes.h>
#include <ctype.h>

#include "jwsc/base/error.h"
#include "jwsc/base/file_io.h"
#include "jwsc/math/complex.h"
#include "jwsc/vector/vector.h"
#include "jwsc/vector/vector_io.h"

#ifdef JWSC_HAVE_DMALLOC
#include <dmalloc.h>
#endif


/** @brief Reads a line of a vector from a file pointer. */
static int read_vector_line(char** line, FILE* fp)
{
    static const uint32_t block_size = 1024;

    uint32_t num_elts;
    uint32_t blocks;

    char c;

    num_elts = 0;
    blocks   = 0;

    while ((c = fgetc(fp)) != EOF && c != '\n') 
    {
        if (num_elts % block_size == 0)
        {
            *line = realloc(*line, (++blocks)*block_size*sizeof(char));
            assert(*line);
        }

        (*line)[ num_elts++ ] = c;
    }

    if (num_elts % block_size == 0)
    {
        *line = realloc(*line, (blocks*block_size+1)*sizeof(char));
        assert(*line);
    }
    (*line)[ num_elts ] = '\0';

    return (c == EOF);
}


/** @brief Counts the number of space-separated vector columns in a line. */
static uint32_t count_vector_line_columns(const char* line)
{
    uint32_t count;
    uint32_t done;
    uint32_t in_column;

    const char* c;

    count     = 0;
    done      = 0;
    in_column = 0;

    c = line;

    while(!done)
    {
        if (*c == '\0')
        {
            done = 1;
        }
        else if (isspace(*c))
        {
            in_column = 0;
        }
        else if (!in_column)
        {
            count++;
            in_column = 1;
        }
        c++;
    }

    return count;
}


/** @brief Increments a pointer to a vector line to the next column. */
static const char* advance_line_column_ptr(const char* col)
{
    while (*col != ' ' && *col != '\0') col++;

    while (*col == ' ' && *col != '\0') col++;

    return col;
}




/**
 * @name read_vector
 *
 * Reads a vector from a file.
 * @{
 */

/**
 * Format of the vector in the file is expected to be one line containing 
 * all the elements. Blank lines and lines beginning with '#' are skipped. For 
 * example,
 *
 * @code
 * # Comment.
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v_out   Result parameter. If @em *v_out is NULL, a vector is
 *                 allocated; otherwise its space is re-used.
 * @param  fname   Name of the file to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *v_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the vector in the file is 
 *   not formatted properly.
 */
Error* read_vector_u32(Vector_u32** v_out, const char* fname)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "r")) == NULL)
    {
        free_vector_u32(*v_out); *v_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    if ((e = read_vector_fp_u32(v_out, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        free_vector_u32(*v_out); *v_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the vector in the file is expected to be one line containing 
 * all the elements. Blank lines and lines beginning with '#' are skipped. For 
 * example,
 *
 * @code
 * # Comment.
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v_out   Result parameter. If @em *v_out is NULL, a vector is
 *                 allocated; otherwise its space is re-used.
 * @param  fname   Name of the file to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *v_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the vector in the file is 
 *   not formatted properly.
 */
Error* read_vector_i32(Vector_i32** v_out, const char* fname)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "r")) == NULL)
    {
        free_vector_i32(*v_out); *v_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    if ((e = read_vector_fp_i32(v_out, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        free_vector_i32(*v_out); *v_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the vector in the file is expected to be one line containing 
 * all the elements. Blank lines and lines beginning with '#' are skipped. For 
 * example,
 *
 * @code
 * # Comment.
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v_out   Result parameter. If @em *v_out is NULL, a vector is
 *                 allocated; otherwise its space is re-used.
 * @param  fname   Name of the file to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *v_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the vector in the file is 
 *   not formatted properly.
 */
Error* read_vector_i64(Vector_i64** v_out, const char* fname)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "r")) == NULL)
    {
        free_vector_i64(*v_out); *v_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    if ((e = read_vector_fp_i64(v_out, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        free_vector_i64(*v_out); *v_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the vector in the file is expected to be one line containing 
 * all the elements. Blank lines and lines beginning with '#' are skipped. For 
 * example,
 *
 * @code
 * # Comment.
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v_out   Result parameter. If @em *v_out is NULL, a vector is
 *                 allocated; otherwise its space is re-used.
 * @param  fname   Name of the file to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *v_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the vector in the file is 
 *   not formatted properly.
 */
Error* read_vector_f(Vector_f** v_out, const char* fname)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "r")) == NULL)
    {
        free_vector_f(*v_out); *v_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    if ((e = read_vector_fp_f(v_out, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        free_vector_f(*v_out); *v_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the vector in the file is expected to be one line containing 
 * all the elements. Blank lines and lines beginning with '#' are skipped. For 
 * example,
 *
 * @code
 * # Comment.
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v_out   Result parameter. If @em *v_out is NULL, a vector is
 *                 allocated; otherwise its space is re-used.
 * @param  fname   Name of the file to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *v_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the vector in the file is 
 *   not formatted properly.
 */
Error* read_vector_d(Vector_d** v_out, const char* fname)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "r")) == NULL)
    {
        free_vector_d(*v_out); *v_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    if ((e = read_vector_fp_d(v_out, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        free_vector_d(*v_out); *v_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the vector in the file is expected to be one line containing 
 * all the elements. Blank lines and lines beginning with '#' are skipped. For 
 * example,
 *
 * @code
 * # Comment.
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v_out   Result parameter. If @em *v_out is NULL, a vector is
 *                 allocated; otherwise its space is re-used.
 * @param  fname   Name of the file to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *v_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the vector in the file is 
 *   not formatted properly.
 */
Error* read_vector_cf(Vector_cf** v_out, const char* fname)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "r")) == NULL)
    {
        free_vector_cf(*v_out); *v_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    if ((e = read_vector_fp_cf(v_out, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        free_vector_cf(*v_out); *v_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the vector in the file is expected to be one line containing 
 * all the elements. Blank lines and lines beginning with '#' are skipped. For 
 * example,
 *
 * @code
 * # Comment.
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v_out   Result parameter. If @em *v_out is NULL, a vector is
 *                 allocated; otherwise its space is re-used.
 * @param  fname   Name of the file to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *v_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the vector in the file is 
 *   not formatted properly.
 */
Error* read_vector_cd(Vector_cd** v_out, const char* fname)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "r")) == NULL)
    {
        free_vector_cd(*v_out); *v_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    if ((e = read_vector_fp_cd(v_out, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        free_vector_cd(*v_out); *v_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/*@}*/




/**
 * @name read_vector_fp
 *
 * Reads a vector from an opened file.
 * @{
 */

/**
 * Format of the vector in the file is expected to be one line containing 
 * all the elements. Blank lines and lines beginning with '#' are skipped. For 
 * example,
 *
 * @code
 * # Comment.
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v_out   Result parameter. If @em *v_out is NULL, a vector is
 *                 allocated; otherwise its space is re-used.
 * @param  fp      File to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *v_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the vector in the file is 
 *   not formatted properly.
 */
Error* read_vector_fp_u32(Vector_u32** v_out, FILE* fp)
{
    static const uint32_t block_size = 1024;

    uint32_t num_elts, num_cols;
    uint32_t elt, col;
    uint32_t blocks;

    char* line = NULL;
    const char* line_col;

    uint32_t  val;
    uint32_t* elts   = NULL;
    uint32_t* v_elts = NULL;

    if (skip_fp_spaces_and_comments(fp) == 0)
    {
        free_vector_u32(*v_out); *v_out = NULL;
        return JWSC_EIO("Nothing to read");
    }

    num_cols = 0;
    num_elts = 0;
    blocks   = 0;

    read_vector_line(&line, fp);

    if ((num_cols = count_vector_line_columns(line)) == 0)
    {
        free(line);
        free_vector_u32(*v_out); *v_out = NULL;
        return JWSC_EIO("Vector to read not formatted properly");
    }

    line_col = line;

    for (col = 0; col < num_cols; col++)
    {
        if (sscanf(line_col, "%d", &val) != 1)
        {
            free(line);
            free(elts);
            free_vector_u32(*v_out); *v_out = NULL;
            return JWSC_EIO("Vector to read not formatted properly");
        }

        line_col = advance_line_column_ptr(line_col);

        if (num_elts % block_size == 0)
        {
            elts = realloc(elts, (++blocks)*block_size*sizeof(uint32_t));
            assert(elts);
        }

        elts[ num_elts++ ] = val;
    }

    create_vector_u32(v_out, num_elts);
    v_elts = (*v_out)->elts;

    for (elt = 0; elt < num_elts; elt++)
    {
        v_elts[ elt ] = elts[ elt ];
    }

    free(line);
    free(elts);

    return NULL;
}

/**
 * Format of the vector in the file is expected to be one line containing 
 * all the elements. Blank lines and lines beginning with '#' are skipped. For 
 * example,
 *
 * @code
 * # Comment.
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v_out   Result parameter. If @em *v_out is NULL, a vector is
 *                 allocated; otherwise its space is re-used.
 * @param  fp      File to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *v_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the vector in the file is 
 *   not formatted properly.
 */
Error* read_vector_fp_i32(Vector_i32** v_out, FILE* fp)
{
    static const uint32_t block_size = 1024;

    uint32_t num_elts, num_cols;
    uint32_t elt, col;
    uint32_t blocks;

    char* line = NULL;
    const char* line_col;

    int32_t  val;
    int32_t* elts   = NULL;
    int32_t* v_elts = NULL;

    if (skip_fp_spaces_and_comments(fp) == 0)
    {
        free_vector_i32(*v_out); *v_out = NULL;
        return JWSC_EIO("Nothing to read");
    }

    num_cols = 0;
    num_elts = 0;
    blocks   = 0;

    read_vector_line(&line, fp);

    if ((num_cols = count_vector_line_columns(line)) == 0)
    {
        free(line);
        free_vector_i32(*v_out); *v_out = NULL;
        return JWSC_EIO("Vector to read not formatted properly");
    }

    line_col = line;

    for (col = 0; col < num_cols; col++)
    {
        if (sscanf(line_col, "%d", &val) != 1)
        {
            free(line);
            free(elts);
            free_vector_i32(*v_out); *v_out = NULL;
            return JWSC_EIO("Vector to read not formatted properly");
        }

        line_col = advance_line_column_ptr(line_col);

        if (num_elts % block_size == 0)
        {
            elts = realloc(elts, (++blocks)*block_size*sizeof(int32_t));
            assert(elts);
        }

        elts[ num_elts++ ] = val;
    }

    create_vector_i32(v_out, num_elts);
    v_elts = (*v_out)->elts;

    for (elt = 0; elt < num_elts; elt++)
    {
        v_elts[ elt ] = elts[ elt ];
    }

    free(line);
    free(elts);

    return NULL;
}

/**
 * Format of the vector in the file is expected to be one line containing 
 * all the elements. Blank lines and lines beginning with '#' are skipped. For 
 * example,
 *
 * @code
 * # Comment.
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v_out   Result parameter. If @em *v_out is NULL, a vector is
 *                 allocated; otherwise its space is re-used.
 * @param  fp      File to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *v_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the vector in the file is 
 *   not formatted properly.
 */
Error* read_vector_fp_i64(Vector_i64** v_out, FILE* fp)
{
    static const uint32_t block_size = 1024;

    uint32_t num_elts, num_cols;
    uint32_t elt, col;
    uint32_t blocks;

    char* line = NULL;
    const char* line_col;

    int64_t  val;
    int64_t* elts   = NULL;
    int64_t* v_elts = NULL;

    if (skip_fp_spaces_and_comments(fp) == 0)
    {
        free_vector_i64(*v_out); *v_out = NULL;
        return JWSC_EIO("Nothing to read");
    }

    num_cols = 0;
    num_elts = 0;
    blocks   = 0;

    read_vector_line(&line, fp);

    if ((num_cols = count_vector_line_columns(line)) == 0)
    {
        free(line);
        free_vector_i64(*v_out); *v_out = NULL;
        return JWSC_EIO("Vector to read not formatted properly");
    }

    line_col = line;

    for (col = 0; col < num_cols; col++)
    {
        if (sscanf(line_col, "%" SCNd64, &val) != 1)
        {
            free(line);
            free(elts);
            free_vector_i64(*v_out); *v_out = NULL;
            return JWSC_EIO("Vector to read not formatted properly");
        }

        line_col = advance_line_column_ptr(line_col);

        if (num_elts % block_size == 0)
        {
            elts = realloc(elts, (++blocks)*block_size*sizeof(int64_t));
            assert(elts);
        }

        elts[ num_elts++ ] = val;
    }

    create_vector_i64(v_out, num_elts);
    v_elts = (*v_out)->elts;

    for (elt = 0; elt < num_elts; elt++)
    {
        v_elts[ elt ] = elts[ elt ];
    }

    free(line);
    free(elts);

    return NULL;
}

/**
 * Format of the vector in the file is expected to be one line containing 
 * all the elements. Blank lines and lines beginning with '#' are skipped. For 
 * example,
 *
 * @code
 * # Comment.
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v_out   Result parameter. If @em *v_out is NULL, a vector is
 *                 allocated; otherwise its space is re-used.
 * @param  fp      File to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *v_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the vector in the file is 
 *   not formatted properly.
 */
Error* read_vector_fp_f(Vector_f** v_out, FILE* fp)
{
    static const uint32_t block_size = 1024;

    uint32_t num_elts, num_cols;
    uint32_t elt, col;
    uint32_t blocks;

    char* line = NULL;
    const char* line_col;

    float  val;
    float* elts   = NULL;
    float* v_elts = NULL;

    if (skip_fp_spaces_and_comments(fp) == 0)
    {
        free_vector_f(*v_out); *v_out = NULL;
        return JWSC_EIO("Nothing to read");
    }

    num_cols = 0;
    num_elts = 0;
    blocks   = 0;

    read_vector_line(&line, fp);

    if ((num_cols = count_vector_line_columns(line)) == 0)
    {
        free(line);
        free_vector_f(*v_out); *v_out = NULL;
        return JWSC_EIO("Vector to read not formatted properly");
    }

    line_col = line;

    for (col = 0; col < num_cols; col++)
    {
        if (sscanf(line_col, "%f", &val) != 1)
        {
            free(line);
            free(elts);
            free_vector_f(*v_out); *v_out = NULL;
            return JWSC_EIO("Vector to read not formatted properly");
        }

        line_col = advance_line_column_ptr(line_col);

        if (num_elts % block_size == 0)
        {
            elts = realloc(elts, (++blocks)*block_size*sizeof(float));
            assert(elts);
        }

        elts[ num_elts++ ] = val;
    }

    create_vector_f(v_out, num_elts);
    v_elts = (*v_out)->elts;

    for (elt = 0; elt < num_elts; elt++)
    {
        v_elts[ elt ] = elts[ elt ];
    }

    free(line);
    free(elts);

    return NULL;
}

/**
 * Format of the vector in the file is expected to be one line containing 
 * all the elements. Blank lines and lines beginning with '#' are skipped. For 
 * example,
 *
 * @code
 * # Comment.
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v_out   Result parameter. If @em *v_out is NULL, a vector is
 *                 allocated; otherwise its space is re-used.
 * @param  fp      File to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *v_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the vector in the file is 
 *   not formatted properly.
 */
Error* read_vector_fp_d(Vector_d** v_out, FILE* fp)
{
    static const uint32_t block_size = 1024;

    uint32_t num_elts, num_cols;
    uint32_t elt, col;
    uint32_t blocks;

    char* line = NULL;
    const char* line_col;

    double  val;
    double* elts   = NULL;
    double* v_elts = NULL;

    if (skip_fp_spaces_and_comments(fp) == 0)
    {
        free_vector_d(*v_out); *v_out = NULL;
        return JWSC_EIO("Nothing to read");
    }

    num_cols = 0;
    num_elts = 0;
    blocks   = 0;

    read_vector_line(&line, fp);

    if ((num_cols = count_vector_line_columns(line)) == 0)
    {
        free(line);
        free_vector_d(*v_out); *v_out = NULL;
        return JWSC_EIO("Vector to read not formatted properly");
    }

    line_col = line;

    for (col = 0; col < num_cols; col++)
    {
        if (sscanf(line_col, "%lf", &val) != 1)
        {
            free(line);
            free(elts);
            free_vector_d(*v_out); *v_out = NULL;
            return JWSC_EIO("Vector to read not formatted properly");
        }

        line_col = advance_line_column_ptr(line_col);

        if (num_elts % block_size == 0)
        {
            elts = realloc(elts, (++blocks)*block_size*sizeof(double));
            assert(elts);
        }

        elts[ num_elts++ ] = val;
    }

    create_vector_d(v_out, num_elts);
    v_elts = (*v_out)->elts;

    for (elt = 0; elt < num_elts; elt++)
    {
        v_elts[ elt ] = elts[ elt ];
    }

    free(line);
    free(elts);

    return NULL;
}

/**
 * Format of the vector in the file is expected to be one line containing 
 * all the elements. Blank lines and lines beginning with '#' are skipped. For 
 * example,
 *
 * @code
 * # Comment.
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v_out   Result parameter. If @em *v_out is NULL, a vector is
 *                 allocated; otherwise its space is re-used.
 * @param  fp      File to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *v_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the vector in the file is 
 *   not formatted properly.
 */
Error* read_vector_fp_cf(Vector_cf** v_out, FILE* fp)
{
    static const uint32_t block_size = 1024;

    uint32_t num_elts, num_cols;
    uint32_t elt, col;
    uint32_t blocks;

    char* line = NULL;
    const char* line_col;

    Complex_f  val;
    Complex_f* elts   = NULL;
    Complex_f* v_elts = NULL;

    if (skip_fp_spaces_and_comments(fp) == 0)
    {
        free_vector_cf(*v_out); *v_out = NULL;
        return JWSC_EIO("Nothing to read");
    }

    num_cols = 0;
    num_elts = 0;
    blocks   = 0;

    read_vector_line(&line, fp);

    if ((num_cols = count_vector_line_columns(line)) == 0)
    {
        free(line);
        free_vector_cf(*v_out); *v_out = NULL;
        return JWSC_EIO("Vector to read not formatted properly");
    }

    line_col = line;

    for (col = 0; col < num_cols; col++)
    {
        if (sscanf(line_col, "%f %f", &(val.r), &(val.i)) != 2)
        {
            free(line);
            free(elts);
            free_vector_cf(*v_out); *v_out = NULL;
            return JWSC_EIO("Vector to read not formatted properly");
        }

        line_col = advance_line_column_ptr(line_col);
        line_col = advance_line_column_ptr(line_col);

        if (num_elts % block_size == 0)
        {
            elts = realloc(elts, (++blocks)*block_size*sizeof(Complex_f));
            assert(elts);
        }

        elts[ num_elts++ ] = val;
    }

    create_vector_cf(v_out, num_elts);
    v_elts = (*v_out)->elts;

    for (elt = 0; elt < num_elts; elt++)
    {
        v_elts[ elt ] = elts[ elt ];
    }

    free(line);
    free(elts);

    return NULL;
}

/**
 * Format of the vector in the file is expected to be one line containing 
 * all the elements. Blank lines and lines beginning with '#' are skipped. For 
 * example,
 *
 * @code
 * # Comment.
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v_out   Result parameter. If @em *v_out is NULL, a vector is
 *                 allocated; otherwise its space is re-used.
 * @param  fp      File to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *v_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the vector in the file is 
 *   not formatted properly.
 */
Error* read_vector_fp_cd(Vector_cd** v_out, FILE* fp)
{
    static const uint32_t block_size = 1024;

    uint32_t num_elts, num_cols;
    uint32_t elt, col;
    uint32_t blocks;

    char* line = NULL;
    const char* line_col;

    Complex_d  val;
    Complex_d* elts   = NULL;
    Complex_d* v_elts = NULL;

    if (skip_fp_spaces_and_comments(fp) == 0)
    {
        free_vector_cd(*v_out); *v_out = NULL;
        return JWSC_EIO("Nothing to read");
    }

    num_cols = 0;
    num_elts = 0;
    blocks   = 0;

    read_vector_line(&line, fp);

    if ((num_cols = count_vector_line_columns(line)) == 0)
    {
        free(line);
        free_vector_cd(*v_out); *v_out = NULL;
        return JWSC_EIO("Vector to read not formatted properly");
    }

    line_col = line;

    for (col = 0; col < num_cols; col++)
    {
        if (sscanf(line_col, "%lf %lf", &(val.r), &(val.i)) != 2)
        {
            free(line);
            free(elts);
            free_vector_cd(*v_out); *v_out = NULL;
            return JWSC_EIO("Vector to read not formatted properly");
        }

        line_col = advance_line_column_ptr(line_col);
        line_col = advance_line_column_ptr(line_col);

        if (num_elts % block_size == 0)
        {
            elts = realloc(elts, (++blocks)*block_size*sizeof(Complex_d));
            assert(elts);
        }

        elts[ num_elts++ ] = val;
    }

    create_vector_cd(v_out, num_elts);
    v_elts = (*v_out)->elts;

    for (elt = 0; elt < num_elts; elt++)
    {
        v_elts[ elt ] = elts[ elt ];
    }

    free(line);
    free(elts);

    return NULL;
}

/*@}*/




/**
 * @name write_vector
 *
 * Writes a vector to a file.
 * @{
 */

/**
 * Format of the vector written is one line containing all the elements. For 
 * example,
 *
 * @code
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v      Vector to write.
 * @param  fname  Name of the file to write the vector to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_vector_u32(const Vector_u32* v, const char* fname)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return JWSC_EIO(strerror(errno));
    }

    if ((e = write_vector_fp_u32(v, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the vector written is one line containing all the elements. For 
 * example,
 *
 * @code
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v      Vector to write.
 * @param  fname  Name of the file to write the vector to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_vector_i32(const Vector_i32* v, const char* fname)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return JWSC_EIO(strerror(errno));
    }

    if ((e = write_vector_fp_i32(v, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the vector written is one line containing all the elements. For 
 * example,
 *
 * @code
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v      Vector to write.
 * @param  fname  Name of the file to write the vector to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_vector_i64(const Vector_i64* v, const char* fname)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return JWSC_EIO(strerror(errno));
    }

    if ((e = write_vector_fp_i64(v, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the vector written is one line containing all the elements. For 
 * example,
 *
 * @code
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v      Vector to write.
 * @param  fname  Name of the file to write the vector to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_vector_f(const Vector_f* v, const char* fname)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return JWSC_EIO(strerror(errno));
    }

    if ((e = write_vector_fp_f(v, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the vector written is one line containing all the elements. For 
 * example,
 *
 * @code
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v      Vector to write.
 * @param  fname  Name of the file to write the vector to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_vector_d(const Vector_d* v, const char* fname)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return JWSC_EIO(strerror(errno));
    }

    if ((e = write_vector_fp_d(v, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the vector written is one line containing all the elements. For 
 * example,
 *
 * @code
 * 1 2 3 4 5 6
 * @endcode
 *
 * @param  v      Vector to write.
 * @param  fname  Name of the file to write the vector to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_vector_cf(const Vector_cf* v, const char* fname)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return JWSC_EIO(strerror(errno));
    }

    if ((e = write_vector_fp_cf(v, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the vector written is one line containing all the elements. For 
 * example,
 *
 * @code
 * 1 2 3 4 5 6
 * @endcode
 *
 * @param  v      Vector to write.
 * @param  fname  Name of the file to write the vector to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_vector_cd(const Vector_cd* v, const char* fname)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return JWSC_EIO(strerror(errno));
    }

    if ((e = write_vector_fp_cd(v, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/*@}*/




/**
 * @name write_vector
 *
 * Writes a vector to a file.
 * @{
 */

/**
 * Format of the vector written is one line containing all the elements. For 
 * example,
 *
 * @code
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v      Vector to write.
 * @param  fmt    Printf style format for each element of the vector.
 * @param  fname  Name of the file to write the vector to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_formatted_vector_u32
(
    const Vector_u32* v, 
    const char*       fmt,
    const char*       fname
)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return JWSC_EIO(strerror(errno));
    }

    if ((e = write_formatted_vector_fp_u32(v, fmt, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the vector written is one line containing all the elements. For 
 * example,
 *
 * @code
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v      Vector to write.
 * @param  fmt    Printf style format for each element of the vector.
 * @param  fname  Name of the file to write the vector to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_formatted_vector_i32
(
    const Vector_i32* v, 
    const char*       fmt,
    const char*       fname
)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return JWSC_EIO(strerror(errno));
    }

    if ((e = write_formatted_vector_fp_i32(v, fmt, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the vector written is one line containing all the elements. For 
 * example,
 *
 * @code
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v      Vector to write.
 * @param  fmt    Printf style format for each element of the vector.
 * @param  fname  Name of the file to write the vector to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_formatted_vector_i64
(
    const Vector_i64* v, 
    const char*       fmt,
    const char*       fname
)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return JWSC_EIO(strerror(errno));
    }

    if ((e = write_formatted_vector_fp_i64(v, fmt, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the vector written is one line containing all the elements. For 
 * example,
 *
 * @code
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v      Vector to write.
 * @param  fmt    Printf style format for each element of the vector.
 * @param  fname  Name of the file to write the vector to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_formatted_vector_f
(
    const Vector_f* v, 
    const char*     fmt,
    const char*     fname
)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return JWSC_EIO(strerror(errno));
    }

    if ((e = write_formatted_vector_fp_f(v, fmt, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the vector written is one line containing all the elements. For 
 * example,
 *
 * @code
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v      Vector to write.
 * @param  fmt    Printf style format for each element of the vector.
 * @param  fname  Name of the file to write the vector to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_formatted_vector_d
(
    const Vector_d* v, 
    const char*     fmt,
    const char*     fname
)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return JWSC_EIO(strerror(errno));
    }

    if ((e = write_formatted_vector_fp_d(v, fmt, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the vector written is one line containing all the elements. For 
 * example,
 *
 * @code
 * 1 2 3 4 5 6
 * @endcode
 *
 * @param  v      Vector to write.
 * @param  fmt    Printf style format for each element of the vector.
 * @param  fname  Name of the file to write the vector to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_formatted_vector_cf
(
    const Vector_cf* v, 
    const char*      fmt,
    const char*      fname
)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return JWSC_EIO(strerror(errno));
    }

    if ((e = write_formatted_vector_fp_cf(v, fmt, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the vector written is one line containing all the elements. For 
 * example,
 *
 * @code
 * 1 2 3 4 5 6
 * @endcode
 *
 * @param  v      Vector to write.
 * @param  fmt    Printf style format for each element of the vector.
 * @param  fname  Name of the file to write the vector to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_formatted_vector_cd
(
    const Vector_cd* v, 
    const char*      fmt,
    const char*      fname
)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return JWSC_EIO(strerror(errno));
    }

    if ((e = write_formatted_vector_fp_cd(v, fmt, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/*@}*/




/**
 * @name write_vector_fp
 *
 * Writes a vector to an opened file.
 * @{
 */

/**
 * Format of the vector written is one line containing all the elements. For 
 * example,
 *
 * @code
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v   Vector to write.
 * @param  fp  File to write the vector to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_vector_fp_u32(const Vector_u32* v, FILE* fp)
{
    return write_formatted_vector_fp_u32(v, "%u", fp);
}

/**
 * Format of the vector written is one line containing all the elements. For 
 * example,
 *
 * @code
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v   Vector to write.
 * @param  fp  File to write the vector to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_vector_fp_i32(const Vector_i32* v, FILE* fp)
{
    return write_formatted_vector_fp_i32(v, "%d", fp);
}

/**
 * Format of the vector written is one line containing all the elements. For 
 * example,
 *
 * @code
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v   Vector to write.
 * @param  fp  File to write the vector to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_vector_fp_i64(const Vector_i64* v, FILE* fp)
{
#ifdef JWSC_HAVE_64_BIT_LONG_INT
    return write_formatted_vector_fp_i64(v, "%ld", fp);
#else
    return write_formatted_vector_fp_i64(v, "%lld", fp);
#endif
}

/**
 * Format of the vector written is one line containing all the elements. For 
 * example,
 *
 * @code
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v   Vector to write.
 * @param  fp  File to write the vector to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_vector_fp_f(const Vector_f* v, FILE* fp)
{
    return write_formatted_vector_fp_f(v, "%f", fp);
}

/**
 * Format of the vector written is one line containing all the elements. For 
 * example,
 *
 * @code
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v   Vector to write.
 * @param  fp  File to write the vector to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_vector_fp_d(const Vector_d* v, FILE* fp)
{
    return write_formatted_vector_fp_d(v, "%.8e", fp);
}

/**
 * Format of the vector written is one line containing all the elements. For 
 * example,
 *
 * @code
 * 1 2 3 4 5 6
 * @endcode
 *
 * @param  v   Vector to write.
 * @param  fp  File to write the vector to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_vector_fp_cf(const Vector_cf* v, FILE* fp)
{
    return write_formatted_vector_fp_cf(v, "%f %f", fp);
}

/**
 * Format of the vector written is one line containing all the elements. For 
 * example,
 *
 * @code
 * 1 2 3 4 5 6
 * @endcode
 *
 * @param  v   Vector to write.
 * @param  fp  File to write the vector to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_vector_fp_cd(const Vector_cd* v, FILE* fp)
{
    return write_formatted_vector_fp_cd(v, "%lf %lf", fp);
}

/*@}*/




/**
 * @name write_formatted_vector_fp
 *
 * Writes a formatted vector to an opened file.
 * @{
 */

/**
 * Format of the vector written is one line containing all the elements. For 
 * example,
 *
 * @code
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v    Vector to write.
 * @param  fmt  Printf style format for each element of the vector.
 * @param  fp   File to write the vector to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_formatted_vector_fp_u32
(
    const Vector_u32* v, 
    const char*       fmt, 
    FILE*             fp
)
{
    uint32_t num_elts, elt;

    num_elts = v->num_elts;
    for (elt = 0; elt < num_elts; elt++)
    {
        fprintf(fp, fmt, v->elts[ elt ]);
        fprintf(fp, " ");
    }
    fprintf(fp, "\n");

    return NULL;
}

/**
 * Format of the vector written is one line containing all the elements. For 
 * example,
 *
 * @code
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v    Vector to write.
 * @param  fmt  Printf style format for each element of the vector.
 * @param  fp   File to write the vector to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_formatted_vector_fp_i32
(
    const Vector_i32* v, 
    const char*       fmt,
    FILE*             fp
)
{
    uint32_t num_elts, elt;

    num_elts = v->num_elts;
    for (elt = 0; elt < num_elts; elt++)
    {
        fprintf(fp, fmt, v->elts[ elt ]);
        fprintf(fp, " ");
    }
    fprintf(fp, "\n");

    return NULL;
}

/**
 * Format of the vector written is one line containing all the elements. For 
 * example,
 *
 * @code
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v    Vector to write.
 * @param  fmt  Printf style format for each element of the vector.
 * @param  fp   File to write the vector to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_formatted_vector_fp_i64
(
    const Vector_i64* v, 
    const char*       fmt,
    FILE*             fp
)
{
    uint32_t num_elts, elt;

    num_elts = v->num_elts;
    for (elt = 0; elt < num_elts; elt++)
    {
        fprintf(fp, fmt, v->elts[ elt ]);
        fprintf(fp, " ");
    }
    fprintf(fp, "\n");

    return NULL;
}

/**
 * Format of the vector written is one line containing all the elements. For 
 * example,
 *
 * @code
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v    Vector to write.
 * @param  fmt  Printf style format for each element of the vector.
 * @param  fp   File to write the vector to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_formatted_vector_fp_f
(
    const Vector_f* v, 
    const char*     fmt,
    FILE*           fp
)
{
    uint32_t num_elts, elt;

    num_elts = v->num_elts;
    for (elt = 0; elt < num_elts; elt++)
    {
        fprintf(fp, fmt, v->elts[ elt ]);
        fprintf(fp, " ");
    }
    fprintf(fp, "\n");

    return NULL;
}

/**
 * Format of the vector written is one line containing all the elements. For 
 * example,
 *
 * @code
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v    Vector to write.
 * @param  fmt  Printf style format for each element of the vector.
 * @param  fp   File to write the vector to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_formatted_vector_fp_d
(
    const Vector_d* v, 
    const char*     fmt,
    FILE*           fp
)
{
    uint32_t num_elts, elt;

    num_elts = v->num_elts;
    for (elt = 0; elt < num_elts; elt++)
    {
        fprintf(fp, fmt, v->elts[ elt ]);
        fprintf(fp, " ");
    }
    fprintf(fp, "\n");

    return NULL;
}

/**
 * Format of the vector written is one line containing all the elements. For 
 * example,
 *
 * @code
 * 1 2 3 4 5 6
 * @endcode
 *
 * @param  v    Vector to write.
 * @param  fmt  Printf style format for each element of the vector.
 * @param  fp   File to write the vector to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_formatted_vector_fp_cf
(
    const Vector_cf* v, 
    const char*      fmt,
    FILE*            fp
)
{
    uint32_t num_elts, elt;

    num_elts = v->num_elts;
    for (elt = 0; elt < num_elts; elt++)
    {
        fprintf(fp, fmt, v->elts[ elt ].r, v->elts[ elt ].i);
        fprintf(fp, " ");
    }
    fprintf(fp, "\n");

    return NULL;
}

/**
 * Format of the vector written is one line containing all the elements. For 
 * example,
 *
 * @code
 * 1 2 3 4 5 6
 * @endcode
 *
 * @param  v    Vector to write.
 * @param  fmt  Printf style format for each element of the vector.
 * @param  fp   File to write the vector to.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_formatted_vector_fp_cd
(
    const Vector_cd* v, 
    const char*      fmt,
    FILE*            fp
)
{
    uint32_t num_elts, elt;

    num_elts = v->num_elts;
    for (elt = 0; elt < num_elts; elt++)
    {
        fprintf(fp, fmt, v->elts[ elt ].r, v->elts[ elt ].i);
        fprintf(fp, " ");
    }
    fprintf(fp, "\n");

    return NULL;
}

/*@}*/




/**
 * @name read_vector_with_header
 *
 * Reads a vector with a header from a file.
 * @{
 */

/**
 * Format of the vector in the file is expected to be a line containing the 
 * number of elements followed by a line containing all the elements. Blank
 * lines and lines beginning with '#' are skipped. For example,
 *
 * @code
 * # Comment
 * num_elts=9
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v_out   Result parameter. If @em *v_out is NULL, a vector is
 *                 allocated; otherwise its space is re-used.
 * @param  fname   Name of the file to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *v_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the vector in the file is 
 *   not formatted properly.
 */
Error* read_vector_with_header_u32(Vector_u32** v_out, const char* fname)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "r")) == NULL)
    {
        free_vector_u32(*v_out); *v_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    if ((e = read_vector_with_header_fp_u32(v_out, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        free_vector_u32(*v_out); *v_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the vector in the file is expected to be a line containing the 
 * number of elements followed by a line containing all the elements. Blank
 * lines and lines beginning with '#' are skipped. For example,
 *
 * @code
 * # Comment
 * num_elts=9
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v_out   Result parameter. If @em *v_out is NULL, a vector is
 *                 allocated; otherwise its space is re-used.
 * @param  fname   Name of the file to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *v_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the vector in the file is 
 *   not formatted properly.
 */
Error* read_vector_with_header_i32(Vector_i32** v_out, const char* fname)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "r")) == NULL)
    {
        free_vector_i32(*v_out); *v_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    if ((e = read_vector_with_header_fp_i32(v_out, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        free_vector_i32(*v_out); *v_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the vector in the file is expected to be a line containing the 
 * number of elements followed by a line containing all the elements. Blank
 * lines and lines beginning with '#' are skipped. For example,
 *
 * @code
 * # Comment
 * num_elts=9
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v_out   Result parameter. If @em *v_out is NULL, a vector is
 *                 allocated; otherwise its space is re-used.
 * @param  fname   Name of the file to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *v_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the vector in the file is 
 *   not formatted properly.
 */
Error* read_vector_with_header_i64(Vector_i64** v_out, const char* fname)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "r")) == NULL)
    {
        free_vector_i64(*v_out); *v_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    if ((e = read_vector_with_header_fp_i64(v_out, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        free_vector_i64(*v_out); *v_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the vector in the file is expected to be a line containing the 
 * number of elements followed by a line containing all the elements. Blank
 * lines and lines beginning with '#' are skipped. For example,
 *
 * @code
 * # Comment
 * num_elts=9
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v_out   Result parameter. If @em *v_out is NULL, a vector is
 *                 allocated; otherwise its space is re-used.
 * @param  fname   Name of the file to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *v_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the vector in the file is 
 *   not formatted properly.
 */
Error* read_vector_with_header_f(Vector_f** v_out, const char* fname)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "r")) == NULL)
    {
        free_vector_f(*v_out); *v_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    if ((e = read_vector_with_header_fp_f(v_out, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        free_vector_f(*v_out); *v_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the vector in the file is expected to be a line containing the 
 * number of elements followed by a line containing all the elements. Blank
 * lines and lines beginning with '#' are skipped. For example,
 *
 * @code
 * # Comment
 * num_elts=9
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v_out   Result parameter. If @em *v_out is NULL, a vector is
 *                 allocated; otherwise its space is re-used.
 * @param  fname   Name of the file to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *v_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the vector in the file is 
 *   not formatted properly.
 */
Error* read_vector_with_header_d(Vector_d** v_out, const char* fname)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "r")) == NULL)
    {
        free_vector_d(*v_out); *v_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    if ((e = read_vector_with_header_fp_d(v_out, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        free_vector_d(*v_out); *v_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the vector in the file is expected to be a line containing the 
 * number of elements followed by a line containing all the elements. Blank
 * lines and lines beginning with '#' are skipped. For example,
 *
 * @code
 * # Comment
 * num_elts=3
 * 1 2 3 4 5 6
 * @endcode
 *
 * @param  v_out   Result parameter. If @em *v_out is NULL, a vector is
 *                 allocated; otherwise its space is re-used.
 * @param  fname   Name of the file to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *v_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the vector in the file is 
 *   not formatted properly.
 */
Error* read_vector_with_header_cf(Vector_cf** v_out, const char* fname)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "r")) == NULL)
    {
        free_vector_cf(*v_out); *v_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    if ((e = read_vector_with_header_fp_cf(v_out, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        free_vector_cf(*v_out); *v_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the vector in the file is expected to be a line containing the 
 * number of elements followed by a line containing all the elements. Blank
 * lines and lines beginning with '#' are skipped. For example,
 *
 * @code
 * # Comment
 * num_elts=3
 * 1 2 3 4 5 6
 * @endcode
 *
 * @param  v_out   Result parameter. If @em *v_out is NULL, a vector is
 *                 allocated; otherwise its space is re-used.
 * @param  fname   Name of the file to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *v_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the vector in the file is 
 *   not formatted properly.
 */
Error* read_vector_with_header_cd(Vector_cd** v_out, const char* fname)
{
    FILE*  fp;
    Error* e;

    if ((fp = fopen(fname, "r")) == NULL)
    {
        free_vector_cd(*v_out); *v_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    if ((e = read_vector_with_header_fp_cd(v_out, fp)) != NULL)
    {
        fclose(fp);
        return e;
    }

    if (fclose(fp) != 0)
    {
        free_vector_cd(*v_out); *v_out = NULL;
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/*@}*/




/**
 * @name read_vector_with_header_fp
 *
 * Reads a vector with a header from an opened file.
 * @{
 */

/**
 * Format of the vector in the file is expected to be a line containing the 
 * number of elements followed by a line containing all the elements. Blank
 * lines and lines beginning with '#' are skipped. For example,
 *
 * @code
 * # Comment
 * num_elts=9
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v_out  Result parameter. If @em *v_out is NULL, a vector is
 *                allocated; otherwise its space is re-used.
 * @param  fp     File to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *v_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the vector in the file is 
 *   not formatted properly.
 */
Error* read_vector_with_header_fp_u32(Vector_u32** v_out, FILE* fp)
{
    uint32_t    num_elts, elt;
    Vector_u32* v;

    if (skip_fp_spaces_and_comments(fp) == 0)
    {
        free_vector_u32(*v_out); *v_out = NULL;
        return JWSC_EIO("No vector to read in the file");
    }

    if (fscanf(fp, "num_elts=%u\n", &num_elts) != 1)
    {
        free_vector_u32(*v_out); *v_out = NULL;
        return JWSC_EIO("File to read vector from not formatted properly");
    }

    create_vector_u32(v_out, num_elts);
    v = *v_out;

    for (elt = 0; elt < num_elts; elt++)
    {
        if (fscanf(fp, "%u", &(v->elts[ elt ])) != 1)
        {
            free_vector_u32(*v_out); *v_out = NULL;
            return JWSC_EIO("File to read vector from not formatted properly");
        }
    }

    return NULL;
}

/**
 * Format of the vector in the file is expected to be a line containing the 
 * number of elements followed by a line containing all the elements. Blank
 * lines and lines beginning with '#' are skipped. For example,
 *
 * @code
 * # Comment
 * num_elts=9
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v_out  Result parameter. If @em *v_out is NULL, a vector is
 *                allocated; otherwise its space is re-used.
 * @param  fp     File to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *v_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the vector in the file is 
 *   not formatted properly.
 */
Error* read_vector_with_header_fp_i32(Vector_i32** v_out, FILE* fp)
{
    uint32_t    num_elts, elt;
    Vector_i32* v;

    if (skip_fp_spaces_and_comments(fp) == 0)
    {
        free_vector_i32(*v_out); *v_out = NULL;
        return JWSC_EIO("No vector to read in the file");
    }

    if (fscanf(fp, "num_elts=%u\n", &num_elts) != 1)
    {
        free_vector_i32(*v_out); *v_out = NULL;
        return JWSC_EIO("File to read vector from not formatted properly");
    }

    create_vector_i32(v_out, num_elts);
    v = *v_out;

    for (elt = 0; elt < num_elts; elt++)
    {
        if (fscanf(fp, "%d", &(v->elts[ elt ])) != 1)
        {
            free_vector_i32(*v_out); *v_out = NULL;
            return JWSC_EIO("File to read vector from not formatted properly");
        }
    }

    return NULL;
}

/**
 * Format of the vector in the file is expected to be a line containing the 
 * number of elements followed by a line containing all the elements. Blank
 * lines and lines beginning with '#' are skipped. For example,
 *
 * @code
 * # Comment
 * num_elts=9
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v_out  Result parameter. If @em *v_out is NULL, a vector is
 *                allocated; otherwise its space is re-used.
 * @param  fp     File to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *v_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the vector in the file is 
 *   not formatted properly.
 */
Error* read_vector_with_header_fp_i64(Vector_i64** v_out, FILE* fp)
{
    uint32_t    num_elts, elt;
    Vector_i64* v;

    if (skip_fp_spaces_and_comments(fp) == 0)
    {
        free_vector_i64(*v_out); *v_out = NULL;
        return JWSC_EIO("No vector to read in the file");
    }

    if (fscanf(fp, "num_elts=%u\n", &num_elts) != 1)
    {
        free_vector_i64(*v_out); *v_out = NULL;
        return JWSC_EIO("File to read vector from not formatted properly");
    }

    create_vector_i64(v_out, num_elts);
    v = *v_out;

    for (elt = 0; elt < num_elts; elt++)
    {
        if (fscanf(fp, "%" SCNd64, &(v->elts[ elt ])) != 1)
        {
            free_vector_i64(*v_out); *v_out = NULL;
            return JWSC_EIO("File to read vector from not formatted properly");
        }
    }

    return NULL;
}

/**
 * Format of the vector in the file is expected to be a line containing the 
 * number of elements followed by a line containing all the elements. Blank
 * lines and lines beginning with '#' are skipped. For example,
 *
 * @code
 * # Comment
 * num_elts=9
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v_out  Result parameter. If @em *v_out is NULL, a vector is
 *                allocated; otherwise its space is re-used.
 * @param  fp     File to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *v_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the vector in the file is 
 *   not formatted properly.
 */
Error* read_vector_with_header_fp_f(Vector_f** v_out, FILE* fp)
{
    uint32_t  num_elts, elt;
    Vector_f* v;

    if (skip_fp_spaces_and_comments(fp) == 0)
    {
        free_vector_f(*v_out); *v_out = NULL;
        return JWSC_EIO("No vector to read in the file");
    }

    if (fscanf(fp, "num_elts=%u\n", &num_elts) != 1)
    {
        free_vector_f(*v_out); *v_out = NULL;
        return JWSC_EIO("File to read vector from not formatted properly");
    }

    create_vector_f(v_out, num_elts);
    v = *v_out;

    for (elt = 0; elt < num_elts; elt++)
    {
        if (fscanf(fp, "%f", &(v->elts[ elt ])) != 1)
        {
            free_vector_f(*v_out); *v_out = NULL;
            return JWSC_EIO("File to read vector from not formatted properly");
        }
    }

    return NULL;
}

/**
 * Format of the vector in the file is expected to be a line containing the 
 * number of elements followed by a line containing all the elements. Blank
 * lines and lines beginning with '#' are skipped. For example,
 *
 * @code
 * # Comment
 * num_elts=9
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v_out  Result parameter. If @em *v_out is NULL, a vector is
 *                allocated; otherwise its space is re-used.
 * @param  fp     File to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *v_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the vector in the file is 
 *   not formatted properly.
 */
Error* read_vector_with_header_fp_d(Vector_d** v_out, FILE* fp)
{
    uint32_t  num_elts, elt;
    Vector_d* v;

    if (skip_fp_spaces_and_comments(fp) == 0)
    {
        free_vector_d(*v_out); *v_out = NULL;
        return JWSC_EIO("No vector to read in the file");
    }

    if (fscanf(fp, "num_elts=%u\n", &num_elts) != 1)
    {
        free_vector_d(*v_out); *v_out = NULL;
        return JWSC_EIO("File to read vector from not formatted properly");
    }

    create_vector_d(v_out, num_elts);
    v = *v_out;

    for (elt = 0; elt < num_elts; elt++)
    {
        if (fscanf(fp, "%lf", &(v->elts[ elt ])) != 1)
        {
            free_vector_d(*v_out); *v_out = NULL;
            return JWSC_EIO("File to read vector from not formatted properly");
        }
    }

    return NULL;
}

/**
 * Format of the vector in the file is expected to be a line containing the 
 * number of elements followed by a line containing all the elements. Blank
 * lines and lines beginning with '#' are skipped. For example,
 *
 * @code
 * # Comment
 * num_elts=3
 * 1 2 3 4 5 6
 * @endcode
 *
 * @param  v_out  Result parameter. If @em *v_out is NULL, a vector is
 *                allocated; otherwise its space is re-used.
 * @param  fp     File to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *v_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the vector in the file is 
 *   not formatted properly.
 */
Error* read_vector_with_header_fp_cf(Vector_cf** v_out, FILE* fp)
{
    uint32_t   num_elts, elt;
    Vector_cf* v;

    if (skip_fp_spaces_and_comments(fp) == 0)
    {
        free_vector_cf(*v_out); *v_out = NULL;
        return JWSC_EIO("No vector to read in the file");
    }

    if (fscanf(fp, "num_elts=%u\n", &num_elts) != 1)
    {
        free_vector_cf(*v_out); *v_out = NULL;
        return JWSC_EIO("File to read vector from not formatted properly");
    }

    create_vector_cf(v_out, num_elts);
    v = *v_out;

    for (elt = 0; elt < num_elts; elt++)
    {
        if (fscanf(fp, "%f %f", &(v->elts[ elt ].r), &(v->elts[ elt ].i)) != 2)
        {
            free_vector_cf(*v_out); *v_out = NULL;
            return JWSC_EIO("File to read vector from not formatted properly");
        }
    }

    return NULL;
}

/**
 * Format of the vector in the file is expected to be a line containing the 
 * number of elements followed by a line containing all the elements. Blank
 * lines and lines beginning with '#' are skipped. For example,
 *
 * @code
 * # Comment
 * num_elts=3
 * 1 2 3 4 5 6
 * @endcode
 *
 * @param  v_out  Result parameter. If @em *v_out is NULL, a vector is
 *                allocated; otherwise its space is re-used.
 * @param  fp     File to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned and 
 * @em *v_out is freed and set to NULL.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   Either there was a problem reading the file, or the vector in the file is 
 *   not formatted properly.
 */
Error* read_vector_with_header_fp_cd(Vector_cd** v_out, FILE* fp)
{
    uint32_t   num_elts, elt;
    Vector_cd* v;

    if (skip_fp_spaces_and_comments(fp) == 0)
    {
        free_vector_cd(*v_out); *v_out = NULL;
        return JWSC_EIO("No vector to read in the file");
    }

    if (fscanf(fp, "num_elts=%u\n", &num_elts) != 1)
    {
        free_vector_cd(*v_out); *v_out = NULL;
        return JWSC_EIO("File to read vector from not formatted properly");
    }

    create_vector_cd(v_out, num_elts);
    v = *v_out;

    for (elt = 0; elt < num_elts; elt++)
    {
        if (fscanf(fp, "%lf %lf", &(v->elts[ elt ].r), &(v->elts[ elt ].i)) !=2)
        {
            free_vector_cd(*v_out); *v_out = NULL;
            return JWSC_EIO("File to read vector from not formatted properly");
        }
    }

    return NULL;
}

/*@}*/




/**
 * @name write_vector_with_header
 *
 * Writes a vector with a header to a file.
 * @{
 */

/**
 * Format of the vector written is a line containing the number of elements 
 * followed by a line containing all the elements. For example,
 *
 * @code
 * # Comment.
 * num_elts=9
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v      Vector to write.
 * @param  fname  Name of the file to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_vector_with_header_u32(const Vector_u32* v, const char* fname)
{
    FILE* fp;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return JWSC_EIO(strerror(errno));
    }

    write_vector_with_header_fp_u32(v, fp);

    if (fclose(fp) != 0)
    {
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the vector written is a line containing the number of elements 
 * followed by a line containing all the elements. For example,
 *
 * @code
 * # Comment.
 * num_elts=9
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v      Vector to write.
 * @param  fname  Name of the file to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_vector_with_header_i32(const Vector_i32* v, const char* fname)
{
    FILE* fp;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return JWSC_EIO(strerror(errno));
    }

    write_vector_with_header_fp_i32(v, fp);

    if (fclose(fp) != 0)
    {
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the vector written is a line containing the number of elements 
 * followed by a line containing all the elements. For example,
 *
 * @code
 * # Comment.
 * num_elts=9
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v      Vector to write.
 * @param  fname  Name of the file to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_vector_with_header_i64(const Vector_i64* v, const char* fname)
{
    FILE* fp;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return JWSC_EIO(strerror(errno));
    }

    write_vector_with_header_fp_i64(v, fp);

    if (fclose(fp) != 0)
    {
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the vector written is a line containing the number of elements 
 * followed by a line containing all the elements. For example,
 *
 * @code
 * # Comment.
 * num_elts=9
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v      Vector to write.
 * @param  fname  Name of the file to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_vector_with_header_f(const Vector_f* v, const char* fname)
{
    FILE* fp;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return JWSC_EIO(strerror(errno));
    }

    write_vector_with_header_fp_f(v, fp);

    if (fclose(fp) != 0)
    {
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the vector written is a line containing the number of elements 
 * followed by a line containing all the elements. For example,
 *
 * @code
 * # Comment.
 * num_elts=9
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v      Vector to write.
 * @param  fname  Name of the file to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_vector_with_header_d(const Vector_d* v, const char* fname)
{
    FILE* fp;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return JWSC_EIO(strerror(errno));
    }

    write_vector_with_header_fp_d(v, fp);

    if (fclose(fp) != 0)
    {
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the vector written is a line containing the number of elements 
 * followed by a line containing all the elements. For example,
 *
 * @code
 * # Comment.
 * num_elts=3
 * 1 2 3 4 5 6
 * @endcode
 *
 * @param  v      Vector to write.
 * @param  fname  Name of the file to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_vector_with_header_cf(const Vector_cf* v, const char* fname)
{
    FILE* fp;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return JWSC_EIO(strerror(errno));
    }

    write_vector_with_header_fp_cf(v, fp);

    if (fclose(fp) != 0)
    {
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the vector written is a line containing the number of elements 
 * followed by a line containing all the elements. For example,
 *
 * @code
 * # Comment.
 * num_elts=3
 * 1 2 3 4 5 6
 * @endcode
 *
 * @param  v      Vector to write.
 * @param  fname  Name of the file to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_vector_with_header_cd(const Vector_cd* v, const char* fname)
{
    FILE* fp;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return JWSC_EIO(strerror(errno));
    }

    write_vector_with_header_fp_cd(v, fp);

    if (fclose(fp) != 0)
    {
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/*@}*/




/**
 * @name write_formatted_vector_with_header
 *
 * Writes a formatted vector with a header to a file.
 * @{
 */

/**
 * Format of the vector written is a line containing the number of elements 
 * followed by a line containing all the elements. For example,
 *
 * @code
 * # Comment.
 * num_elts=9
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v      Vector to write.
 * @param  fmt    Printf style format for each element of the vector.
 * @param  fname  Name of the file to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_formatted_vector_with_header_u32
(
    const Vector_u32* v, 
    const char*       fmt,
    const char*       fname
)
{
    FILE* fp;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return JWSC_EIO(strerror(errno));
    }

    write_formatted_vector_with_header_fp_u32(v, fmt, fp);

    if (fclose(fp) != 0)
    {
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the vector written is a line containing the number of elements 
 * followed by a line containing all the elements. For example,
 *
 * @code
 * # Comment.
 * num_elts=9
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v      Vector to write.
 * @param  fmt    Printf style format for each element of the vector.
 * @param  fname  Name of the file to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_formatted_vector_with_header_i32
(
    const Vector_i32* v, 
    const char*       fmt,
    const char*       fname
)
{
    FILE* fp;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return JWSC_EIO(strerror(errno));
    }

    write_formatted_vector_with_header_fp_i32(v, fmt, fp);

    if (fclose(fp) != 0)
    {
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the vector written is a line containing the number of elements 
 * followed by a line containing all the elements. For example,
 *
 * @code
 * # Comment.
 * num_elts=9
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v      Vector to write.
 * @param  fmt    Printf style format for each element of the vector.
 * @param  fname  Name of the file to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_formatted_vector_with_header_i64
(
    const Vector_i64* v, 
    const char*       fmt,
    const char*       fname
)
{
    FILE* fp;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return JWSC_EIO(strerror(errno));
    }

    write_formatted_vector_with_header_fp_i64(v, fmt, fp);

    if (fclose(fp) != 0)
    {
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the vector written is a line containing the number of elements 
 * followed by a line containing all the elements. For example,
 *
 * @code
 * # Comment.
 * num_elts=9
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v      Vector to write.
 * @param  fmt    Printf style format for each element of the vector.
 * @param  fname  Name of the file to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_formatted_vector_with_header_f
(
    const Vector_f* v, 
    const char*     fmt,
    const char*     fname
)
{
    FILE* fp;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return JWSC_EIO(strerror(errno));
    }

    write_formatted_vector_with_header_fp_f(v, fmt, fp);

    if (fclose(fp) != 0)
    {
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the vector written is a line containing the number of elements 
 * followed by a line containing all the elements. For example,
 *
 * @code
 * # Comment.
 * num_elts=9
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v      Vector to write.
 * @param  fmt    Printf style format for each element of the vector.
 * @param  fname  Name of the file to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_formatted_vector_with_header_d
(
    const Vector_d* v, 
    const char*     fmt,
    const char*     fname
)
{
    FILE* fp;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return JWSC_EIO(strerror(errno));
    }

    write_formatted_vector_with_header_fp_d(v, fmt, fp);

    if (fclose(fp) != 0)
    {
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the vector written is a line containing the number of elements 
 * followed by a line containing all the elements. For example,
 *
 * @code
 * # Comment.
 * num_elts=3
 * 1 2 3 4 5 6
 * @endcode
 *
 * @param  v      Vector to write.
 * @param  fmt    Printf style format for each element of the vector.
 * @param  fname  Name of the file to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_formatted_vector_with_header_cf
(
    const Vector_cf* v, 
    const char*      fmt,
    const char*      fname
)
{
    FILE* fp;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return JWSC_EIO(strerror(errno));
    }

    write_formatted_vector_with_header_fp_cf(v, fmt, fp);

    if (fclose(fp) != 0)
    {
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/**
 * Format of the vector written is a line containing the number of elements 
 * followed by a line containing all the elements. For example,
 *
 * @code
 * # Comment.
 * num_elts=3
 * 1 2 3 4 5 6
 * @endcode
 *
 * @param  v      Vector to write.
 * @param  fmt    Printf style format for each element of the vector.
 * @param  fname  Name of the file to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_formatted_vector_with_header_cd
(
    const Vector_cd* v, 
    const char*      fmt,
    const char*      fname
)
{
    FILE* fp;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return JWSC_EIO(strerror(errno));
    }

    write_formatted_vector_with_header_fp_cd(v, fmt, fp);

    if (fclose(fp) != 0)
    {
        return JWSC_EIO(strerror(errno));
    }

    return NULL;
}

/*@}*/




/**
 * @name write_vector_with_header_fp
 *
 * Writes a vector with a header to an opened file.
 * @{
 */

/**
 * Format of the vector written is a line containing the number of elements 
 * followed by a line containing all the elements. For example,
 *
 * @code
 * # Comment.
 * num_elts=9
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v   Vector to write.
 * @param  fp  Name of the file to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_vector_with_header_fp_u32(const Vector_u32* v, FILE* fp)
{
    return write_formatted_vector_with_header_fp_u32(v, "%u", fp);
}

/**
 * Format of the vector written is a line containing the number of elements 
 * followed by a line containing all the elements. For example,
 *
 * @code
 * # Comment.
 * num_elts=9
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v   Vector to write.
 * @param  fp  Name of the file to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_vector_with_header_fp_i32(const Vector_i32* v, FILE* fp)
{
    return write_formatted_vector_with_header_fp_i32(v, "%d", fp);
}

/**
 * Format of the vector written is a line containing the number of elements 
 * followed by a line containing all the elements. For example,
 *
 * @code
 * # Comment.
 * num_elts=9
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v   Vector to write.
 * @param  fp  Name of the file to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_vector_with_header_fp_i64(const Vector_i64* v, FILE* fp)
{
#ifdef JWSC_HAVE_64_BIT_LONG_INT
    return write_formatted_vector_with_header_fp_i64(v, "%ld", fp);
#else
    return write_formatted_vector_with_header_fp_i64(v, "%lld", fp);
#endif
}

/**
 * Format of the vector written is a line containing the number of elements 
 * followed by a line containing all the elements. For example,
 *
 * @code
 * # Comment.
 * num_elts=9
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v   Vector to write.
 * @param  fp  Name of the file to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_vector_with_header_fp_f(const Vector_f* v, FILE* fp)
{
    return write_formatted_vector_with_header_fp_f(v, "%f", fp);
}

/**
 * Format of the vector written is a line containing the number of elements 
 * followed by a line containing all the elements. For example,
 *
 * @code
 * # Comment.
 * num_elts=9
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v   Vector to write.
 * @param  fp  Name of the file to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_vector_with_header_fp_d(const Vector_d* v, FILE* fp)
{
    return write_formatted_vector_with_header_fp_d(v, "%.8e", fp);
}

/**
 * Format of the vector written is a line containing the number of elements 
 * followed by a line containing all the elements. For example,
 *
 * @code
 * # Comment.
 * num_elts=3
 * 1 2 3 4 5 6
 * @endcode
 *
 * @param  v   Vector to write.
 * @param  fp  Name of the file to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_vector_with_header_fp_cf(const Vector_cf* v, FILE* fp)
{
    return write_formatted_vector_with_header_fp_cf(v, "%f %f", fp);
}

/**
 * Format of the vector written is a line containing the number of elements 
 * followed by a line containing all the elements. For example,
 *
 * @code
 * # Comment.
 * num_elts=3
 * 1 2 3 4 5 6
 * @endcode
 *
 * @param  v   Vector to write.
 * @param  fp  Name of the file to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_vector_with_header_fp_cd(const Vector_cd* v, FILE* fp)
{
    return write_formatted_vector_with_header_fp_cd(v, "%lf %lf", fp);
}

/*@}*/




/**
 * @name write_formatted_vector_with_header_fp
 *
 * Writes a formatted vector with a header to an opened file.
 * @{
 */

/**
 * Format of the vector written is a line containing the number of elements 
 * followed by a line containing all the elements. For example,
 *
 * @code
 * # Comment.
 * num_elts=9
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v    Vector to write.
 * @param  fmt  Printf style format for each element of the vector.
 * @param  fp   Name of the file to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_formatted_vector_with_header_fp_u32
(
    const Vector_u32* v, 
    const char*       fmt,
    FILE*             fp
)
{
    fprintf(fp, "num_elts=%u\n", v->num_elts);

    return write_formatted_vector_fp_u32(v, fmt, fp);
}

/**
 * Format of the vector written is a line containing the number of elements 
 * followed by a line containing all the elements. For example,
 *
 * @code
 * # Comment.
 * num_elts=9
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v    Vector to write.
 * @param  fmt  Printf style format for each element of the vector.
 * @param  fp   Name of the file to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_formatted_vector_with_header_fp_i32
(
    const Vector_i32* v, 
    const char*       fmt,
    FILE*             fp
)
{
    fprintf(fp, "num_elts=%u\n", v->num_elts);

    return write_formatted_vector_fp_i32(v, fmt, fp);
}

/**
 * Format of the vector written is a line containing the number of elements 
 * followed by a line containing all the elements. For example,
 *
 * @code
 * # Comment.
 * num_elts=9
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v    Vector to write.
 * @param  fmt  Printf style format for each element of the vector.
 * @param  fp   Name of the file to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_formatted_vector_with_header_fp_i64
(
    const Vector_i64* v, 
    const char*       fmt,
    FILE*             fp
)
{
    fprintf(fp, "num_elts=%u\n", v->num_elts);

    return write_formatted_vector_fp_i64(v, fmt, fp);
}

/**
 * Format of the vector written is a line containing the number of elements 
 * followed by a line containing all the elements. For example,
 *
 * @code
 * # Comment.
 * num_elts=9
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v    Vector to write.
 * @param  fmt  Printf style format for each element of the vector.
 * @param  fp   Name of the file to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_formatted_vector_with_header_fp_f
(
    const Vector_f* v, 
    const char*     fmt,
    FILE*           fp
)
{
    fprintf(fp, "num_elts=%u\n", v->num_elts);

    return write_formatted_vector_fp_f(v, fmt, fp);
}

/**
 * Format of the vector written is a line containing the number of elements 
 * followed by a line containing all the elements. For example,
 *
 * @code
 * # Comment.
 * num_elts=9
 * 1 2 3 4 5 6 7 8 9
 * @endcode
 *
 * @param  v    Vector to write.
 * @param  fmt  Printf style format for each element of the vector.
 * @param  fp   Name of the file to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_formatted_vector_with_header_fp_d
(
    const Vector_d* v, 
    const char*     fmt,
    FILE*           fp
)
{
    fprintf(fp, "num_elts=%u\n", v->num_elts);

    return write_formatted_vector_fp_d(v, fmt, fp);
}

/**
 * Format of the vector written is a line containing the number of elements 
 * followed by a line containing all the elements. For example,
 *
 * @code
 * # Comment.
 * num_elts=3
 * 1 2 3 4 5 6
 * @endcode
 *
 * @param  v    Vector to write.
 * @param  fmt  Printf style format for each element of the vector.
 * @param  fp   Name of the file to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_formatted_vector_with_header_fp_cf
(
    const Vector_cf* v, 
    const char*      fmt,
    FILE*            fp
)
{
    fprintf(fp, "num_elts=%u\n", v->num_elts);

    return write_formatted_vector_fp_cf(v, fmt, fp);
}

/**
 * Format of the vector written is a line containing the number of elements 
 * followed by a line containing all the elements. For example,
 *
 * @code
 * # Comment.
 * num_elts=3
 * 1 2 3 4 5 6
 * @endcode
 *
 * @param  v    Vector to write.
 * @param  fmt  Printf style format for each element of the vector.
 * @param  fp   Name of the file to read the vector from.
 *
 * @return On success, NULL is returned; otherwise an Error is returned.
 * - @link Error_types::ERROR_IO ERROR_IO @endlink 
 *   There was a problem writing to the file.
 */
Error* write_formatted_vector_with_header_fp_cd
(
    const Vector_cd* v, 
    const char*      fmt,
    FILE*            fp
)
{
    fprintf(fp, "num_elts=%u\n", v->num_elts);

    return write_formatted_vector_fp_cd(v, fmt, fp);
}

/*@}*/
