/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Declarations for I/O operations on a vector.
 */


#ifndef VECTOR_IO_H 
#define VECTOR_IO_H 


#include <jwsc/config.h>

#include <stdlib.h>
#include <stdio.h>
#include <inttypes.h>

#include "jwsc/base/error.h"
#include "jwsc/math/complex.h"
#include "jwsc/vector/vector.h"


#ifdef __cplusplus
namespace jwsc {
extern "C" {
#endif


/**
 * @name read_vector
 *
 * Reads a vector from a file.
 * @{
 */

/** @brief Reads a 32-bit unsigned integer vector. */
Error* read_vector_u32(Vector_u32** v_out, const char* fname);

/** @brief Reads a 32-bit signed integer vector. */
Error* read_vector_i32(Vector_i32** v_out, const char* fname);

/** @brief Reads a 64-bit signed integer vector. */
Error* read_vector_i64(Vector_i64** v_out, const char* fname);

/** @brief Reads a single precision floating point vector. */
Error* read_vector_f(Vector_f** v_out, const char* fname);

/** @brief Reads a double precision floating point vector. */
Error* read_vector_d(Vector_d** v_out, const char* fname);

/** @brief Reads a single precision complex floating point vector. */
Error* read_vector_cf(Vector_cf** v_out, const char* fname);

/** @brief Reads a double precision complex floating point vector. */
Error* read_vector_cd(Vector_cd** v_out, const char* fname);

/*@}*/




/**
 * @name read_vector_fp
 *
 * Reads a vector from an opened file.
 * @{
 */

/** @brief Reads a 32-bit unsigned integer vector. */
Error* read_vector_fp_u32(Vector_u32** v_out, FILE* fp);

/** @brief Reads a 32-bit signed integer vector. */
Error* read_vector_fp_i32(Vector_i32** v_out, FILE* fp);

/** @brief Reads a 64-bit signed integer vector. */
Error* read_vector_fp_i64(Vector_i64** v_out, FILE* fp);

/** @brief Reads a single precision floating point vector. */
Error* read_vector_fp_f(Vector_f** v_out, FILE* fp);

/** @brief Reads a double precision floating point vector. */
Error* read_vector_fp_d(Vector_d** v_out, FILE* fp);

/** @brief Reads a single precision complex floating point vector. */
Error* read_vector_fp_cf(Vector_cf** v_out, FILE* fp);

/** @brief Reads a double precision complex floating point vector. */
Error* read_vector_fp_cd(Vector_cd** v_out, FILE* fp);

/*@}*/




/**
 * @name write_vector
 *
 * Writes a vector to a file.
 * @{
 */

/** @brief Writes a 32-bit unsigned integer vector. */
Error* write_vector_u32(const Vector_u32* v, const char* fname);

/** @brief Writes a 32-bit signed integer vector. */
Error* write_vector_i32(const Vector_i32* v, const char* fname);

/** @brief Writes a 64-bit signed integer vector. */
Error* write_vector_i64(const Vector_i64* v, const char* fname);

/** @brief Writes a single precision floating point vector. */
Error* write_vector_f(const Vector_f* v, const char* fname);

/** @brief Writes a double precision floating point vector. */
Error* write_vector_d(const Vector_d* v, const char* fname);

/** @brief Writes a single precision complex floating point vector. */
Error* write_vector_cf(const Vector_cf* v, const char* fname);

/** @brief Writes a double precision complex floating point vector. */
Error* write_vector_cd(const Vector_cd* v, const char* fname);

/*@}*/




/**
 * @name write_formatted_vector
 *
 * Writes a formatted vector to a file.
 * @{
 */

/** @brief Writes a formatted 32-bit unsigned integer vector. */
Error* write_formatted_vector_u32
(
    const Vector_u32* v, 
    const char*       fmt,
    const char*       fname
);

/** @brief Writes a formatted 32-bit signed integer vector. */
Error* write_formatted_vector_i32
(
    const Vector_i32* v, 
    const char*       fmt,
    const char*       fname
);

/** @brief Writes a formatted 64-bit signed integer vector. */
Error* write_formatted_vector_i64
(
    const Vector_i64* v, 
    const char*       fmt,
    const char*       fname
);

/** @brief Writes a formatted single precision floating point vector. */
Error* write_formatted_vector_f
(
    const Vector_f* v, 
    const char*     fmt,
    const char*     fname
);

/** @brief Writes a formatted double precision floating point vector. */
Error* write_formatted_vector_d
(
    const Vector_d* v, 
    const char*     fmt,
    const char*     fname
);

/** @brief Writes a formatted single precision complex floating point vector. */
Error* write_formatted_vector_cf
(
    const Vector_cf* v, 
    const char*      fmt,
    const char*      fname
);

/** @brief Writes a formatted double precision complex floating point vector. */
Error* write_formatted_vector_cd
(
    const Vector_cd* v, 
    const char*      fmt,
    const char*      fname
);

/*@}*/




/**
 * @name write_vector_fp
 *
 * Writes a vector to an opened file.
 * @{
 */

/** @brief Writes a 32-bit unsigned integer vector. */
Error* write_vector_fp_u32(const Vector_u32* v, FILE* fp);

/** @brief Writes a 32-bit signed integer vector. */
Error* write_vector_fp_i32(const Vector_i32* v, FILE* fp);

/** @brief Writes a 64-bit signed integer vector. */
Error* write_vector_fp_i64(const Vector_i64* v, FILE* fp);

/** @brief Writes a single precision floating point vector. */
Error* write_vector_fp_f(const Vector_f* v, FILE* fp);

/** @brief Writes a double precision floating point vector. */
Error* write_vector_fp_d(const Vector_d* v, FILE* fp);

/** @brief Writes a single precision complex floating point vector. */
Error* write_vector_fp_cf(const Vector_cf* v, FILE* fp);

/** @brief Writes a double precision complex floating point vector. */
Error* write_vector_fp_cd(const Vector_cd* v, FILE* fp);

/*@}*/




/**
 * @name write_formatted_vector_fp
 *
 * Writes a vector to an opened file.
 * @{
 */

/** @brief Writes a formatted 32-bit unsigned integer vector. */
Error* write_formatted_vector_fp_u32
(
    const Vector_u32* v, 
    const char*       fmt,
    FILE*             fp
);

/** @brief Writes a formatted 32-bit signed integer vector. */
Error* write_formatted_vector_fp_i32
(
    const Vector_i32* v, 
    const char*       fmt,
    FILE*             fp
);

/** @brief Writes a formatted 64-bit signed integer vector. */
Error* write_formatted_vector_fp_i64
(
    const Vector_i64* v, 
    const char*       fmt,
    FILE*             fp
);

/** @brief Writes a formatted single precision floating point vector. */
Error* write_formatted_vector_fp_f
(
    const Vector_f* v, 
    const char*     fmt,
    FILE*           fp
);

/** @brief Writes a formatted double precision floating point vector. */
Error* write_formatted_vector_fp_d
(
    const Vector_d* v, 
    const char*     fmt,
    FILE*           fp
);

/** @brief Writes a formatted single precision complex floating point vector. */
Error* write_formatted_vector_fp_cf
(
    const Vector_cf* v, 
    const char*      fmt,
    FILE*            fp
);

/** @brief Writes a formatted double precision complex floating point vector. */
Error* write_formatted_vector_fp_cd
(
    const Vector_cd* v, 
    const char*      fmt,
    FILE*            fp
);

/*@}*/




/**
 * @name read_vector_with_header
 *
 * Reads a vector with a header from a file.
 * @{
 */

/** @brief Reads a 32-bit unsigned integer vector. */
Error* read_vector_with_header_u32(Vector_u32** v_out, const char* fname);

/** @brief Reads a 32-bit signed integer vector. */
Error* read_vector_with_header_i32(Vector_i32** v_out, const char* fname);

/** @brief Reads a 64-bit signed integer vector. */
Error* read_vector_with_header_i64(Vector_i64** v_out, const char* fname);

/** @brief Reads a single precision floating point vector. */
Error* read_vector_with_header_f(Vector_f** v_out, const char* fname);

/** @brief Reads a double precision floating point vector. */
Error* read_vector_with_header_d(Vector_d** v_out, const char* fname);

/** @brief Reads a single precision complex floating point vector. */
Error* read_vector_with_header_cf(Vector_cf** v_out, const char* fname);

/** @brief Reads a double precision complex floating point vector. */
Error* read_vector_with_header_cd(Vector_cd** v_out, const char* fname);

/*@}*/




/**
 * @name read_vector_with_header_fp
 *
 * Reads a vector with a header from an opened file.
 * @{
 */

/** @brief Reads a 32-bit unsigned integer vector. */
Error* read_vector_with_header_fp_u32(Vector_u32** v_out, FILE* fp);

/** @brief Reads a 32-bit signed integer vector. */
Error* read_vector_with_header_fp_i32(Vector_i32** v_out, FILE* fp);

/** @brief Reads a 64-bit signed integer vector. */
Error* read_vector_with_header_fp_i64(Vector_i64** v_out, FILE* fp);

/** @brief Reads a single precision floating point vector. */
Error* read_vector_with_header_fp_f(Vector_f** v_out, FILE* fp);

/** @brief Reads a double precision floating point vector. */
Error* read_vector_with_header_fp_d(Vector_d** v_out, FILE* fp);

/** @brief Reads a single precision complex floating point vector. */
Error* read_vector_with_header_fp_cf(Vector_cf** v_out, FILE* fp);

/** @brief Reads a double precision complex floating point vector. */
Error* read_vector_with_header_fp_cd(Vector_cd** v_out, FILE* fp);

/*@}*/




/**
 * @name write_vector_with_header
 *
 * Writes a vector with a header to a file.
 * @{
 */

/** @brief Writes a 32-bit unsigned integer vector. */
Error* write_vector_with_header_u32(const Vector_u32* v, const char* fname);

/** @brief Writes a 32-bit signed integer vector. */
Error* write_vector_with_header_i32(const Vector_i32* v, const char* fname);

/** @brief Writes a 64-bit signed integer vector. */
Error* write_vector_with_header_i64(const Vector_i64* v, const char* fname);

/** @brief Writes a single precision floating point vector. */
Error* write_vector_with_header_f(const Vector_f* v, const char* fname);

/** @brief Writes a double precision floating point vector. */
Error* write_vector_with_header_d(const Vector_d* v, const char* fname);

/** @brief Writes a single precision complex floating point vector. */
Error* write_vector_with_header_cf(const Vector_cf* v, const char* fname);

/** @brief Writes a double precision complex floating point vector. */
Error* write_vector_with_header_cd(const Vector_cd* v, const char* fname);

/*@}*/




/**
 * @name write_formatted_vector_with_header
 *
 * Writes a formatted vector with a header to a file.
 * @{
 */

/** @brief Writes a formatted 32-bit unsigned integer vector. */
Error* write_formatted_vector_with_header_u32
(
    const Vector_u32* v, 
    const char*       fmt,
    const char*       fname
);

/** @brief Writes a formatted 32-bit signed integer vector. */
Error* write_formatted_vector_with_header_i32
(
    const Vector_i32* v, 
    const char*       fmt,
    const char*       fname
);

/** @brief Writes a formatted 64-bit signed integer vector. */
Error* write_formatted_vector_with_header_i64
(
    const Vector_i64* v, 
    const char*       fmt,
    const char*       fname
);

/** @brief Writes a formatted single precision floating point vector. */
Error* write_formatted_vector_with_header_f
(
    const Vector_f* v, 
    const char*     fmt,
    const char*     fname
);

/** @brief Writes a formatted double precision floating point vector. */
Error* write_formatted_vector_with_header_d
(
    const Vector_d* v, 
    const char*     fmt,
    const char*     fname
);

/** @brief Writes a formatted single precision complex floating point vector. */
Error* write_formatted_vector_with_header_cf
(
    const Vector_cf* v, 
    const char*      fmt,
    const char*      fname
);

/** @brief Writes a formatted double precision complex floating point vector. */
Error* write_formatted_vector_with_header_cd
(
    const Vector_cd* v, 
    const char*      fmt,
    const char*      fname
);

/*@}*/




/**
 * @name write_vector_with_header_fp
 *
 * Writes a vector with a header to an opened file.
 * @{
 */

/** @brief Writes a 32-bit unsigned integer vector. */
Error* write_vector_with_header_fp_u32(const Vector_u32* v, FILE* fp);

/** @brief Writes a 32-bit signed integer vector. */
Error* write_vector_with_header_fp_i32(const Vector_i32* v, FILE* fp);

/** @brief Writes a 64-bit signed integer vector. */
Error* write_vector_with_header_fp_i64(const Vector_i64* v, FILE* fp);

/** @brief Writes a single precision floating point vector. */
Error* write_vector_with_header_fp_f(const Vector_f* v, FILE* fp);

/** @brief Writes a double precision floating point vector. */
Error* write_vector_with_header_fp_d(const Vector_d* v, FILE* fp);

/** @brief Writes a single precision complex floating point vector. */
Error* write_vector_with_header_fp_cf(const Vector_cf* v, FILE* fp);

/** @brief Writes a double precision complex floating point vector. */
Error* write_vector_with_header_fp_cd(const Vector_cd* v, FILE* fp);

/*@}*/




/**
 * @name write_formatted_vector_with_header_fp
 *
 * Writes a formatted vector with a header to an opened file.
 * @{
 */

/** @brief Writes a formatted 32-bit unsigned integer vector. */
Error* write_formatted_vector_with_header_fp_u32
(
    const Vector_u32* v, 
    const char*       fmt,
    FILE*             fp
);

/** @brief Writes a formatted 32-bit signed integer vector. */
Error* write_formatted_vector_with_header_fp_i32
(
    const Vector_i32* v, 
    const char*       fmt,
    FILE*             fp
);

/** @brief Writes a formatted 64-bit signed integer vector. */
Error* write_formatted_vector_with_header_fp_i64
(
    const Vector_i64* v, 
    const char*       fmt,
    FILE*             fp
);

/** @brief Writes a formatted single precision floating point vector. */
Error* write_formatted_vector_with_header_fp_f
(
    const Vector_f* v, 
    const char*     fmt,
    FILE*           fp
);

/** @brief Writes a formatted double precision floating point vector. */
Error* write_formatted_vector_with_header_fp_d
(
    const Vector_d* v, 
    const char*     fmt,
    FILE*           fp
);

/** @brief Writes a formatted single precision complex floating point vector. */
Error* write_formatted_vector_with_header_fp_cf
(
    const Vector_cf* v, 
    const char*      fmt,
    FILE*            fp
);

/** @brief Writes a formatted double precision complex floating point vector. */
Error* write_formatted_vector_with_header_fp_cd
(
    const Vector_cd* v, 
    const char*      fmt,
    FILE*            fp
);

/*@}*/


#ifdef __cplusplus
}
}
#endif


#endif
