/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Declarations for mathematical vector operations.
 */


#ifndef VECTOR_MATH_H 
#define VECTOR_MATH_H 


#include <jwsc/config.h>

#include <stdlib.h>

#include "jwsc/base/error.h"
#include "jwsc/math/complex.h"
#include "jwsc/vector/vector.h"


#ifdef __cplusplus
namespace jwsc {
extern "C" {
#endif


/**
 * @name add_scalar_to_vector
 *
 * Adds a scalar to a vector.
 * @{
 */

/** @brief Adds a scalar to a single precision vector. */
void add_scalar_to_vector_f(Vector_f** v_out, const Vector_f* v_in, float a);

/** @brief Adds a scalar to a double precision vector. */
void add_scalar_to_vector_d(Vector_d** v_out, const Vector_d* v_in, double a);

/** @brief Adds a scalar to a single precision complex vector. */
void add_scalar_to_vector_cf
(
    Vector_cf**      v_out, 
    const Vector_cf* v_in, 
    Complex_f        z
);

/** @brief Adds a scalar to a double precision complex vector. */
void add_scalar_to_vector_cd
(
    Vector_cd**      v_out, 
    const Vector_cd* v_in, 
    Complex_d       z
);

/*@}*/




/** 
 * @name multiply_vector_by_scalar
 *
 * Multiplies a vector by a scalar.
 * @{
 */

/** @brief Multiplies a single precision vector by a scalar */
void multiply_vector_by_scalar_f
(
    Vector_f**      v_out, 
    const Vector_f* v_in, 
    float           a
);

/** @brief Multiplies a double precision vector by a scalar. */
void multiply_vector_by_scalar_d
(
    Vector_d**      v_out, 
    const Vector_d* v_in, 
    double          a
);

/** @brief Multiplies a single precision complex vector by a scalar. */
void multiply_vector_by_scalar_cf
(
    Vector_cf**      v_out, 
    const Vector_cf* v_in, 
    Complex_f        z
);

/** @brief Multiplies a double precision complex vector by a scalar. */
void multiply_vector_by_scalar_cd
(
    Vector_cd**      v_out, 
    const Vector_cd* v_in, 
    Complex_d        z
);

/*@}*/




/** 
 * @name calc_vector_dot_prod
 *
 * Calculates the dot product of two real vectors.
 * @{
 */

/** @brief Calculates the dot product of two single precision vectors. */
Error* calc_vector_dot_prod_f
(
    float*          dp_out,
    const Vector_f* v_1, 
    const Vector_f* v_2
);

/** @brief Calculates the dot product of two double precision vectors. */
Error* calc_vector_dot_prod_d
(
    double*         dp_out,
    const Vector_d* v_1, 
    const Vector_d* v_2
);

/*@}*/




/** 
 * @name calc_3d_vector_cross_prod
 *
 * Calculates the cross product of two R^3 vectors.
 * @{
 */

/** @brief Calculates the cross product of two single precision R^3 vectors. */
Error* calc_3d_vector_cross_prod_f
(
    Vector_f**      v_out, 
    const Vector_f* v_1, 
    const Vector_f* v_2
);

/** @brief Calculates the cross product of two double precision R^3 vectors. */
Error* calc_3d_vector_cross_prod_d
(
    Vector_d**      v_out, 
    const Vector_d* v_1, 
    const Vector_d* v_2
);

/*@}*/




/** 
 * @name calc_vector_asum
 *
 * Calculates the absolute sum of a real vector.
 * @{
 */

/** @brief Calculates the absolute value sum of a single precision vector. */
void calc_vector_asum_f(float* sum_out, const Vector_f* v);

/** @brief Calculates the absolute value sum of a double precision vector. */
void calc_vector_asum_d(double* sum_out, const Vector_d* v);

/*@}*/




/** 
 * @name calc_vector_mag
 *
 * Calculates the magnitude of a real vector.
 * @{
 */

/** @brief Calculates the magnitude of a single precision vector. */
void calc_vector_mag_f(float* mag_out, const Vector_f* v);

/** @brief Calculates the magnitude of a double precision vector. */
void calc_vector_mag_d(double* mag_out, const Vector_d* v);

/*@}*/




/** 
 * @name normalize_vector_sum
 *
 * Normalizes a real vector by its sum.
 * @{
 */

/** @brief Normalizes a single precision vector by its sum. */
void normalize_vector_sum_f(Vector_f** v_out, const Vector_f* v_in);

/** @brief Normalizes a double precision vector by its sum. */
void normalize_vector_sum_d(Vector_d** v_out, const Vector_d* v_in);

/*@}*/




/** 
 * @name normalize_vector_mag
 *
 * Normalizes a real vector by its magnitude.
 * @{
 */

/** @brief Normalizes a single precision vector by its magnitude. */
void normalize_vector_mag_f(Vector_f** v_out, const Vector_f* v_in);

/** @brief Normalizes a double precision vector by its magnitude. */
void normalize_vector_mag_d(Vector_d** v_out, const Vector_d* v_in);

/*@}*/




/** 
 * @name add_vectors
 *
 * Adds two vectors element-wise.
 * @{
 */

/** @brief Adds two single precision vectors element-wise. */
Error* add_vectors_f
(
    Vector_f**      v_out, 
    const Vector_f* v_1, 
    const Vector_f* v_2
);

/** @brief Adds two double precision vectors element-wise. */
Error* add_vectors_d
(
    Vector_d**      v_out, 
    const Vector_d* v_1, 
    const Vector_d* v_2
);

/** @brief Adds two single precision complex vectors element-wise. */
Error* add_vectors_cf
(
    Vector_cf**      v_out, 
    const Vector_cf* v_1, 
    const Vector_cf* v_2
);

/** @brief Adds two double precision complex vectors element-wise. */
Error* add_vectors_cd
(
    Vector_cd**      v_out, 
    const Vector_cd* v_1, 
    const Vector_cd* v_2
);

/*@}*/




/** 
 * @name subtract_vectors
 *
 * Subtracts two vectors element-wise.
 * @{
 */

/** @brief Subtracts two single precision vectors element-wise. */
Error* subtract_vectors_f
(
    Vector_f**      v_out, 
    const Vector_f* v_1, 
    const Vector_f* v_2
);

/** @brief Subtracts two double precision vectors element-wise. */
Error* subtract_vectors_d
(
    Vector_d**      v_out, 
    const Vector_d* v_1, 
    const Vector_d* v_2
);

/** @brief Subtracts two single precision complex vectors element-wise. */
Error* subtract_vectors_cf
(
    Vector_cf**      v_out, 
    const Vector_cf* v_1, 
    const Vector_cf* v_2
);

/** @brief Subtracts two double precision complex vectors element-wise. */
Error* subtract_vectors_cd
(
    Vector_cd**      v_out, 
    const Vector_cd* v_1, 
    const Vector_cd* v_2
);

/*@}*/


#ifdef __cplusplus
}
}
#endif


#endif
