/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <math.h>

#include "jwsc/base/error.h"
#include "jwsc/vector/vector.h"
#include "jwsc/matrix/matrix.h"
#include "jwsc/prob/mv_pdf.h"


static const char* usage = "mv-gauss-test <N> [<seed>]";


int main(int argc, const char** argv)
{
    Vector_d*   x      = NULL;
    Vector_d*   mu     = NULL;
    Matrix_d*   sigma  = NULL;
    Error*      e;

    uint32_t n, row, col;
    uint32_t N;
    uint32_t seed;
    double   p;
    double   r;

    if (argc < 2)
    {
        fprintf(stderr, "usage: %s\n", usage);
        return EXIT_FAILURE;
    }

    sscanf(argv[1], "%u", &N);
    if (argc > 2)
    {
        sscanf(argv[2], "%u", &seed);
    }
    else
    {
        seed = 0;
    }

    srand(seed);

    create_random_vector_d(&x, N, 0, 1);
    create_random_vector_d(&mu, N, 0, 1);
    create_zero_matrix_d(&sigma, N, N);

    fprintf(stdout, "\n");

    fprintf(stdout, "x = \n");
    for (n = 0; n < N; n++)
    {
        fprintf(stdout, "%6.3f ", x->elts[ n ]);
    }
    fprintf(stdout, "\n");

    fprintf(stdout, "\n");

    fprintf(stdout, "mu = \n");
    for (n = 0; n < N; n++)
    {
        fprintf(stdout, "%6.3f ", mu->elts[ n ]);
    }
    fprintf(stdout, "\n");

    fprintf(stdout, "\n");

#if defined JWSC_HAVE_LAPACK
    fprintf(stdout, "sigma = \n");
    for (row = 0; row < N; row++)
    {
        for (col = 0; col < N; col++)
        {
            r = (double)rand() / RAND_MAX;
            if (col > row)
            {
                sigma->elts[ row ][ col ] = pow(r, col - row + 9);
                sigma->elts[ col ][ row ] = pow(r, col - row + 9);
            }
            else if (row == col)
            {
                sigma->elts[ row ][ col ] = r;
            }

            fprintf(stdout, "%9.6f ", sigma->elts[ row ][ col ]);
        }
        fprintf(stdout, "\n");
    }

    fprintf(stdout, "\n");

    if ((e = mv_gaussian_pdf_d(&p, mu, sigma, x)) != NULL)
    {
        print_error_msg_exit(NULL, e->msg);
    }
    fprintf(stdout, "p(x; mu, sigma) = %.3e\n", p);

    fprintf(stdout, "\n");
#else
    fprintf(stderr, "No lapack\n");
#endif

    free_matrix_d(sigma);
    free_vector_d(mu);
    free_vector_d(x);

    return EXIT_SUCCESS;
}
