/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <math.h>

#include "jwsc/base/error.h"
#include "jwsc/vector/vector.h"
#include "jwsc/vector/vector_math.h"
#include "jwsc/matrix/matrix.h"
#include "jwsc/matrix/matrix_math.h"
#include "jwsc/matrix/matrix_io.h"
#include "jwsc/prob/mv_pdf.h"
#include "jwsc/stat/stat.h"


static const char* usage = "sample-mv-gauss-test <D> <N> [<seed>]";


int main(int argc, const char** argv)
{
    Matrix_d*   x          = NULL;
    Vector_d*   y          = NULL;
    Vector_d*   mu         = NULL;
    Matrix_d*   sigma      = NULL;
    Vector_d*   mu_hat     = NULL;
    Matrix_d*   sigma_hat  = NULL;
    Error*      e;

    uint32_t n, N;
    uint32_t d, D;
    uint32_t row, col;
    uint32_t seed;
    double   r;

#if defined JWSC_HAVE_LAPACK
    if (argc < 2)
    {
        fprintf(stderr, "usage: %s\n", usage);
        return EXIT_FAILURE;
    }

    sscanf(argv[1], "%u", &D);
    sscanf(argv[2], "%u", &N);
    if (argc > 3)
    {
        sscanf(argv[3], "%u", &seed);
    }
    else
    {
        seed = 0;
    }

    srand(seed);

    create_random_vector_d(&mu, D, 0, 1);
    create_zero_matrix_d(&sigma, D, D);

    fprintf(stdout, "\n");
    fprintf(stdout, "mu = \n");
    for (d = 0; d < D; d++)
    {
        fprintf(stdout, "%6.4f ", mu->elts[ d ]);
    }
    fprintf(stdout, "\n");

    fprintf(stdout, "\n");
    fprintf(stdout, "sigma = \n");
    for (row = 0; row < D; row++)
    {
        for (col = 0; col < D; col++)
        {
            r = (double)rand() / RAND_MAX;
            if (col > row)
            {
                sigma->elts[ row ][ col ] = pow(r, col - row + 9);
                sigma->elts[ col ][ row ] = pow(r, col - row + 9);
            }
            else if (row == col)
            {
                sigma->elts[ row ][ col ] = r;
            }

            fprintf(stdout, "%6.4f ", sigma->elts[ row ][ col ]);
        }
        fprintf(stdout, "\n");
    }
    fprintf(stdout, "\n");

    /*
    create_matrix_d(&x, N, D);
    for (n = 0; n < N; n++)
    {
        if ((e = sample_mv_gaussian_pdf_d(&y, mu, sigma)) != NULL)
        {
            print_error_msg_exit(NULL, e->msg);
        }

        for (d = 0; d < D; d++)
        {
            x->elts[n][d] = y->elts[d];
        }
    }
    */
    if ((e = set_sample_mv_gaussian_pdf_d(&x, N, mu, sigma)) != NULL)
    {
        print_error_msg_exit(NULL, e->msg);
    }

    //assert(write_matrix_fp_d(x, stdout) == NULL);

    mv_sample_stats_d(&mu_hat, &sigma_hat, x);

    subtract_vectors_d(&mu, mu, mu_hat);
    subtract_matrices_d(&sigma, sigma, sigma_hat);

    fprintf(stdout, "\n");
    fprintf(stdout, "mu - mu_hat = \n");
    for (d = 0; d < D; d++)
    {
        fprintf(stdout, "%6.4f ", mu->elts[ d ]);
    }
    fprintf(stdout, "\n");

    fprintf(stdout, "\n");
    fprintf(stdout, "sigma - sigma_hat = \n");
    for (n = 0; n < D; n++)
    {
        for (d = 0; d < D; d++)
        {
            fprintf(stdout, "%6.4f ", sigma->elts[ n ][ d ]);
        }
        fprintf(stdout, "\n");
    }
    fprintf(stdout, "\n");

    free_matrix_d(sigma);
    free_vector_d(mu);
    free_matrix_d(sigma_hat);
    free_vector_d(mu_hat);
    free_matrix_d(x);
    free_vector_d(y);
#else
    fprintf(stderr, "No lapack\n");
#endif

    return EXIT_SUCCESS;
}
